/******************************************/ 
/* Carbon-compatible section of OroborOSX */
/* (c) Adrian Umpleby Jan 2002            */
/*                                        */
/* Original oroborus window manager code  */
/* (c) Ken Lynch Jan 2001                 */
/*                                        */
/* This file started life as main.c from  */
/* Apple's prototype Carbon application   */
/*                                        */
/* Distributed under the GPL              */ 
/* See LICENSE for more details           */ 
/******************************************/ 
 
/* macevents.c */

#include <string.h>
#include "main.h"
#include "hackcmds.h"

int	idleCount=0, noIdleMax=3;
long	iconbar_WaitTicks=100, gGotMenuClick=0, lastMovedTen=0, time_to_leave=0, count_time_start=0;


static OSStatus LowRunAppleScript(const void* text, long textLength,
                                    AEDesc *resultData) {
    ComponentInstance theComponent;
    AEDesc scriptTextDesc;
    OSStatus err;
    OSAID scriptID, resultID;

        /* set up locals to a known state */
    theComponent = NULL;
    AECreateDesc(typeNull, NULL, 0, &scriptTextDesc);
    scriptID = kOSANullScript;
    resultID = kOSANullScript;

        /* open the scripting component */
    theComponent = OpenDefaultComponent(kOSAComponentType,
                    typeAppleScript);
    if (theComponent == NULL) { err = paramErr; goto bail; }

        /* put the script text into an aedesc */
    err = AECreateDesc(typeChar, text, textLength, &scriptTextDesc);
    if (err != noErr) goto bail;

        /* compile the script */
    err = OSACompile(theComponent, &scriptTextDesc,
                    kOSAModeNull, &scriptID);
    if (err != noErr) goto bail;

        /* run the script */
    err = OSAExecute(theComponent, scriptID, kOSANullScript,
                    kOSAModeNull, &resultID);

        /* collect the results - if any */
    if (resultData != NULL) {
        AECreateDesc(typeNull, NULL, 0, resultData);
        if (err == errOSAScriptError) {
            OSAScriptError(theComponent, kOSAErrorMessage,
                        typeChar, resultData);
        } else if (err == noErr && resultID != kOSANullScript) {
            OSADisplay(theComponent, resultID, typeChar,
                        kOSAModeNull, resultData);
        }
    }
bail:
    AEDisposeDesc(&scriptTextDesc);
    if (scriptID != kOSANullScript) OSADispose(theComponent, scriptID);
    if (resultID != kOSANullScript) OSADispose(theComponent, resultID);
    if (theComponent != NULL) CloseComponent(theComponent);
    printf("LowRunAppleScript returned: %d\n",err);
    return err;
}

    /* SimpleRunAppleScript compiles and runs the AppleScript in
    the c-style string provided as a parameter.  The result returned
    indicates the success of the operation. */
static OSStatus SimpleRunAppleScript(const char* theScript) {
    return LowRunAppleScript(theScript, strlen(theScript), NULL);
}

void OpenXDarwinPrefs()
{
    SimpleRunAppleScript(
        "tell application \"XDarwin\"\n"
	"   set visible of window named \"Preferences\" to true\n"
	"   set visible of window named \"Voorkeuren\" to true\n"
	"   set visible of window named \"Preferencias\" to true\n"
	"   set visible of window named \"Preferenze\" to true\n"
	"   set visible of window named \"Instllningar\" to true\n"
	"   set visible of window named \"Prfrences\" to true\n"
	"   set visible of window named \"Einstellungen\" to true\n"
//	"   set visible of window named \"tX-[\" to true\n"
//	"   set visible of window named \"X$\" to true\n"
	"   activate\n"
        "end tell\n");
    SendSimpleHackCommand(HACKCMD_SHOWINGPREFS,1);
}

static OSErr QuitAppleEventHandler( const AppleEvent *appleEvt, AppleEvent* reply, UInt32 refcon )
{
    int doquit=0;
    
    sprintf(lastsub,"TRACE1: QuitAppleEventHandler\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: QuitAppleEventHandler\n");
#endif

    if(launched_xdarwin==0) {
        Bring_OOSX_To_Front();	// should put a bouncing dock icon type notification here...
        // Then again, maybe OK, since if we're not front, user probably told us to quit via dock menu...?
        // But if the user says not to quit then the original front process should really come in front again...
        doquit=ShowAlert(QUITWM_NAME,1,1);	// confirmation dialog if we did not start XDarwin
    } else if(client_count==0)	// don't bother with confirmation if there's no clients
        doquit=1;
    else {
        Bring_OOSX_To_Front();	// should put a bouncing dock icon type notification here...
        doquit=ShowAlert(QUITX11_NAME,1,1);	// confirmation dialog if we started XDarwin
    }
    
    if(doquit) {
        Allow_HUPs(0);
        quit();
        gQuitFlag =  GOTQUITAE_ERRNO;
    }
    
    return noErr;
}

static OSErr RAPPAppleEventHandler( const AppleEvent *appleEvt, AppleEvent* reply, UInt32 refcon )
{
    sprintf(lastsub,"TRACE1: RAPPAppleEventHandler\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: RAPPAppleEventHandler\n");
#endif

#ifdef DEBUG
    printf("Got a 'rapp' AppleEvent\n");
#endif
    if(gInvisible!=0) {
        gInvisible=0;
        map_all_clients();
    }
    raise_all_clients();
    
    DestroyBadWindows((WindowRef) 0);
    
    return noErr;
}

static OSErr ODOCAppleEventHandler( const AppleEvent *appleEvt, AppleEvent* reply, UInt32 refcon )
{
    OSType	HLEclass, HLEmsgID;
    OSErr	myErr;
    long	i, ni;
    int		namelen;
    DescType	typeCode;
    Size	dataSize;
    AEDescList	myDesc, newDesc;
    AEKeyword	theAEKeyword;
    Boolean	changed, dummy;
    FSSpec	mySpec;
    AliasHandle	myAlias;
    pstr_t	*pStr;
    char	*cStr;
    
    sprintf(lastsub,"TRACE1: ODOCAppleEventHandler\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: ODOCAppleEventHandler\n");
#endif

    HLEmsgID = appleEvt->descriptorType;
    HLEclass = reply->descriptorType;
    
    myErr=AECountItems(appleEvt,&ni);
    
    printf("  Number of initial items: %d\n",ni);
    
    if(ni>0) {
        myErr=AESizeOfNthItem(appleEvt, 1, &typeCode, &dataSize);
        printf("  typeCode,dataSize: %d,%d\n",typeCode,dataSize);
        myErr=AEGetNthDesc(appleEvt,1,typeCode,&theAEKeyword,&myDesc);
        HLEclass=(OSType) theAEKeyword;
        HLEmsgID=myDesc.descriptorType;
        printf("  HLEclass,HLEmsgID: %d,%d\n",HLEclass,HLEmsgID);
        myErr=AECountItems(&myDesc,&ni);
        printf("  Number of actual items: %d\n",ni);
        
        for(i=1;i<=ni;i++) {
            myErr=AESizeOfNthItem(&myDesc,i,&typeCode,&dataSize);
            printf("  item %d   typeCode,dataSize: %d,%d\n",i,typeCode,dataSize);
            myErr=AEGetNthDesc(&myDesc,i,typeCode,&theAEKeyword,&newDesc);
            myAlias=(AliasHandle) newDesc.dataHandle;
            printf("  item %d   myAlias: %#lx\n",i,myAlias);
            myErr=ResolveAlias(NULL,myAlias,&mySpec,&changed);
            if(myErr==noErr) {
// Use the horrible hack to get C string out of a Pascal string
// -there's probably a macro somewhere to do this 'legally'...
                pStr=&(mySpec.name);
                namelen=pStr->len;
                cStr=&(pStr->thestr);
                cStr[namelen]=0;
                printf("  item %d  name: %s\n",i,cStr);
// Now look for that item in the Launch menu, and run the associated script
                myErr=LaunchMenuItem(cStr,1,0,0);
                if(myErr!=0)
                    printf("%s warning: could not launch item %s -got error %d\n",PROGNAME,cStr,myErr);
            } else
                printf("  item %d  got an error: %d\n",i,myErr);
        }
    }
    
    return noErr;
}

void InitAEHandlers()
{
    int	err;
    
// Install the quit AppleEvent handler
    err = AEInstallEventHandler( kCoreEventClass, kAEQuitApplication, NewAEEventHandlerUPP((AEEventHandlerProcPtr)QuitAppleEventHandler), 0, false );
    if (err != noErr) {
        ClearUpSockets();
        ExitToShell();
    }
    
// Install the rapp AppleEvent handler
    err = AEInstallEventHandler( kCoreEventClass, kAEReopenApplication, NewAEEventHandlerUPP((AEEventHandlerProcPtr)RAPPAppleEventHandler), 0, false );
    if (err != noErr) {
        ClearUpSockets();
        ExitToShell();
    }

// Install the odoc AppleEvent handler
    err = AEInstallEventHandler( kCoreEventClass, kAEOpenDocuments, NewAEEventHandlerUPP((AEEventHandlerProcPtr)ODOCAppleEventHandler), 0, false );
    if (err != noErr) {
        ClearUpSockets();
        ExitToShell();
    }
        
}

void EventLoop()
{
    int		dummy;
    Boolean	gotEvent;
    EventRecord	event;
    char	script[100];
        
    sprintf(lastsub,"TRACE1: EventLoop\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: EventLoop\n");
#endif

    gQuitFlag = 0;

// Make sure we're all in front once started...
    Restart_XDarwin();
    Bring_To_Front(0);
    Stop_XDarwin();

    idleCount=0;
    noIdleMax=2;
// Leave longer gaps between background mouse tracking if tracers are on
#ifdef SHOW_TRACE1
    noIdleMax=10;
#endif
#ifdef SHOW_TRACE2
    noIdleMax=20;
#endif
#ifdef SHOW_TRACE3
    noIdleMax=50;
#endif

// all ready, so make sure the idle loop is called, and that HUPs are allowed...
    gInited=1;
    Allow_HUPs(0);
    SendSimpleHackCommand(HACKCMD_ALLOWFRONT,0);
    
// xinerama should have been checked by now, so add screens to Options menu
    printf("heads: %d  screen: %d\n",count_xinerama_heads(),iconbar_screen);
    if( iconbar_screen>count_xinerama_heads() )
        iconbar_screen=count_xinerama_heads();
    else if( iconbar_screen < 0 )
        iconbar_screen=0;
    Setup_Options_Menu(itemScreenListStart + iconbar_screen);

// and launch the items in the StartupItems file
    Launch_Startup_Items();

    do
    {
        if(TickCount()>lastOKTicks+600) {
            lastOKTicks=TickCount();
// Write to console for the moment, to try and keep track of when it hangs
//            sprintf(script,"echo -n 'reRead: %d    ';date",reReadOnNextFront);
#ifdef SHOW_10S_LOG
            sprintf(script,"echo -n '%s still OK on  ';date;/usr/bin/touch /var/tmp/oroborosx.check.%s",PROGNAME,randomID);
#else
            sprintf(script,"/usr/bin/touch /var/tmp/oroborosx.check.%s",randomID);
#endif
            system(script);
// Do we still need to recheck the Launch Menu Items folder...?
/*            if(gX11InFront==1) {
                if(reReadOnNextFront<0) {
                    Create_Launch_Menu();
                    reReadOnNextFront=-reReadOnNextFront;
                }
            } else if(reReadOnNextFront>0) {
                Create_Launch_Menu();
                reReadOnNextFront--;
            }
*/
            check_osx_windows_bounds();
        }

        
    sprintf(lastsub,"TRACE1: Calling GetNextEvent/usleep in Eventloop\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: Calling GetNextEvent/usleep in Eventloop\n");
#endif

        Stop_XDarwin();
        dummy=0;
        gotEvent=FALSE;
        while(gotEvent==FALSE && dummy<5 && mouse_mode==MOUSE_NORMAL) {
            usleep(10000);
            gotEvent = GetNextEvent(everyEvent,&event);
            dummy++;
        }
//        gotEvent = WaitNextEvent(everyEvent,&event,2,nil);
        Restart_XDarwin();

    sprintf(lastsub,"TRACE1: Back from GetNextEvent/usleep in Eventloop\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: Back from GetNextEvent/usleep in Eventloop\n");
#endif

        /*
        if(gX11InFront!=0 || mouse_mode!=MOUSE_NORMAL) {
            if( moving==2 )
                gotEvent = WaitNextEvent(everyEvent,&event,2,nil);
            else
                gotEvent = WaitNextEvent(everyEvent,&event,6,nil);
        } else if( moving==2 )
            gotEvent = WaitNextEvent(everyEvent,&event,3,nil);
        else
            gotEvent = WaitNextEvent(everyEvent,&event,8,nil);
        */
        if(gQuitFlag==0) {
            if (event.what!=0)
                DoEvent(&event);
            if(gQuitFlag==0) {
                if(idleCount==noIdleMax || mouse_mode!=MOUSE_NORMAL) {
                    DoIdle(1);
                    idleCount=0;
                } else {
                    DoIdle(0);
                    idleCount++;
                }
            }
        }
        
    } while (gQuitFlag==0);
    
    SavePrefs();
    
    dummy=sleep(1);
    //system("/bin/sleep 1");
    sprintf(script,"/bin/rm -f /var/tmp/oroborosx.*.%s",randomID);
    system(script);
    ClearUpSockets();
    ExitToShell();					
}

void Allow_HUPs(int allow)
{
    char script[55];
    
    if(alertWin!=NULL) allow=0;
    
    if(allow==0)
        sprintf(script,"/usr/bin/touch /var/tmp/oroborosx.menus.%s",randomID);
    else
        sprintf(script,"/bin/rm -rf /var/tmp/oroborosx.menus.%s",randomID);
        
    system(script);
}

void Start_Time_Check(long addon)
{
    count_time_start=TickCount();
    time_to_leave=count_time_start+addon;
}

int Is_Time_Up()
{
    if(TickCount()>time_to_leave) return 1;
    if(TickCount()<count_time_start) return 1;	// just to cover case of wrap-around
    return 0;
}

void DoIdle(int trackMouse)
{
    int			moving;
    short		myErr;
    Point		mspt;
    ProcessSerialNumber	frontPSN;
    Boolean		x11InFront;
    Boolean		same, cursorInMenubar=FALSE;
    WindowRef		currentFrontWin;

#ifdef DEBUG
    ProcessInfoRec	procInfo;
    pstr_t		myName;
    pstr_t		procName;
#endif

    sprintf(lastsub,"TRACE1: DoIdle (gX11InFront,trackMouse: %d,%d)\n",gX11InFront,trackMouse);
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: DoIdle (gX11InFront,trackMouse: %d,%d)\n",gX11InFront,trackMouse);
#endif

    myErr=GetFrontProcess(&frontPSN);
    if(myErr!=noErr)
        return;
        
    myErr = GetCurrentProcess(&myPSN);
    if(myErr!=noErr)
        return;
        
#ifdef DEBUG
    procInfo.processAppSpec=NULL;
    procInfo.processName=(char *) &myName;
    myErr=GetProcessInformation(&myPSN,&procInfo);
    if(myErr!=noErr)
        return;
    myName.thestr[(int) myName.len]=0;

    procInfo.processAppSpec=NULL;
    procInfo.processName=(char *) &procName;
    myErr=GetProcessInformation(&frontPSN,&procInfo);
    if(myErr!=noErr)
        return;
    procName.thestr[(int) procName.len]=0;
#endif

    myErr=SameProcess(&xPSN,&frontPSN,&x11InFront);
    if(myErr!=noErr)
        return;

    if(x11InFront)
        xPSN=frontPSN;
        
    myErr=SameProcess(&myPSN,&frontPSN,&same);
    if(myErr!=noErr)
        return;

    GetGlobalMouse(&mspt);
    cursorInMenubar=FALSE;
    
    if(x11InFront || same)
        if(mspt.h>=gMainRect.left && mspt.v>=gMainRect.top && mspt.v<gMainRect.top+22 && mspt.h<=gMainRect.right) {
            cursorInMenubar=TRUE;
            left_client();
        }

// Fix for extra items on far right of menubar in OSX10.1 -only works for holding down button,
// not for click-release in menubar item...
    if(cursorInMenubar) {
        if( Button() )
// gives an extra 30 ticks after button release before allowing switch to XDarwin
            gGotMenuClick=TickCount()+30;
        else
            gGotMenuClick=0;
    } else if (gGotMenuClick!=0) {
        if( Button() ) {
// gives an extra 30 ticks after button release before allowing switch to XDarwin
            gGotMenuClick=TickCount()+30;
            cursorInMenubar=TRUE;
        } else if(gGotMenuClick>TickCount()) {
            cursorInMenubar=TRUE;
        } else
            gGotMenuClick=0;
    }
    
    currentFrontWin=FrontWindow();

    if( IsWindowVisible(currentFrontWin) ) {
        if(gInvisible!=0) {
            gInvisible=0;
            if(DestroyBadWindows(currentFrontWin)) {
                printf("%s info: Front window visible, but removed\n",PROGNAME);
                currentFrontWin=FrontWindow();
            } else
                printf("%s info: Front window is visible again\n",PROGNAME);
            map_all_clients();
        }
    } else {
        if(gInvisible==0) {
            printf("%s info: Front window is invisible\n",PROGNAME);
            gInvisible=1;
            unmap_all_clients();
        }
    }
            
    if(x11InFront) {
#ifdef SHOW_FRONT
        printf("FRONT: XDarwin\n");
#endif
// Make sure OroborOSX application is in front if the about window is frontmost...
        if(currentFrontWin==aboutWin || currentFrontWin==alertWin) {
            Bring_OOSX_To_Front();
            same=TRUE;
            x11InFront=FALSE;
            forceOOSXInFront=FALSE;
// Also needs to be frontmost if the cursor is in the menubar - but we still want to
// act as if XDarwin is in front (i.e. keep the current X11 focus window)
        } else if(cursorInMenubar) {
            Bring_OOSX_To_Front();
            forceOOSXInFront=FALSE;
        } else if(forceOOSXInFront) {
            Bring_OOSX_To_Front();
            forceOOSXInFront=FALSE;
        }
    } else if(same) {
#ifdef SHOW_FRONT
        printf("FRONT: %s\n",PROGNAME);
#endif
// But if the OroborOSX application is in front, and the about window is not frontmost...
        if(currentFrontWin!=aboutWin && currentFrontWin!=alertWin) {
// ...then we must make sure the right window has the focus (before XDarwin comes in front)
            if(currentFrontWin!=lastFrontWin) {
                MoveWindow(currentFrontWin,-5000,-5000,false);
#ifdef SHOW_EVENTS
                printf("New window in front: %#lx %#lx\n",currentFrontWin,lastFrontWin);
#endif
// tell the window manager that a new X11 window has come in front
		Restart_XDarwin();
                change_focus((long) currentFrontWin);
		Stop_XDarwin();
                lastFrontWin=currentFrontWin;
            }
// ...and then we want XDarwin in front
// (unless the cursor is in the menubar, in which case we pretend it is)
            if(!cursorInMenubar) {
		Restart_XDarwin();
                Bring_X_To_Front(0);
            	Stop_XDarwin();
            } else if(forceOOSXInFront)
                forceOOSXInFront=FALSE;
            same=FALSE;
            x11InFront=TRUE;
        }
#ifdef SHOW_FRONT
    } else {
        printf("FRONT: unknown\n");
#endif
    }

// Check to see if the frontmost window has changed
    if(currentFrontWin!=lastFrontWin) {
        if(currentFrontWin==aboutWin || currentFrontWin==alertWin) {
// Disable the window menu if the about window is frontmost
            Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
        } else {
// And now make sure the frontmost window is off the screen
            MoveWindow(currentFrontWin,-5000,-5000,false);
#ifdef SHOW_EVENTS
            printf("New window in front: %#lx %#lx\n",currentFrontWin,lastFrontWin);
#endif
            if(x11InFront) {
// tell the window manager that a new X11 window has come in front
                Restart_XDarwin();
                change_focus((long) currentFrontWin);
                Stop_XDarwin();
            }
        }
        lastFrontWin=currentFrontWin;
    }

    real_x=mspt.h-diffscrn_x;
    real_y=mspt.v-diffscrn_y;
    
    moving=0;
    if(real_x>lastrl_x+3)
        moving=2;
    else if(real_y>lastrl_y+3)
        moving=2;
    else if(real_x<lastrl_x-3)
        moving=2;
    else if(real_y<lastrl_y-3)
        moving=2;
    else if(real_x!=lastrl_x)
        moving=1;
    else if(real_y!=lastrl_y)
        moving=1;
        
    if(moving!=0) {
#ifdef SHOW_MOUSE
        printf("Mouse moved: %d,%d  (from %d,%d)\n",real_x,real_y,lastrl_x,lastrl_y);
#endif
        lastMovedTicks=TickCount();
        lastMovedTen=lastMovedTicks;
        lastrl_x=real_x;
        lastrl_y=real_y;
    } else if(TickCount()>lastMovedTen+10) {
        lastMovedTen=TickCount();
        moving=10;
    }

    if(x11InFront) {
    
	if(gX11InFront==0) {
#ifdef DEBUG
            printf("%s has come to front\n",procName.thestr);
#endif
            gX11InFront=1;
            Restart_XDarwin();
            come_to_front();
            Stop_XDarwin();
        }

//        printf("Calling do_event_loop with 2000 or 1000...\n");
        if( moving==0 )
            do_event_loop(2000,0);
        else
            do_event_loop(1000,moving);

    } else {
    
        if(gX11InFront!=0) {
#ifdef DEBUG
            if(same)
                printf("%s: I am in front!\n",procName.thestr);
            else
                printf("%s has come to front\n",procName.thestr);
#endif
            gX11InFront=0;
            Restart_XDarwin();
            gone_to_back();
            Stop_XDarwin();
        }
        
//        printf("Calling do_event_loop with 4000 or 2000...\n");
        if( moving==0 )
            do_event_loop(4000,0);
        else
            do_event_loop(2000,moving);
    }
    
    if(gQuitFlag==0)
        if(lastIconbarOutTicks!=0)
            if(TickCount()>lastIconbarOutTicks+iconbar_WaitTicks) {
                Restart_XDarwin();
                hide_all_iconbars();
            	Stop_XDarwin();
            }
}

void DoEvent(EventRecord *event)
{
    short		part;
    Boolean		hit;
    char		key;
    Rect		tempRect;
    WindowRef		whichWindow;
    long		menusel;
    Point		mouse;
    ControlHandle	control;
        
    sprintf(lastsub,"TRACE1: DoEvent (type: %d)\n",event->what);
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: DoEvent (type: %d)\n",event->what);
#endif

    switch (event->what) 
    {
        case mouseDown:
#ifdef SHOW_EVENTS
            printf("Got mouse down...");
#endif
            part = FindWindow(event->where, &whichWindow);
            switch (part)
            {
                case inMenuBar:  /* process a moused menu command */
                    SendSimpleHackCommand(HACKCMD_ALLOWFRONT,1); // X11 window click must activate XDarwin
                    Allow_HUPs(0);
                    menusel=MenuSelect(event->where);
                    SendSimpleHackCommand(HACKCMD_ALLOWFRONT,0);
                    if(alertWin==NULL) Allow_HUPs(1);
#ifdef SHOW_EVENTS
                    printf(" in menubar (%#lx)...\n",menusel);
#endif
                    if(menusel!=0)
                        DoMenuCommand(menusel,0,0);
                    break;
                    
                case inSysWindow:
                    break;
                
                case inContent:
#ifdef SHOW_EVENTS
                    printf(" in window (%#lx)...\n",whichWindow);
#endif
                    if(whichWindow != FrontWindow()) 
                        SelectWindow(whichWindow);
                    if(whichWindow==alertWin) {
                        SetPort(GetWindowPort(whichWindow));
                        mouse = event->where;    //get the mouse location
                        GlobalToLocal(&mouse);   //convert to local coordinates
                        part = FindControl(mouse, whichWindow, &control);
                        printf("   ...got control part: %d\n",part);
                    }
                    break;
                
                case inDrag:	/* pass screenBits.bounds */
#ifdef SHOW_EVENTS
                    printf(" in window title (%#lx)...\n",whichWindow);
#endif
                    Restart_XDarwin();
                    if(showingAboutWin)
                        if(whichWindow==aboutWin)
                            if(gX11InFront==1) {
                                gX11InFront=0;
                                gone_to_back();
                                }
                    GetRegionBounds(GetGrayRgn(), &tempRect);
                    DragWindow(whichWindow, event->where, &tempRect);
                    Stop_XDarwin();
                    break;
                    
                case inGrow:
                    printf(" in window resize widget (%#lx)...\n",whichWindow);
                    break;
                    
                case inGoAway:
#ifdef SHOW_EVENTS
                    printf(" in window close widget (%#lx)...\n",whichWindow);
#endif
                    Restart_XDarwin();
                    hit = TrackBox(whichWindow, event->where, part);
                    Stop_XDarwin();
                    if (hit) 
                    {
                        if(whichWindow==aboutWin) {
#ifdef DEBUG
                            printf("Hiding about window\n");
#endif
                            showingAboutWin=0;
                            DisposeWindow(aboutWin);
                            aboutWin=NULL;
                            if(showingNoFocusWin!=0) {
                                ShowWindow(noFocusWin);
                                SelectWindow(noFocusWin);
                            }
                        } else {
#ifdef DEBUG
                            printf("Closing window: %#lx (%#lx)\n",whichWindow,aboutWin);
#endif
                            DisposeWindow(whichWindow); }
/*                        quit();
                        system("/bin/sleep 1");
                        ExitToShell();*/
                    }
                    break;
                    
                case inZoomIn:
                case inZoomOut:
                    printf(" in window zoom widget (%#lx)...\n",whichWindow);
                    Restart_XDarwin();
                    hit = TrackBox(whichWindow, event->where, part);
                    Stop_XDarwin();
                    if (hit) 
                    {
                        SetPort(GetWindowPort(whichWindow));   // window must be current port
                        EraseRect(GetWindowPortBounds(whichWindow, &tempRect));   // inval/erase because of ZoomWindow bug
                        ZoomWindow(whichWindow, part, true);
                        InvalWindowRect(whichWindow, GetWindowPortBounds(whichWindow, &tempRect));	
                    }
                    break;
            }
            break;
		
            case keyDown:
            case autoKey:
                key = event->message & charCodeMask;
                if((event->modifiers & cmdKey)==0) {
//                    if(passwordWin!=NULL && key==65 && passwordWin==alertWin) {
//                        printf("Got rid of password window\n");
//                        DisposeWindow(passwordWin);
//                        passwordWin=NULL;
//                        alertWin=NULL
//                    }
                } else if (event->what == keyDown) {
                    // deal with the shift-key menu items
                    // ANNOYING!!! - why won't this work automatically?!?
                    if((event->modifiers & shiftKey)!=0) {
#ifdef SHOW_EVENTS
                        printf("Got shift-command-key %d\n",key);
#endif
                        if(key==47)		// "OroborOSX Help" in Help menu
                            DoMenuCommand((long) 0,menuHelp,itemHelp);
                        else if(key==102)	// "Auto-focus" in Window menu
                            DoMenuCommand((long) 0,menuProperties,itemAutoFocus);
                        else if(key==114)	// "Auto-raise" in Window menu
                            DoMenuCommand((long) 0,menuProperties,itemAutoRaise);
                        else if(cycleOff==0) {
                            if(key==96)		// "Next" in Window menu -as shift backquote
                                DoMenuCommand((long) 0,menuWindow,itemNext);
                            else if(key==126)	// "Next" in Window menu again -as tilde, +shift held down
                                DoMenuCommand((long) 0,menuWindow,itemNext);
                        }
                    } else {
#ifdef SHOW_EVENTS
                        printf("Got command-key: %d\n",key);
#endif
                        if(key==96) key=126; // "Send To Back" in Window menu
                        DoMenuCommand(MenuKey(key),0,0);
                    }
                }
            case activateEvt:	       /* if you needed to do something special */
                break;
                    
            case updateEvt:
                    DrawWindow((WindowRef) event->message);
                    break;
                        
            case kHighLevelEvent:
#ifdef SHOW_EVENTS
                    printf("Got an AppleEvent...\n");
#endif
                    AEProcessAppleEvent( event );
                    break;
		
            case diskEvt:
                    break;
                
            default:
#ifdef SHOW_EVENTS
                    printf("TRACE1: Got an unknown event: %d\n",event->what);
#endif
                    break;
    }
        
    sprintf(lastsub,"TRACE1: leaving DoEvent\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: leaving DoEvent\n");
#endif
}

