/******************************************/ 
/* Carbon-compatible section of OroborOSX */
/* (c) Adrian Umpleby Jan 2002            */
/*                                        */
/* Original oroborus window manager code  */
/* (c) Ken Lynch Jan 2001                 */
/*                                        */
/* Distributed under the GPL              */ 
/* See LICENSE for more details           */ 
/******************************************/ 
 
/* menus.c */

#include "hackcmds.h"
#include "main.h"

char		*nextItem=NULL, *aboutThemeItem=NULL;
int		themeItemsCount=8, launchItemsCount=17;
// guess I should really make sure the Launch menu is not too long...
int		launch_counter[250];	// should be enough items, though   :-)

/* Lookup table iconbar_steps -> menu item. */
int steps_menu_items[][2] = {{ICONSTEPS_FAST,   itemFastIconbars  },
                            {  ICONSTEPS_MED,   itemMedIconbars   },
                            {  ICONSTEPS_SLOW,  itemSlowIconbars  }};
extern int iconbar_steps;


void Edit_Startup_Items()
{
    char	*script=NULL;
    int		err;

    sprintf(lastsub,"TRACE1: Edit_Startup_Items\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: Edit_Startup_Items\n");
#endif

#ifdef DEBUG
    printf("About to malloc script in Edit_Startup_Items: %s\n",Display_Name);
#endif
    script=malloc(45+strlen(Display_Name));
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif
    if(script==NULL)
    {
	if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Edit_Startup_Items.\n", PROGNAME);
            return;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in Edit_Startup_Items.\n", PROGNAME);
            ShowAlert(WARNMEM_NAME,1,0);
            return;
        }
    }

// crude editing method - just open up the file in xedit for the user to fiddle with...
    sprintf(script,"Contents/Scripts/EditStartups '%s'",Display_Name);
    printf("%s info: Running: %s\n",PROGNAME,script);
    err=system(script);
    if(err==NOLMIFLDR_STAT)
        ShowAlert(NOLMIFLDR_NAME,1,0);
    else if(err==NOTEXIST_STAT)
        ShowAlert(NOTEXISTLAUNCH_NAME,1,0);
    else if(err!=0) {
        printf("Script returned with error: %d\n",err);
        ShowAlert(WARNLAUNCH_NAME,1,0);
    }

#ifdef DEBUG
    printf("About to free script in Edit_Startup_Items...\n");
#endif
    free(script);
#ifdef DEBUG
    printf("script freed OK\n");
#endif
}

int RunX11App(char *name, int namelen, char *args, int arglen, int inhibit)
{
    char	*script=NULL;
    int		err;

    sprintf(lastsub,"TRACE1: RunX11App (name,args: %s %s)\n",name,args);
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: RunX11App (name,args: %s %s)\n",name,args);
#endif

#ifdef DEBUG
    printf("About to malloc script in RunX11App: %d\n",namelen);
#endif
    script=malloc(namelen+arglen+strlen(Display_Name)+50);
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif

    if(script==NULL)
    {
	if(on_cmd_line || inhibit) {
            fprintf(stderr, "%s: Memory allocation failed in RunX11App.\n", PROGNAME);
            return -1;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in RunX11App.\n", PROGNAME);
            ShowAlert(WARNMEM_NAME,1,0);
            return -1;
        }
    }
    
    if (arglen==0)
        sprintf(script,"Contents/LaunchScripts/generic '%s' '%s'", Display_Name, name);
    else
        sprintf(script,"Contents/LaunchScripts/generic '%s' '%s' %s", Display_Name, name, args);

//#ifdef DEBUG
    printf("%s info: Running: %s\n",PROGNAME,script);
//#endif

// run script to launch X11 application...
    err=system(script);
    if(inhibit==0) {
        if(err==NOLMIFLDR_STAT)
            ShowAlert(NOLMIFLDR_NAME,1,0);
        else if(err==NOTEXIST_STAT)
            ShowAlert(NOTEXISTLAUNCH_NAME,1,0);
        else if(err!=0) {
            printf("Script returned with error: %d\n",err);
            ShowAlert(WARNLAUNCH_NAME,1,0); }
    }
    
#ifdef DEBUG
    printf("About to free script in RunX11App...\n");
#endif
    free(script);
#ifdef DEBUG
    printf("script freed OK\n");
#endif
    return err;
}

int RunX11Script(int item, int inhibit)
{
    MenuRef	menu;
    Str255	itemStr;
    int		namelen, cntr, err=0;
    pstr_t	*pStr;
    char	*cStr;
    char	*script=NULL;

    sprintf(lastsub,"TRACE1: RunX11Script (item: %d)\n",item);
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: RunX11Script (item: %d)\n",item);
#endif

    menu = GetMenuHandle( menuLaunch );
    GetMenuItemText(menu,item,itemStr);
    
// hack to get the string out of the pascal string
// - there's probably a macro to do this 'legally' somewhere!
    pStr=&itemStr;
    namelen=pStr->len;
    cStr=&(pStr->thestr);
    cStr[namelen]=0;
    
    // Make sure cStr has any single quotes escaped
    script=Escape_Quotes(cStr,250);
    if(script) {
      cStr=script;
      err=1; }
    namelen=strlen(cStr);

#ifdef DEBUG
    printf("Got item: (%d) %s\n",namelen,cStr);    
    printf("About to malloc script in RunX11Script: %d\n",namelen);
#endif
    script=malloc(namelen+strlen(Display_Name)+65);
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif
    if(script==NULL)
    {
	if(on_cmd_line || inhibit) {
            fprintf(stderr, "%s: Memory allocation failed in RunX11Script.\n", PROGNAME);
            if(err) free(cStr);
            return -1;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in RunX11Script.\n", PROGNAME);
            ShowAlert(WARNMEM_NAME,1,0);
            if(err) free(cStr);
            return -1;
        }
    }
    
    // counters start at zero... (oops! never mind)
    cntr=launch_counter[item-itemUserItems-1]+1;
    // Tell the LaunchList script to launch the item
    if(cntr==1)
        sprintf(script,"Contents/Scripts/LaunchList launch '%s.x11app' '%s'",cStr,Display_Name);
    else
        sprintf(script,"Contents/Scripts/LaunchList launch '%s.x11app' '%s' '%d'",cStr,Display_Name,cntr);

    if(err) free(cStr);

//#ifdef DEBUG
    printf("%s info: Running: %s\n",PROGNAME,script);
//#endif

    err=system(script);
    if(err==0)
        launch_counter[item-itemUserItems-1]=cntr;
    else if(inhibit==0) {
        if(err==NOTEXIST_STAT)
            ShowAlert(NOTEXISTLAUNCH_NAME,1,0);
        else {
            printf("Script returned with error: %d\n",err);
            ShowAlert(WARNLAUNCH_NAME,1,0); }
    }
        
#ifdef DEBUG
    printf("About to free script in RunX11Script...\n");
#endif
    free(script);
#ifdef DEBUG
    printf("script freed OK\n");
#endif
    return err;
}

/*
 * Rather 'non-Macish' way to get a list of files in the Themes folder,
 * and then use it to create the user items for the Themes submenu
 */
void Create_Themes_Menu(int setThemeDir)
{
    MenuRef	menu;
    int		i, buflen, err, jjj;
    char	*script=NULL, *buf=NULL;
    FILE	*file;
    OSStatus	stat;
    pstr_t	*pStr;
    char	*cStr;
    Str255	itemStr;

//    menu = GetMenuHandle( menuThemes );
    stat = GetMenuItemHierarchicalMenu( GetMenuHandle(menuOptions), itemThemeSubmenu, &menu );
    
    sprintf(lastsub,"TRACE1: Create_Themes_Menu (set, menu: %d, %#lx)\n",setThemeDir,menu);
    subtrace();
#ifdef SHOW_TRACE1    
    printf("TRACE1: Create_Themes_Menu (menu: %#lx)\n",menu);
#endif

#ifdef DEBUG    
    printf("About to malloc script & buf: 55 & 256\n");
#endif
    script=malloc(55);
    buf=malloc(256);
#ifdef DEBUG    
    printf("script,buf: %#lx,%#lx\n",(long) script, (long) buf);
#endif

    if(script==NULL || buf==NULL)
    {
        if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Create_Themes_Menu.\n", PROGNAME);
            ClearUpSockets();
            exit(1);
        } else {
            fprintf(stderr, "%s: Memory allocation failed in Create_Themes_Menu.\n", PROGNAME);
            ShowAlert(FATALMEM_NAME,0,0);
            gQuitFlag=FATALMEM_ERRNO;
            return;
        }
    }
    
    sprintf(script,"Contents/Scripts/ThemeList '%s'",randomID);
    printf("%s info: Running: %s\n",PROGNAME,script);
// run script to get list of items...
    err=system(script);
    if(err==NOTHEMEFLDR_STAT)
        ShowAlert(NOTHEMEFLDR_NAME,1,0);

// This is the file that the script should have just created
    sprintf(script,"/tmp/themelist.%s",randomID);
#ifdef DEBUG
    printf("%s info: Themelist file: '%s'\n",PROGNAME,script);
#endif

// Now check to see if the file exists
    if((file=fopen(script, "r"))) 
    {
// Now that we know the file exists, we start fiddling the menu entries
        themeItemsCount=CountMenuItems(menu);
        for(i=itemUserThemes;i<themeItemsCount;i++)
            DeleteMenuItem(menu,itemUserThemes+1);
        themeItemsCount=itemUserThemes;
        
// hack to easily convert the string to a pascal string
        buf++;
        
// now extract the lines from file (max 254 characters)
        while(fgets(buf, 254, file)) 
        {
// only use a line if it does not start with a hash and it is not blank
// (remember that a line includes the newline at the end of it)
            buflen=strlen(buf);
            if(buflen<2) {
                printf("Short line: (%d) %s",buflen,buf);
            } else if(strncmp(buf,"#",1)) {
                printf("%s info: Got an item: %s  ",PROGNAME,buf);
                // now hack it to a pascal string
                buf--;
                buf[0]=buflen;
                AppendMenu(menu,buf);
                buf++;
                themeItemsCount++;
            } else
                printf("Comment: %s",buf); // shouldn't be any of these anyway...?
        } 
// close the file and get rid of it
        fclose(file);
        sprintf(script,"/bin/cat /tmp/themelist.%s",randomID);
        system(script);
        sprintf(script,"/bin/rm -f /tmp/themelist.%s",randomID);
        printf("%s info: Removing file: %s\n",PROGNAME,script);
        system(script);
        
// make sure that we undo the hack before freeing this...
        buf--;

    } else
        printf("%s warning: themelist file not found: '%s'\n",PROGNAME,script);
    
// and make sure we have the right number of menu items
    themeItemsCount=CountMenuItems(menu);

    if(themeItemsCount<=itemUserThemes) {
        printf("%s: No Themes found!! Could end up fatal...\n",PROGNAME);
        ShowAlert(NOTHEMES_NAME,1,0);
        return;
    }
    
#ifdef DEBUG
    printf("About to free buf and script in Create_Themes_Menu...\n");
#endif
    free(buf);
    free(script);
#ifdef DEBUG
    printf("buf and script freed OK\n");
#endif

// search for the current theme, and set it
    if(setThemeDir!=0) {
        printf("%s info: searching for item %s...\n",PROGNAME,gThemeDir);
        setThemeDir=0;
        if(gThemeDir==NULL) {
            printf("%s info: Create_Themes_Menu with NULL theme dir set\n",PROGNAME);
        } else {
            for(i=itemUserThemes;i<themeItemsCount;i++) {
                GetMenuItemText(menu,i+1,itemStr);
                pStr=&itemStr;
                buflen=pStr->len;
                cStr=&(pStr->thestr);
                cStr[buflen]=0;
                printf("%s info: Got item %d: %s\n",PROGNAME,i+1,cStr);
                if(buflen==strlen(gThemeDir))
                    if(!strncmp(cStr,gThemeDir,buflen))
                        setThemeDir=i+1;
            }
        }
        if(setThemeDir==0) {
            if(gThemeDir!=NULL) {
                printf("%s warning: Could not find current theme in menu: %s\n",PROGNAME,gThemeDir);
                ShowAlert(NOCURTHEME_NAME,1,0);
                gThemeDir=NULL;
            } else {
                printf("%s warning: No theme set; using first in menu...\n",PROGNAME);
                setThemeDir=itemUserThemes+1;
            }
        }
        if(setThemeDir!=0)
            Setup_Themes_Menu(setThemeDir);
    }
}

/*
 * Rather 'non-Macish' way to get a list of files in the Launch Menu Items folder,
 * and then use it to create the user items for the launch menu
 */
void Create_Launch_Menu()
{
    MenuRef	menu;
    int		i, buflen, err;
    char	*script=NULL, *buf=NULL;
    FILE	*file;
    
    menu = GetMenuHandle( menuLaunch );

    sprintf(lastsub,"TRACE1: Create_Launch_Menu\n");
    subtrace();
#ifdef SHOW_TRACE1    
    printf("TRACE1: Create_Launch_Menu\n");
#endif

#ifdef DEBUG    
    printf("About to malloc script & buf: 55 & 256\n");
#endif
    script=malloc(55);
    buf=malloc(256);
#ifdef DEBUG    
    printf("script,buf: %#lx,%#lx\n",(long) script, (long) buf);
#endif

    if(script==NULL || buf==NULL)
    {
        if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Create_Launch_Menu.\n", PROGNAME);
            ClearUpSockets();
            exit(1);
        } else {
// need a nice alert here...
            fprintf(stderr, "%s: Memory allocation failed in Create_Launch_Menu.\n", PROGNAME);
            ShowAlert(FATALMEM_NAME,0,0);
            gQuitFlag=FATALMEM_ERRNO;
            return;
        }
    }
    
    sprintf(script,"Contents/Scripts/LaunchList '%s'",randomID);
    printf("%s info: Running: %s\n",PROGNAME,script);
// run script to get list of items...
    err=system(script);
    if(err==NOLMIFLDR_STAT)
        ShowAlert(NOLMIFLDR_NAME,1,0);

// This is the file that the script should have just created
    sprintf(script,"/tmp/launchlist.%s",randomID);
//#ifdef DEBUG
    printf("%s info: Launchlist file: '%s'\n",PROGNAME,script);
//#endif

// Now check to see if the file exists
    if((file=fopen(script, "r"))) 
    {
// Now that we know the file exists, we start fiddling the menu entries
        launchItemsCount=CountMenuItems(menu);
        for(i=itemUserItems;i<launchItemsCount;i++) {
            DeleteMenuItem(menu,itemUserItems+1); }
        launchItemsCount=itemUserItems;
        
// hack to easily convert the string to a pascal string
        buf++;
        
// now extract the lines from file (max 254 characters)
        while(fgets(buf, 254, file)) 
        {
// only use a line if it does not start with a hash and it is not blank
// (remember that a line includes the newline at the end of it)
            buflen=strlen(buf);
            if(buflen<2) {
                printf("Short line: (%d) %s",buflen,buf);
            } else if(strncmp(buf,"#",1)) {
                printf("%s info: Got an item: %s  ",PROGNAME,buf);
                // now hack it to a pascal string
                buf--;
                buf[0]=buflen;
                AppendMenu(menu,buf);
                buf++;
                // just have to reset these counters to zero, I think...
                // Without storing all the item names and checking to see how they have changed
                // it's difficult to keep track of which counters go where if the item list changes...
                launch_counter[launchItemsCount-itemUserItems]=0;
                launchItemsCount++;
            } else
                printf("Comment: %s",buf); // shouldn't be any of these anyway...?
        } 
// close the file and get rid of it
        fclose(file);
        sprintf(script,"/bin/rm -f /tmp/launchlist.%s",randomID);
        printf("%s info: Removing file: %s\n",PROGNAME,script);
        system(script);
        
// make sure that we undo the hack before freeing this...
        buf--;

    } else
        printf("%s warning: launchlist file not found: '%s'\n",PROGNAME,script);
    
#ifdef DEBUG
    printf("About to free buf and script in Create_Launch_Menu...\n");
#endif
    free(buf);
    free(script);
#ifdef DEBUG
    printf("buf and script freed OK\n");
#endif
    
// and make sure we have the right number of menu items
    launchItemsCount=CountMenuItems(menu);
}

void MakeMenu()		/* Put up a menu */
{
    Handle	menuBar;
    MenuRef	menu;
    long	response;
    OSErr	err;
    OSStatus	myStat;
    IBNibRef	theNIBRef;
    int		(*i)[2];
    
    sprintf(lastsub,"TRACE1: MakeMenu\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: MakeMenu\n");
#endif

    myStat = CreateNibReference(CFSTR("menubar"),&theNIBRef);
    
    myStat = SetMenuBarFromNib(theNIBRef,CFSTR("MainMenu"));
        
    DisposeNibReference(theNIBRef);

    gHiding_Iconbars=1;
    gInterleaving_Windows=0;
    
//    iconbarpos=ICONBARPOS_TOPLEFT;
//    Setup_Options_Menu(itemTopLeft);
    Setup_Options_Menu(itemTopLeft + iconbarpos);

//    Setup_Options_Menu(itemMedIconbars);
    /* Set up the right iconbar_steps item. If none apply, don't set any. */
    for (i = steps_menu_items; i < steps_menu_items+sizeof(steps_menu_items); ++i) {
        if ((*i)[0] == iconbar_steps) {
            Setup_Options_Menu((*i)[1]);
            break;
        }
    }
        
// can't do Options menu items for the screens yet -xinerama is not ready. Done later, in EventLoop
//    Setup_Options_Menu(itemScreenListStart + iconbar_screen);

    Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    Create_Launch_Menu();
//    Create_Themes_Menu(0);
//    Setup_Themes_Menu(itemUserThemes+1);
    if (NULL == gThemeDir) {
        Create_Themes_Menu(0);        // Pick a default theme.
        Setup_Themes_Menu(itemUserThemes+1);
    } else {
        Create_Themes_Menu(1);        // Use the stored theme.
    }
}

void Setup_Themes_Menu(int which)
{
    MenuRef	menu;
    int		cnt,i,namelen,kjkasjdhk;
    Str255	itemStr;
    pstr_t	*pStr;
    char	*cStr;
    OSStatus	stat;
    
    sprintf(lastsub,"TRACE3: Setup_Themes_Menu (which: %d)\n",which);
    subtrace();
//#ifdef SHOW_TRACE3
    printf("TRACE3: Setup_Themes_Menu (which: %d)\n",which);
//#endif

//    menu=GetMenuHandle( menuThemes );
    stat = GetMenuItemHierarchicalMenu( GetMenuHandle(menuOptions), itemThemeSubmenu, &menu );
    cnt=CountMenuItems(menu);
    
    if(gUseShapedTheme==0) {
//        printf("Removed mark from item %d (out of %d) on menu %#lx\n",itemToggleShaped,cnt,menu);
        CheckMenuItem(menu, itemToggleShaped, false );
    } else {
//        printf("Put checkmark on item %d (out of %d) on menu %#lx\n",itemToggleShaped,cnt,menu);
        CheckMenuItem(menu, itemToggleShaped, true );
    }
    
    if(which>itemUserThemes) {
        for(i=itemUserThemes+1;i<=cnt;i++)
            CheckMenuItem( menu, i, false );
        SetItemMark( menu, which, diamondMark );
        //printf("Put diamond mark on item: %d (out of %d)\n",which,cnt);
        // now get the name of the item and put it into gThemeDir
        GetMenuItemText(menu,which,itemStr);
        pStr=&itemStr;
        namelen=pStr->len;
        cStr=&(pStr->thestr);
        cStr[namelen]=0;
        if(gThemeDir!=NULL) free(gThemeDir);
        gThemeDir=strdup(cStr);
        printf("%s info: Set new theme directory from menu item: %s\n",PROGNAME,gThemeDir);
        GetMenuItemText(menu,itemAboutThisTheme,itemStr);
        pStr=&itemStr;
        namelen=pStr->len;
        cStr=&(pStr->thestr);
        cStr[namelen]=0;
        if(aboutThemeItem==NULL)
            aboutThemeItem=strdup(cStr);
        sprintf(cStr,"%s %s",aboutThemeItem,gThemeDir);
        namelen=strlen(cStr);
        pStr->len=namelen;
        SetMenuItemText(menu,itemAboutThisTheme,itemStr);
    }
}

void Setup_Options_Menu(int which)
{
    MenuRef	menu;
    int		cnt,i,namelen;
    Str255	itemStr;
    pstr_t	*pStr;
    char	*cStr;

    sprintf(lastsub,"TRACE3: Setup_Options_Menu (which: %d %d)\n",which,can_send_hackcmds);
    subtrace();
//#ifdef SHOW_TRACE3
    printf("TRACE3: Setup_Options_Menu (which: %d %d)\n",which,can_send_hackcmds);
//#endif

    menu = GetMenuHandle( menuOptions );
    
    if(gQuietStartup==0)
        CheckMenuItem( menu, itemQuietStartup, false );
    else
        CheckMenuItem( menu, itemQuietStartup, true );
    
    if(gHiding_Iconbars==0 && can_send_hackcmds)
        CheckMenuItem( menu, itemHideIconbars, false );
    else {
        CheckMenuItem( menu, itemHideIconbars, true );
    }
    
    if(gMinMethod==minmethod_dock && can_send_hackcmds)
        CheckMenuItem( menu, itemMinToDock, true );
    else {
        CheckMenuItem( menu, itemMinToDock, false );
    }
    
    if(gInterleaving_Windows!=0 && can_send_hackcmds!=0)
        CheckMenuItem( menu, itemInterleave, true );
    else
        CheckMenuItem( menu, itemInterleave, false );
            
    if(can_send_hackcmds==0) {
        DisableMenuItem( menu, itemHideIconbars );
        DisableMenuItem( menu, itemMinToDock );
        DisableMenuItem( menu, itemInterleave );
        DisableMenuItem( menu, itemBGWinDimmed );
        DisableMenuItem( menu, itemBGWinTranslucent );
    } else {
        EnableMenuItem( menu, itemHideIconbars );
        EnableMenuItem( menu, itemMinToDock );
        EnableMenuItem( menu, itemInterleave );
        EnableMenuItem( menu, itemBGWinDimmed );
        EnableMenuItem( menu, itemBGWinTranslucent );
        
        CheckMenuItem( menu, itemBGWinDimmed, false );
        CheckMenuItem( menu, itemBGWinTranslucent, false );

        if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED)
            CheckMenuItem( menu, itemBGWinDimmed, true );
        if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT)
            CheckMenuItem( menu, itemBGWinTranslucent, true );
    }
    
//    if(gMenubarSpace==0)
//        CheckMenuItem( menu, itemLeaveMenubar, false );
//    else
//        CheckMenuItem( menu, itemLeaveMenubar, true );

    if(shrinking_iconbars==0)
        CheckMenuItem( menu, itemFancyIconbars, false );
    else
        CheckMenuItem( menu, itemFancyIconbars, true );
    
    if(which>=itemScreenListStart) {
        cnt=CountMenuItems(menu);
        if(cnt>itemScreenListStart) {
            for(i=itemScreenListStart+1;i<=cnt;i++)
                DeleteMenuItem( menu, itemScreenListStart+1 ); }
        cnt=count_xinerama_heads();
        printf("%s info: Screens for Options menu: %d (%d)\n",PROGNAME,cnt,itemScreenListStart);
        if(cnt>1) {
            GetMenuItemText(menu,itemScreenListStart,itemStr);
// hack (again) to get the string out of the pascal string
// - there's probably a macro to do this 'legally' somewhere!
            pStr=&itemStr;
            namelen=pStr->len;
            cStr=&(pStr->thestr);
            for(i=1;i<cnt;i++) {
// strip off the screen number from the string
                cStr[namelen-1]=0;
                printf("  adding %d\n",i);
                sprintf(cStr,"%s%d",cStr,i);
                pStr->len=strlen(cStr);
                AppendMenu(menu,itemStr);
            }
        }
        cnt+=itemScreenListStart;
        if(which<=cnt) {
            for(i=itemScreenListStart;i<cnt;i++)
                CheckMenuItem( menu, i, false );
            SetItemMark( menu, which, diamondMark );
        }
    } else if(which>=itemTopLeft) {
        CheckMenuItem( menu, itemTopLeft, false );
        CheckMenuItem( menu, itemTopRight, false );
        CheckMenuItem( menu, itemBotLeft, false );
        CheckMenuItem( menu, itemBotRight, false );
        SetItemMark( menu, which, diamondMark );
    } else if(which>=itemSlowIconbars) {
        CheckMenuItem( menu, itemSlowIconbars, false );
        CheckMenuItem( menu, itemMedIconbars, false );
        CheckMenuItem( menu, itemFastIconbars, false );
        SetItemMark( menu, which, diamondMark );
    } 
}

void Setup_Properties_Menu_Translucency(OSXWinInfo *winfo)
{
    MenuRef	menu, propmenu;
    OSStatus	stat;

    
    menu = GetMenuHandle( menuWindow );
    stat = GetMenuItemHierarchicalMenu( menu, itemProperties, &propmenu );
    
    CheckMenuItem(propmenu,itemTranslucent0,false);
    CheckMenuItem(propmenu,itemTranslucent1,false);
    CheckMenuItem(propmenu,itemTranslucent2,false);
    
    if(winfo==NULL || can_send_hackcmds==0) {
        DisableMenuItem(propmenu,itemTranslucent0);
        DisableMenuItem(propmenu,itemTranslucent1);
        DisableMenuItem(propmenu,itemTranslucent2);
    } else {
        EnableMenuItem(propmenu,itemTranslucent0);
        EnableMenuItem(propmenu,itemTranslucent1);
        EnableMenuItem(propmenu,itemTranslucent2);
        if(winfo->translucency==ACTIVE_TRANSLUCENT0)
            SetItemMark(propmenu,itemTranslucent0,diamondMark);
        else if(winfo->translucency==ACTIVE_TRANSLUCENT1)
            SetItemMark(propmenu,itemTranslucent1,diamondMark);
        else if(winfo->translucency==ACTIVE_TRANSLUCENT2)
            SetItemMark(propmenu,itemTranslucent2,diamondMark);
    }
}

void Setup_Window_Menu(int has_frame, int has_close, int has_min, int has_zoom, int is_float, int can_af, int can_ar, int first_click, int click_raise, int live_resize)
{
    MenuRef	menu, propmenu;
    OSStatus	stat;
    long	nextwin;
    Str255	itemStr;
    pstr_t	*pStr;
    char	*cStr;
    int		namelen;
    
    sprintf(lastsub,"TRACE3: Setup_Window_Menu (has_frame,can_af,can_ar: %d,%d,%d)\n",has_frame,can_af,can_ar);
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: Setup_Window_Menu (has_frame,can_af,can_ar: %d,%d,%d)\n",has_frame,can_af,can_ar);
#endif

    menu = GetMenuHandle( menuWindow );
    stat = GetMenuItemHierarchicalMenu( menu, itemProperties, &propmenu );

    if(has_frame==0) {
        EnableMenuItem(menu,0);
        DisableMenuItem(menu,itemWindowshade);
        DisableMenuItem(menu,itemMinimise);
        DisableMenuItem(menu,itemZoom);
        DisableMenuItem(menu,itemClose);
        DisableMenuItem(menu,itemLower);
        DisableMenuItem(menu,itemNext);
        cycleOff=1;
        DisableMenuItem(menu,itemProperties);
        DisableMenuItem(propmenu,0);
        CheckMenuItem(propmenu,itemFirstClick,false);
        CheckMenuItem(propmenu,itemAutoFocus,false);
        CheckMenuItem(propmenu,itemAutoRaise,false);
        CheckMenuItem(propmenu,itemClickRaise,false);
        CheckMenuItem(propmenu,itemPopup,false);
        CheckMenuItem(propmenu,itemNoframes,false);
        CheckMenuItem(propmenu,itemLiveResize,false);
//        DisableMenuItem(menu,0);
        DrawMenuBar();
        return;
    }
    
    EnableMenuItem(menu,0);

    EnableMenuItem(menu,itemWindowshade);
    EnableMenuItem(menu,itemMinimise);
    EnableMenuItem(menu,itemZoom);
    EnableMenuItem(menu,itemClose);
    EnableMenuItem(menu,itemLower);
    
        GetMenuItemText(menu,itemNext,itemStr);
        pStr=&itemStr;
        namelen=pStr->len;
        cStr=&(pStr->thestr);
        cStr[namelen]=0;
        if(nextItem==NULL)
            printf("%s info: Starting out nextItem as %s\n",PROGNAME,cStr);
        if(nextItem==NULL)
            nextItem=strdup(cStr);
        nextwin=get_next_osxwinref(0);
        if(nextwin==0) {
            sprintf(cStr,"%s",nextItem);
            DisableMenuItem(menu,itemNext);
            DisableMenuItem(menu,itemLower);
            cycleOff=1;
        } else {
            cycleOff=0;
            GetWTitle((WindowRef) nextwin,itemStr);
            pStr=&itemStr;
            namelen=pStr->len;
            cStr=&(pStr->thestr);
            cStr[namelen]=0;
            cStr=strdup(pStr->thestr);
            sprintf(pStr->thestr,"%s:  \"%s\"",nextItem,cStr);
            free(cStr);
            cStr=&(pStr->thestr);
            EnableMenuItem(menu,itemNext);
        }
#ifdef DEBUG
        printf("Setting next item to %s\n",cStr);
#endif
        namelen=strlen(cStr);
        pStr->len=namelen;
        SetMenuItemText(menu,itemNext,itemStr);
        
    EnableMenuItem(menu,itemProperties);

    if(has_close==0)
        DisableMenuItem(menu,itemClose);
    
    if(has_min==0)
        DisableMenuItem(menu,itemMinimise);

    if(has_zoom==0)
        DisableMenuItem(menu,itemZoom);

    EnableMenuItem(propmenu,0);

    EnableMenuItem(propmenu,itemFloat);
    EnableMenuItem(propmenu,itemAutoFocus);
    EnableMenuItem(propmenu,itemAutoRaise);

    if(is_float==0)
        DisableMenuItem(propmenu,itemFloat);
    else if(is_float==1)
        CheckMenuItem(propmenu,itemFloat,true);
    else if(is_float==-1)
        CheckMenuItem(propmenu,itemFloat,false);
        
    if(can_ar==0) {
        CheckMenuItem(propmenu,itemAutoRaise,false);
        DisableMenuItem(propmenu,itemAutoRaise);
    } else if(can_ar==1)
        CheckMenuItem(propmenu,itemAutoRaise,true);
    else if(can_ar==-1)
        CheckMenuItem(propmenu,itemAutoRaise,false);

    if(first_click==0)
        CheckMenuItem(propmenu,itemFirstClick,false);
    else
        CheckMenuItem(propmenu,itemFirstClick,true);

    if(click_raise==0)
        CheckMenuItem(propmenu,itemClickRaise,false);
    else
        CheckMenuItem(propmenu,itemClickRaise,true);

    if(live_resize==0)
        CheckMenuItem(propmenu,itemLiveResize,false);
    else
        CheckMenuItem(propmenu,itemLiveResize,true);

    if(can_af==0) {
        CheckMenuItem(propmenu,itemAutoFocus,false);
        DisableMenuItem(propmenu,itemAutoFocus);
        DisableMenuItem(propmenu,itemAutoRaise);
        DisableMenuItem(propmenu,itemClickRaise);
        DisableMenuItem(propmenu,itemFirstClick);
    } else if(can_af==1) {
        CheckMenuItem(propmenu,itemAutoFocus,true);
        DisableMenuItem(propmenu,itemFirstClick);
        EnableMenuItem(propmenu,itemClickRaise);
    } else if(can_af==-1) {
        CheckMenuItem(propmenu,itemAutoFocus,false);
        DisableMenuItem(propmenu,itemAutoRaise);
        DisableMenuItem(propmenu,itemClickRaise);
        EnableMenuItem(propmenu,itemFirstClick);
    }
    
    DrawMenuBar();
}

void Launch_Startup_Items()
{
    char *script=NULL;
    char buf[255];
    FILE *file;
    int buflen, err, gotbad=0;

    sprintf(lastsub,"TRACE1: Launch_Startup_Items\n");
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: Launch_Startup_Items\n");
#endif

#ifdef DEBUG
    printf("About to malloc script in Launch_Startup_Items: 60\n");
#endif
    script=malloc(60);
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif
    if(script==NULL)
    {
	if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Launch_Startup_Items.\n", PROGNAME);
            return;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in Launch_Startup_Items.\n", PROGNAME);
            ShowAlert(WARNMEM_NAME,1,0);
            return;
        }
    }

// Tell the EditStartups script to make a temporary copy of the StartupItems file
// (This lets the script do the work of finding out if it's the user's personal one)
    sprintf(script,"Contents/Scripts/EditStartups 'copy' '%s'",randomID);
    printf("%s info: Running: %s\n",PROGNAME,script);
    err=system(script);
    if(err==NOLMIFLDR_STAT)
        ShowAlert(NOLMIFLDR_NAME,1,0);
    else if(err==NOTEXIST_STAT)
        ShowAlert(NOTEXISTLAUNCH_NAME,1,0);
    else if(err!=0) {
        printf("Script returned with error: %d\n",err);
        ShowAlert(WARNLAUNCH_NAME,1,0);
    }
    
// This is the name of the temporary file
    sprintf(script,"/tmp/StartupItems.%s",randomID);
#ifdef DEBUG
    printf("Temporary StartupItems file: %s\n",script);
#endif
    
// Check to see if the file exists
    if((file=fopen(script, "r"))) 
    {        
// and extract each line from the file (max 254 characters for a line)
        while(fgets(buf, 254, file)) 
        {
// only use a line if it does not start with a hash and it is not blank
// (remember that a line includes the newline at the end of it)
            buflen=strlen(buf);
            if(buflen<2) {
#ifdef DEBUG
                printf("Short line: (%d)  %s",buflen,buf);
#endif
            } else if(strncmp(buf,"#",1)) {
// Now try to launch it (must be the same as an item in the Launch menu)
                printf("%s info: Got an item: (%d)  %s",PROGNAME,buflen,buf);
                if(LaunchMenuItem(buf,0,1,1)!=0) gotbad++;
#ifdef DEBUG
            } else {
                printf("Comment: %s",buf);
#endif
            }
        } 
// close the file
        fclose(file);
// and remove it
        sprintf(script,"/bin/rm -rf /tmp/StartupItems.%s",randomID);
//#ifdef DEBUG
        printf("%s info: removing temporary file: %s\n",PROGNAME,script);
//#endif
        system(script);
        if(gotbad==1)
            ShowAlert(LAUNCHFAIL1_NAME,1,0);
        else if(gotbad>1)
            ShowAlert(LAUNCHFAILS_NAME,1,0);
    } else
        printf("%s warning: temporary StartupItems file not found: '%s'\n",PROGNAME,script);
    
#ifdef DEBUG
    printf("About to free script in Launch_Startup_Items...\n");
#endif
    free(script);
#ifdef DEBUG
    printf("script freed OK\n");
#endif
}

void Edit_Menu_Items(char *scriptname)
{
    char	*script=NULL;
    int		err;

    sprintf(lastsub,"TRACE1: Edit_Menu_Items: %s\n",scriptname);
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: Edit_Menu_Items: %s\n",scriptname);
#endif

#ifdef DEBUG
    printf("About to malloc script in Edit_Menu_Items: 45\n");
#endif
    script=malloc(25+strlen(scriptname));
#ifdef DEBUG
    printf("script: %#lx\n",(long) script);
#endif
    if(script==NULL)
    {
	if(on_cmd_line) {
            fprintf(stderr, "%s: Memory allocation failed in Edit_Menu_Items.\n", PROGNAME);
            return;
        } else {
            fprintf(stderr, "%s: Memory allocation failed in Edit_Menu_Items.\n", PROGNAME);
            ShowAlert(WARNMEM_NAME,1,0);
            return;
        }
    }

// crude editing method - the LaunchList/ThemeList script just opens up the Launch Menu Items
// folder for the user to fiddle with...
    sprintf(script,"Contents/Scripts/%s open",scriptname);
    printf("%s info: Running: %s\n",PROGNAME,script);
    err=system(script);
    if(err==NOLMIFLDR_STAT)
        ShowAlert(NOLMIFLDR_NAME,1,0);
    else if(err==NOTHEMEFLDR_STAT)
        ShowAlert(NOTHEMEFLDR_NAME,1,0);
    
#ifdef DEBUG
    printf("About to free script in Edit_Menu_Items...\n");
#endif
    free(script);
#ifdef DEBUG
    printf("script freed OK\n");
#endif

// Once OroborOSX is brought in front, it will recheck the Launch Menu Items folder
// and continue to do so every ten seconds for about a minute - hopefully this will
// allow it to catch any changes that are made to the items in the folder...
// switched off -use "Rebuild Launch Menu" item in the Launch menu...
// or the "Rebuild Themes Menu" item in the Themes menu...
//    reReadOnNextFront=-6;
    lastOKTicks=TickCount();
}

int LaunchItem(int item, int inhibit)
{
    char argstr[128];
    int arglen=0, err=0;
    
    sprintf(lastsub,"TRACE2: LaunchItem (item: %d)\n",item);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: LaunchItem (item: %d)\n",item);
#endif

    sprintf(argstr,"");
    
    switch (item)
    {
        case itemEditStart:
            Edit_Startup_Items();
            break;
            
        case itemNewItem:
            printf("How did you do that!! -not yet available.\n");
            break;
            
        case itemEditItems:
            Edit_Menu_Items("LaunchList");
            break;
            
        case itemRebuildMenu:
            Create_Launch_Menu();
            break;
            
        case itemxinitrc:
            err=RunX11App("xinitrc",7,"",0,inhibit);
            break;
            
// These have all been replaced by x11app items in the Launch Menu Items folder...
/*
        case itemxterm:
            if(launch_count_xterm==1)
                sprintf(argstr,"-ls -sb -sl 5000 -name \"`/bin/hostname`\"");
            else
                sprintf(argstr,"-ls -sb -sl -name \"`/bin/hostname` (%d)\"",launch_count_xterm);
            arglen=strlen(argstr);
            RunX11App("xterm",5,argstr,arglen);
            launch_count_xterm++;
            break;
        
        case itemxedit:
            if(launch_count_xedit==1)
                sprintf(argstr,"-name \"xedit\"");
            else
                sprintf(argstr,"-name \"xedit (%d)\"",launch_count_xedit);
            arglen=strlen(argstr);
            RunX11App("xedit",5,argstr,arglen);
            launch_count_xedit++;
            break;
        
        case itemxbiff:
            if(launch_count_xbiff==1)
                sprintf(argstr,"-geometry 50x50-0+0 -file /var/mail/`/usr/bin/whoami` -title \"mail:`/usr/bin/whomai`\"");
            else
                sprintf(argstr,"-geometry 50x50-0+0 -file /var/mail/`/usr/bin/whoami` -title \"mail:`/usr/bin/whomai` (%d)\"",launch_count_xbiff);
            arglen=strlen(argstr);
            RunX11App("xbiff",5,argstr,arglen);
            launch_count_xbiff++;
            break;
        
        case itemxclock:
            if(launch_count_xclock==1)
                sprintf(argstr,"-geometry 70x70-0-0 -title \"xclock\"");
            else
                sprintf(argstr,"-geometry 70x70-0-0 -title \"xclock (%d)\"",launch_count_xclock);
            arglen=strlen(argstr);
            RunX11App("xclock",6,argstr,arglen);
            launch_count_xclock++;
            break;
        
        case itemxeyes:
            if(launch_count_xeyes==1)
                sprintf(argstr,"-geometry 100x60-50+0 +shape -fg black -bg skyblue -outline black -center white -title \"xeyes\"");
            else
                sprintf(argstr,"-geometry 100x60-50+0 +shape -fg black -bg skyblue -outline black -center white -title \"xeyes (%d)\"",launch_count_xeyes);
            arglen=strlen(argstr);
            RunX11App("xeyes",5,argstr,arglen);
            launch_count_xeyes++;
            break;
        
        case itemxcalc:
            if(launch_count_xcalc==1)
                sprintf(argstr,"-title \"Calculator\"");
            else
                sprintf(argstr,"-title \"Calculator (%d)\"",launch_count_xcalc);
            arglen=strlen(argstr);
            RunX11App("xcalc",5,argstr,arglen);
            launch_count_xcalc++;
            break;
*/            
        default:
            if(item>itemUserItems)
                err=RunX11Script(item,inhibit);
            else {
                printf("What menu item was that? %d\n",item);
                err=-1234;
            }
            break;
    }
    
    return err;
}

/*
 * Launch an item from the Launch menu given its name
 */
 
int LaunchMenuItem(char *itemname, int hasext, int hasnl, int inhibit)
{
    MenuRef	menu;
    Str255	itemStr;
    int		namelen, item, itemlen, err=0;
    pstr_t	*pStr;
    char	*cStr;

    itemlen=strlen(itemname);
    
    sprintf(lastsub,"TRACE2: LaunchMenuItem (itemname,itemlen,hasext: %s , %d,%d)\n",itemname,itemlen,hasext);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: LaunchMenuItem (itemname,itemlen,hasext: %s , %d,%d)\n",itemname,itemlen,hasext);
#endif

// Strip off the .x11app extension, if necessary
    if(hasext!=0) {
        if(itemlen<8) return -2;
        itemlen-=7; }
        
// And strip off the newline, if necessary
    if(hasnl!=0) {
        if(itemlen<2) return -2;
        itemlen-=1; }
    
    menu = GetMenuHandle( menuLaunch );

// Loop through the items in the Launch menu, looking for a match
    for(item=itemxinitrc;item<=launchItemsCount;item++) {
        GetMenuItemText(menu,item,itemStr);
    
// use that nasty hack again, to get the string out of the pascal string
// - there's probably a macro to do this 'legally' somewhere!
        pStr=&itemStr;
        namelen=pStr->len;
        cStr=&(pStr->thestr);
        cStr[namelen]=0;
                
// Now compare the strings, and if they match Launch that item
        if(namelen==itemlen)
            if(!strncmp(cStr,itemname,itemlen)) {
                err=LaunchItem(item,inhibit);
                return err; }
    }
    
// failed to find item
    return -1;
}

void Change_Theme(int toggleShape, int themeNum)
{
    if(themeNum==0)
        gUseShapedTheme=toggleShape;
    Setup_Themes_Menu(themeNum);
    reload_config();
}

void DoMenuCommand(long menuResult, short menuID, short menuItem)
{
//    short	menuID;		/* the resource ID of the selected menu */
//    short	menuItem;	/* the item number of the selected menu */
    
    sprintf(lastsub,"TRACE1: DoMenuCommand (menuResult: %#lx)\n",menuResult);
    subtrace();
#ifdef SHOW_TRACE1
    printf("TRACE1: DoMenuCommand (menuResult: %#lx)\n",menuResult);
#endif

    if(menuID==0 || menuItem==0) {
        if(menuResult==0) return;
        menuID = HiWord(menuResult);    /* use macros to get item & menu number */
        menuItem = LoWord(menuResult);
    }
    
#ifdef SHOW_EVENTS
    printf("Got menuID, menuItem: %d, %d (%#lx)\n",menuID,menuItem,menuResult);
#endif
    switch (menuID) 
    {
        case menuApplication:
            switch (menuItem) 
            {
                case itemAbout:
                    DoAboutBox();
                    break;
                    
/*                case itemQuit:
                    quit();
                    system("/bin/sleep 1");
                    ExitToShell();
                    break;
*/				
                default:
                    break;
            }
            break;
        
        case menuLaunch:
            LaunchItem(menuItem,0);
            break;
            
        case menuWindow:
            Restart_XDarwin();
            manipulate_focus_client(menuItem);
            Stop_XDarwin();
            break;
            
        case menuProperties:
            Restart_XDarwin();
            manipulate_focus_client(menuItem+100);
            Stop_XDarwin();
            break;
		
        case menuEdit:
            break;

        case menuOptions:
            if(menuItem==itemQuietStartup) {
                gQuietStartup = 1-gQuietStartup;
                Setup_Options_Menu(0);
                SavePrefs();
            } else if(menuItem==itemInterleave) {
                gInterleaving_Windows=1-gInterleaving_Windows;
                Setup_Options_Menu(0);
/*                passwordWin=MakeWindow(passwordWintype);
                if(passwordWin==NULL)
                    printf("%s warning: Failed to show password window!\n",PROGNAME);
                else
                    DoPasswordWindow();*/
                SavePrefs();
            } else if(menuItem==itemXDarwinPrefs) {
                Restart_XDarwin();
                OpenXDarwinPrefs();
                Stop_XDarwin();
            } else {
                Restart_XDarwin();
                change_options(menuItem);
                Stop_XDarwin();
                SavePrefs();
            }
            break;

        case menuHelp:
            if(menuItem==itemHelp)
                ShowOroborOSXHelp();
            else if(menuItem==itemVers)
                ShowHistoryFile();
                //ShowGeneralFile("Contents/Resources/HISTORY.rtf");
            else if(menuItem==itemLicense)
                ShowLicenseFile();
                //ShowGeneralFile("Contents/Resources/LICENSE.rtf");
            break;

        case menuThemes:
            switch (menuItem)
            {
                case itemAboutThemes:
                    ShowThemesInfoFile();
                    break;
                    
                case itemAddTheme:
                    break;
                    
                case itemEditThemeList:
                    Edit_Menu_Items("ThemeList");
                    break;
                    
                case itemRebuildThemes:
                    Create_Themes_Menu(1);
                    break;
                    
                case itemToggleShaped:
                    Change_Theme(1-gUseShapedTheme,0);
                    SavePrefs();
                    break;
                
                case itemAboutThisTheme:
                    ShowAboutThisTheme();
                    break;
                    
                default:
                    if(menuItem>itemUserThemes) {
                        Change_Theme(gUseShapedTheme,menuItem);
                        SavePrefs();
                    } else
                        printf("How on earth did you get that Themes menu item?! %d\n",menuItem);
                    break;
            }
            break;
    }
    HiliteMenu(0);	// unhighlight what MenuSelect (or MenuKey) hilited
    Restart_XDarwin();
    force_ungrab();	// Try to force ungrabbing of cursor (sometimes gets grabbed on menu select)
    Stop_XDarwin();
}

