/******************************************/ 
/* Carbon-compatible section of OroborOSX */
/* (c) Adrian Umpleby Jan 2002            */
/*                                        */
/* Original oroborus window manager code  */
/* (c) Ken Lynch Jan 2001                 */
/*                                        */
/* Distributed under the GPL              */ 
/* See LICENSE for more details           */ 
/******************************************/ 
 
/* osxwindows.c */

#include "main.h"

typedef struct _BadWin BadWin; 

struct _BadWin
{
    long	winref;
    BadWin	*next;
}; 

BadWin	*badwin_list=NULL;
int	badwin_count=0;

WindowRef MakeAlertWindow(const char *name)
{
    IBNibRef	theNIBRef;
    OSErr	err;
    OSStatus	myStat;
    WindowRef	myWindow=NULL;

    sprintf(lastsub,"TRACE3: MakeAlertWindow (%s)\n",name);
    subtrace();
//#ifdef SHOW_TRACE3
    printf("TRACE3: MakeAlertWindow (%s)\n",name);
//#endif

    if(name==NULL) {
        printf("%s warning: got NULL for name in MakeAlertWindow! (%s)\n",PROGNAME,name);
        return NULL;
    }

    myStat = CreateNibReference(CFSTR("alerts"),&theNIBRef);
    myStat = CreateWindowFromNib(theNIBRef,CFStringCreateWithCString(NULL,name,CFStringGetSystemEncoding()),&myWindow);

    if(myWindow==NULL) return NULL;

    ShowWindow(myWindow);    
    DisposeNibReference(theNIBRef);

    return myWindow;
}

WindowRef MakeWindow(int windowtype)
{
    WindowRef	myWindow=NULL;
    OSErr	err;
    OSStatus	myStat;
    IBNibRef	theNIBRef;
    Rect	winrct;
    
    sprintf(lastsub,"TRACE3: MakeWindow (type: %d)\n",windowtype);
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: MakeWindow (type: %d)\n",windowtype);
#endif

    if(windowtype==x11Wintype || windowtype==noFocusWintype) {
/* Only set size for the x11 windows and the 'no active window' window */    
        winrct.left=1;
        winrct.right=2;
        winrct.top=21;
        winrct.bottom=22;
    }
    
    myStat = CreateNibReference(CFSTR("window"),&theNIBRef);
    
    switch(windowtype)
    {
/* put up the information window */
    case waiting1Wintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Waiting1"),&myWindow);
        break;
    case waiting2Wintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Waiting2"),&myWindow);
        break;
    case waiting3Wintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Waiting3"),&myWindow);
        break;
/* special window for when all X11 windows are iconbars */    
    case noFocusWintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("NoFocusWin"),&myWindow);
        break;
/* put up the about window */
    case aboutWintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("About"),&myWindow);
        break;
/* put up the interleaving dialog */
    case passwordWintype:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Password"),&myWindow);
        break;
/* make a 'dummy' window, so that it appears in dock icon's window list */
    default:
        myStat = CreateWindowFromNib(theNIBRef,CFSTR("Window"),&myWindow);
        HideWindow(noFocusWin);
        showingNoFocusWin=0;
        break;
    }
    
    if(myWindow==NULL) return NULL;
    
/* Only set title if it's an x11 window */    
    if(windowtype==x11Wintype) {
        myStat=SetWindowBounds(myWindow,kWindowContentRgn,&winrct);
    } else if(windowtype==noFocusWintype)
        myStat=SetWindowBounds(myWindow,kWindowContentRgn,&winrct);
    
    ShowWindow(myWindow);
    
    DisposeNibReference(theNIBRef);

/* make sure it is not on the screen (if it's one of the x11 windows etc.) */    
    if(windowtype==x11Wintype || windowtype==noFocusWintype)
        MoveWindow(myWindow,-5000,-5000,false);
    
    return myWindow;
}

void MakeX11Window(OSXWinInfo *winfo, char *title, char *class)
{
    WindowRef	winref;
    
    sprintf(lastsub,"TRACE2: MakeX11Window (winfo,title,class: %#lx,  %s  %s)\n",winfo,title,class);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: MakeX11Window (winfo,title,class: %#lx,  %s  %s)\n",winfo,title,class);
#endif

    winref=MakeWindow(x11Wintype);
    
    if(winref==0) {
        printf("%s warning: Failed to create X11 window: %s  (%#lx)\n",PROGNAME,title,winref);
        return;
    }
    
    CheckOSXWinTitle(winfo,title,class);
    
#ifdef DEBUG
    printf("Created X11 window: %s  (%#lx)\n",title,winref);
#endif

    winfo->winref=(long) winref;
}

int DestroyBadWindows(WindowRef checkinlist)
{
    BadWin	*w, *wo;
    int		i;
    int		wasinlist=0;

    if(badwin_count==0) return 0;
    if(badwin_list==NULL) return 0;
    
    w=badwin_list;
    
    for(i=0;i<badwin_count;i++) {
#ifdef DEBUG
        printf("destroying window %d: %#lx\n",i,w->winref);
#endif
        if(w->winref==(long) checkinlist) wasinlist=1;
        ShowWindow((WindowRef) w->winref);
        SelectWindow((WindowRef) w->winref);
        DisposeWindow((WindowRef) w->winref);
        wo=w;
        w=w->next;
        free(wo);
    }
    
    badwin_list=NULL;
    badwin_count=0;
    
    return wasinlist;
}

void AddToBadWindowList(WindowRef win)
{
    BadWin	*w;
    
    w=malloc(sizeof(BadWin)+2);
    
#ifdef DEBUG
    printf("bad window %d: %#lx, %#lx\n",badwin_count+1,w,win);
#endif

    if(badwin_list==NULL) {
        badwin_list=w;
        badwin_count=0;
        w->winref=(long) win;
    } else {
        w->next=badwin_list;
        badwin_list=w;
        w->winref=(long) win;
    }
    
    badwin_count++;
}

void Check_OSX_Window_Bounds(OSXWinInfo *winfo)
{
    long	winref=0;
    
    if(winfo!=NULL);
        winref=winfo->winref;
    if(winref==0) return;

// don't bother checking bounds -just make sure it's offscreen...
    MoveWindow((WindowRef) winref,-5000,-5000,false);
}

void ReorderOSXWin(OSXWinInfo *winfo)	/* Put a window at the bottom of the dock icon window list */
{
    long	winref=0;
    
    if(winfo!=NULL);
        winref=winfo->winref;
    
    sprintf(lastsub,"TRACE2: ReorderOSXWindow (winref: %d)\n",winref);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: ReorderOSXWindow (winref: %d)\n",winref);
#endif

    if(winref==0) return;

// this is enough to put the window at the bottom of the list
    HideWindow((WindowRef) winref);
    ShowWindow((WindowRef) winref);
    
// just to make sure it's offscreen...
    MoveWindow((WindowRef) winref,-5000,-5000,false);
}

void CheckOSXWinTitle(OSXWinInfo *winfo, char *title, char *class)  // Check title in dock icon window list
{
    pstr_t	ptitle;
    char	*osxtitle;
    long	winref=0;
    
    if(winfo!=NULL);
        winref=winfo->winref;
    
//    sprintf(lastsub,"TRACE3: CheckOSXWinTitle (winref: %d)\n",winref);
//    subtrace();
//#ifdef SHOW_TRACE3
//    printf("TRACE3: CheckOSXWinTitle (winref: %d)\n",winref);
//#endif

    if(winref==0) return;
    
    osxtitle=&ptitle.thestr;
    if(title==NULL && class==NULL)
      sprintf(osxtitle,"?:  Untitled");
    else if(title==NULL)
      sprintf(osxtitle,"%s:  Untitled",class);
    else if(class==NULL)
      sprintf(osxtitle,"%s",title);
    else
      sprintf(osxtitle,"%s:  %s",class,title);
      
    ptitle.len=strlen(osxtitle);
//    strncpy((char *) &ptitle.thestr,title,ptitle.len);
    
    MoveWindow((WindowRef) winref,-5000,-5000,false);
    SetWTitle((WindowRef) winref, (char *) &ptitle);
}

void DestroyOSXWindow(OSXWinInfo *winfo)
{
    long	winref=0;
    
    if(winfo!=NULL);
        winref=winfo->winref;
    
    sprintf(lastsub,"TRACE3: DestroyOSXWindow (winref: %d)\n",winref);
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: DestroyOSXWindow (winref: %d)\n",winref);
#endif
    
    if(winref==0) return;

    if(gInvisible==0) {
        // This disposes of the OroborOSX window associated with the X11 window
        DisposeWindow((WindowRef) winref);
    } else {
        // This is to work around what appears to be a bug in the OSX (Carbon) window manager
        // -windows can't be disposed of (or hidden) while the app itself is hidden; they don't
        // disappear from the Window menu, and trying to bring up the list of windows from the
        // dock icon crashes the app... (YUK!)
        CheckOSXWinTitle(winfo,"<dead window>",NULL);
        AddToBadWindowList((WindowRef) winfo->winref);
    }
}

/*
 * X11 window has moved or changed size -update info
 * input...
 * winfo: pointer to OSWinInfo structure, which is filled in with the relevant information
 * x,y: X11 co-ords of top-left of frame
 * w,h: width and height of frame
 * wox,woy: top-left widget co-ords relative to top-left of frame
 * wendox,wendoy: bottom-right widget co-ords relative to top-left of frame
 */
void SetOSXWinPos(OSXWinInfo *winfo,int x, int y, int w, int h, int wox, int woy, int wendox, int wendoy) 
{
    long	winref=0;
    
    if(winfo!=NULL)
        winref=winfo->winref;

//    sprintf(lastsub,"TRACE2: SetOSXWinPos (winref: %d x,y: %d,%d w,h: %d,%d wsx,wsy: %d,%d wex,wey: %d,%d)\n",winref,x,y,w,h,wox,woy,wendox,wendoy);
//    subtrace();
//#ifdef SHOW_TRACE2
//    printf("TRACE2: SetOSXWinPos (winref: %d x,y: %d,%d w,h: %d,%d wsx,wsy: %d,%d wex,wey: //%d,%d)\n",winref,x,y,w,h,wox,woy,wendox,wendoy);
//#endif

    if(winref==0) return;
        
    winfo->x=x;
    winfo->y=y;
    winfo->width=w;
    winfo->height=h;
    
    winfo->widgets_x=x+wox;
    winfo->widgets_y=y+woy;
    winfo->widgets_w=x+wendox;
    winfo->widgets_h=y+wendoy;
}

void HideOSXWin(OSXWinInfo *winfo)	// called when an X11 window is minimised
{
    long	winref=0;
    
    if(winfo!=NULL);
        winref=winfo->winref;
    
//    sprintf(lastsub,"TRACE2: HideOSXWin (winref: %d)\n",winref);
//    subtrace();
//#ifdef SHOW_TRACE2
//    printf("TRACE2: HideOSXWin (winref: %d)\n",winref);
//#endif

    if(winref==0) return;
    
// note: don't actually use this since it removes it from the window list in the dock icon
//    HideWindow((WindowRef) winref);

// But this hook might be useful at some point, so that main can do something when a window is minimised
}

void FocusOSXWin(OSXWinInfo *winfo)
{
    long	winref=0;
    
    sprintf(lastsub,"TRACE2: FocusOSXWin (winfo: %#lx)\n",winfo);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: FocusOSXWin (winfo,winref: %#lx,%d)\n",winfo,winref);
#endif

    if(winfo==NULL) {
#ifdef SHOW_TRACE2
        printf("  FocusOSXWin got NULL\n");
#endif
        if(alertWin!=NULL) {
            Bring_OOSX_To_Front();
            ShowWindow(alertWin);
            SelectWindow(alertWin);
            HideWindow(noFocusWin);
            lastFrontWin=alertWin;
            showingNoFocusWin=1;
            Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
            Setup_Properties_Menu_Translucency(NULL);
        } else if(showingAboutWin) {
            ShowWindow(aboutWin);
            SelectWindow(aboutWin);
            HideWindow(noFocusWin);
            lastFrontWin=aboutWin;
            showingNoFocusWin=1;
#ifdef DEBUG
            printf("Selected window: %#lx\n",aboutWin);
#endif
            Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
            Setup_Properties_Menu_Translucency(NULL);
        } else {
            ShowWindow(noFocusWin);
            showingNoFocusWin=1;
            MoveWindow(noFocusWin,-5000,-5000,false);
            SelectWindow(noFocusWin);
#ifdef DEBUG
            printf("Selected window: %#lx\n",noFocusWin);
#endif
            lastFrontWin=noFocusWin;
            Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
            Setup_Properties_Menu_Translucency(NULL);
        }
        return;
    }
    
    HideWindow(noFocusWin);
    showingNoFocusWin=0;
    
    winref=winfo->winref;
#ifdef SHOW_TRACE2
    printf("  FocusOSXWin (winfo,winref: %#lx,%d)\n",winfo,winref);
#endif

    Setup_Properties_Menu_Translucency(winfo);
    
    ShowWindow((WindowRef) winref);
    SelectWindow((WindowRef) winref);
    lastFrontWin=(WindowRef) winref;
#ifdef DEBUG
    printf("Selected window: %#lx\n",winref);
#endif
}

void DrawWindow(WindowRef window)
{
    Rect		tempRect;
    GrafPtr		curPort;
	
    sprintf(lastsub,"TRACE3: DrawWindow (window: %#lx)\n",window);
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: DrawWindow (window: %#lx)\n",window);
#endif

    GetPort(&curPort);
    if(window!=aboutWin && window!=passwordWin && window!=alertWin)
        MoveWindow(window,-5000,-5000,false);
    SetPort(GetWindowPort(window));
    BeginUpdate(window);
    EraseRect(GetWindowPortBounds(window, &tempRect));
    DrawControls(window);
    DrawGrowIcon(window);
    EndUpdate(window);
    SetPort(curPort);
}

pascal OSStatus OKEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
//        printf("Got into OKEventHandler\n");
        gotControlEvent=OK_CTRL_ID;
        return noErr;
}

pascal OSStatus CNCLEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
//        printf("Got into CNCLEventHandler\n");
        gotControlEvent=CNCL_CTRL_ID;
        return noErr;
}

pascal OSStatus RDMEEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
//        printf("Got into RDMEEventHandler\n");
        //gotControlEvent=RDME_CTRL_ID;
        ShowOroborOSXHelp();
        return noErr;
}

pascal OSStatus LCNCEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
//        printf("Got into lcncEventHandler\n");
        //gotControlEvent=LCNC_CTRL_ID;
        ShowLicenseFile();
//        ShowGeneralFile("Contents/Resources/LICENSE.rtf");
        return noErr;
}

pascal OSStatus HISTEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
//        printf("Got into HISTEventHandler\n");
        //gotControlEvent=HIST_CTRl_ID;
        ShowHistoryFile();
//        ShowGeneralFile("Contents/Resources/History.rtf");
        return noErr;
}

int ShowAlert(const char *name, int useidle, int extrabuts)
{
    OSStatus		myStat;
    EventRef		outEvent;
    EventTargetRef 	theTarget;
    long		startPWTicks;
//    EventTypeSpec	controlSpec = { kEventClassCommand, kEventCommandProcess }; // event class, event kind
    EventTypeSpec	controlSpec = { kEventClassControl, kEventControlHit }; // event class, event kind
    ControlRef 		okControl, rdmeControl, histControl, lcncControl, cnclControl;
    ControlID 		controlID;
    int			oldQuitFlag;
    EventHandlerRef	okRef, rdmeRef, histRef, lcncRef, cnclRef;
    
    if(alertWin!=NULL) {
        printf("%s warning: Alert already being shown?!? (%s)\n",PROGNAME,name);
        return 0;
    }
    
    alertWin=MakeAlertWindow(name);
    if(alertWin==NULL) {
        printf("%s warning: Failed to create Alert window! (%s)",PROGNAME,name);
        return 0;
    }
#ifdef DEBUG
    printf("Created Alert window: %#lx (%s)",alertWin,name);
#endif

    Allow_HUPs(0);

    //Find the ok control and install an event handler on it.
    controlID.signature = 'OK  ';
    controlID.id = OK_CTRL_ID;
    GetControlByID(alertWin, &controlID, &okControl);

    InstallEventHandler(GetControlEventTarget(okControl),
                        NewEventHandlerUPP( OKEventHandler ),
                        1, &controlSpec, 0, &okRef);
//    InstallWindowEventHandler(alertWin,NewEventHandlerUPP(okEventHandler),1,&controlSpec,0,&outRef);

    if(extrabuts==3) {
        // install handlers for the three extra buttons
        controlID.signature = 'RDME';
        controlID.id = RDME_CTRL_ID;
        GetControlByID(alertWin, &controlID, &rdmeControl);
        InstallEventHandler(GetControlEventTarget(rdmeControl),
                            NewEventHandlerUPP( RDMEEventHandler ),
                            1, &controlSpec, 0, &rdmeRef);
                            
        controlID.signature = 'LCNC';
        controlID.id = LCNC_CTRL_ID;
        GetControlByID(alertWin, &controlID, &lcncControl);
        InstallEventHandler(GetControlEventTarget(lcncControl),
                            NewEventHandlerUPP( LCNCEventHandler ),
                            1, &controlSpec, 0, &lcncRef);

        controlID.signature = 'HIST';
        controlID.id = HIST_CTRL_ID;
        GetControlByID(alertWin, &controlID, &histControl);
        InstallEventHandler(GetControlEventTarget(histControl),
                            NewEventHandlerUPP( HISTEventHandler ),
                            1, &controlSpec, 0, &histRef);
    } else if(extrabuts==1) {
        // install handler for the extra cancel button
        controlID.signature = 'CNCL';
        controlID.id = CNCL_CTRL_ID;
        GetControlByID(alertWin, &controlID, &cnclControl);
        InstallEventHandler(GetControlEventTarget(cnclControl),
                            NewEventHandlerUPP( CNCLEventHandler ),
                            1, &controlSpec, 0, &cnclRef);
    }

    gotControlEvent=0;
    oldQuitFlag=gQuitFlag;
    gQuitFlag=0;
    theTarget = GetEventDispatcherTarget();
    myStat=noErr;
    startPWTicks=TickCount();
    if(gInited==0) useidle=0;
#ifdef DEBUG
    printf("Entering event loop... %d\n",startPWTicks);
#endif
    while(myStat==noErr) {
        if(useidle!=0) DoIdle(1);
        myStat=ReceiveNextEvent(0,NULL,kEventDurationForever,TRUE,&outEvent);
        if(myStat==noErr) {
//            printf("Got event... %d\n",TickCount()-startPWTicks);
            SendEventToEventTarget(outEvent, theTarget);
            ReleaseEvent(outEvent);
        } else
            printf("Got an error: %d\n",myStat);
        if(gotControlEvent!=0 || gQuitFlag!=0) myStat=1;
//        if(TickCount()>startPWTicks+3600) myStat=1;
    }
    RemoveEventHandler(okRef);
    if(extrabuts==3) {
        RemoveEventHandler(lcncRef);
        RemoveEventHandler(histRef);
        RemoveEventHandler(rdmeRef);
    } else if(extrabuts==1) {
        RemoveEventHandler(cnclRef);
    }
#ifdef DEBUG
    printf("Removing alert window...\n");
#endif
    DisposeWindow(alertWin);
    alertWin=NULL;
    Allow_HUPs(1);
    if(gQuitFlag==0) gQuitFlag=oldQuitFlag;
    // The button IDs, returned in gotControlEvent, start at 100.
    // 100 is the OK button (so return zero), 101 is cancel (so return 1).
    // The extra three buttons do not remove the alert window, so don't worry about them... 
    return gotControlEvent-100;
}

void DoPasswordWindow()
{
    OSStatus		myStat;
    EventRef		outEvent;
    EventTargetRef 	theTarget;
    long		startPWTicks;
//    EventTypeSpec	controlSpec = { kEventClassCommand, kEventCommandProcess }; // event class, event kind
    EventTypeSpec	controlSpec = { kEventClassControl, kEventControlHit }; // event class, event kind
    ControlRef 		okControl, cnclControl;
    ControlID 		controlID;
    int			oldQuitFlag, err;
    EventHandlerRef	okRef, cnclRef;
    char		*passwdText;
    Size		passwdSize;
    
    Allow_HUPs(0);

    //Find the ok control and install an event handler on it.
    controlID.signature = 'OK  ';
    controlID.id = OK_CTRL_ID;
    GetControlByID(passwordWin, &controlID, &okControl);

    InstallEventHandler(GetControlEventTarget(okControl),
                        NewEventHandlerUPP( OKEventHandler ),
                        1, &controlSpec, 0, &okRef);
//    InstallWindowEventHandler(alertWin,NewEventHandlerUPP(okEventHandler),1,&controlSpec,0,&outRef);

    //Find the cancel control and install an event handler on it.
    controlID.signature = 'CNCL';
    controlID.id = CNCL_CTRL_ID;
    GetControlByID(passwordWin, &controlID, &cnclControl);

    InstallEventHandler(GetControlEventTarget(cnclControl),
                        NewEventHandlerUPP( CNCLEventHandler ),
                        1, &controlSpec, 0, &cnclRef);
//    InstallWindowEventHandler(alertWin,NewEventHandlerUPP(okEventHandler),1,&controlSpec,0,&outRef);

    gotControlEvent=0;
    oldQuitFlag=gQuitFlag;
    gQuitFlag=0;
    if(alertWin==NULL)
        alertWin=passwordWin;
    theTarget = GetEventDispatcherTarget();
    myStat=noErr;
    startPWTicks=TickCount();
    
// get ready to read the password text
    controlID.signature = 'PWTX';
    controlID.id = PWTX_CTRL_ID;
    GetControlByID(passwordWin, &controlID, &cnclControl);
    
//    printf("Entering event loop... %d\n",startPWTicks);
    while(myStat==noErr) {

        if(gInited!=0) DoIdle(1);
        myStat=ReceiveNextEvent(0,NULL,kEventDurationForever,TRUE,&outEvent);

        if(myStat==noErr) {
//            printf("Got event... %d\n",TickCount()-startPWTicks);
            SendEventToEventTarget(outEvent, theTarget);
            ReleaseEvent(outEvent);
        } else
            printf("Got an error: %d\n",myStat);

        // check for OK button
        if(gotControlEvent==OK_CTRL_ID) {
            GetControlDataSize(cnclControl,0,kControlEditTextPasswordTag,&passwdSize);
#ifdef DEBUG
            printf("size of data: %#lx\n",passwdSize);
#endif
            // carry on if there's nothing in the edit box
            if(passwdSize==0) {
                gotControlEvent=0;
            }
        }
        
        if(gotControlEvent!=0 || gQuitFlag!=0) myStat=1;
//        if(TickCount()>startPWTicks+3600) myStat=1;
    }
    
    RemoveEventHandler(okRef);
    RemoveEventHandler(cnclRef);

    if(gotControlEvent!=OK_CTRL_ID)
    {
        if(alertWin==passwordWin)
            alertWin=NULL;
        DisposeWindow(passwordWin);
        passwordWin=NULL;
        if(gotControlEvent!=CNCL_CTRL_ID) {
            fprintf(stderr,"%s warning: failed password window: %d (%d)n",PROGNAME,myStat,gotControlEvent);
            ShowAlert(PASSWDFAIL_NAME,1,0);
#ifdef DEBUG
        } else {
            printf("Removed password window (got cancel)...\n");
#endif
        }
        Allow_HUPs(1);

    }
    else
    {
                
// Get password from text-edit field
        passwdText=malloc(passwdSize+5);    
        if(passwdText!=NULL)
            GetControlData(cnclControl,0,kControlEditTextPasswordTag,passwdSize+4,passwdText,&passwdSize);
//    success = CFStringGetCString(passwdCFStr, passwdText, 100, CFStringGetSystemEncoding()); 
     
#ifdef DEBUG
        printf("Removing password window (got OK)...\n");
#endif
        if(alertWin==passwordWin)
            alertWin=NULL;
        DisposeWindow(passwordWin);
        passwordWin=NULL;

        if(passwdText==NULL) {
            fprintf(stderr,"%s warning: could not allocate memory for password text\n",PROGNAME);
            ShowAlert(WARNMEM_NAME,1,0);
        } else if(passwdSize<1 || passwdSize>100) {
            fprintf(stderr,"%s warning: found strange password length: %d\n",PROGNAME,passwdSize);
            ShowAlert(PASSWINFAIL_NAME,1,0);
        } else {
            passwdText[passwdSize]=0;
#ifdef DEBUG
            printf("Got OK in password window... %s\n",passwdText);
#endif
            if(setenv("OOSX_TMP_PASSWD",passwdText,1)!=0) {
                fprintf(stderr,"%s warning: failed to set passwd environment variable\n",PROGNAME);
                ShowAlert(PASSWINFAIL_NAME,1,0);
            } else {
                if(gInterleaving_Windows==0)
                    err = system("Contents/Scripts/GetXpid interleave");
                else
                    err = system("Contents/Scripts/GetXpid uninterleave");
                unsetenv("OOSX_TMP_PASSWD");
                if(err==0) {
                    gInterleaving_Windows=1-gInterleaving_Windows;
                    Setup_Options_Menu(0);
                    if(gInterleaving_Windows==0)
                        ShowAlert(INTLVOFFOK_NAME,1,0);
                    else
                        ShowAlert(INTLVONOK_NAME,1,0);
                } else if(err==PASSWDFAIL_ERRNO) {
                    ShowAlert(PASSWDFAIL_NAME,1,0);
                } else if(err==LOSTXDARWIN_ERRNO) {
                    ShowAlert(LOSTXDARWIN_NAME,1,0);
                } else if(err==INTLVON_ERRNO) {
                    gInterleaving_Windows=1;
                    Setup_Options_Menu(0);
                    ShowAlert(INTLVON_NAME,1,0);
                } else if(err==INTLVOFF_ERRNO) {
                    gInterleaving_Windows=0;
                    Setup_Options_Menu(0);
                    ShowAlert(INTLVOFF_NAME,1,0);
                } else {
                    fprintf(stderr,"%s warning: interleaving script failed (err: %d)\n",PROGNAME,err);
                    ShowAlert(INTLVFAIL_NAME,1,0);
                }
            }
        }
        
        if(passwdText!=NULL)
            free(passwdText);
    }
    
    if(gQuitFlag==0) gQuitFlag=oldQuitFlag;
    Allow_HUPs(1);
}

void DoAboutBox(void)
{
    sprintf(lastsub,"TRACE3: DoAboutBox\n");
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: DoAboutBox\n");
#endif

//   (void) Alert(kAboutBox, nil);  / simple alert dialog box
    aboutWin=(WindowRef) MakeWindow(aboutWintype);
    if(aboutWin==NULL) return;
    
    gX11InFront=0;
    Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    gone_to_back();

    if(lastFrontWin==noFocusWin) {
        HideWindow(noFocusWin);
        showingNoFocusWin=1;
    }
        
    ShowWindow(aboutWin);
    SelectWindow(aboutWin);
#ifdef DEBUG
    printf("Showing about window: %#lx\n",aboutWin);
#endif
    showingAboutWin=1;
    lastFrontWin=aboutWin;
}

