/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 

/* common.h */

/*
 * This contains the info that both the 'oroborus part' and thhe 'Carbon part'
 * of OroborOSX need to know about
 */
 
#ifndef COMMON_H 
#define COMMON_H 

#include <unistd.h>

#include "definitions.h"
#include "errors.h"
#include "trace.h"

#define PROGNAME		"OroborOSX"

#define BGWIN_NOTHING		0
#define BGWIN_DIMMED		1
#define BGWIN_TRANSLUCENT	2

// Translucency given as a percentage (i.e. 100 is opaque)
#define	ACTIVE_TRANSLUCENT0	100
#define ACTIVE_TRANSLUCENT1	85
#define ACTIVE_TRANSLUCENT2	65

#define INACTIVE_TRANSLUCENT0	85
#define INACTIVE_TRANSLUCENT1	80
#define INACTIVE_TRANSLUCENT2	60

#define MOUSE_STOPPED	-1
#define MOUSE_NORMAL	0
#define MOUSE_MOVE	1
#define MOUSE_RESIZE	2
#define MOUSE_CLOSE	3
#define MOUSE_ZOOM	4
#define MOUSE_CLOSE_OUT	5
#define MOUSE_ZOOM_OUT	6
#define MOUSE_OVERCLOSE	7
#define MOUSE_OVERZOOM	8
#define MOUSE_ICON	9
#define MOUSE_ICON_OUT	10
#define MOUSE3_TITLE	20

#define ICONBARPOS_TOPLEFT	0
#define ICONBARPOS_TOPRIGHT	1
#define ICONBARPOS_BOTLEFT	2
#define ICONBARPOS_BOTRIGHT	3

// number of steps for iconbars
#define ICONSTEPS_FAST 4
#define ICONSTEPS_MED 7
#define ICONSTEPS_SLOW 20

#define VIEWSTATE_HIDDEN	0
#define VIEWSTATE_VISIBLE	1
#define VIEWSTATE_SHADED	2
#define VIEWSTATE_ICONBAR	3

#define	menuApplication		128	/* Application menu ID */
#define	itemAbout		1

#define menuLaunch		129	/* Launch menu ID */
#define itemEditStart		1
#define itemNewItem		3
#define itemEditItems		4
#define itemRebuildMenu		5
#define itemxinitrc		7
// points to divider item in Launch menu before the items added from the Launch Menu Items folder
#define itemUserItems		8
// These have now been replaced by x11app items in the Launch menu Items folder...
/*
#define itemxterm		9
#define itemxedit		10
#define itemxbiff		11
#define itemxclock		12
#define itemxeyes		13
#define itemxcalc		14
#define itemUserItems		15
// The rest of these are not actually used - they are just to remind of the default for the user part
#define itemETerm		16
#define itemGimp		17
#define itemNEdit		18
*/

#define	menuEdit		130	/* Edit menu ID */

#define	menuOptions		131	/* Options menu ID */
#define itemQuietStartup	1
#define itemXDarwinPrefs	2
#define itemInterleave		3
#define itemBGWinDimmed		5
#define itemBGWinTranslucent	6
#define itemThemeSubmenu	8
#define itemLeaveMenubar	8
#define itemMinToDock		10
#define itemFancyIconbars	11
#define itemHideIconbars	12
#define itemSlowIconbars	14
#define itemMedIconbars		15
#define itemFastIconbars	16
#define itemTopLeft		18
#define itemTopRight		19
#define itemBotLeft		20
#define itemBotRight		21
#define itemScreenListStart	23

#define menuThemes		132	/* Themes submenu ID */
#define itemAboutThemes		1
#define itemAddTheme		3
#define itemEditThemeList	4
#define itemRebuildThemes	5
#define itemToggleShaped	7
#define itemAboutThisTheme	9
// points to the divider in the Themes submenu before the items added from the Themes folder
#define itemUserThemes		10
// These items are not actually used -they are replaced by the themes that are actually found...
#define itemEau			11
#define itemGreyphite		12
#define itemNextish		13

#define	menuWindow		133	/* Window menu ID */
#define itemWindowshade		1
#define	itemMinimise		2
#define	itemZoom		3
#define	itemClose		5
#define itemProperties		7
#define itemLower		9
#define itemNext		10

#define menuProperties		135	/* Properties submenu ID */
#define itemFirstClick		1
#define itemAutoFocus		3
#define itemAutoRaise		4
#define itemClickRaise		5
#define itemFloat		7
#define itemPopup		8
#define itemNoframes		9
#define itemLiveResize		10
#define itemTranslucent0	12
#define itemTranslucent1	13
#define itemTranslucent2	14
// not used any more
#define itemSendToDock		113
#define itemGetFromDock		114

#define	menuHelp		134	/* Help menu ID */
#define itemHelp		1
#define itemVers		2
#define itemLicense		3

#define minmethod_iconbars	0
#define minmethod_dock		1
#define minmethod_icons		2

/***** Preferences *****/

/* Iconbar */
#define kPrefIconbarPos			"IconbarPos"
#define kPrefIconbarScreen		"IconbarScreen"
#define kPrefIconbarSteps		"IconbarSteps"
#define kPrefIconbarShrinking		"IconbarShrinking"
#define kPrefMinMethod			"MinimiseMethod"
/* Themes */
#define kPrefThemeName			"ThemeName"
#define kPrefThemeShaped		"ThemeShaped"
/* Misc */
#define kPrefInterleavingWindows	"InterleavingWindows"
#define kPrefBGWinSetting		"InactiveWindowBehaviour"
#define kPrefQuietStartup		"QuietStartup"


// Extra window info structure
typedef struct 
{
    long winref;
    int x, y, width, height;
    int widgets_x, widgets_y, widgets_w, widgets_h;
    int view_state, translucency, inactive_translucency, drawalpha;
} OSXWinInfo; 

/* globals */
extern int realPathLen;
extern char realPath[512];
extern int on_cmd_line, shrinking_iconbars, iconbar_screen, live_resizing;
extern int gQuitFlag, gX11InFront, gInterleaving_Windows, gMinMethod;
extern int gHiding_Iconbars, gMenubarSpace, gUseShapedTheme, gBGWinSetting, gX11MouseStuck;
extern int mouse_mode, real_x, real_y, iconbarpos, diffscrn_x, diffscrn_y, launched_xdarwin;
extern long lastBackTicks, lastIconbarOutTicks, client_count;
extern char randomID[20];
extern char *Display_Name, *xpidstr, *xverstr, *gThemeDir;
extern char *lastsub15, *lastsub14, *lastsub13, *lastsub12, *lastsub11;
extern char *lastsub10, *lastsub9, *lastsub8, *lastsub7, *lastsub6;
extern char *lastsub5, *lastsub4, *lastsub3, *lastsub2, *lastsub1;

/* main.c */
extern void ClearUpSockets();
extern char *Escape_Quotes(char *instr, int which);
extern void Bring_To_Front(int quick);
extern void Bring_X_To_Front(int quick);
extern void Bring_OOSX_To_Front();
extern void will_hide_all_iconbars();
extern void Compare_Screens(int scrnx, int scrny, int x11x, int x11y, int allowset);
extern void Stop_XDarwin();
extern void Restart_XDarwin();
/* Prefs */
extern int	getIntPref(const char* id, int defaultVal);
extern void	setIntPref(const char* id, int newVal);
extern char*	getStringPref(const char* id, const char* defaultVal);	/* Returns newly malloc'ed memory; caller should free. */
extern void	setStringPref(const char* id, const char* newVal);

/* macevents.c */
void Start_Time_Check(long);
int Is_Time_Up();
void Allow_HUPs(int);

/* menus.c */
extern void Setup_Options_Menu(int which);
extern void Setup_Window_Menu(int has_frame, int has_close, int has_min, int has_zoom, int is_float, int can_af, int can_ar, int first_click, int click_raise, int live_resize);
extern void Setup_Properties_Menu_Translucency(OSXWinInfo *winfo);
extern long GetCarbonTickCount();

/* osxwindows.c */
extern int ShowAlert(const char *name, int useidle, int extrabuts);
extern void Check_OSX_Window_Bounds(OSXWinInfo *winfo);
extern void MakeX11Window(OSXWinInfo *winfo, char *title, char *class);
extern void FocusOSXWin(OSXWinInfo *winfo);
extern void DestroyOSXWindow(OSXWinInfo *winfo);
extern void CheckOSXWinTitle(OSXWinInfo *winfo, char *title, char *class);
extern void HideOSXWin(OSXWinInfo *winfo);
extern void ReorderOSXWin(OSXWinInfo *winfo);
extern void SetOSXWinPos(OSXWinInfo *winfo,int x, int y, int w, int h, int wox, int woy, int wendox, int wendoy);

/* oroborus.c */
extern void quit();
extern int orob_main(int argc, char *argv[], char *dpy_str);

/* x11events.c */ 
extern void do_event_loop(int sleepus,int trackMouse);
extern void force_ungrab();

/* client.c */
extern void left_client();
extern void gone_to_back();
extern void come_to_front();
extern void change_focus(long winref);
extern void hide_all_iconbars();
extern void manipulate_focus_client(int action);
extern void change_options(int newattr);
extern void raise_all_clients();
extern void map_all_clients();
extern void unmap_all_clients();
extern void check_osx_windows_bounds();
extern void list_all_windows(); 
extern void list_all_clients(); 
extern long get_next_osxwinref(int);

/* config.c */ 
extern void reload_config(); 

/* misc.c */
extern int count_xinerama_heads();

/* xdarwinhack.c */
extern void SendSimpleHackCommand(int cmd, int param);

#endif

