#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include "communication.h"

int main_socket=0, xdarwin_socket=0;

//#define SOCK_PATH "/tmp/echo_server_socket"

int init_listen_socket(char *pathstr)
{
    int s, sret, dummy, len;
    struct timeval delay;
    struct sockaddr_un local;
    char script[100];

    if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
        perror("socket");
        return -1;
    }

//    printf("Setting TIMEO...\n");

    delay.tv_sec=1;
    delay.tv_usec=0;
    dummy = setsockopt(s,SOL_SOCKET,SO_RCVTIMEO,(void *)&delay,sizeof(delay));
//    printf("Got dummy=%d\n",dummy);

    sprintf(script,"rm -f %s",pathstr);
    system(script);

    local.sun_family = AF_UNIX;
    strcpy(local.sun_path, pathstr);
    unlink(local.sun_path);
    len = strlen(local.sun_path) + sizeof(local.sun_family) + 1;
//    printf("sun_path: %s\n",local.sun_path);
    if(bind(s, (struct sockaddr *)&local, len) == -1) {
        perror("bind");
        return -2;
    }
    
//    printf("Going into listen...\n");

    dummy = listen(s,5);
    if (dummy == -1) {
        perror("listen");
        return -3;
    }
//    printf("Got dummy=%d\n",dummy);

    sret = fcntl(s,F_GETFL,dummy);
//    printf("Got flags: %d\n",sret);
    sret |= O_NONBLOCK;
    dummy = fcntl(s,F_SETFL,sret);
//    printf("Set flags to %d, result: %d\n",sret,dummy);
    sret = 0;

    return s;
}

int check_listen_accept(int s)
{
    int t, sret;
    struct sockaddr_un remote;
    
    t = sizeof(remote);
    sret = accept(s, (struct sockaddr *)&remote, &t);
    if(sret==-1) {
        if(errno==EAGAIN) {
            printf("Server waiting for accept on socket... (%d)\n",sret);
            sret=0;
        } else
            perror("accept (in server)");
    }
    
//    printf("Returning %d from check_listen_accept\n",sret);
    return sret;
}

void end_listen_socket(int s, char *pathstr)
{
    char	script[100];
    
    close(s);
    sprintf(script,"rm -f %s",pathstr);
    system(script);
}

int check_listen_recv(int s)
{
    int n=0, rcvlen;
    
    n = recv(s, &rcvlen, 4, 0);
    if(n<0) {
        if(errno==EAGAIN)
            return -1;
        perror("recv");	// got some error
        return n-2;
    } else if(n==0)
        return -2;	// socket disappeared, so signal end
    if(rcvlen==0)
        return -2;	// client has indicated it wants to end
    if(rcvlen<0)
        return rcvlen-2;	// something weird going on here...
        
    return rcvlen;
}

int get_bytes_from_socket(int s, int rcvlen, char *str)
{
    int n;
    
    while(n>0) {
        n = recv(s, str, rcvlen, 0);
        if( n < 0 ) {
            if( errno==EAGAIN) {
                printf("waiting for more... %d\n",n);
                    n=1;
            } else {
                perror("recv");
                return -1;
            }
        } else if(n==0) {
            printf("No connection...\n");
            perror("recv");
            return -1;
        } else {
            if( n<rcvlen )
                printf("Did not get all? %d %d\n",n,rcvlen);
            return n;
        }
    }
}

int check_listen_get_bytes(int s, char *str)
{
    int rcvlen, gotlen;
    
    rcvlen = check_listen_recv(s);
    if(rcvlen==0) {
        printf("Got zero...\n");
        return 0;
    } else if(rcvlen==-1) {
        return -1;
    } else if(rcvlen==-2) {
        return -2;
    } else if(rcvlen<0) {
        printf("check_listen_get_bytes got error from check_listen_recv: %d\n",rcvlen);
        return rcvlen-3;
    } else {
//        printf("Expecting %d bytes\n",rcvlen);
        gotlen = get_bytes_from_socket(s,rcvlen,str);
        if(gotlen!=rcvlen) {
            printf("Something went wrong: %d %d\n",rcvlen,gotlen);
            return -3;
        } else {
            str[rcvlen]=0;
        }
        return rcvlen;
    }
}

int check_listen_accept_name(int s, char *str)
{
    int sret, rcvlen;
    
    sret = check_listen_accept(s);
    
    if(sret==0) return 0;
    
    if(sret<0) {
        printf("check_listen_accept_name got error from check_listen_accept: %d\n",sret);
        return sret;
    }

    rcvlen = 0;
    while(rcvlen==0) {
        rcvlen = check_listen_get_bytes(sret,str);
        if(rcvlen<0) {
            printf("check_listen_accept_name got error from check_listen_get_bytes: %d\n",rcvlen);
            return rcvlen;
        }
    }
    return sret;
}
