/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 

#include "common.h"

#ifndef OROBORUS_H 
#define OROBORUS_H 
 
#include <X11/Xlib.h> 
#include <X11/Xutil.h> 
#include <X11/Xatom.h> 
#include <X11/Xmd.h> 
#include <X11/xpm.h> 
#include <X11/cursorfont.h> 

#ifdef I18N_SUPPORT 
# include <X11/Xlocale.h> 
#endif /* I18N */ 
 
#ifdef SHAPE_SUPPORT 
#include <X11/extensions/shape.h> 
#endif /* SHAPE */ 
 
#ifdef XINERAMA_SUPPORT
#ifndef EMULATE_XINERAMA 
#include <X11/extensions/Xinerama.h> 
#endif
#endif /* XINERAMA */  
 
#include <stdlib.h> 
#include <stdio.h> 
#include <string.h> 
#include <sys/wait.h> 
#include <sys/stat.h> 
#include <time.h>
#include <signal.h> 
#include <unistd.h> 
 
/* 
 * 
 * Various defines 
 * 
 */ 

#define NORMAL_CURS		XC_top_left_arrow	// is there a more Mac-like one than this?
#define RESIZE_CURS             XC_sizing
#define MOVE_CURS		XC_fleur 
#define DBL_CLICK		500
#define MOVE_MOTION_GAP		80
#define RESIZE_MOTION_GAP	250
#define FIRST_MOTION_GAP	300
#define BG_FIRST_MOTION_GAP	600
#define TAB_RAISE_TIMEOUT	1000 
 
/* 
 * 
 * Various flags 
 * 
 */ 

/* Icon-bar modes */
#define NOT_ICONBAR		0
#define ICONBAR_VISIBLE		1
#define ICONBAR_RAISED		2
#define ICONBAR_HIDDEN		3

/* e_m_r (entered/moving/resizing) flags */
#define IS_ENTERED		0x0001
#define IS_MOVING		0x0002
#define IS_RESIZING		0x0004
#define SHOWING_SIZE		0x0100
#define SHOWING_POSITION	0x0200
/* used to mask out the bits that can change (in redraw_client_title) from the bits that won't */
#define NOCHANGE_SHOWFLAGS	0x00ff
#define CHANGE_SHOWFLAGS	0xff00
/* used to unset particular bits of the e_m_r flag */
#define NOT_ENTERED		0xffff - IS_ENTERED
#define NOT_MOVING		0xffff - IS_MOVING
#define NOT_RESIZING		0xffff - IS_RESIZING
#define NOT_SHOWING_SIZEPOS	0xffff - SHOWING_SIZE - SHOWING_POSITION

/* Gravitate modes */ 
#define APPLY_GRAVITY		1 
#define REMOVE_GRAVITY		-1 
 
/* Window areas  -with lastenter check */ 
#define WINDOW_LE		1 
#define TITLE			2 
#define OSXWINREF		3
#define FRAME_LE		4
/* Window areas  -without lastenter check */ 
#define WINDOW			-1 
#define FRAME			-4
 
/* Window placement */ 
#define CENTER_MOUSE		0 
#define CENTER_ROOT		1 
 
/* Focus styles */ 
#define FOCUS_CLICK		0 
#define FOCUS_SLOPPY		1 
#define SLOPPY_CHILDREN		2
#define SLOPPY_SIBLINGS		4
#define SLOPPY_PARENT		8

/* Per window 'auto' properties */
#define AUTO_NONE		0
#define AUTO_FOCUS		1
#define AUTO_RAISE		2
/* some related flags: raise an autofocus window when clicked, allow first click to nonautofocus window */
#define AUTO_CLICKRAISE		4
#define NONAUTO_FIRSTCLICK	8
/* special flag for live resizing */
#define LIVE_RESIZING		128
 
/* Frame states */ 
#define FRAME_ACTIVE		0 
#define FRAME_INACTIVE		1
#define FRAME_ACTIVE_SHADED	2
#define FRAME_INACTIVE_SHADED	3
#define FRAME_ACTIVE_ICONBAR	4
#define FRAME_INACTIVE_ICONBAR	5
 
/* Button states */ 
#define BUTTON_UNPRESSED	0 
#define BUTTON_PRESSED		2 
#define BUTTON_OVER		3
#define BUTTON_OVERBACK		4
#define BUTTON_PRESSEDBACK	5
#define BUTTON_ACTIVEOFF	6
#define BUTTON_INACTIVEOFF	7
 
/* Text alignments */ 
#define ALIGNMENT_LEFT		0 
#define ALIGNMENT_MIDDLE	1 
#define ALIGNMENT_RIGHT		2 
 
/* Modality States */
#define MODELESS		0
#define APP_MODAL		1
#define SYS_MODAL		2
#define ANY_MODAL               3
#define HAS_APP_MODAL_CHILD	4
#define HAS_SYS_MODAL_CHILD	8
#define HAS_ANY_MODAL_CHILD	12
#define APP_FLOATING		16
#define SYS_FLOATING		32
#define ANY_FLOATING		48
#define HAS_APP_FLOATING	64
#define HAS_SYS_FLOATING	128
#define HAS_ANY_FLOATING	192

/* 
 * 
 * Simplify some long calculations 
 * 
 */ 
 
/* Gravity */ 
#define GRAVITY		(c->size->flags&PWinGravity?c->size->win_gravity:NorthWestGravity) 
 
/* Title size and position */ 
#define TH	((c->has_frame || (c->is_icon!=NOT_ICONBAR))?config.top.height:0) 
#define TW	((c->has_frame || (c->is_icon!=NOT_ICONBAR))?(FW-TRW-TLW):0) 
#define TX	(TLW) 
#define TY	0 
 
/* Frame size and position */ 
#define FW	((c->is_icon!=NOT_ICONBAR)?c->icon_w+LW+RW:c->width+LW+RW) 
#define FH	(((get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED)==WIN_STATE_SHADED || (c->is_icon!=NOT_ICONBAR))?TH:c->height+TH+BH) 
#define FX	((c->is_icon!=NOT_ICONBAR)?c->icon_x-LW:c->x-LW) 
#define FY	((c->is_icon!=NOT_ICONBAR)?c->icon_y-TH:c->y-TH) 
 
/* Top left border size and position */ 
#define TLW	((c->is_icon==NOT_ICONBAR)?config.top_left.width:config.iconbar_left_width)
#define TLH	((c->is_icon==NOT_ICONBAR)?config.top_left.height:config.iconbar_left_height) 
#define TLX	0 
#define TLY	0 
 
/* Top right border size and position */ 
#define TRW	((c->is_icon==NOT_ICONBAR)?config.top_right.width:config.iconbar_right_width) 
#define TRH	((c->is_icon==NOT_ICONBAR)?config.top_right.height:config.iconbar_right_height) 
#define TRX	(FW-TRW)
#define TRY	0 
 
/* Left border size and position */ 
#define LW	((c->has_frame && (c->is_icon==NOT_ICONBAR))?config.left.width:0) 
#define LH	(FH-TLH-BLH) 
#define LX	0 
#define LY	(TLH) 
 
/* Right border size and position */ 
#define RW	((c->has_frame && (c->is_icon==NOT_ICONBAR))?config.right.width:0) 
#define RH	(FH-TRH-RSH) 
#define RX	(c->width+LW)
#define RY	(TRH) 
 
/* Bottom left border size and position */ 
#define BLW	(config.bottom_left.width) 
#define BLH	(config.bottom_left.height) 
#define BLY	(c->height+TH+BH-BLH) 
#define BLX	0 
 
/* Bottom border size and position */ 
#define BX	(BLW)
#define BY	(c->height+TH) 
#define BH	(c->has_frame?config.bottom.height:0) 
#define BW	(FW-BLW-config.resize.width) 
 
/* Resize size and position */ 
#define RSH	(config.resize.height) 
#define RSW	(config.resize.width) 
#define RSX	(FW-RSW)
#define RSY	(c->height+TH+BH-RSH) 
 
/* OLD! Close button size and position  
#define CX	(LW+c->width-config.close.width+1) 
#define CY	((TH-config.close.height)/2) 
#define CW	(config.close.width) 
#define CH	(config.close.height) */
 
/* Close button size and position */ 
//#define CX      (LW+6)
//#define CY      ((TH-config.close.height)/2)
#define CX      (config.close.x>0?config.close.x-1:FW+config.close.x-config.close.width+1)
#define CY      (config.close.y>0?config.close.y-1:FH+config.close.y-config.close.height+1)
#define CW      (config.close.width)
#define CH      (config.close.height)

/* Maximise button size and position */ 
//#define HX	(LW+6+config.close.width+config.icon.width) 
//#define HY	((TH-config.zoom.height)/2) 
#define HX      (config.zoom.x>0?config.zoom.x-1:FW+config.zoom.x-config.zoom.width+1)
#define HY      (config.zoom.y>0?config.zoom.y-1:FH+config.zoom.y-config.zoom.height+1)
#define HW	(config.zoom.width) 
#define HH	(config.zoom.height) 

/* 'Iconise' button size and position */
//#define IX	(LW+6+config.close.width) 
//#define IY	((TH-config.icon.height)/2) 
#define IX      (config.icon.x>0?config.icon.x-1:FW+config.icon.x-config.icon.width+1)
#define IY      (config.icon.y>0?config.icon.y-1:FH+config.icon.y-config.icon.height+1)
#define IW	(config.icon.width) 
#define IH	(config.icon.height) 

/*
#define HX      (LW+c->width-config.zoom.width+1)
#define HY      ((TH-config.zoom.height)/2)
#define HW      (config.zoom.width)
#define HH      (config.zoom.height)
*/
 
/* 
 * 
 * Atoms and hints 
 * 
 */ 
 
/* ICCCM atoms */ 
#define WM_STATE			0 
#define WM_CHANGE_STATE			1 
#define WM_DELETE			2 
#define WM_PROTOCOLS			3 
#define ICCCM_ATOM_COUNT		4 
 
/* Motif atoms */ 
#define MOTIF_WM_HINTS			0 
#define MOTIF_ATOM_COUNT		1 
 
/* GNOME atoms */ 
#define WIN_STATE			0 
#define WIN_LAYER			1 
#define WIN_CLIENT_LIST			2 
#define WIN_DESKTOP_BUTTON_PROXY	3 
#define WIN_HINTS			4 
#define WIN_SUPPORTING_WM_CHECK		5 
#define WIN_WORKSPACE			6 
#define WIN_WORKSPACE_COUNT		7 
#define GNOME_ATOM_COUNT		8 

/* hints  -already defined in Xutil.h
#define InputHint			(1L<<0)
#define StateHint			(1L<<1)
#define IconPixmapHint			(1L<<2)
#define IconWindowHint			(1L<<3)
#define IconPositionHint		(1L<<4)
#define IconMaskHint			(1L<<5)
#define WindowGroupHint			(1L<<6)
*/

/* Mwm hints */ 
#define MWM_HINTS_DECORATIONS		(1L<<1) 
#define MWM_HINTS_INPUT_MODE		(1L<<2)
#define MWM_DECOR_ALL			(1L<<0) 
#define MWM_DECOR_BORDER		(1L<<1) 
#define MWM_DECOR_TITLE			(1L<<3) 
#define MWM_DECOR_MAXIMIZE		(1L<<6)
#define MWM_INPUT_MODELESS			0
#define MWM_INPUT_PRIMARY_APPLICATION_MODAL	1
#define MWM_INPUT_SYSTEM_MODAL			2
#define MWM_INPUT_FULL_APPLICATION_MODAL	3

/* GNOME hints */ 
#define WIN_HINTS_SKIP_FOCUS		(1<<0) 
#define WIN_HINTS_SKIP_TASKBAR		(1<<2) 
#define WIN_STATE_STICKY		(1<<0) 
#define WIN_STATE_MAXIMIZED_VERT	(1<<2) 
#define WIN_STATE_MAXIMIZED_HORIZ	(1<<3) 
#define WIN_STATE_MAXIMIZED		(WIN_STATE_MAXIMIZED_VERT|WIN_STATE_MAXIMIZED_HORIZ) 
#define WIN_STATE_SHADED		(1<<5) 
#define WIN_LAYER_DESKTOP		0 
#define WIN_LAYER_BELOW			2 
#define WIN_LAYER_NORMAL		4 
#define WIN_LAYER_ONTOP			6 
 
/* 
 * 
 * Structures 
 * 
 */ 
 
/* Mwm hints structure */
typedef struct 
{
// had troble with CARD32 on some platforms - 'long' seems to work OK on everything I've tried...
//  CARD32 flags;
  long flags;
//  CARD32 functions; 
  long functions; 
//  CARD32 decorations;
  long decorations;
//  CARD32 inputMode; 
  long inputMode; 
//  CARD32 status;
  long status;
} PropMwmHints; 
 
/* Client structure */ 
typedef struct _Client Client; 
 
struct _Client 	// -various bits added to this for OroborOSX
{ 
  Window window, frame, title;
  Window top_left, top_right, left, right, bottom_left, bottom, resize, close, icon, zoom; 
/* Window trans   ...don't know what this was for? */
  char *name, *clientcmd, *clientclass, *clientgroup;
  XSizeHints *size;
  int hack_id, done_icon;
  int is_icon, is_dimmed, is_docked, is_e_m_r; // e/m/r = entered/moving/resizing
  int has_focus, has_frame, can_max, ignore_unmap, modal_state, sloppy_family, auto_properties, is_shaped;
  int name_x, name_y, name_w, name_h;
  int x, y, width, height, old_x, old_y, old_width, old_height;
  int icon_x, icon_y, icon_w, icon_h, icontxt_x, icontxt_y, icontxt_w, icontxt_h; 
  int zoom_state, icon_state, close_state;
  int child_count;
  long dock_timer, undock_timer, resize_timer;
  OSXWinInfo osxwinfo;
  Client *parent;
  Client *first_child;
  Client *prev_sibling;
  Client *next_sibling;
  Client *next; 
  Client *prev; 
}; 
 
/* Frame part structure */ 
typedef struct 
{ 
// bumped up to 6 'slots' and 3 masks for extra frame states (active/inactive and normal/shaded/iconbar)
  Pixmap pixmap[6], mask[3]; 
  int width, height; 
} FramePart; 

typedef struct 
{ 
// bumped up to 8 'slots' for extra window widget states (active/inactive and normal/over/pressed/off)
  Pixmap pixmap[8], mask[1]; 
// added x,y co-ords so buttons can move
  int x, y, width, height;
} FrameWidget; 
 
#ifdef KEYBOARD_SUPPORT 
/* Key structure */ 
typedef struct 
{ 
  int keycode; 
  int modifier; 
} Key; 
 
/* Keys structure */ 
typedef struct 
{ 
  Key next_window, prev_window, maximize, shade, hide, close, move_up, move_down, move_left, move_right, resize_up, resize_down, resize_left, resize_right; 
 
#ifdef LAYER_SUPPORT 
  Key raise_layer, lower_layer; 
#endif /* LAYER */ 
 
#ifdef WORKSPACE_SUPPORT 
  Key sticky, add_ws, del_ws; 
  Key next_ws, prev_ws, ws_1, ws_2, ws_3, ws_4, ws_5, ws_6, ws_7, ws_8, ws_9; 
  Key move_next_ws, move_prev_ws, move_ws_1, move_ws_2, move_ws_3, move_ws_4, move_ws_5, move_ws_6, move_ws_7, move_ws_8, move_ws_9; 
#endif /* WORKSPACE */ 
} Keys; 
#endif /* KEYBOARD */ 
 
/* Config structure */ 
typedef struct 
{ 
  char *theme_dir;
  FramePart top_left, top, title_left, title, title_right, top_right;
  FramePart left, right, bottom_left, bottom, resize; 
  FrameWidget close, icon, zoom;
  int iconbar_left_width, iconbar_left_height;
  int iconbar_right_width, iconbar_right_height;
  int text_align, active_titlebar_alpha, inactive_titlebar_alpha; 
  GC font_gc[2]; 
  XFontStruct *font; 
 
#ifdef I18N_SUPPORT 
  XFontSet fontset; 
#endif /* I18N */ 
 
  int window_placement, focus_style, raise_on_focus, focus_new, continuous_resize; 
  int top_margin, bottom_margin, left_margin, right_margin; 
 
#ifdef KEYBOARD_SUPPORT 
  Keys keys;
#endif /* KEYBOARD */ 
 
#ifdef WORKSPACE_SUPPORT 
  int workspace_count; 
#endif /* WORKSPACE */ 
} Config; 

#ifdef EMULATE_XINERAMA
  typedef struct
  {
    int screen;
    short x_org;
    short y_org;
    short width;
    short height;
  } XineramaScreenInfo;
#endif 
 
/* 
 * 
 * Global variables 
 * 
 */ 
 
/* oroborus.c */ 
extern int nexticon_x, nexticon_y;
extern Display *dpy; 
extern Window root, check_win, lastenter, lastfocus; 
// flag so we know if the focus window has a menu down - keeps window frame of focus window looking active
extern int hacking_focus;
extern Cursor move_curs, resize_curs, normal_curs; 
extern int screen; 
extern Client *overicon;	// indicates which iconbar (if any) the cursor is currently over
extern Client *client_list; 
extern Config config; 
extern Atom icccm[ICCCM_ATOM_COUNT], motif[MOTIF_ATOM_COUNT], gnome[GNOME_ATOM_COUNT]; 
extern char *rcfile;
extern int iconbar_state, inhibit_errors;
extern Client *lastc, *lastfocus_client, *last_client_entered_hack, *current_resize_client;
extern int replay_if_drag;	// not properly used at the moment

#ifdef SHAPE_SUPPORT 
extern int shape, shape_event; 
#endif /* SHAPE */ 
 
#ifdef WORKSPACE_SUPPORT 
extern int workspace; 
#endif /* WORKSPACE */ 
 
#ifdef XINERAMA_SUPPORT 
extern int xinerama; 
extern int xinerama_heads;
#ifndef EMULATE_XINERAMA
extern XineramaScreenInfo *xinerama_screens;
#else
#define EMULATED_SCREENS 3
extern XineramaScreenInfo xinerama_screens[EMULATED_SCREENS]; 
#endif
#endif /* XINERAMA */ 

#ifdef I18N_SUPPORT
extern int use_fontset;
#endif /* I18N_SUPPORT */
 
/* 
 * 
 * Functions 
 * 
 */ 
 
/* client.c */
extern Client *check_for_undock(int hackid);
extern int undock_client(Client *);
extern void dock_client(Client *);
extern void gravitate_client(Client *, int); 
extern Window create_frame_part(Window win, int x, int y, int w, int h, int mapit);
extern void create_new_client(Window, int); 
extern Client *get_client_from_window(Window, int); 
extern Client *get_focus_client(); 
extern Client *get_next_client(Client *);
extern Client *get_prev_client(Client *);
extern Client *get_top_most_client(int, int); 
extern Client *get_lowest_entered_client(int get_icons);
extern Client *get_bottom_most_client(int, int); 
extern Client *get_client_container(int, int, int);
extern Client *check_client_modality(Client *orgc, int raiseit);
extern void raise_client(Client *); 
extern void force_raise_client(Client *); 
extern void entered_client(Client *, int);
extern void set_focus_client(Client *);
extern void lower_client(Client *); 
extern void show_client(Client *); 
extern void toggle_minimise_client(Client *, int); 
extern void minmax_client(Client *); 
extern void configure_client(Client *, unsigned int, XWindowChanges *, int, int); 
extern void move_client(Client *, int, int, int); 
extern void resize_client(Client *, int, int, int); 
extern void toggle_client_maximized(Client *); 
extern void toggle_client_shaded(Client *); 
extern void set_client_layer(Client *, int); 
extern int get_client_layer(Client *); 
extern void delete_client(Client *, int); 
extern void close_client(Client *); 
extern void redraw_client_frame(Client *); 
extern void redraw_client_buttons(Client *); 
extern void redraw_client_title(Client *); 
extern void compress_iconbars();
extern int still_in_auto_client();

#ifdef SHAPE_SUPPORT 
extern void set_client_shape(Client *, int); 
#endif /* SHAPE */ 
 
#ifdef WORKSPACE_SUPPORT 
extern void toggle_client_sticky(Client *); 
extern void move_client_to_workspace(Client *, int); 
#endif /* WORKSPACE */ 
 
/* hints.c */ 
extern PropMwmHints *get_mwm_hints(Window); 
extern void set_wm_state(Window, int); 
extern long get_wm_state(Window); 
extern void set_gnome_hint(Window, int, long); 
extern long get_gnome_hint(Window, int); 
extern void update_gnome_client_list(); 
  
/* config.c */ 
extern void load_config(); 
extern void free_config(); 
extern time_t config_modified(char *, char *); 
 
/* misc.c */ 
extern int handle_xerror(Display *, XErrorEvent *); 
extern void add_dir_to_font_path(char *);
extern void rem_dir_from_font_path(char *);
extern void get_mouse_xy(int *, int *); 
extern Window get_pointer_win(Window); 
extern Window get_window_parent(Window); 
extern int screen_width(int); 
extern int screen_height(int); 
extern int screen_x(int); 
extern int screen_y(int);
extern int get_icontab_width();
extern void MyIncreaseXServerGrab();
extern void MyDecreaseXServerGrab();
extern int MyXServerGrabLevel();

#ifdef XINERAMA_SUPPORT 
extern int find_head(int, int); 
#endif /* XINERAMA */ 
 
#ifdef I18N_SUPPORT 
extern Status XGetWMNameWrapper(Display *, Window, XTextProperty *); 
extern Status XFetchNameWrapper(Display *, Window, char **); 
extern XFontSet XLoadQueryFontWrapper(Display *, const char *); 
#endif /* I18N */ 
 
/* pixmaps.c */ 
extern void load_pixmaps(char *, XpmColorSymbol *, int); 
 
#ifdef KEYBOARD_SUPPORT 
/* keyboard.c */
extern void init_keyboard(); 
extern void parse_key(char *, Key *); 
extern void ungrab_keys(Window); 
extern void grab_keys(Window); 
#endif /* KEYBOARD */ 
 
#ifdef WORKSPACE_SUPPORT 
/* workspace.c */ 
extern void change_to_workspace(int); 
extern void update_workspace_count(); 
#endif /* WORKSPACE */ 

/* xdarwinhack.c */
extern void SendHackTitleToDockCommand(Client *c,int todock);
extern void SendHackCommand(Client *c, int nparam, int orgcmd, int p1, int p2, int p3, int p4, int p5);

#endif /* OROBORUS_H */ 

