/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 

#include "hackcmds.h"
#include "oroborus.h"
#include "communication.h"

int mouse_x, mouse_y, lastx=-100, lasty=-100, lastcnt=0, lastbadenter=0, got_button=False, check_ptrwin=0;
double lastmv=0.0;
Client *lastc=NULL, *last_button1_client=NULL;
int replay_x, replay_y, replay_if_drag=0, orgmove_x, orgmove_y, replay_motion=0;//, got_replay_pointer=0;
XEvent replay_button_event;
XEvent replay_motion_event;
Window replay_win;
Time last_button1_time, replay_time, last_motion_time;
long last_gotbutton_time=0,next_get_cursorpos_time=0;
#ifdef KEYBOARD_SUPPORT
int tab_client_state;
Time last_tab_time;
#endif

#ifdef KEYBOARD_SUPPORT

/*
 * 
 * Handle key press event
 *
 */

void event_key_press(XKeyEvent *ev)
{
  Client *c=NULL;

  sprintf(lastsub,"TRACE1: event_handle_key_press\n");
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_handle_key_press\n");
#endif

  c=get_client_from_window(ev->window, WINDOW);

  ev->state=ev->state & (Mod1Mask|ControlMask|ShiftMask);
  
  /* If we have an active client */
  if(c!=NULL)
  {
    /* Maximize client */
    if(ev->keycode==config.keys.maximize.keycode && ev->state==config.keys.maximize.modifier)
      toggle_client_maximized(c);
      
    /* Shade client */
    if(ev->keycode==config.keys.shade.keycode && ev->state==config.keys.shade.modifier)
      toggle_client_shaded(c);

    /* Hide client */
    if(ev->keycode==config.keys.hide.keycode && ev->state==config.keys.hide.modifier)
      minmax_client(c);
     
    /* Close client */ 
    if(ev->keycode==config.keys.close.keycode && ev->state==config.keys.close.modifier) 
      close_client(c); 
       
    /* Move client up */ 
    if(ev->keycode==config.keys.move_up.keycode && ev->state==config.keys.move_up.modifier)
    {
      mouse_mode=MOUSE_STOPPED; 
      move_client(c, c->x, c->y-10, 0); 
    }
       
    /* Move client down */ 
    if(ev->keycode==config.keys.move_down.keycode && ev->state==config.keys.move_down.modifier)
    {
      mouse_mode=MOUSE_STOPPED; 
      move_client(c, c->x, c->y+10, 0); 
    }
       
    /* Move client left */ 
    if(ev->keycode==config.keys.move_left.keycode && ev->state==config.keys.move_left.modifier)
    {
      mouse_mode=MOUSE_STOPPED; 
      move_client(c, c->x-10, c->y, 0); 
    }
       
    /* Move client right */ 
    if(ev->keycode==config.keys.move_right.keycode && ev->state==config.keys.move_right.modifier) 
    {
      mouse_mode=MOUSE_STOPPED;
      move_client(c, c->x+10, c->y, 0); 
    }
 
    if(!(get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_MAXIMIZED) && !(get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)) 
    { 
      if(ev->keycode==config.keys.resize_up.keycode && ev->state==config.keys.resize_up.modifier) 
      {
        mouse_mode=MOUSE_STOPPED;
        resize_client(c, c->width, c->height-(c->size->flags & (PResizeInc?c->size->height_inc:10)), False); 
      }
      if(ev->keycode==config.keys.resize_down.keycode && ev->state==config.keys.resize_down.modifier) 
      {
        mouse_mode=MOUSE_STOPPED;
        resize_client(c, c->width, c->height+(c->size->flags & (PResizeInc?c->size->height_inc:10)), False);
      } 
      if(ev->keycode==config.keys.resize_left.keycode && ev->state==config.keys.resize_left.modifier) 
      {
        mouse_mode=MOUSE_STOPPED;
        resize_client(c, c->width-(c->size->flags & (PResizeInc?c->size->width_inc:10)), c->height, False);
      } 
      if(ev->keycode==config.keys.resize_right.keycode && ev->state==config.keys.resize_right.modifier) 
      {
        mouse_mode=MOUSE_STOPPED;
        resize_client(c, c->width+(c->size->flags & (PResizeInc?c->size->width_inc:10)), c->height, False); 
      }
    } 
     
#ifdef LAYER_SUPPORT 
    /* Raise client layer */ 
    if(ev->keycode==config.keys.raise_layer.keycode && ev->state==config.keys.raise_layer.modifier) 
      set_client_layer(c, get_client_layer(c)+1); 
       
    /* Lower client layer */ 
    if(ev->keycode==config.keys.lower_layer.keycode && ev->state==config.keys.lower_layer.modifier) 
      set_client_layer(c, get_client_layer(c)-1); 
#endif /* LAYER */ 
 
#ifdef WORKSPACE_SUPPORT 
    /* Toggle sticky */ 
    if(ev->keycode==config.keys.sticky.keycode && ev->state==config.keys.sticky.modifier) 
      toggle_client_sticky(c); 
       
    /* Move window to next workspace */ 
    if(ev->keycode==config.keys.move_next_ws.keycode && ev->state==config.keys.move_next_ws.modifier) 
      move_client_to_workspace(c, workspace+1); 
 
    /* Move window to prev workspace */ 
    if(ev->keycode==config.keys.move_prev_ws.keycode && ev->state==config.keys.move_prev_ws.modifier) 
      move_client_to_workspace(c, workspace-1); 
       
    /* Move window to specific workspace */ 
    if(ev->keycode==config.keys.move_ws_1.keycode && ev->state==config.keys.move_ws_1.modifier) 
      move_client_to_workspace(c, 0);      
    if(ev->keycode==config.keys.move_ws_2.keycode && ev->state==config.keys.move_ws_2.modifier) 
      move_client_to_workspace(c, 1); 
    if(ev->keycode==config.keys.move_ws_3.keycode && ev->state==config.keys.move_ws_3.modifier) 
      move_client_to_workspace(c, 2); 
    if(ev->keycode==config.keys.move_ws_4.keycode && ev->state==config.keys.move_ws_4.modifier) 
      move_client_to_workspace(c, 3); 
    if(ev->keycode==config.keys.move_ws_5.keycode && ev->state==config.keys.move_ws_5.modifier) 
      move_client_to_workspace(c, 4); 
    if(ev->keycode==config.keys.move_ws_6.keycode && ev->state==config.keys.move_ws_6.modifier) 
      move_client_to_workspace(c, 5); 
    if(ev->keycode==config.keys.move_ws_7.keycode && ev->state==config.keys.move_ws_7.modifier) 
      move_client_to_workspace(c, 6); 
    if(ev->keycode==config.keys.move_ws_8.keycode && ev->state==config.keys.move_ws_8.modifier) 
      move_client_to_workspace(c, 7); 
    if(ev->keycode==config.keys.move_ws_9.keycode && ev->state==config.keys.move_ws_9.modifier) 
      move_client_to_workspace(c, 8); 
#endif /* WORKSPACE */ 
  } 
 
  /* Next window */ 
  if(ev->keycode==config.keys.next_window.keycode && ev->state==config.keys.next_window.modifier) 
  { 
    Client *c2=NULL; 
 
    /* Check time of last tab */ 
    if(ev->time-last_tab_time<TAB_RAISE_TIMEOUT) 
    { 
     /* If window was hidden then hide it again 
     if(tab_client_state==IconicState && c!=NULL) */
     if((tab_client_state!=0) && (c!=NULL) )
        minmax_client(c); 
    } 
    last_tab_time=ev->time; 
    if(c)
      c2=get_next_client(c);
    else if(client_list)
      c2=get_next_client(client_list->prev); 
    if(c2!=NULL) 
    { 
/*      tab_client_state=get_wm_state(c2->window); */
      tab_client_state=c2->is_icon;
      if(tab_client_state!=0)
        minmax_client(c2);
      raise_client(c2); 
      set_focus_client(c2); 
    } 
  }

  /* Previous window */ 
  if(ev->keycode==config.keys.prev_window.keycode && ev->state==config.keys.prev_window.modifier) 
  { 
    Client *c2=NULL; 
 
    /* Check time of last tab */ 
    if(ev->time-last_tab_time<TAB_RAISE_TIMEOUT) 
    { 
     /* If window was hidden then hide it again 
     if(tab_client_state==IconicState && c!=NULL)  */
     if(tab_client_state!=0 && c!=NULL)
        minmax_client(c); 
    } 
    last_tab_time=ev->time; 
    if(c)
      c2=get_prev_client(c);
    else if(client_list)
      c2=get_prev_client(client_list->next); 
    if(c2!=NULL) 
    { 
/*      tab_client_state=get_wm_state(c2->window); */
      tab_client_state=c2->is_icon;
      if(tab_client_state!=0)
        minmax_client(c2);
      raise_client(c2);
      set_focus_client(c2); 
    }
  } 
 
#ifdef WORKSPACE_SUPPORT   
  /* Switch to next workspace */ 
  if(ev->keycode==config.keys.next_ws.keycode && ev->state==config.keys.next_ws.modifier) 
    change_to_workspace(workspace+1); 
 
  /* Switch to prev workspace */ 
  if(ev->keycode==config.keys.prev_ws.keycode && ev->state==config.keys.prev_ws.modifier) 
    change_to_workspace(workspace-1); 
 
  /* Add new workspace */ 
  if(ev->keycode==config.keys.add_ws.keycode && ev->state==config.keys.add_ws.modifier) 
    set_gnome_hint(root, WIN_WORKSPACE_COUNT, config.workspace_count+1); 
 
  /* Delete last workspace */ 
  if(ev->keycode==config.keys.del_ws.keycode && ev->state==config.keys.del_ws.modifier) 
    set_gnome_hint(root, WIN_WORKSPACE_COUNT, config.workspace_count-1); 
 
  /* Switch to specific workspace */ 
  if(ev->keycode==config.keys.ws_1.keycode && ev->state==config.keys.ws_1.modifier) 
    change_to_workspace(0); 
  if(ev->keycode==config.keys.ws_2.keycode && ev->state==config.keys.ws_2.modifier) 
    change_to_workspace(1); 
  if(ev->keycode==config.keys.ws_3.keycode && ev->state==config.keys.ws_3.modifier) 
    change_to_workspace(2); 
  if(ev->keycode==config.keys.ws_4.keycode && ev->state==config.keys.ws_4.modifier) 
    change_to_workspace(3); 
  if(ev->keycode==config.keys.ws_5.keycode && ev->state==config.keys.ws_5.modifier) 
    change_to_workspace(4); 
  if(ev->keycode==config.keys.ws_6.keycode && ev->state==config.keys.ws_6.modifier) 
    change_to_workspace(5); 
  if(ev->keycode==config.keys.ws_7.keycode && ev->state==config.keys.ws_7.modifier) 
    change_to_workspace(6); 
  if(ev->keycode==config.keys.ws_8.keycode && ev->state==config.keys.ws_8.modifier) 
    change_to_workspace(7); 
  if(ev->keycode==config.keys.ws_9.keycode && ev->state==config.keys.ws_9.modifier) 
    change_to_workspace(8); 
#endif /* WORKSPACE */ 
} 
#endif /* KEYBOARD */ 
 
/* 
 * 
 * Handle button press events 
 * 
 */ 

void event_button_press(XButtonEvent *ev) 
{ 
  Client *c=NULL; 
  Client *win=NULL;
  Window ptr_win; 
  int replay=False, focus=True, raise=True, xinfront=True;
//  unsigned int stateorg; 
//  XWindowChanges wc; 

  sprintf(lastsub,"TRACE1: event_button_press (window,lastfocus: %#lx,%#lx,%d)\n",ev->window,lastfocus,ev->button); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_button_press (window,lastfocus: %#lx,%#lx,%d)\n",ev->window,lastfocus,ev->button); 
#endif
 
  /*got_button=True;*/

/*
This "replay_if_drag" stuff is totally inoperative, so forget it for now...

  if(replay_if_drag==2) {
#ifdef SHOW_EVENTS
    printf("   got event again: replaying...\n");
#endif
    // got click event for the second time, so try to replay it to the window
    // does not work for some reason...?
    XAllowEvents(dpy, ReplayPointer, CurrentTime);
    replay_if_drag=0;
    return;
  } else if(replay_if_drag==1) {
#ifdef SHOW_EVENTS
    printf("   allowing events (sync) on button down...\n");
#endif
    replay_if_drag=0;
    XAllowEvents(dpy, SyncPointer, CurrentTime);
  }
*/

//  replay_button_event=*(XEvent *)ev;

//  win=get_focus_client();
  win=lastfocus_client;
  if(gX11InFront==0)
    win=NULL;
  else
    xinfront=False;
  
  c=get_client_from_window(ev->window, FRAME); 

  ev->state=ev->state & (Mod1Mask|ControlMask|ShiftMask|Mod2Mask);
  current_resize_client=NULL;
 
  /* Clicked on root */ 
  if(c==NULL) 
  { 
#ifdef DEBUG
    printf("Button pressed in root %d %d %d\n",ev->x_root,ev->y_root,ev->state);
#endif
/*    if(ev->button==1)
        c=check_for_undock(ev->x_root,ev->y_root);
        if(c!=NULL) {
            XAllowEvents(dpy, SyncPointer, CurrentTime);
            c->dock_timer=0;
            c->undock_timer=0;
            undock_client(c);
            return;
        }*/
    /* Proxy root click */ 
//    c=get_focus_client();
    // don't want any window to have focus...
    gone_to_back();
    XSendEvent(dpy, check_win, False, SubstructureNotifyMask, (XEvent *)ev);
  } 
  /* Clicked somewhere on a client */ 
  else 
  { 
#ifdef DEBUG
    printf("Button pressed in client %#lx, %d %d\n",c,c->is_docked,ev->state);
#endif
/*    if(c->is_docked && ev->button==1) {
        XAllowEvents(dpy, SyncPointer, CurrentTime);
        c->dock_timer=0;
        c->undock_timer=0;
        undock_client(c);
        return;
    }*/
    
    ptr_win=get_pointer_win(c->frame); 
 
    if( c->is_icon!=NOT_ICONBAR )
    {
        // clicked on an iconbar -don't wait for mouse up, just go ahead and restore it
      minmax_client(c);
    }
    
    /* If we want to resize */ 
    else if((ptr_win==c->resize && ((ev->button==Button1 && (ev->state==0 || ev->state==Mod1Mask)) || (ev->button==Button2 && ev->state==0))) || (ev->button==Button2 && ev->state==(Mod1Mask|ControlMask))) 
    { 
//      printf("Unmapping window in buttonpress\n");
//      c->ignore_unmap++; 
//      XUnmapWindow(dpy, c->window); 
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime); 
         
      /* Set mouse_x and mouse_y to position of cursor relative to southeast corner of window */ 
      mouse_x=FW-(ev->x_root-FX); 
      mouse_y=FH-(ev->y_root-FY); 
      mouse_mode=MOUSE_RESIZE;
      current_resize_client=c;
// start the timer for motion events
      if(gX11InFront==0)
        last_motion_time=ev->time+BG_FIRST_MOTION_GAP;
      else
        last_motion_time=ev->time+FIRST_MOTION_GAP;
        
      if((ev->button==Button1 && ev->state==Mod1Mask) || (ev->button==Button2 && ev->state==0)) {
        raise=False;
        focus=False;
        xinfront=False;
      }
    } 
 
    /* If we want to maximise */ 
    else if(ptr_win==c->zoom && ev->button==Button1 && ev->state==0) 
    { 
      raise=False;
      focus=False;
      xinfront=False;
// This next bit of code makes it behave like OS X windows -i.e. it comes straight in front as soon
// as maximise is pressed, rather than waiting for the user to 'let go' of the button -but I don't like it!!
//      raise_client(c);
//      set_focus_client(c);
//      win=c;
//      raise=True;
//      focus=True;
//      xinfront=true;
      if(c==win)
         c->zoom_state=BUTTON_PRESSED;
      if(c!=win)
         c->zoom_state=BUTTON_PRESSEDBACK;
      mouse_mode=MOUSE_ZOOM; 
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
// start the timer for motion events (don't need to bother with such a large gap at the start)
      last_motion_time=ev->time; 
      redraw_client_buttons(c); 
    } 
 
    /* If we want to 'iconbar-ise' */ 
    else if(ptr_win==c->icon && ev->button==Button1 && ev->state==0) 
    { 
      if(c==win)
         c->icon_state=BUTTON_PRESSED; 
      if(c!=win)
         c->icon_state=BUTTON_PRESSEDBACK;
      mouse_mode=MOUSE_ICON; 
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
// start the timer for motion events (don't need to bother with such a large gap at the start)
      last_motion_time=ev->time; 
      raise=False;
      focus=False;
      xinfront=False;
      redraw_client_buttons(c); 
    } 
 
    /* If we want to close */ 
    else if(ptr_win==c->close && ev->button==Button1 && ev->state==0) 
    { 
      if(c==win)
         c->close_state=BUTTON_PRESSED;
      if(c!=win)
         c->close_state=BUTTON_PRESSEDBACK;
      mouse_mode=MOUSE_CLOSE; 
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
      raise=False;
      focus=False;
      xinfront=False;
// start the timer for motion events (don't need to bother with such a large gap at the start)
      last_motion_time=ev->time; 
      redraw_client_buttons(c); 
    } 
 
    /* If we want to windowshade */ 
    else if(ptr_win!=c->window && ev->time-last_button1_time<DBL_CLICK && c==last_button1_client && ((ev->button==Button1 && (ev->state==0 || ev->state==Mod1Mask)) || (ev->button==Button2 && ev->state==0))) 
    { 
        // got a double-click on the title-bar
#ifdef DEBUG
        printf("Got button down again in titlebar...\n");
#endif
        toggle_client_shaded(c);
        last_button1_time=0; 
        last_button1_client=NULL;
        raise=False;
        focus=False;
        xinfront=False;
    }
     
    /* If we want to move */ 
    else if((ptr_win!=c->window && ((ev->button==Button1 && (ev->state==0 || ev->state==Mod1Mask)) || (ev->button==Button2 && ev->state==0))) || (ev->button==Button1 && ev->state==(Mod1Mask|ControlMask))) 
    { 
#ifdef DEBUG
        printf("Got button down in titlebar...\n");
#endif
        XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, move_curs, CurrentTime); 
 
        /* Set mouse_x and mouse_y to position of cursor relative to root window */ 
        mouse_x=ev->x_root; 
        mouse_y=ev->y_root;
        orgmove_x=mouse_x;
        orgmove_y=mouse_y;
        mouse_mode=MOUSE_MOVE; 
        if((ev->button==Button1 || ev->button==Button2) && (ev->state==0 || ev->state==Mod1Mask)) {
          last_button1_time=ev->time;
          last_button1_client=c;
          if(ev->state==Mod1Mask || ev->button==Button2) {
            raise=False;
            focus=False;
            xinfront=False;
          }
        } else {
          last_button1_time=0;
          last_button1_client=NULL;
          raise=False;
          focus=False;
          xinfront=False;
        }
// start the timer for motion events
        if(gX11InFront==0)
          last_motion_time=ev->time+BG_FIRST_MOTION_GAP;
        else
          last_motion_time=ev->time+FIRST_MOTION_GAP;
    } 
 
    /* If we want to 'iconbar-ise' */ 
    else if(ptr_win!=c->window && ev->button==Button3 && ev->state==0) 
    {
      mouse_mode=MOUSE3_TITLE; 
      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
//      minmax_client(c); 
      raise=False;
      focus=False;
      xinfront=False;
    }

    /* If we want to send to back */ 
    else if((ptr_win!=c->window && ev->button==Button3 && ev->state==ShiftMask) || (ev->button==Button3 && ev->state==(Mod1Mask|ControlMask))) 
    { 
      lower_client(c); 
      raise=False; 
      focus=False; 
      xinfront=False;
    } 
 
#ifdef WORKSPACE_SUPPORT 
    /* If we want to make window sticky */ 
    else if(ptr_win!=c->window && ev->button==Button2 && ev->state==Mod1Mask) 
      toggle_client_sticky(c); 
#endif /* WORKSPACE */ 
 
#ifdef LAYER_SUPPORT 
    /* If we want to move to a higher layer */ 
    else if(ptr_win!=c->window && ev->button==Button1 && ev->state==Mod1Mask) 
      set_client_layer(c, get_client_layer(c)+1); 
 
    /* If we want to send to a lower layer */ 
    else if(ptr_win!=c->window && ev->button==Button3 && ev->state==Mod1Mask) 
      set_client_layer(c, get_client_layer(c)-1); 
#endif /* LAYER */ 
    else if( gX11InFront==0 ) {
      // left-click in a window should just bring XDarwin/OroborOSX in front, since it isn't already
      if(ev->button==Button1 && ev->state==0) {
//This "replay_if_drag" stuff is totally inoperative, so forget it for now...
//        replay_if_drag=1;
#ifdef DEBUG
        printf("Got button in window -is but1\n");
#endif
        raise=True;
        focus=True;
        xinfront=True;
      } else {
        replay=True;
        raise=False;
        focus=False;
        xinfront=False;
#ifdef DEBUG
        printf("Got button in window -not but1\n");
#endif
      }
//    } else if( get_focus_client()!=c ) {
    } else if( lastfocus_client!=c ) {
      // left-click in a window should just bring it in front, since it isn't already
      if(ev->button==Button1 && ev->state==0) {
//This "replay_if_drag" stuff is totally inoperative, so forget it for now...
//        replay_if_drag=1;
        raise=True;
        focus=True;
        // if this window is set to accept the first click, then allow it go through
        if((c->auto_properties & NONAUTO_FIRSTCLICK)==NONAUTO_FIRSTCLICK)
            replay=True;
      } else {
        replay=True;
        raise=False;
        focus=False;
      }
      xinfront=False;
    } else {
      // click in focus window, so allow it through
      replay=True;
      raise=False;
      focus=False;
      xinfront=False;
      // if it is an autofocus window set to raise on click, then raise it
      if((c->auto_properties & AUTO_CLICKRAISE)==AUTO_CLICKRAISE)
        raise=True;
    }
 
    if(c->is_icon==NOT_ICONBAR) {
      if(raise) force_raise_client(c); 
      if(xinfront)
        Bring_X_To_Front(1);
      if(raise) raise_client(c); 
      if(focus) set_focus_client(c); 
      if(xinfront)
        Bring_To_Front(0);
    }

    if(replay)  {
#ifdef SHOW_EVENTS
      printf("  allowing events for replay on button down... %d,%d,%d,%d \n",(ev->state&ShiftMask),(ev->state&ControlMask),(ev->state&Mod1Mask),(ev->state&Mod2Mask));
#endif
      // Event can be sent through to window
//      XAllowEvents(dpy, SyncPointer, CurrentTime);
//      XSendEvent(dpy, c->window, False, ButtonPressMask, (XEvent *) &replay_button_event);
      last_gotbutton_time = GetCarbonTickCount();
      XAllowEvents(dpy, ReplayPointer, CurrentTime);
//      got_replay_pointer=2;
    } else if(replay_if_drag==1) {
//      XGrabPointer(dpy, c->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, None, CurrentTime); 
 
    /* Set mouse_x and mouse_y to position of cursor relative to root window */ 
      replay_x=ev->x_root; 
      replay_y=ev->y_root; 
#ifdef SHOW_EVENTS
      printf("  Starting drag check: %d,%d\n",replay_x,replay_y);
#endif
      replay_time=ev->time;
      replay_win=c->frame;
      XAllowEvents(dpy, SyncPointer, CurrentTime);
// try to get around the cursor bug...
      if(xinfront) {
        XUndefineCursor(dpy, c->frame);
        XDefineCursor(dpy, c->frame, normal_curs);
      }
// This is not working, so forget it for now...
      replay_if_drag=0;
    } else {
#ifdef SHOW_EVENTS
      printf("  allowing events (sync) on button down...\n");
#endif
      XAllowEvents(dpy, SyncPointer, CurrentTime);
// try to get around the cursor bug...
      if(xinfront) {
        XUndefineCursor(dpy, c->frame);
        XDefineCursor(dpy, c->frame, normal_curs);
      }
    }
  } 
} 

Client *check_for_hackid(int hackid)
{
    Client *c;
    int i;
    
    c=client_list;
    if(c==NULL) return NULL;
    if(hackid==0) return NULL;
    
    for(i=0;i<client_count;i++) {
        if(hackid==c->hack_id)
            return c;
        c=c->next;
    }
    return NULL;
}

/* 
 * 
 * Handle button release events 
 * 
 */ 

void event_button_release(XButtonEvent *ev) 
{ 
  Client *c=NULL; 
  int dx, dy;
 
  sprintf(lastsub,"TRACE1: event_button_release (window,lastfocus: %#lx,%#lx,%d)\n",ev->window,lastfocus,ev->button);
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_button_release (window,lastfocus: %#lx,%#lx,%d)\n",ev->window,lastfocus,ev->button);
#endif 
 
/*
This "replay_if_drag" stuff is totally inoperative, so forget it for now...

  if(replay_if_drag!=0) {
    replay_if_drag=0;
#ifdef SHOW_EVENTS
    printf("   allowing events (sync) on button release...\n");
#endif
  }
*/

 
    if(ev->button==5) {
#ifdef DEBUG
        printf("-------- Got button 5 at %d %d\n",ev->x_root,ev->y_root);
#endif
        check_ptrwin = 1;
        if(ev->y_root==100) {
            c=check_for_undock(ev->x_root);
            if(c!=NULL) {
//                XAllowEvents(dpy, SyncPointer, CurrentTime);
                c->dock_timer=0;
                c->undock_timer=0;
                undock_client(c);
//                XUngrabPointer(dpy, CurrentTime); 
                return;
            }
        } else if(ev->y_root==101) {
            /*if(mouse_mode!=MOUSE_NORMAL) {
#ifdef SHOW_EVENTS
                printf("-------- Ignoring entered at %d %d\n",ev->x_root,ev->y_root);
#endif
                return;
            }*/
            c=check_for_hackid(ev->x_root);
            if(c!=NULL) {
#ifdef SHOW_EVENTS
                printf("-------- entered client %#lx,%#lx,%d\n",c,last_client_entered_hack,c->is_e_m_r);
#endif
//                XAllowEvents(dpy, SyncPointer, CurrentTime);
//                XUngrabPointer(dpy, CurrentTime);
                c->is_e_m_r |= IS_ENTERED;
                if(c->is_docked && GetCarbonTickCount()>c->dock_timer) {
                    c->dock_timer=0;
                    c->undock_timer=0;
                    undock_client(c);
                }
/*                c=get_lowest_entered_client(1);
                if(c!=last_client_entered_hack) {
                    //entered_client(c,1);
                    last_client_entered_hack=c;
                    if(c==NULL)
                        lastenter=0;
                    else
                        lastenter=c->window;
                }*/
                return;
            }
        } else if(ev->y_root==102) {
            /*if(mouse_mode!=MOUSE_NORMAL) {
#ifdef SHOW_EVENTS
                printf("-------- Ignoring exited at %d %d\n",ev->x_root,ev->y_root);
#endif
                return;
            }*/
            c=check_for_hackid(ev->x_root);
            if(c!=NULL) {
#ifdef SHOW_EVENTS
                printf("-------- exited client %#lx,%#lx,%d\n",c,last_client_entered_hack,c->is_e_m_r);
#endif
//                XAllowEvents(dpy, SyncPointer, CurrentTime);
//                XUngrabPointer(dpy, CurrentTime); 
                if(c->is_docked && GetCarbonTickCount()>c->dock_timer) {
                    c->dock_timer=0;
                    c->undock_timer=0;
                    undock_client(c);
                }
                c->is_e_m_r &= NOT_ENTERED;
/*                c=get_lowest_entered_client(1);
                if(c!=last_client_entered_hack) {
                    //entered_client(c,1);
                    last_client_entered_hack=c;
                    if(c==NULL)
                        lastenter=0;
                    else
                        lastenter=c->window;
                }*/
                return;
            }
        }
        return;
    }
        
  current_resize_client=NULL;
  c=get_client_from_window(ev->window, FRAME);
  
  /* Release was on root window */ 
  if(c==NULL) 
  { 
#ifdef DEBUG
    printf("Button released in root %d %d\n",ev->x_root,ev->y_root);
#endif
    /* Proxy root release */ 
    XSendEvent(dpy, check_win, False, SubstructureNotifyMask, (XEvent *)ev); 
  } 
  /* Clicked somewhere on a client */ 
  else 
  { 
#ifdef DEBUG
    printf("Button released in client %#lx, %d\n",c,c->is_docked);
#endif
/*    if(c->is_docked && ev->button==1) {
        XAllowEvents(dpy, SyncPointer, CurrentTime);
        c->dock_timer=0;
        c->undock_timer=0;
        undock_client(c);
        return;
    }*/

    if(c->is_icon!=NOT_ICONBAR)
    {
      /* minmax_client(c); */
    }
    else
    {
      c->zoom_state=BUTTON_UNPRESSED; 
      c->icon_state=BUTTON_UNPRESSED;
      c->close_state=BUTTON_UNPRESSED; 
      redraw_client_buttons(c); 
       
      /* Release of button 3 occured on titlebar */
      if(mouse_mode==MOUSE3_TITLE && ev->button==3) {
        if(get_pointer_win(c->frame)!=c->window) {
          XUngrabPointer(dpy, CurrentTime); 
          mouse_mode=MOUSE_NORMAL;
          got_button=False;
          toggle_minimise_client(c,1); // perform the alternative minimisation method
          return;
        }
      /* Release occured on maximise button */
      } else if(mouse_mode==MOUSE_ZOOM || mouse_mode==MOUSE_ZOOM_OUT) {
        if(get_pointer_win(c->frame)==c->zoom) {
          raise_client(c);
          set_focus_client(c);
          if(gX11InFront==0)
            Bring_To_Front(0);
          toggle_client_maximized(c);
        }
      }
      /* Release occured on iconise button */ 
      else if(mouse_mode==MOUSE_ICON || mouse_mode==MOUSE_ICON_OUT) {
        if(get_pointer_win(c->frame)==c->icon)
          toggle_minimise_client(c,0);
      }
      /* Release occured on close button */ 
      else if(mouse_mode==MOUSE_CLOSE || mouse_mode==MOUSE_CLOSE_OUT) {
        if(get_pointer_win(c->frame)==c->close) 
          close_client(c);
      }
      /* If we are resizing */
      else if(mouse_mode==MOUSE_RESIZE) 
      { 
        /* Only allow resizing if not shaded */ 
        if((get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)==0) {
          dx=(ev->x_root-FX-LW-RW)+mouse_x; 
          dy=(ev->y_root-FY-TH-BH)+mouse_y; 
//#ifdef DEBUG
//          printf("Mapping window & Setting w,h in release: %d,%d\n",dx,dy);
//#endif
          // make sure configure event is sent
//          XMapWindow(dpy, c->window); 
          mouse_mode=MOUSE_NORMAL;
          c->is_e_m_r &= NOT_RESIZING;
          resize_client(c, dx, dy, False);
        }
      }
      /* If moving window */ 
      else if(mouse_mode==MOUSE_MOVE) 
      { 
        dx=orgmove_x-ev->x_root;
        dy=orgmove_y-ev->y_root;
        if((ev->time-last_button1_time<DBL_CLICK) && (last_button1_client==c) && dx<6 && dy<6 && dx>-6 && dy>-6) {
          last_button1_time=ev->time;
          last_button1_client=c;
#ifdef DEBUG
          printf("Got fast button up for MOUSE_MOVE...\n");
#endif
        } else {
#ifdef DEBUG
          printf("Got slow button up for MOUSE_MOVE...\n");
#endif
          dx=ev->x_root-mouse_x; 
          dy=ev->y_root-mouse_y; 
          mouse_x=ev->x_root; 
          mouse_y=ev->y_root;
          mouse_mode=MOUSE_NORMAL;
          c->is_e_m_r &= NOT_MOVING;
          move_client(c, c->x + dx, c->y + dy, 0);
//          redraw_client_frame(c);	// make sure it is redrawn translucent
        }
      }
      else if(ev->time-last_button1_time<DBL_CLICK && (last_button1_client==c))
      {
        if(get_pointer_win(c->frame)==c->title) {
#ifdef DEBUG
          printf("Got fast button up in titlebar (not MOUSE_MOVE)...\n");
#endif
          toggle_client_shaded(c);
        } else
#ifdef DEBUG
          printf("Got fast button up (not MOUSE_MOVE, nor title)...\n");
#endif
        last_button1_time=0; 
        last_button1_client=NULL;
      }
//      if(mouse_mode==MOUSE_RESIZE)
//        resize_client(c, c->width, c->height, False); 
    }

  } 

  XUngrabPointer(dpy, CurrentTime); 
  XAllowEvents(dpy, SyncPointer, CurrentTime);
  mouse_mode=MOUSE_NORMAL;
  
  got_button=False;
} 

/*
 * Attempt to sort out a problem that sometimes appears, where the X11 cursor looks like
 * it has frozen (i.e. been grabbed) after a menu item has been chosen that did not work?
 * Don't know why it does this - but this attempt does not fix it either...
 */
 
void force_ungrab()
{
  XUngrabPointer(dpy, CurrentTime); 
  XAllowEvents(dpy, SyncPointer, CurrentTime);
  mouse_mode=MOUSE_NORMAL;
  
  got_button=False;
}

/* 
 * 
 * Handle motion notify events 
 * 
 */ 

void event_motion_notify(XMotionEvent *ev) 
{ 
  Client *c=NULL; 
  Client *win=NULL;
  int height, width, dx=0, dy=0;
  Window w1, w2; 
  int wx, wy, mask, x, y; 
//  int now_x, now_y; 
  XEvent e, *ep=NULL;
  XMotionEvent ee;
 
  sprintf(lastsub,"TRACE1: event_motion_notify (window,time,lastfocus: %#lx,%#lx,%#lx)\n",ev->window,ev->time,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_motion_notify (window,time,lastfocus: %#lx,%#lx,%#lx)\n",ev->window,ev->time,lastfocus); 
#endif 
 
  check_ptrwin=1;
  if(replay_motion!=0) return;
  
  // hack an XEvent pointer to the XMotionEvent
  ep=(XEvent *) &ee;
  
  /* If there are other motion events on this window, use the latest one! (well, maybe not...) */
  if(XCheckTypedWindowEvent(dpy, ev->window, MotionNotify,&e)) 
  { 
// Why not actually use the latest motion event!?
/*    if(e.time>ev->time)
      *ev=e;
#ifdef DEBUG 
    printf("Got another motion event: %d\n",e.time);
#endif
    while(XCheckTypedWindowEvent(dpy, ev->window, MotionNotify,(XEvent *) &e)) {
// Why not actually use the latest motion event!?
      if(e.time>ev->time)
        *ev=e;
//      dx++;
#ifdef DEBUG 
      printf("Got more motion events: %d\n",e.time);
#endif
    }
hmmmm... probably because it doesn't work, for some strange reason...??? */
    *ep=e;
    while(XCheckTypedWindowEvent(dpy, ev->window, MotionNotify,&e)) 
      *ep=e;
//    XPutBackEvent(dpy,ep);
    replay_motion=0;
    *ev=ee;
//    return; 
  }
  
/*
  if(!got_button) return;
*/
  if(gX11InFront==0)
    win=NULL;
  else
    win=lastfocus_client;
//    win=get_focus_client();

  c=get_client_from_window(ev->window, FRAME); 
  if(c!=NULL) 
  { 
    /* If there is a release event on this window, ignore this motion */ 
    if(XCheckTypedWindowEvent(dpy, ev->window, ButtonRelease,&e)) 
    { 
#ifdef DEBUG
      printf("Got button release in motion (%d)\n",mouse_mode);
#endif
      XPutBackEvent(dpy,&e); 
      replay_motion=0;
      return; 
    }
    /* If resizing */ 
    if(mouse_mode==MOUSE_RESIZE) 
    { 
      /* Don't allow resizing if maximized or shaded 
      if(get_gnome_hint(c->window, WIN_STATE) & (WIN_STATE_MAXIMIZED|WIN_STATE_SHADED)) */ 
      if(get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)
        return;
 
      // don't want to get swamped by these events, so only allow so many per second
      // (this also helps spurious motion events which occur when really only clicking...)
//      if(ev->time-last_motion_time<RESIZE_MOTION_GAP) {
//#ifdef DEBUG
//        printf("Not time yet for motion (MOUSE_RESIZE) -put it back (%d,%d)\n",ev->x_root,ev->y_root);
//#endif
//        replay_motion_event=*(XEvent *)ev;
//        last_motion_time-=20; 
//        replay_motion=1;
//        return; }
      replay_motion=0;

      width=(ev->x_root-FX-LW-RW)+mouse_x; 
      height=(ev->y_root-FY-TH-BH)+mouse_y;
      if(c->width!=width || c->height!=height) {
#ifdef DEBUG
        printf("Setting w,h in motion: %d,%d  (%d,%d)\n",width,height,c->width,c->height);
#endif
        c->is_e_m_r |= IS_RESIZING;
        resize_client(c, width, height, !config.continuous_resize); 
//      resize_client(c, width, height, True); 
        // reset the timer for motion events
        c->is_e_m_r &= NOT_RESIZING;
        last_motion_time=ev->time; 
      }
    } 
 
    /* If moving window */ 
    else if(mouse_mode==MOUSE_MOVE) 
    { 
      // don't want to get swamped by these events, so only allow so many per second
      // (this also helps spurious motion events which occur when really only clicking...)
//      if(ev->time-last_motion_time<MOVE_MOTION_GAP) {
//#ifdef DEBUG
//        printf("Not time yet for motion (MOUSE_MOVE) -put it back (%d,%d)\n",ev->x_root,ev->y_root);
//#endif
//        replay_motion_event=*(XEvent *)ev;
//        replay_motion=1;
//        last_motion_time-=20; 
//        return; }

      dx=ev->x_root-mouse_x; 
      dy=ev->y_root-mouse_y; 
      mouse_x=ev->x_root; 
      mouse_y=ev->y_root; 
      if(dx!=0 || dy!=0) {
        // reset the timer for motion events
        last_motion_time=ev->time; 
        c->is_e_m_r |= IS_MOVING;
        move_client(c, c->x + dx, c->y + dy, 0);
        c->is_e_m_r &= NOT_MOVING;
      }
      replay_motion=0;
    } 
    else if(mouse_mode==MOUSE_ZOOM)
    // must check for enter/leave of window widgets
    {
      if(get_pointer_win(c->frame)!=c->zoom) {
        mouse_mode=MOUSE_ZOOM_OUT;
        if(c==win)
          c->zoom_state=BUTTON_OVER;
        if(c!=win)
          c->zoom_state=BUTTON_OVERBACK;  
        redraw_client_buttons(c); }
    }   
    else if(mouse_mode==MOUSE_ZOOM_OUT)
    {     
       if(get_pointer_win(c->frame)==c->zoom) {
         mouse_mode=MOUSE_ZOOM;
         if(c==win)
           c->zoom_state=BUTTON_PRESSED;
         if(c!=win)
           c->zoom_state=BUTTON_PRESSEDBACK;
         redraw_client_buttons(c); }
    } 
    else if(mouse_mode==MOUSE_ICON)
    {
      if(get_pointer_win(c->frame)!=c->icon) {
        mouse_mode=MOUSE_ICON_OUT;
        if(c==win)
          c->icon_state=BUTTON_OVER;
        if(c!=win)
          c->icon_state=BUTTON_OVERBACK;
        redraw_client_buttons(c); }
    }
    else if(mouse_mode==MOUSE_ICON_OUT)
    {
       if(get_pointer_win(c->frame)==c->icon) {
         mouse_mode=MOUSE_ICON;
         if(c==win)
           c->icon_state=BUTTON_PRESSED;
         if(c!=win)
           c->icon_state=BUTTON_PRESSEDBACK;
         redraw_client_buttons(c); }
    }
    else if(mouse_mode==MOUSE_CLOSE)
    {
      if(get_pointer_win(c->frame)!=c->close) {
        mouse_mode=MOUSE_CLOSE_OUT;
        if(c==win)
          c->close_state=BUTTON_OVER;
        if(c!=win)
          c->close_state=BUTTON_OVERBACK;
        redraw_client_buttons(c); } 
    } 
    else if(mouse_mode==MOUSE_CLOSE_OUT)
    {
       if(get_pointer_win(c->frame)==c->close) {
         mouse_mode=MOUSE_CLOSE;
         if(c==win)
           c->close_state=BUTTON_PRESSED;
         if(c!=win)
           c->close_state=BUTTON_PRESSEDBACK;
         redraw_client_buttons(c); }
    }
    XQueryPointer(dpy, root, &w1, &w2, &x, &y, &wx, &wy, &mask);
    if(can_send_hackcmds==0) Compare_Screens(real_x,real_y,x,y,2);
  } 
} 
 
/* 
 * 
 * Handle idle time -looks out for cursor over window widgets
 * 
 */ 

void event_idle()
{
  Client *c=NULL;
  Client *win=NULL;
  Client *top=NULL;
  Window ptrwin=0;
  int retval;
//  int now_x, now_y;

#ifdef DEBUG
#ifndef SHOW_IDLE
// Showing all of the debugging junk is a bit much during idle time (background mouse tracking),
// so this makes sure we don't see it unless we specifically ask for it by defining SHOW_IDLE
  return;
#endif
#endif

  sprintf(lastsub,"TRACE1: event_idle\n");
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_idle\n");
#endif

/*
This "replay_if_drag" stuff is totally inoperative, so forget it for now...

  // replaying for drag in background window does not work (for some reason the XAllowEvents with
  // replayPointer does not actually send the event back to the window...)
  if(replay_if_drag==1) {
    // Hmmm... CurrentTime is not actually the value of the current time,
    // so this timer is rubbish anyway...
    if(CurrentTime-replay_time>DBL_CLICK) {
      get_mouse_xy(&now_x,&now_y);
#ifdef SHOW_EVENTS
      printf("  check for drag: %d,%d  (%d,%d)\n",now_x,now_y,replay_x,replay_y);
      printf("  currenttime/replaytime: %#lx,%#lx\n",CurrentTime,replay_time);
#endif
      if(now_x>replay_x+2)
         replay_if_drag=2;
      else if(now_x<replay_x-2)
         replay_if_drag=2;
      else if(now_y>replay_y+2)
         replay_if_drag=2;
      else if(now_y<replay_y-2)
         replay_if_drag=2;
      if(replay_if_drag==2) {
#ifdef SHOW_EVENTS
        printf("  re-sending event for drag...\n");
#endif
        XPutBackEvent(dpy,&replay_button_event);
      }
    } else
#ifdef SHOW_EVENTS
      printf("  waiting for DBL_CLICK time...\n");
#endif
    return;
  }
*/

  if(mouse_mode!=MOUSE_NORMAL)
    return;

/*  XGetInputFocus(dpy, &focus_win, &dummy); */

/*  if(lastfocus_client!=NULL && last_gotbutton_time!=0) {
    retval=XGrabPointer(dpy, lastfocus_client->frame, False, PointerMotionMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, normal_curs, CurrentTime);
    if(retval==GrabSuccess) {
      XUngrabPointer(dpy, CurrentTime);
      if(GetCarbonTickCount()>last_gotbutton_time+10) last_gotbutton_time=0;
      if(inhibiting_info!=0) {
        printf("Got %d from XGrabPointer -uninhibiting info. %d %d %d\n",retval,GrabSuccess,AlreadyGrabbed,GrabFrozen);
        SendSimpleHackCommand(HACKCMD_INHIBITINFO,0);
      }
    } else if(inhibiting_info==0) {
      printf("Got %d from XGrabPointer -inhibiting info. %d %d %d\n",retval,GrabSuccess,AlreadyGrabbed,GrabFrozen);
      SendSimpleHackCommand(HACKCMD_INHIBITINFO,1);
    }
  }
  
  // don't want to do anything odd if the cursor has been grabbed by a client
  if(inhibiting_info!=0) return;*/

  if(can_send_hackcmds==0) {

  ptrwin=get_pointer_win(root);

  if( ptrwin!=lastenter ) {
    if( ptrwin!=lastbadenter || lastenter!=0 ) {
#ifdef SHOW_IDLE
      printf("  discrepancy: %#lx %#lx\n",ptrwin,lastenter );
#endif
      lastenter=ptrwin;
      if(lastenter!=0) {
        c=get_client_from_window(lastenter, FRAME_LE);
        if(c==NULL) {
#ifdef SHOW_IDLE
          printf("  failed to find client: %#lx %#lx\n",lastenter, root );
#endif
          lastenter=0;
          lastbadenter=ptrwin;
        }
      }
      if(c!=NULL && c->is_docked)
        return;
      else
        entered_client(c,True);
    }
  } else if(lastenter!=0)
    c=get_client_from_window(lastenter, FRAME_LE);

  } else if(check_ptrwin==0) {
    c=last_client_entered_hack;
    if(c==NULL) {
        if(lastenter!=0) {
            printf("%s warning: non-zero lastenter with NULL last_client_entered_hack: %#lx\n",PROGNAME,lastenter);
            lastenter = 0;
        }
    } else if(c->frame!=lastenter) {
        printf("%s warning: lastenter does not match with last_client_entered_hack: %#lx,%#lx\n",PROGNAME,lastenter,c->frame);
        lastenter = c->frame;
    }
    ptrwin=lastenter;
  } else {
    check_ptrwin=0;
    ptrwin=get_pointer_win(root);
    if(gX11InFront!=0) {
        c=get_top_most_client(WIN_LAYER_NORMAL,False);
        if(c!=NULL)
            if(c->frame==ptrwin && c!=last_client_entered_hack)
                c->is_e_m_r |= IS_ENTERED;
    }
    c=last_client_entered_hack;
    if(c==NULL) {
        lastenter=0;
        if(ptrwin!=0) {
            c=get_client_from_window(ptrwin,FRAME);
            if(c!=NULL) {
                win=get_lowest_entered_client(1);
                if(win==NULL) {
                    c=NULL;
                } else if(win==c) {
                    entered_client(c,1);
                    lastenter=c->frame;
                    last_client_entered_hack=c;
                } else if((c->is_e_m_r & IS_ENTERED)==0) {
                    c=NULL;
                } else {
                    win=c;
                    entered_client(c,1);
                    lastenter=c->frame;
                    last_client_entered_hack=c;
                }
            }
        }
    } else {
        lastenter=c->frame;
        if((c->is_e_m_r & IS_ENTERED)==0 || ptrwin!=lastenter) {
            c=get_client_from_window(ptrwin,FRAME);
            if(c==NULL) {
                entered_client(NULL,1);
                lastenter=0;
                last_client_entered_hack=NULL;
            } else {
                win=get_lowest_entered_client(1);
                if(win==NULL) {
                    c=NULL;
                    entered_client(NULL,1);
                    lastenter=0;
                    last_client_entered_hack=NULL;
                } else if(win==c) {
                    entered_client(c,1);
                    lastenter=c->frame;
                    last_client_entered_hack=c;
                } else if((c->is_e_m_r & IS_ENTERED)==0) {
                    c=NULL;
                    entered_client(NULL,1);
                    lastenter=0;
                    last_client_entered_hack=NULL;
                } else {
                    win=c;
                    entered_client(c,1);
                    lastenter=c->frame;
                    last_client_entered_hack=c;
                }
            }
        }
    }
  }
  
  ptrwin=0;

  if( hacking_focus ) return;
  
  if(gX11InFront==0)
    win=NULL;
  else
    win=lastfocus_client;
//    win=get_focus_client();
  
  if(win!=NULL && config.focus_style!=FOCUS_SLOPPY) {
    top=get_top_most_client(WIN_LAYER_NORMAL,False);
    if(top!=win) {
      if((win->auto_properties & AUTO_FOCUS)==0) {
        printf("%s info: Noticed that top and focus clients are different...\n",PROGNAME);
        set_focus_client(top);
        win=lastfocus_client;
//        win=get_focus_client();
      }
    }
  }


  if(c!=NULL)
  {
    // if over different client, make sure window widgets of last client are reset
    if(lastc!=c) {
      if(lastc!=NULL) {
        if(lastc->is_icon==NOT_ICONBAR) {
#ifdef SHOW_IDLE
          printf("Resetting buttons of last client %#lx (%#lx)\n",lastc,c);
#endif
          lastc->zoom_state=BUTTON_UNPRESSED;
          lastc->icon_state=BUTTON_UNPRESSED;
          lastc->close_state=BUTTON_UNPRESSED;
          redraw_client_buttons(lastc);
        }
      }
    }
    lastc=c;
    ptrwin=get_pointer_win(c->frame);
    // check if we're over any of the window widgets
    if(ptrwin==c->zoom || ptrwin==c->icon || ptrwin==c->close) {
#ifdef SHOW_IDLE
       printf("Over buttons of current client %#lx (%#lx)\n",c,lastc);
#endif
       if(c->zoom_state==BUTTON_UNPRESSED || c->icon_state==BUTTON_UNPRESSED || c->close_state==BUTTON_UNPRESSED) {
         // set "x - +" glyphs in all of them appropriately
         if( win==c ) {
           c->zoom_state=BUTTON_OVER;
           c->icon_state=BUTTON_OVER;
           c->close_state=BUTTON_OVER;
         } else {
           c->zoom_state=BUTTON_OVERBACK;
           c->icon_state=BUTTON_OVERBACK;
           c->close_state=BUTTON_OVERBACK;
         }
#ifdef SHOW_IDLE
         printf("Lighting up buttons of current client %#lx (%#lx)\n",c,lastc);
#endif
         redraw_client_buttons(c);
       }
    } else if(c->zoom_state!=BUTTON_UNPRESSED || c->icon_state!=BUTTON_UNPRESSED || c->close_state!=BUTTON_UNPRESSED) {
       // one of the widgets had a glyph, so make sure none of them have now
       c->zoom_state=BUTTON_UNPRESSED;
       c->icon_state=BUTTON_UNPRESSED;
       c->close_state=BUTTON_UNPRESSED;
#ifdef SHOW_IDLE
       printf("Switching off buttons of current client %#lx (%#lx)\n",c,lastc);
#endif
       redraw_client_buttons(c);
       lastc=NULL;
    }
  }
  else if(lastc!=NULL)
  {
    // not over a window any more, so make sure last one has no glyphs in window widgets
    if(lastc->zoom_state!=BUTTON_UNPRESSED || lastc->icon_state!=BUTTON_UNPRESSED || lastc->close_state!=BUTTON_UNPRESSED) {
      lastc->zoom_state=BUTTON_UNPRESSED;
      lastc->icon_state=BUTTON_UNPRESSED;
      lastc->close_state=BUTTON_UNPRESSED;
#ifdef SHOW_IDLE
       printf("Switching off buttons of last client %#lx (%#lx)\n",lastc,c);
#endif
      redraw_client_buttons(lastc);
      lastc=NULL;
    }
  }
}


/* 
 * 
 * Handle enter notify events - and also leave notify events
 * 
 */ 

void event_enter_notify(XCrossingEvent *ev, int entering) 
{ 
  Window ptrwin=0;
  Client *c=NULL; 
  XEvent e, ee; 

  check_ptrwin=1;
  if(can_send_hackcmds!=0) return;
    
  sprintf(lastsub,"TRACE1: event_enter_notify (window,lastenter,entering,lastfocus: %#lx,%#lx,%d,%#lx)\n",ev->window,lastenter,entering,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_enter_notify (window,lastenter,entering,lastfocus: %#lx,%#lx,%d,%#lx)\n",ev->window,lastenter,entering,lastfocus); 
#endif
 
  /* If there are other enter/leave notify events, ignore this one */ 
  if(XCheckTypedEvent(dpy, EnterNotify|LeaveNotify, &e))
  {
    ee=e;
    while(XCheckTypedEvent(dpy, EnterNotify|LeaveNotify, &e))
      ee=e;
    XPutBackEvent(dpy, &ee);
#ifdef DEBUG
    printf("Found other events -ignoring enter\n");
#endif
    return;
  }
  
  MyIncreaseXServerGrab();
  if( entering && ev->window ) {
    ptrwin=get_pointer_win(root);
    if(ptrwin!=ev->window) {
      printf("Got enter %#lx, but in window %#lx %#lx\n",ev->window,ptrwin,lastenter);
      MyDecreaseXServerGrab();
      return; }
    c=get_client_from_window(ev->window,FRAME);
#ifdef DEBUG
    if(c==NULL)
       printf("  got NULL for frame - not changing lastenter\n");
#endif
    if(c!=NULL) {
        if(c->is_docked) {
            MyDecreaseXServerGrab();
            return;
        }
//       SendHackCommand(c,2,HACKCMD_HITCHECK,real_x+diffscrn_x,real_y+diffscrn_y,0,0,0);
       if(lastenter!=ev->window)
         entered_client(c,mouse_mode!=MOUSE_STOPPED);
#ifdef DEBUG
       else
         printf("  entered lastenter?!\n");
#endif
       lastenter=ev->window;
    }
  } else {
    if(ev->window==lastenter && lastenter) {
      c=get_client_from_window(ev->window,FRAME);
      if(c!=NULL) {
        if(c->is_docked) {
            MyDecreaseXServerGrab();
            return;
        }
        ptrwin=get_pointer_win(c->frame);
        if(ptrwin) {
#ifdef DEBUG
          printf("  ignoring leave event: [%#lx], %#lx\n",lastenter,ptrwin);
#endif
          MyDecreaseXServerGrab();
          return;
        }
        if(c->is_icon!=NOT_ICONBAR) {
          ptrwin=get_pointer_win(root);
          if(ptrwin) {
            c=get_client_from_window(ptrwin,FRAME);
            if(c!=NULL) {
              if(c->is_icon!=NOT_ICONBAR) {
                MyDecreaseXServerGrab();
                return; }
            }
          }
        }
      }
    }
    ptrwin=get_pointer_win(root);
    if(!ptrwin) {
      lastenter=0;
      entered_client(NULL,mouse_mode!=MOUSE_STOPPED);
#ifdef DEBUG
    } else {
      printf("Got leave, but cursor in window %#lx %#lx\n",ptrwin,lastenter);
#endif
    }
  }
  MyDecreaseXServerGrab();
#ifdef DEBUG
  printf("  set lastenter to [%#lx] %#lx, [%#lx] %#lx\n",lastenter,ev->window,c==NULL?0:c->frame,c==NULL?0:c->window);
#endif

} 
 
/* 
 * 
 * Handle focus in events 
 * 
 */ 

void event_focus_in(XFocusChangeEvent *ev) 
{ 
  Client *c=NULL; 
  XEvent e, ee; 
 
  sprintf(lastsub,"TRACE1: event_focus_in (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_focus_in (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif
  
  check_ptrwin=1;
  return;
  
  /* If there are other focus in events, ignore this one */ 
  MyIncreaseXServerGrab();
  if(XCheckTypedEvent(dpy, FocusIn, &e)) 
  { 
#ifdef DEBUG
    printf("  ignored...\n");
#endif
    ee=e;
    while(XCheckTypedEvent(dpy, FocusIn, &e)) 
      ee=e;
    XPutBackEvent(dpy, &ee); 
  }
  else
  {
    c=get_client_from_window(ev->window, WINDOW); 
    if(c!=NULL) 
    { 
      if(c->is_icon==NOT_ICONBAR)
        set_focus_client(c);
/*      if(c->is_icon==NOT_ICONBAR && ev->window!=lastfocus) {
        if(config.raise_on_focus || config.focus_style==FOCUS_CLICK) 
          raise_client(c); 
        lastfocus=ev->window;
        lastfocus_client=c;
        redraw_client_frame(c);
      }*/
    } 
    else {
      lastfocus=0;
      set_focus_client(get_top_most_client(WIN_LAYER_NORMAL,False)); 
    }
  }
  MyDecreaseXServerGrab();
} 
 
/* 
 * 
 * Handle focus out events 
 * 
 */ 

void event_focus_out(XFocusChangeEvent *ev) 
{ 
  Client *c=NULL; 
 
  sprintf(lastsub,"TRACE1: event_focus_out (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_focus_out (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  check_ptrwin=1;
//  c=get_client_from_window(ev->window, WINDOW); 
//  if(c!=NULL && c==lastfocus_client) {
//    set_focus_client(NULL);
//    lastfocus_client=NULL;
//    redraw_client_frame(c);
//  }
} 
 
/* 
 * 
 * Handle destroy notify events 
 * 
 */  

void event_destroy_notify(XDestroyWindowEvent *ev) 
{ 
  Client *c; 
   
  sprintf(lastsub,"TRACE1: event_destroy_notify (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_destroy_notify (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
//  if(ev->window==check_win) {
//    printf("check_win destroyed! Quitting...\n");
//    exit(0);
//  }
  
  check_ptrwin=1;
  if(ev->window==lastfocus)
    lastfocus=0;

  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) { 
    delete_client(c, 0); 
  } else if(lastenter!=0) {
    c=get_client_from_window(ev->window, FRAME);
    if(c!=NULL) {
#ifdef DEBUG
      printf("  checking lastenter: [%#lx]",c->frame);
#endif
      if(c->frame==lastenter) {
#ifdef DEBUG
         printf("  got lastenter in destroy: [%#lx]",lastenter);
#endif
         lastenter=0;
      }
    }
  }
} 
 
/* 
 * 
 * Handle unmap notify events 
 * 
 */ 

void event_unmap_notify(XUnmapEvent *ev) 
{ 
  Client *c=NULL; 
  XEvent e; 

  sprintf(lastsub,"TRACE1: event_unmap_notify (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: event_unmap_notify (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif
 
  check_ptrwin=1;
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    if(c->ignore_unmap) 
      c->ignore_unmap--; 
    else 
    { 
      delete_client(c, 0);
        // check to see if there are any more unmaps to come...
      while (XCheckTypedEvent(dpy, UnmapNotify, &e)) 
      { 
        c=get_client_from_window(((XUnmapEvent *) &e)->window, WINDOW );
        // if windows have already vanished, the following get_top_most_client might cause BadWindow
        // errors, so go back and deal with this next unmapped window
        if(c!=NULL) {
          if(c->ignore_unmap)
            c->ignore_unmap--;
          else
            delete_client(c, 0);
        }
      }
      if(still_in_auto_client()==0)
        set_focus_client(get_top_most_client(WIN_LAYER_NORMAL,False)); 
    } 
  } 
} 
 
/* 
 * 
 * Handle map request events 
 * 
 */ 
void event_map_request(XMapRequestEvent *ev) 
{ 
  Client *c=NULL; 
 
  sprintf(lastsub,"TRACE1: event_map_request (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_map_request (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  check_ptrwin=1;
  c=get_client_from_window(ev->window, WINDOW); 
  if(c==NULL) 
    create_new_client(ev->window, True); 
  else 
  { 
    raise_client(c);
    if(still_in_auto_client()==0)
      set_focus_client(c); 
  } 
} 
 
/* 
 * 
 * Handle configure request events 
 * 
 */ 

void event_configure_request(XConfigureRequestEvent *ev) 
{ 
  Client *c=NULL; 
  XWindowChanges wc; 
  XEvent e, ee; 
 
  sprintf(lastsub,"TRACE1: event_configure_request (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_configure_request (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  check_ptrwin=1;
  MyIncreaseXServerGrab();
  // If there are other configure requests on this window, ignore this one 
/*  if(XCheckTypedWindowEvent(dpy, ev->window, ConfigureRequest, &e)) 
  { 
    ee=e;
    while(XCheckTypedWindowEvent(dpy, ev->window, ConfigureRequest, &e)) {
      ee=e;
    }
    XPutBackEvent(dpy, &ee);
    MyDecreaseXServerGrab();
    return; 
  }*/ 
  // Actually, can't really do that because they may not just be changes to the size of the
  // window. They might also be changes to the position or to the stacking order.
  // We can't really afford to miss any of them!
  // It may be possible to check ahead for configures that match this one (i.e. look ahead for
  // another resize/move/stacking and ignore this one if one is found), but then we would still
  // need to make sure the *order* of requests is kept the same, so it would mean we would have
  // to put all events, except the current one and all matching ones *before* the last matching,
  // back onto the event queue in the same order we found them...
  
  wc.x=ev->x; 
  wc.y=ev->y; 
  wc.width=ev->width; 
  wc.height=ev->height; 
  wc.sibling=ev->above; 
  wc.stack_mode=ev->detail; 
  wc.border_width=0; 
 
//  XGrabServer(dpy);
  c=get_client_from_window(ev->window, WINDOW);
  if(c!=NULL) printf("doing configure_client from event_configure_request: %#lx %d\n",c,ev->value_mask);
  if(c!=NULL) {
    configure_client(c, ev->value_mask, &wc, c->is_icon!=NOT_ICONBAR, 2); 
  } else 
    XConfigureWindow(dpy, ev->window, ev->value_mask, &wc); 

// If a client is trying to come in front (not advised!) then bring X11 in front
  if (gX11InFront==0 && (ev->value_mask & CWStackMode) && (ev->detail==Above)) 
    Bring_To_Front(0);

  MyDecreaseXServerGrab();
} 
 
/* 
 *  
 * Handle property notify events 
 * 
 */ 

void event_property_notify(XPropertyEvent *ev) 
{ 
  Client *c=NULL; 
  long dummy;
 
  sprintf(lastsub,"TRACE1: event_property_notify (window,lastfocus: %#lx,%#lx,%d)\n",ev->window,lastfocus,ev->atom); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_property_notify (window,lastfocus: %#lx,%#lx,%d)\n",ev->window,lastfocus,ev->atom);
#endif 
 
  check_ptrwin=1;
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    switch(ev->atom) 
    { 
    case XA_WM_NAME: 
      if(c->name) XFree(c->name); 
      XFetchName(dpy, c->window, &c->name); 
      XClearWindow(dpy, c->title); 
      redraw_client_title(c); 
      break; 
    case XA_WM_NORMAL_HINTS: 
      XGetWMNormalHints(dpy, c->window, c->size, &dummy); 
      break; 
    } 
  } 
#ifdef WORKSPACE_SUPPORT 
  else 
  { 
    if(ev->atom==gnome[WIN_WORKSPACE_COUNT]) 
      update_workspace_count(); 
  } 
#endif /* WORKSPACE */ 

  sprintf(lastsub,"TRACE1: leaving event_property_notify...\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: leaving event_property_notify...\n"); 
#endif 
 
} 
 
/* 
 * 
 * Handle client message events 
 * 
 */ 

void event_client_message(XClientMessageEvent *ev) 
{ 
  Client *c=NULL; 
 
  sprintf(lastsub,"TRACE1: event_client_message (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_client_message (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  check_ptrwin=1;
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    /* Client wants to minimise */ 
    if(ev->message_type==icccm[WM_CHANGE_STATE] && ev->format==32 && ev->data.l[0]==IconicState) 
      toggle_minimise_client(c,0);
//      minmax_client(c); 
 
    /* Client wants to toggle shaded state */ 
    if(ev->message_type==gnome[WIN_STATE] && ev->format==32 && (ev->data.l[0] & WIN_STATE_SHADED)) 
      toggle_client_shaded(c); 
       
#ifdef WORKSPACE_SUPPORT 
    /* Client wants to toggle sticky state */ 
    if(ev->message_type==gnome[WIN_STATE] && ev->format==32 && (ev->data.l[0] & WIN_STATE_STICKY)) 
      toggle_client_sticky(c); 
#endif /* WORKSPACE */ 
 
#ifdef LAYER_SUPPORT 
    /* Client wants to change layer */ 
    if(ev->message_type==gnome[WIN_LAYER] && ev->format==32) 
      set_client_layer(c, ev->data.l[0]); 
#endif /* LAYER */ 
  } 
  else if(ev->window==root)  
  { 
#ifdef WORKSPACE_SUPPORT 
    /* Change to new workspace */ 
    if(ev->message_type==gnome[WIN_WORKSPACE] && ev->format==32) 
      change_to_workspace(ev->data.l[0]); 
#endif /* WORKSPACE */ 
  } 
} 
 
#ifdef SHAPE_SUPPORT 
/* 
 * 
 * Handle shape events 
 * 
 */ 

void event_shape(XShapeEvent *ev) 
{ 
  Client *c=NULL; 
  XEvent e, ee; 
 
  sprintf(lastsub,"TRACE1: event_shape (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: event_shape (window,lastfocus: %#lx,%#lx)\n",ev->window,lastfocus); 
#endif 
 
  check_ptrwin=1;
  c=get_client_from_window(ev->window, WINDOW); 
  if(c!=NULL) 
  { 
    c->is_shaped=1;
    /* If there are other shape events on this window, ignore this one */
//    XGrabServer(dpy);
    if(XCheckTypedWindowEvent(dpy, ev->window, shape_event, &e)) 
    {
      ee=e;
      while(XCheckTypedWindowEvent(dpy, ev->window, shape_event, &e))
        ee=e;
      XPutBackEvent(dpy, &ee); 
//      XUngrabServer(dpy);
      return; 
    } 
//    XUngrabServer(dpy);

    set_client_shape(c,0);
    redraw_client_frame(c);	// make sure it is redrawn translucent
//    if(c==get_top_most_client(WIN_LAYER_NORMAL,True)) {
//      printf("Raising after shape\n");
//      raise_client(c); }
  } 
} 
#endif /* SHAPE */ 
 
/* 
 * 
 * Main event loop 
 * 
 */ 

void do_event_loop(int sleepus,int trackMouse) 
{ 
int gotev=0, sleeporg=0, cursorpos_gap=4, rcvlen;
char buf[100];
int *v1, *v2;
long last_get_cursorpos_time=0;
XEvent ev;
Client *c;

  sprintf(lastsub,"TRACE1: do_event_loop (sleep,moved: %d,%d)\n",sleepus,trackMouse); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: do_event_loop (sleep,moved: %d,%d)\n",sleepus,trackMouse); 
#endif 
    sleeporg=sleepus;

//    got_replay_pointer=1;
    
    Start_Time_Check(15);	// only want to stay here for 15 ticks at most
    
    Restart_XDarwin();
    
/*XGetInputFocus(dpy, &lastenter, &dummy);*/
 
    if(gX11InFront==0)
      cursorpos_gap = 8;
    else
      cursorpos_gap = 4;
    
    last_get_cursorpos_time=next_get_cursorpos_time-cursorpos_gap;
    
    if(can_send_hackcmds!=0) {
      if(xdarwin_socket) {
        rcvlen = check_listen_get_bytes(xdarwin_socket,buf);
        if(rcvlen==0) {
            printf("Got zero from XDarwin...\n");
            ClearUpSockets();
            exit(1);
        } else if(rcvlen==-2) {
            printf("XDarwin ended session.\n");
            ClearUpSockets();
            exit(1);
        } else if(rcvlen==-1) {
            ;	// nothing in receive buffer
        } else if(rcvlen<0) {
            printf("Got an error from check_listen_get_bytes: %d\n",rcvlen);
            exit(1);
        } else if(rcvlen!=8) {
            printf("Got strange length from XDarwin: %d (%s)\n",rcvlen,buf);
        } else if(rcvlen!=-1) {
            v1 = (int *) &buf[0];
            v2 = (int *) &buf[4];
//            printf("Received: %d %d\n",*v1,*v2);
// for now, send it through to the horrific hack that was used for v0.8 preview...
            ev.xbutton.button=5;
            ev.xbutton.x_root=*v1;
            ev.xbutton.y_root=*v2;
            event_button_release(&ev.xbutton); 
        }
      }
      if(client_list!=NULL)	// no point tracking if there are no clients!
        if(GetCarbonTickCount()>next_get_cursorpos_time) {
          XMoveWindow(dpy,check_win,HACKCMD_MOUSEMOVED_X,HACKCMD_MOUSEMOVED_Y);
          XMoveWindow(dpy,check_win,-1000,-1000);
          last_get_cursorpos_time=GetCarbonTickCount();
          next_get_cursorpos_time=last_get_cursorpos_time+cursorpos_gap;
          //printf("Asked for cursor pos: %d %d\n",last_get_cursorpos_time,next_get_cursorpos_time);
        }
    }
        
//    while(got_replay_pointer!=0) {
//    while(XPending(dpy)!=0) 
//    while(Is_Time_Up()==0 && XPending(dpy)!=0 && got_replay_pointer==1) 
    while(Is_Time_Up()==0 && XPending(dpy)!=0) 
    { 
      
      XNextEvent(dpy, &ev); 

      gotev = 1;

      switch(ev.type) 
      { 
#ifdef KEYBOARD_SUPPORT 
      case KeyPress:
#ifdef SHOW_EVENTS
        printf("Got a key press event\n");
#endif
        event_key_press(&ev.xkey); 
        break; 
#endif /* KEYBOARD */ 
      case ButtonPress:
#ifdef SHOW_EVENTS
        printf("Got a button down event\n");
#endif
        event_button_press(&ev.xbutton); 
        cursorpos_gap=5;
        next_get_cursorpos_time=GetCarbonTickCount()+5;
        break; 
      case ButtonRelease: 
#ifdef SHOW_EVENTS
        printf("Got a button up event\n");
#endif
        event_button_release(&ev.xbutton); 
        break; 
      case MotionNotify: 
#ifdef SHOW_EVENTS
        printf("Got a motion event\n");
#endif
        event_motion_notify(&ev.xmotion); 
        next_get_cursorpos_time=GetCarbonTickCount()+3;
        cursorpos_gap=3;
        break; 
      case LeaveNotify:
#ifdef SHOW_EVENTS
        printf("Got a leave event\n");
#endif
        event_enter_notify(&ev.xcrossing,False);
        break;
      case EnterNotify: 
#ifdef SHOW_EVENTS
        printf("Got an enter event\n");
#endif
        event_enter_notify(&ev.xcrossing,True);
        break; 
      case FocusIn: 
#ifdef SHOW_EVENTS
        printf("Got a focus in event\n");
#endif
        event_focus_in(&ev.xfocus); 
        break; 
      case FocusOut: 
#ifdef SHOW_EVENTS
        printf("Got a focus out event\n");
#endif
        event_focus_out(&ev.xfocus); 
        break; 
      case DestroyNotify: 
#ifdef SHOW_EVENTS
        printf("Got a destroy event\n");
#endif
        event_destroy_notify(&ev.xdestroywindow); 
        break; 
      case UnmapNotify: 
#ifdef SHOW_EVENTS
        printf("Got an unmap event\n");
#endif
        event_unmap_notify(&ev.xunmap); 
        break; 
      case MapRequest: 
#ifdef SHOW_EVENTS
        printf("Got a map request event\n");
#endif
        event_map_request(&ev.xmaprequest); 
        break; 
      case ConfigureRequest: 
#ifdef SHOW_EVENTS
        printf("Got a configure event\n");
#endif
        event_configure_request(&ev.xconfigurerequest); 
        break; 
      case PropertyNotify: 
        event_property_notify(&ev.xproperty); 
        break; 
      case ClientMessage: 
#ifdef SHOW_EVENTS
        printf("Got a client message event\n");
#endif
        event_client_message(&ev.xclient); 
        break; 
      default:
        gotev = 0;
#ifdef SHAPE_SUPPORT 
        if(ev.type==shape_event) {
          if(shape) {
#ifdef SHOW_EVENTS
            printf("Got a shape event\n");
#endif /* SHOW_EVENTS */
            event_shape((XShapeEvent *)&ev); 
            gotev = 1;
#ifdef SHOW_EVENTS
          } else {
            printf("Got a shape event -but not used...\n");
#endif /* SHOW_EVENTS */
          }
#ifdef SHOW_EVENTS
        } else {
        printf("Got an unknown event, type: %d\n",ev.type);
#endif /* SHOW_EVENTS */
        }
#else
#ifdef SHOW_EVENTS
        printf("Got an unknown event, type: %d\n",ev.type);
#endif /* SHOW_EVENTS */
#endif /* SHAPE */ 
        break; 
      }
      
//    if(gX11InFront==0)
      if(can_send_hackcmds!=0 && client_list!=NULL)	// no point tracking if there are no clients!
        if(GetCarbonTickCount()>next_get_cursorpos_time) {
          XMoveWindow(dpy,check_win,HACKCMD_MOUSEMOVED_X,HACKCMD_MOUSEMOVED_Y);
          XMoveWindow(dpy,check_win,-1000,-1000);
          last_get_cursorpos_time=GetCarbonTickCount();
          next_get_cursorpos_time=last_get_cursorpos_time+cursorpos_gap;
          //printf("Asked for cursor pos: %d %d\n",last_get_cursorpos_time,next_get_cursorpos_time);
        }
    }
    
    if(replay_motion!=0)
      XPutBackEvent(dpy, (XEvent *) &replay_motion_event); 
    replay_motion=0;

    /*if(current_resize_client!=NULL) {
        c=current_resize_client;
        if(c->resize_timer>0) {
            if(c->resize_timer>GetCarbonTickCount()) {
                unsigned int mask = (CWHeight|CWWidth|CWX|CWY);
                XWindowChanges wc;
                XConfigureEvent ce; 
                c->resize_timer=0;
                wc.x=LW;
                wc.y=TH;
                wc.width=c->width;
                wc.height=c->height;
                XConfigureWindow(dpy, c->window, mask, &wc);
                ce.type=ConfigureNotify;
                ce.display=dpy;
                ce.event=c->window; 
                ce.window=c->window; 
                ce.x=c->x; 
                ce.y=c->y; 
                ce.width=c->width; 
                ce.height=c->height;
                ce.border_width=0; 
//		  ce.above=c->frame; 
                ce.above=0; 
                ce.override_redirect=0; 
                XSendEvent(dpy, c->window, False, StructureNotifyMask, (XEvent *)&ce);
                //set_client_shape(c,0);
            }
        }
    }*/
    // only need to check mouse-over stuff if the mouse has moved
    if(trackMouse!=0 || replay_if_drag!=0)
      if(cursorpos_gap>2) event_idle();
//      if(got_replay_pointer<2) event_idle();
      
    gotev = 0;

    if(mouse_mode==MOUSE_STOPPED) mouse_mode=MOUSE_NORMAL;
    
  Stop_XDarwin();
    
#ifdef SHOW_WEIRD
  if(sleepus>100000 || sleepus<10) {
    printf("%s oddity: Got very strange value for sleepus: %d  (was %d)\n",PROGNAME,sleepus,sleeporg);
    show_last_fifteen(); }
#endif 
  if(sleepus>100000 || sleepus<10) sleepus=100000;

// Come back very quickly if our time is up, or if the gap till next cursor request is small,
// or if we are dragging (MOUSE_DRAG), resizing (MOUSE_RESIZE) or a window widget is pressed (etc.)
  if(Is_Time_Up()!=0 || cursorpos_gap<3 || mouse_mode!=MOUSE_NORMAL)
    sleepus = 500;

  sprintf(lastsub,"TRACE1: sleeping in do_event_loop (sleep: %d, mouse: %d)\n",sleepus,mouse_mode); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: sleeping in do_event_loop (sleep: %d, mouse: %d)\n",sleepus,mouse_mode); 
#endif

//  if(got_replay_pointer==2) {
//    got_replay_pointer=3;
////    printf("Staying for two seconds...\n");
//    Start_Time_Check(120);	// only want to stay here for 10 ticks at most
//  } else if(got_replay_pointer==3) {
//    if(Is_Time_Up()!=0) {
//      got_replay_pointer=0;
////      printf("Getting out...\n");
//    }
//  } else if(got_replay_pointer==1) {
//    got_replay_pointer=0;
//  }
  
    // oroborus slept for 5 microseconds! -caused far too much CPU usage...
    // This can happily go to several thousand microseconds without noticeable problems
    // However, we do get out a.s.a.p. if there was already considerable time spent in here (10 ticks)
//  if(got_replay_pointer==2)
//    usleep(10000);
//  else if(Is_Time_Up()==0)

    usleep(sleepus);

  
//  }
  
  sprintf(lastsub,"TRACE1: leaving do_event_loop...\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: leaving do_event_loop...\n"); 
#endif 
 
//  Restart_XDarwin();
} 
