/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 

#include "oroborus.h" 
#include "hackcmds.h"

void SendHackTitleToDockCommand(Client *c, int todock)
{
    unsigned char sent_title[120], *title=NULL;
    int i, l, p1, title_cmd;
    
    if(c==NULL) return;
    if(can_send_hackcmds==0) return;

    if(c->name)
        title = c->name;
    else if(c->clientcmd)
        title = c->clientcmd;
    else if(c->clientclass)
        title = c->clientclass;
    
    if(title) {
        l = strlen(title);
#ifdef DEBUG
        printf("Got '%s' as title...\n",title);
#endif
        if(l>99) {
            for(i=0;i<70;i++)
                sent_title[i]=title[i];
            title[70] = 33; // dot character
            title[71] = 33;
            title[72] = 33;
            for(i=73;i<101;i++)
                sent_title[i]=title[i+l-100];
        } else {
            for(i=0;i<=l;i++)
                sent_title[i]=title[i];
        }

#ifdef DEBUG
        printf("Sending '%s' as title...\n",sent_title);
#endif
    
        MyIncreaseXServerGrab();
        XMoveWindow(dpy, c->frame, HACKCMD_STARTX, HACKCMD_STARTY); // start the hack command sequence

        XMoveWindow(dpy, c->frame, HACKCMD_MINIWIN_TITLE, -1); // Tell XDarwin to start a new title string
        // Send the title string
        title_cmd=HACKCMD_MINIWIN_TITLE1;
        for(i=0;i<strlen(sent_title);i++) {
            p1 = sent_title[i];
            XMoveWindow(dpy, c->frame, title_cmd, p1);
            // alternate position, since X server will not send through command if values are the same
            title_cmd = HACKCMD_MINIWIN_TITLE-(title_cmd-HACKCMD_MINIWIN_TITLE);
        }
    
        XMoveWindow(dpy, c->frame, HACKCMD_MINIWIN_TITLE, 0); // Commit the title string
    
    }
    
    if(todock==0)
        XMoveWindow(dpy, c->frame, FX, FY); // The X server thinks the window has moved, so put it back
    else {
        c->is_docked=1;
        XMoveWindow(dpy, c->frame, HACKCMD_DOCKWIN, 1); // Send it to the dock
        XMoveWindow(dpy, c->frame, -FW-10, FY); // The X server thinks the window has moved, so put it back
    }
    
    MyDecreaseXServerGrab();
}

void SendSimpleHackCommand(int cmd, int param)	// uses check_win to send a command not related to a client
{
    if(can_send_hackcmds==0) return;

    MyIncreaseXServerGrab(); // don't want anything else interfering with the X server while this happens
    XMoveWindow(dpy, check_win, HACKCMD_STARTX, HACKCMD_STARTY);
    XMoveWindow(dpy, check_win, cmd, param);
    XMoveWindow(dpy, check_win, -1000, -1000);
    XFlush(dpy);
    MyDecreaseXServerGrab(); // release the X server, if it was before
    
#ifdef DEBUG
    printf("Sent simple hack command: %d %d\n",cmd,param);
#endif
    if(cmd==HACKCMD_INHIBITINFO)
        inhibiting_info = param;
}

void SendHackCommand(Client *c, int nparam, int orgcmd, int p1, int p2, int p3, int p4, int p5)
{
    int params[5], i, cmd;

    if(c==NULL) return;
    if(can_send_hackcmds==0) return;
    
    params[0]=p1;
    params[1]=p2;
    params[2]=p3;
    params[3]=p4;
    params[4]=p5;
    
    if(c->is_docked!=0 && orgcmd!=HACKCMD_DOCKWIN && orgcmd!=HACKCMD_TRANSLUCENT && orgcmd!=HACKCMD_DIMWIN) {
        printf("Entered SendHackCommand for docked client: %d %d\n",orgcmd,p1);
        return;
    }
    
    if(orgcmd==HACKCMD_SETDRAWALPHA) {
        if(c->osxwinfo.drawalpha==p1)
            return;
        else
            c->osxwinfo.drawalpha=p1;
    } else if(orgcmd==HACKCMD_SETID) {
        if(c->hack_id==0) {
            printf("%s warning: got zero hack_id for client %#lx !!\n",PROGNAME,c);
            return;
        }
        params[0] = c->hack_id;
        nparam = 1;
        printf("Sending hack_id = %d to XDarwin...\n",params[0]);
    }
    
    MyIncreaseXServerGrab(); // don't want anything else interfering with the X server while this happens
    XMoveWindow(dpy, c->frame, HACKCMD_STARTX, HACKCMD_STARTY); // start the hack command sequence
    
    cmd = orgcmd;
    
    // Send the command and the arguments
    for(i=0;i<nparam;i++) {
        XMoveWindow(dpy, c->frame, cmd, params[i]);
        cmd--;
    }
    
//    if(orgcmd!=HACKCMD_DOCKWIN || p1==0) // Don't tell X server to put back if it has been docked
        XMoveWindow(dpy, c->frame, FX, FY); // The X server thinks the window has moved, so put it back

    MyDecreaseXServerGrab(); // release the X server, if it was before

    if(orgcmd==HACKCMD_INHIBITINFO) {
        inhibiting_info = p1;
    } else if(orgcmd==HACKCMD_DOCKWIN && p1<2) {
        c->is_docked = p1;
//        if(p1==0) {
//            XMapWindow(dpy,c->window);
//            set_client_shape(c,0); // need to force a complete redraw of the window if it's undocked
//        } else {
//            c->ignore_unmap++;
//            XUnmapWindow(dpy,c->window); // try to avoid getting any other events from this window
//        }
    } else if(orgcmd==HACKCMD_DIMWIN && p1<2) { // p1>=2 allows for temp drawing of frame without being dim
        c->is_dimmed = p1;
//        set_client_shape(c,0); // need to force a complete redraw of the window if it's dimmed/undimmed
        redraw_client_frame(c); // also need to redraw the frame so it is translucent
    }

}
