/*
 * NSWindow subclass for Mac OS X rootless X server
 */
/* $XFree86: xc/programs/Xserver/hw/darwin/bundle/XWindow.m,v 1.5 2001/11/06 23:18:14 torrey Exp $ */

#import "XWindow.h"
#include "quartzCommon.h"

typedef struct {
    XWindow *window;
} AquaWindowRec;

#define WINREC(rw) ((AquaWindowRec *)rw)

@implementation XWindow

// XWindow MUST NOT autodisplay! Autodisplay can cause a deadlock.
// event thread - autodisplay: locks view hierarchy, then window
// X Server thread - window resize: locks window, then view hierarchy
// Deadlock occurs if each thread gets one lock and waits for the other.

// XWindow MUST defer! Otherwise an assertion fails in
// NSViewHierarchyLock sometimes.

- (id)initWithContentRect:(NSRect)aRect
                   isRoot:(BOOL)isRoot
{
    int style;
    NSRect viewRect = {{0, 0}, {aRect.size.width, aRect.size.height}};
//    style = NSBorderlessWindowMask;
    style = NSMiniaturizableWindowMask;
//    style = NSMiniaturizableWindowMask + NSTitledWindowMask;
    
    self = [super initWithContentRect: aRect
                styleMask: style
                backing: NSBackingStoreBuffered
                defer: YES];
    if (! self) return NULL;

    [self setBackgroundColor:[NSColor clearColor]];  // erase transparent
    if(isRoot)
        [self setAlphaValue:0.01];       // draw nearly transparent to get rid of white flash at startup
    else
        [self setAlphaValue:1.0];       // draw opaque
//    [self setOpaque:YES];           // changed when window is shaped
    [self setOpaque:NO];           // always assume window is shaped

    [self useOptimizedDrawing:YES]; // Has no overlapping sub-views
    [self setAutodisplay:NO];       // MUST NOT autodisplay! See comment above
    [self disableFlushWindow];      // We do all the flushing manually
//    [self enableFlushWindow];      // We don't do all the flushing manually
    [self setHasShadow: !isRoot];   // All windows have shadows except root

    [self setAcceptsMouseMovedEvents:NO]; // MUST be AFTER orderFront?

    mIsRoot=isRoot;
    mView = [[XView alloc] initWithFrame:viewRect isRoot:mIsRoot];
//    [[self contentView] addSubview:mView];
    [self setContentView:mView];
    [self setInitialFirstResponder:mView];
    [self setDelegate:self];
    mWrongThread=0;
    mLastTitlePos=0;
    mMyAlpha=1.0;
    [mView setAutoresizingMask: (NSViewWidthSizable | NSViewHeightSizable)];
//    [mView setAutoresizingMask: (NSViewMinYMargin | NSViewMaxXMargin)];
    
    return self;
}

- (void)dealloc
{
    [mView release];
    [super dealloc];
}

- (char *)bits
{
    return [mView bits];
}

- (void)getBits:(char **)bits
       rowBytes:(int *)rowBytes
          depth:(int *)depth
   bitsPerPixel:(int *)bpp
{
    [mView getBits:bits rowBytes:rowBytes depth:depth bitsPerPixel:bpp];
}

- (void)reshapeRects:(fakeBoxRec *)eraseRects count:(int)count
{
    [mView reshapeRects:eraseRects count:count];
}

- (void)setFrame:(NSRect)newframe display:(BOOL)val
{
//    NSLog(@"In setFrame");
    myFrame=newframe;
    [self myDoResize];
    [self mySetFrame:newframe alloc:YES];
//    NSLog(@"Done setFrame");
}

- (void)mySetFrame:(NSRect)newframe alloc:(BOOL)alloc
{
//    NSLog(@"In mySetFrame");
    myFrame=newframe;
    if(alloc) {
        NSSize newSize = {myFrame.size.width, myFrame.size.height};
        [mView checkAndAllocBitsForSize:newSize isRoot:mIsRoot];
    }
//    NSLog(@"Done mySetFrame");
}

- (void)myDoResize
{
//    fakeBoxRec bounds = {0, 0, frame.size.width, frame.size.height};

    NSSize newSize = {myFrame.size.width, myFrame.size.height};
    NSRect aRect = {{0, 0}, {myFrame.size.width, myFrame.size.height}};
    NSRect curFrame;
    
//    NSLog(@"In myDoResize");
    if([mView getMyHackID]) {
//        NSLog(@"HACKID %d : removing tracking rect in XWindow setFrame... %d",[mView getMyHackID],[mView getMyTrackingRect]);   
        [mView removeTrackingRect:[mView getMyTrackingRect]];
    }
//    curFrame=[mView frame];    
//    NSLog(@"before super setFrame @ %d , %d",(int) curFrame.size.width,(int) curFrame.size.height);
//    sleep(3);
//    [mView copyAllToShapeBits];
    [mView setWaitForDrawRect:YES];
    [self setAutodisplay:YES];
    [super setFrame:myFrame display:YES];
//    while([mView getWaitForDrawRect]) {
//        NSLog(@"Waiting for drawrect in setFrame");
//        usleep(10000);
//    }
    curFrame=[mView frame];    
//    NSLog(@"after super setFrame @ %d , %d",(int) curFrame.size.width,(int) curFrame.size.height);
//    sleep(3);
    [self setAutodisplay:NO];
    if([mView getMyHackID]) {
        [mView setMyTrackingRect:[mView addTrackingRect:aRect owner:mView userData:nil assumeInside:NO]];
//        NSLog(@"HACKID %d : changed tracking rect in XWindow setFrame... %d",[mView getMyHackID],[mView getMyTrackingRect]);
    }
    [mView setWaitForDrawRect:NO];
//    NSLog(@"Done myDoResize");
}

// rects are X-flip and LOCAL coords
- (void)refreshRects:(fakeBoxRec *)rectList count:(int)count;
{
    //if( [mView getMyMiniature] ) return;
    [mView refreshRects:rectList count:count];
}

- (BOOL)isFlipped
{
    return NO;
}

// Deferred windows don't handle mouse moved events very well.
- (void)orderWindow:(NSWindowOrderingMode)place
         relativeTo:(int)otherWindowNumber
{
//    if( mWrongThread ) [self checkThreadFlag];
//    if( [mView getMyMiniature] && mWrongThread!=2) {
//        NSLog(@"HACKID %d : Got orderWindow for miniaturised window",[mView getMyHackID]);
//        return;
//    }
    [super orderWindow:place relativeTo:otherWindowNumber];
//    [self setAcceptsMouseMovedEvents:YES];
}

- (void)sendEvent:(NSEvent *)anEvent
{
    if( mWrongThread ) [self checkThreadFlag];
    if( [mView getMyMiniature] && mWrongThread!=2) {
        NSLog(@"HACKID %d : Got sendEvent for miniaturised window",[mView getMyHackID]);
        return;
    }
    [super sendEvent:anEvent];
//    [self setAcceptsMouseMovedEvents:YES];
}

// XWindow may be frameless, and frameless windows default to
// NO key and NO main.
// update: we *don't* want main or key status after all
- (BOOL)canBecomeMainWindow
{
    return NO;
}

- (BOOL)canBecomeKeyWindow
{
    return NO;
}

- (void)setDim:(BOOL)value update:(BOOL)update
{
    if( [mView getMyMiniature] ) return;
    [mView setDim:value update:update];
}

- (BOOL)shadowChangeInhibited
{
    return [mView shadowChangeInhibited];
}

- (void)setInhibitShadowChange:(BOOL)value
{
    [mView setInhibitShadowChange:value];
}

- (BOOL)getMyMiniature
{
    return [mView getMyMiniature];
}

- (void)setMiniaturization:(BOOL)value
{
    if(value) {
        //[self setBackgroundColor:[NSColor blackColor]];
        [mView setMyMiniature:YES];
//        [self setAcceptsMouseMovedEvents:YES];
        NSLog(@"HACKID %d : No miniaturizing from wrong thread...",[mView getMyHackID]);
//        [self miniaturize:nil];
//        [self setAutodisplay:YES];
        mWrongThread=1;
    } else if( [self isMiniaturized] ) {
//        [self setAcceptsMouseMovedEvents:YES];
        NSLog(@"HACKID %d : No deminiaturizing from wrong thread...",[mView getMyHackID]);
        mWrongThread=2;
    } else {
        mWrongThread=0;
        [mView setMyMiniature:NO];
    }
}

- (void)setMyMiniature:(BOOL)value
{
    [mView setMyMiniature:value];
}

- (void)checkThreadFlag
{
    if(mWrongThread==1) {
        NSLog(@"HACKID %d : Miniaturizing from right thread...?",[mView getMyHackID]);
//        [self setAcceptsMouseMovedEvents:YES];
//        [self setAutodisplay:YES];
//        [self enableFlushWindow];      // We don't do all the flushing manually
        [self miniaturize:nil];
    } else if(mWrongThread==2) {
        NSLog(@"HACKID %d : Deminiaturizing from right thread...?",[mView getMyHackID]);
//        [self setAcceptsMouseMovedEvents:YES];
//        [self refreshWholeWindow];
        [self deminiaturize:nil];
    }
    mWrongThread=0;
}

- (int)getThreadFlag
{
    return mWrongThread;
}

- (void)setThreadFlag:(int)value
{
    mWrongThread=value;
}

- (void) refreshWholeWindow
{
    fakeBoxRec bounds = {0, 0, [mView frame].size.width,
                         [mView frame].size.height};

//    [mView display];
    [mView refreshRects:&bounds count:1];
}

- (void)windowDidDeminiaturize:(NSNotification *)notification
{
    [self setMyMiniature:NO];
    xwinHackID[hackIDPos] = [mView getMyHackID];
    //NSLog(@"HACKID %d : <<<<<<<< Sending undock after demin notification\n",xwinHackID[hackIDPos]);
    QuartzMessageMainThread(kQuartzWinDidUndock+hackIDPos);
    hackIDPos++;
    if(hackIDPos>=kDockPosMax) hackIDPos=0;
}

- (void)windowWillMiniaturize:(NSNotification *)notification
{
    //NSLog(@"HACKID %d : ***** Got will miniaturize notification...",[mView getMyHackID]);
}

- (void)setMyDrawAlpha:(int)alphaval
{
    [mView setMyDrawAlpha:alphaval];
}

- (void)setMyTranslucentAlpha:(int)alphaval
{
    [mView setMyTranslucentAlpha:alphaval];
}

- (void)setMyTranslucentRed:(int)redval
{
    [mView setMyTranslucentRed:redval];
}

- (void)setMyTranslucentGreen:(int)greenval
{
    [mView setMyTranslucentGreen:greenval];
}

- (void)setMyTranslucentBlue:(int)blueval
{
    [mView setMyTranslucentBlue:blueval];
}

- (void)setMyHackID:(int)hackid
{
    [mView setMyHackID:hackid];
}

- (int)getMyHackID;
{
    return [mView getMyHackID];
}

- (BOOL)willHitWindow:(NSPoint)aPt
{
    NSView	*aView;
    
    aView=[mView hitTest:aPt];
    if(!aView) return NO;
    if(aView!=mView) return NO;
    return YES;
}

- (void)setMyTitlePart:(int)v
{
    if(v==-1) {
        mLastTitlePos = 0;
        //NSLog(@"starting new miniwin title");
    } else if(v==0) {
        mTitle[mLastTitlePos] = 0;
        //NSLog(@"committing miniwin title: %s",mTitle);
        [self setMiniwindowTitle:[NSString stringWithCString:mTitle]];
        mLastTitlePos = 0;
    } else {
        //NSLog(@"setting miniwin title at %d: %#lx",mLastTitlePos,v);
        mTitle[mLastTitlePos] = v & 0xff;
        mLastTitlePos++;
    }
}

- (void)setMyAlphaValue:(float)val
{
    mMyAlpha=val;
    [self setAlphaValue:val];
}

- (void)resetAlphaValue
{
    [self setAlphaValue:mMyAlpha];
}

//- (void)mySetHasShadow:(BOOL)value
//{
//    if(value) {
//        [self makeKeyWindow];
//        [self makeMainWindow];
//    }
//    [mView lockFocus];
//    [self setHasShadow:value];
//    [self setHasShadow:!value];
//    [self setHasShadow:value];
//    [[NSGraphicsContext currentContext] flushGraphics];
//    [mView unlockFocus];
//}

@end
