/*
 * Rootless implementation for Mac OS X Aqua environment
 */
/* $XFree86: xc/programs/Xserver/hw/darwin/bundle/rootlessAquaImp.m,v 1.11 2002/01/17 02:44:27 torrey Exp $ */

#include "rootlessAquaImp.h"
#include "XWindow.h"
#include "fakeBoxRec.h"
#include "quartzCommon.h"
#include "pseudoramiX.h"

extern void ErrorF(const char *, ...);

typedef struct {
    XWindow *window;
} AquaWindowRec;

static int	hitcheckptx;
int		can_come_in_front=3;	// start off with the default XDarwin behaviour

#define WINREC(rw) ((AquaWindowRec *)rw)


// Multihead note: When rootless mode uses PseudoramiX, the
// X server only sees one screen; only PseudoramiX itself knows
// about all of the screens.

int AquaDisplayCount()
{
    aquaNumScreens = [[NSScreen screens] count];

    if (noPseudoramiXExtension) {
        return aquaNumScreens;
    } else {
        return 1; // only PseudoramiX knows about the rest
    }
}

void AquaScreenInit(int index, int *x, int *y, int *width, int *height,
                    int *rowBytes, unsigned long *bps, unsigned long *spp,
                    int *bpp)
{
    *bps = 8;
    *spp = 3;
    *bpp = 32;

    if (noPseudoramiXExtension) {
        NSScreen *screen = [[NSScreen screens] objectAtIndex:index];
        NSRect frame = [screen frame];

        // set x, y so (0,0) is top left of main screen
        *x = NSMinX(frame);
        *y = NSHeight([[NSScreen mainScreen] frame]) - NSHeight(frame) -
            NSMinY(frame);

        *width =  NSWidth(frame);
        *height = NSHeight(frame);
        *rowBytes = (*width) * (*bpp) / 8;

        // Shift the usable part of main screen down to avoid the menu bar.
        if (NSEqualRects(frame, [[NSScreen mainScreen] frame])) {
            *y      += aquaMenuBarHeight;
            *height -= aquaMenuBarHeight;
        }

    } else {
        int i;
        NSRect unionRect = NSMakeRect(0, 0, 0, 0);
        NSArray *screens = [NSScreen screens];

        // Get the union of all screens (minus the menu bar on main screen)
        for (i = 0; i < [screens count]; i++) {
            NSScreen *screen = [screens objectAtIndex:i];
            NSRect frame = [screen frame];
            frame.origin.y = [[NSScreen mainScreen] frame].size.height -
                             frame.size.height - frame.origin.y;
            if (NSEqualRects([screen frame], [[NSScreen mainScreen] frame])) {
                frame.origin.y    += aquaMenuBarHeight;
                frame.size.height -= aquaMenuBarHeight;
            }
            unionRect = NSUnionRect(unionRect, frame);
        }

        // Use unionRect as the screen size for the X server.
        *x = unionRect.origin.x;
        *y = unionRect.origin.y;
        *width = unionRect.size.width;
        *height = unionRect.size.height;
        *rowBytes = (*width) * (*bpp) / 8;

        // Tell PseudoramiX about the real screens.
        // InitOutput() will move the big screen to (0,0),
        // so compensate for that here.
        for (i = 0; i < [screens count]; i++) {
            NSScreen *screen = [screens objectAtIndex:i];
            NSRect frame = [screen frame];
            int j;

            // Skip this screen if it's a mirrored copy of an earlier screen.
            for (j = 0; j < i; j++) {
                if (NSEqualRects(frame, [[screens objectAtIndex:j] frame])) {
                    ErrorF("PseudoramiX screen %d is a mirror of screen %d.\n",
                           i, j);
                    break;
                }
            }
            if (j < i) continue; // this screen is a mirrored copy

            frame.origin.y = [[NSScreen mainScreen] frame].size.height -
                             frame.size.height - frame.origin.y;

            if (NSEqualRects([screen frame], [[NSScreen mainScreen] frame])) {
                frame.origin.y    += aquaMenuBarHeight;
                frame.size.height -= aquaMenuBarHeight;
            }

            ErrorF("PseudoramiX screen %d added: %dx%d @ (%d,%d).\n", i,
                   (int)frame.size.width, (int)frame.size.height,
                   (int)frame.origin.x, (int)frame.origin.y);

            frame.origin.x -= unionRect.origin.x;
            frame.origin.y -= unionRect.origin.y;

            ErrorF("PseudoramiX screen %d placed at X11 coordinate (%d,%d).\n",
                   i, (int)frame.origin.x, (int)frame.origin.y);

            PseudoramiXAddScreen(frame.origin.x, frame.origin.y,
                                 frame.size.width, frame.size.height);
        }
    }
}

void *AquaNewWindow(void *upperw, int x, int y, int w, int h, int isRoot)
{
    AquaWindowRec *winRec = (AquaWindowRec *)malloc(sizeof(AquaWindowRec));
    NSRect frame = NSMakeRect(x, NSHeight([[NSScreen mainScreen] frame]) -
                              y - h, w, h);

    winRec->window = [[XWindow alloc] initWithContentRect:frame isRoot:isRoot];

    if (upperw) {
        AquaWindowRec *upperRec = WINREC(upperw);
        int uppernum = [upperRec->window windowNumber];
        [winRec->window orderWindow:NSWindowBelow relativeTo:uppernum];
//        [winRec->window setHasShadow:NO];
    } else {
        [winRec->window orderFront:nil];
//        [winRec->window setHasShadow:YES];
    }

    // fixme hide root for now
    if (isRoot) [winRec->window orderOut:nil];

    return winRec;
}

void AquaDestroyWindow(void *rw)
{
    AquaWindowRec *winRec = WINREC(rw);

    if( [winRec->window getMyMiniature] ) {
        ErrorF("HACKID %d : Got AquaDestroyWindow for miniaturized window\n",[winRec->window getMyHackID]);
//        return;
    }
    
    [winRec->window release];
}

void AquaHackCommand(void *rw, int x, int y)
{
    AquaWindowRec *winRec = WINREC(rw);
    float z;

//    NSLog(@"Got hack command: %d %d",x,y);
    if(x==-23400) {
        z=y;
        z=z*0.01;
        [winRec->window setMyAlphaValue:z];
    } else if(x==-23401) {
        if(y==1) {
            xwinToDock[toDockPos]=winRec;
            //ErrorF("<<<<<<<< Set xwinToDock at: %#lx\n",winRec);
            [winRec->window setMiniaturization:YES];
            QuartzMessageMainThread(kQuartzWinToDock+toDockPos);
            toDockPos++;
            if(toDockPos>=kDockPosMax) toDockPos=0;
        } else if(y==0) {
            [winRec->window setMyMiniature:NO];
        } else {
            xwinToUndock[toUndockPos]=winRec;
            //ErrorF("<<<<<<<< Set xwinToUndock at: %#lx\n",winRec);
            [winRec->window setMiniaturization:NO];
            QuartzMessageMainThread(kQuartzWinToUndock+toUndockPos);
            toUndockPos++;
            if(toUndockPos>=kDockPosMax) toUndockPos=0;
        }
    } else if(x==-23402) {
        if(y==0)
            [winRec->window setDim:NO update:YES];
        else if(y==1)
            [winRec->window setDim:YES update:YES];
        else if(y==2)
            [winRec->window setDim:NO update:NO];
        else
            [winRec->window setDim:YES update:NO];
        // want to prevent annoying flicker when shadow changes (hmmm... doesn't work...)
        // since OroborOSX will immediately call set_client_shape right after this hack command,
        // and since the shape will not change, there's no need to change the shadow.
        if(y<2)
            [winRec->window setInhibitShadowChange:YES];
    } else if(x==-23403) {
        if (! [NSApp isActive])
            [NSApp activateIgnoringOtherApps:YES]; // make sure XDarwin is frontmost app
        [NSApp arrangeInFront:nil]; // make sure all windows come in front
/*    } else if(x==-23404) {
        if(y==0)
            [winRec->window mySetHasShadow:NO];
        else
            [winRec->window mySetHasShadow:YES];
*/
    } else if(x==-23405) {
        [winRec->window setMyDrawAlpha:y];
    } else if(x==-23406) {
        [winRec->window setMyTranslucentAlpha:y];
    } else if(x==-23407) {
        [winRec->window setMyTranslucentRed:y];
    } else if(x==-23408) {
        [winRec->window setMyTranslucentGreen:y];
    } else if(x==-23409) {
        [winRec->window setMyTranslucentBlue:y];
    } else if(x==-23410) {
        [NSApp myHide];
    } else if(x==-23411) {
        hitcheckptx = y;
        ErrorF("======== got X: %d ",hitcheckptx);
    } else if(x==-23412) {
        NSPoint hitCheckPt = NSMakePoint(hitcheckptx,NSHeight([[NSScreen mainScreen] frame]) - y);
        NSPoint aPt = [winRec->window convertScreenToBase:hitCheckPt];
        ErrorF(", got Y: %d \n",y);
        if([winRec->window willHitWindow:aPt])
            ErrorF("+++++++++++++++++++++ %d,%d will hit window\n",(int)(aPt.x),(int)(aPt.y));
        else
            ErrorF("--------------------- %d,%d will not hit window\n",(int)(aPt.x),(int)(aPt.y));
    } else if(x==-23413) {
        [winRec->window setMyTitlePart:y];
    } else if(x==-23414) {
        [winRec->window setMyTitlePart:y];
    } else if(x==-23415) {
        [winRec->window setMyTitlePart:y];
    } else if(x==-23416) {
        oroborosx_info_inhibit=y;
    } else if(x==-23417) {
        //ErrorF("Setting can_come_in_front to %d\n",y);
        can_come_in_front=y;
    } else if(x==-23418) {
        [NSApp deactivate];
    } else if(x==-23450) {
        if(y==0)
            [[winRec->window contentView] setInhibitFull:NO];
        else
            [[winRec->window contentView] setInhibitFull:YES];
    } else if(x==-23449) {
        if(y==0)
            [[winRec->window contentView] setInhibitRefresh:NO];
        else
            [[winRec->window contentView] setInhibitRefresh:YES];
    } else if(x==-23490) {
        int dummy=0;
        if(y==0) {
            if(oroborosx_socket) {
                NSLog(@"Closing OroborOSX socket: %d",oroborosx_socket);
                dummy = close(oroborosx_socket);
                oroborosx_socket=0;
            }
        } else {
            if(oroborosx_socket!=0) {
                NSLog(@"Closing OroborOSX socket and reconnecting... (%d)",oroborosx_socket);
                dummy = close(oroborosx_socket);
                oroborosx_socket=0;
            }
            oroborosx_socket = init_socket_connect("/tmp/.OroborOSX_main_socket","XDarwin");
            if( oroborosx_socket<=0 ) {
                NSLog(@"Failed to connect to OroborOSX socket: %d",oroborosx_socket);
                oroborosx_socket=0;
            } else
                NSLog(@"Created OroborOSX socket: %d",oroborosx_socket);
        }
    } else if(x==-23495) {
        if(y==0) {
            oroborosx_info_inhibit=1;
            can_come_in_front=-1;
            NSLog(@"OroborOSX told me to quit...");
            QuartzMessageMainThread(kQuartzOroborOSXQuitting);
        } else {
            NSLog(@"OroborOSX told me it's quitting...");
            oroborosx_info_inhibit=1;
            can_come_in_front=3;
        }
    } else if(x==-23498) {
        showing_prefs = y;
    } else if(x==-23499) {
        [winRec->window setMyHackID:y];
        NSLog(@"======== set new window HackID to %d",y);
    }

}

void AquaMoveWindow(void *rw, int x, int y)
{
    AquaWindowRec *winRec = WINREC(rw);
    NSPoint topLeft = NSMakePoint(x, NSHeight([[NSScreen mainScreen] frame]) -
                                  y);

    if( [winRec->window getMyMiniature] ) {
        ErrorF("HACKID %d : Ignoring AquaMoveWindow for miniaturized window\n",[winRec->window getMyHackID]);
        return;
    }
    
    [winRec->window setFrameTopLeftPoint:topLeft];
}

void AquaStartResizeWindow(void *rw, int x, int y, int w, int h)
{
    AquaWindowRec *winRec = WINREC(rw);
    NSRect frame = NSMakeRect(x, NSHeight([[NSScreen mainScreen] frame]) -
                              y - h, w, h);

    if( [winRec->window getMyMiniature] ) {
        ErrorF("HACKID %d : Ignoring AquaStartResizeWindow for miniaturized window\n",[winRec->window getMyHackID]);
        return;
    }

    [[winRec->window contentView] setWaitForDrawRect:YES];

    [winRec->window mySetFrame:frame alloc:NO];

            xwinToDock[toDockPos]=winRec;
            //ErrorF("<<<<<<<< Set xwinToSize at: %#lx\n",winRec);
            QuartzMessageMainThread(kQuartzWinToSize+toDockPos);
            toDockPos++;
            if(toDockPos>=kDockPosMax) toDockPos=0;

    while([[winRec->window contentView] getWaitForDrawRect]) {
        //NSLog(@"Waiting for drawrect in AquaStartResizeWindow");
        usleep(10000);
    }
    [winRec->window mySetFrame:frame alloc:YES];
}

void AquaFinishResizeWindow(void *rw, int x, int y, int w, int h)
{
    // refresh everything? fixme yes for testing
    fakeBoxRec box = {0, 0, w, h};
    AquaWindowRec *winRec = WINREC(rw);

//    if( [winRec->window getMyMiniature] ) {
//        ErrorF("HACKID %d : Ignoring AquaFinishResizeWindow for miniaturized window\n",[winRec->window getMyHackID]);
//        return;
//    }

//    [winRec->window refreshRects:&box count:1];
}

void AquaUpdateRects(void *rw, fakeBoxRec *rects, int count)
{
    AquaWindowRec *winRec = WINREC(rw);

    if( [winRec->window getMyMiniature] ) {
        ErrorF("HACKID %d : Got AquaUpdateRects for miniaturized window\n",[winRec->window getMyHackID]);
//        return;
    }

    [winRec->window refreshRects:rects count:count];
}

// fixme is this upperw or lowerw?
void AquaRestackWindow(void *rw, void *upperw)
{
    AquaWindowRec *winRec = WINREC(rw);

    if( [winRec->window getMyMiniature] ) {
        ErrorF("HACKID %d : Ignoring AquaRestackWindow for miniaturized window\n",[winRec->window getMyHackID]);
        return;
    }
    
    //ErrorF("In AquaRestackWindow\n");
    if (upperw) {
        AquaWindowRec *upperRec = WINREC(upperw);
        int uppernum = [upperRec->window windowNumber];
        [winRec->window orderWindow:NSWindowBelow relativeTo:uppernum];
//        [winRec->window setHasShadow:NO];
    } else {
//        [winRec->window setHasShadow:YES];
        [winRec->window orderFront:nil];
    }
    // [winRec->window setAcceptsMouseMovedEvents:YES];
    // fixme prefer to orderFront whenever possible - pass upperw, not lowerw
}

// rects are the areas not part of the new shape
void AquaReshapeWindow(void *rw, fakeBoxRec *rects, int count)
{
    AquaWindowRec *winRec = WINREC(rw);

    if( [winRec->window getMyMiniature] ) {
        ErrorF("HACKID %d : Got AquaReshapeWindow for miniaturized window\n",[winRec->window getMyHackID]);
        //return;
    }

    // make transparent if window is now shaped
    // transparent windows never go back to opaque
    if (count > 0) {
        [winRec->window setOpaque:NO];
    }
    [winRec->window reshapeRects:rects count:count];
//    [[winRec->window contentView] reshapeRects:rects count:count];

    if ( [winRec->window shadowChangeInhibited] ) {
        ErrorF("Shadow inhibit is on during a reshape...\n");
        [winRec->window setInhibitShadowChange:NO];
    } else if (! [winRec->window isOpaque]) {
        // force update of window shadow
//        if( [winRec->window hasShadow] ) {
        [winRec->window setHasShadow:NO];
        [winRec->window setHasShadow:YES];
//        } else {
//          [winRec->window mySetHasShadow:YES];
//          [winRec->window mySetHasShadow:NO];
//        }
//        [[NSGraphicsContext currentContext] flushGraphics];
    }
}

void AquaGetPixmap(void *rw, char **bits,
                   int *rowBytes, int *depth, int *bpp)
{
    AquaWindowRec *winRec = WINREC(rw);

    [winRec->window getBits:bits rowBytes:rowBytes depth:depth
                    bitsPerPixel:bpp];
//    ErrorF("HackID %d asked for pixmap bits\n",[winRec->window getMyHackID]);
}
