/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Jan 2002       */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h" 
#include "hackcmds.h"

// number of steps for iconbars -now defined in common.h
//#define ICONSTEPS_FAST 4
//#define ICONSTEPS_MED 7
//#define ICONSTEPS_SLOW 20

// not currently used -would be for hiding/showing of iconbars, but it's too slow...
#define SLIDESTEP 45

// Initial value for number of iconbar steps -use the medium one (really needs to read a pref file)
int iconbar_steps=-1, icontab_width=5, iconbar_width=100; // iconbar_steps now read from prefs
//int iconbar_steps=ICONSTEPS_MED, icontab_width=5, iconbar_width=100;
Client *focus_before_hide;

/* 
 * 
 * Handle different window gravities 
 * 
 */ 

void gravitate_client(Client *c, int grav) 
{ 
  int dx=0, dy=0, orgicon; 
 
  sprintf(lastsub,"TRACE3: gravitate_client (client: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: gravitate_client (client: %#lx)\n",c); 
#endif 
 
  // temporarily switch off iconbar flag, so defines are correct
  orgicon=c->is_icon;
  c->is_icon=NOT_ICONBAR;
  
  switch(GRAVITY) 
  { 
  case NorthWestGravity: 
    dx=LW; 
    dy=TH; 
    break; 
  case NorthGravity: 
    dx=(LW+RW)/2; 
    dy=TH; 
    break; 
  case NorthEastGravity: 
    dx=-RW; 
    dy=TH; 
    break; 
  case WestGravity: 
    dx = LW; 
    dy=(TH+BH)/2; 
    break; 
  case EastGravity: 
    dx=-RW; 
    dy=(TH+BH)/2; 
    break;  
  case SouthWestGravity: 
    dx=LW; 
    dy=-BH; 
    break; 
  case SouthGravity: 
    dx=(LW+RW)/2; 
    dy=-BH; 
    break; 
  case SouthEastGravity: 
    dx=-RW; 
    dy=-BH; 
    break; 
  case CenterGravity: 
    dx=(LW+RW)/2; 
    dy=(TH+BH)/2; 
    break; 
  } 
  c->x=c->x+(dx*grav); 
  c->y=c->y+(dy*grav);
  
  // switch back to original iconbar state
  c->is_icon=orgicon;
} 

/*
 * Remove this client from parent's sibling list
 */

void remove_sibling_client(Client *c)
{
  Client *cp, *cnxt, *cprv;
  
  sprintf(lastsub,"TRACE3: remove_sibling_client (client: %#lx)\n",c);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: remove_sibling_client (client: %#lx)\n",c);
#endif

  cp=c->parent;
  if(!cp) return;
  
  cnxt=cp->first_child;

/* This should never really happen! */
  if(cnxt==NULL) {
#ifdef SHOW_WEIRD
    fprintf(stderr,"%s warning: parent lost children!\n",PROGNAME);
#endif
    cp->child_count=0;
    return;
  }
  
/* This should never really happen! */
  if(cp->child_count!=0)
    cp->child_count--;
#ifdef SHOW_WEIRD
  else
    fprintf(stderr,"%s warning: parent has inconsistent child_count!\n",PROGNAME);
#endif

  if(cnxt==c)
    cp->first_child=c->next_sibling;
    
  cnxt=c->next_sibling;
  cprv=c->prev_sibling;
  
  if(cnxt)
    cnxt->prev_sibling=cprv;
  if(cprv)
    cprv->next_sibling=cnxt;
}

/*
 * Add this client to parent's sibling list
 */

void add_sibling_client(Client *c)
{
  Client *cp, *cnxt;
  
  sprintf(lastsub,"TRACE3: add_sibling_client (client: %#lx)\n",c);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: add_sibling_client (client: %#lx)\n",c);
#endif

  cp=c->parent;
  if(!cp) return;
  
  cnxt=cp->first_child;
  
  if(cnxt) {
    while(cnxt->next_sibling) {
      cnxt=cnxt->next_sibling;
    }
    cnxt->next_sibling=c;
    c->prev_sibling=cnxt;
    c->next_sibling=NULL;
  } else {
    cp->first_child=c;
    c->next_sibling=NULL;
    c->prev_sibling=NULL;
  }
  cp->child_count++;
}

/* 
 * 
 * Create a new frame part 
 * 
 */ 

Window create_frame_part(Window win, int x, int y, int w, int h, int mapit) 
{ 
  Window new;   
 
  sprintf(lastsub,"TRACE3: create_frame_part %#lx (%#lx)\n",win,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: create_frame_part %#lx (%#lx)\n",win,lastfocus); 
#endif 
  
  if(w<1 || h<1) return 0;
  
  new=XCreateSimpleWindow(dpy, win, x, y, w, h, 0, 0, 0);

  XSelectInput(dpy, new, ButtonPressMask|ButtonReleaseMask); 
  if( mapit ) XMapWindow(dpy, new); 
  return new; 
} 

void reorder_dockicon_winlist(Client *c)
{
  long i;
  int classlen;
  Client *ac;
  
  sprintf(lastsub,"TRACE2: reorder_dockicon_winlist %#lx (%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: reorder_dockicon_winlist %#lx (%#lx)\n",c,lastfocus); 
#endif

  if(c->osxwinfo.winref==0) return;
  if(c->clientgroup==NULL) return;
  
  classlen=strlen(c->clientgroup);
// Loop through the clients, looking for matches with class of latest client
  ac=client_list;
  for(i=0;i<client_count;i++) {
    if(ac->osxwinfo.winref!=0)
      if(ac->clientgroup)
        if(strlen(ac->clientgroup)==classlen)
          if(!strncmp(ac->clientgroup,c->clientgroup,classlen))
            ReorderOSXWin(&(ac->osxwinfo));
    ac=ac->next;
  }
}

int still_in_auto_client()
{
  Client *c;
  Window ptrwin;
  
  c = lastfocus_client;
  
  if(lastfocus_client==NULL) return 0;
  if((lastfocus_client->auto_properties & AUTO_FOCUS)==0) return 0;
  
  ptrwin=get_pointer_win(root);
  
  if(lastfocus_client->frame!=ptrwin) return 0;
  
  printf("Still in autofocus client...\n");
  return 1;
}

int create_new_hackid()
{
    Client *ac;
    int i, maxpos, minid, maxid=0, gotanid=0, tryid=0;
    
    maxpos = XDisplayWidth(dpy, screen)-1;
    minid = maxpos;
    
    ac = client_list;
    if(ac==NULL) return maxpos;
    
    for(i=0;i<client_count;i++) {
        if(ac->hack_id > maxid) maxid = ac->hack_id;
        if(ac->hack_id < minid) minid = ac->hack_id;
        ac=ac->next;
    }
    
    if(minid>1) return minid-1;
    if(maxid<maxpos) return maxid+1;
    
    tryid = minid;
    while(gotanid==0) {
        tryid++;
        if(tryid>maxpos) return 0;
        ac = client_list;
        for(i=0;i<client_count;i++) {
            if(ac->hack_id==tryid) break;
            ac=ac->next;
        }
        if(ac->hack_id!=tryid) gotanid=tryid;
    }
    
    return gotanid;
}

/* 
 * 
 * Create a new client frame 
 * 
 */ 

void create_new_client(Window w, int reposition) 
{ 
  Client *c=NULL; 
  XWindowAttributes attr; 
  XSetWindowAttributes wattr; 
  PropMwmHints *mhints=NULL; 
  XWMHints *hints=NULL; 
  long dummy; 
  XID tmpwin;
  XWindowChanges wc; 
  int x, y, xmax, ymax, xmin, ymin, head=0, showfront=0; 
  XClassHint xch;
 
  sprintf(lastsub,"TRACE1: create_new_client %#lx (%#lx)\n",w,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: create_new_client %#lx (%#lx)\n",w,lastfocus); 
#endif 
 
  if((c=malloc(sizeof *c))==NULL)
  {
    if(on_cmd_line)
        fprintf(stderr, "%s: Unable to allocate memory for new client.\n", PROGNAME);
    ClearUpSockets();
    exit(1);
  }
 
  /* Initialize client values */ 
  c->window=w; 
  c->ignore_unmap=0; 
  c->zoom_state=BUTTON_UNPRESSED; 
  c->icon_state=BUTTON_UNPRESSED;
  c->close_state=BUTTON_UNPRESSED; 
  c->can_max=True; 
  c->has_frame=True; 
  c->has_focus=False; 
// assume it isn't an iconbar
  c->is_icon=NOT_ICONBAR;
// start with some "bad" values...
  c->name_x=-1;
  c->name_y=-1;
  c->name_w=-1;
  c->name_h=-1;
  c->icon_x=-10000;
  c->icon_y=-10000;
  c->icon_h=-1;
  c->icon_w=-1;
  c->icontxt_x=-1;
  c->icontxt_y=-1;
  c->icontxt_w=-1;
  c->icontxt_h=-1;
// default to a modeless window
  c->modal_state=MODELESS;
  c->sloppy_family=FOCUS_CLICK;
  c->auto_properties=AUTO_NONE;
// set up initial sibling/child list
  c->child_count=0;
  c->parent=NULL;
  c->first_child=NULL;
  c->next_sibling=NULL;
  c->prev_sibling=NULL;
// initialise the OS X window identifier
  c->osxwinfo.winref=0;
  c->osxwinfo.view_state=VIEWSTATE_VISIBLE;
// Start the string info pointing to nothing
  c->clientcmd=NULL;
  c->clientclass=NULL;
  c->clientgroup=NULL;
  c->is_shaped=0;
  c->osxwinfo.translucency=100;
  c->osxwinfo.inactive_translucency=85;
  c->is_dimmed=0;
  c->is_docked=0;
  c->is_e_m_r=0;
  c->dock_timer=0;
  c->undock_timer=0;
  c->osxwinfo.drawalpha=255;
  c->hack_id=create_new_hackid();
  
  MyIncreaseXServerGrab();
  /* Set event mask for client */ 
  XSelectInput(dpy, c->window, PropertyChangeMask|FocusChangeMask); 
 
  if(gQuitFlag!=0) {
    MyDecreaseXServerGrab();
    return;
  }
  
#ifdef I18N_SUPPORT 
  if(use_fontset)
    /* call wrapper function for I18N */ 
    XFetchNameWrapper(dpy, c->window, &c->name); 
  else
    XFetchName(dpy, c->window, &c->name);
#else 
  XFetchName(dpy, c->window, &c->name); 
#endif /* I18N */ 

  if(gQuitFlag!=0) {
    MyDecreaseXServerGrab();
    return;
  }

  c->size=XAllocSizeHints(); 
  XGetWMNormalHints(dpy, w, c->size, &dummy); 
 
  if(gQuitFlag!=0) {
    MyDecreaseXServerGrab();
    return;
  }

  /* Get size and position of client */ 
  XGetWindowAttributes(dpy, c->window, &attr); 
   
  /* Handle clients with border set > 0 and adjust position accordingly */ 
  if(attr.border_width>0) 
  { 
    switch(GRAVITY) 
    { 
    case SouthEastGravity: 
      attr.x=attr.x+attr.border_width*2; 
      attr.y=attr.y+attr.border_width*2; 
      break; 
    case SouthWestGravity: 
    case SouthGravity: 
      attr.y=attr.y+attr.border_width*2; 
      break; 
    case NorthEastGravity: 
    case EastGravity: 
      attr.x=attr.x+attr.border_width*2; 
      break; 
    } 
  } 
  c->x=attr.x; 
  c->y=attr.y; 
  c->width=attr.width; 
  c->height=attr.height; 
 
#ifdef SHOW_HINTS
  printf("==== NEW WINDOW: %#lx      hints...\n",c->window);
#endif
  head=XGetClassHint(dpy,c->window,&xch);
  if(head!=0) {
#ifdef SHOW_HINTS
    printf("  res_name: %s\n",xch.res_name);
    printf("  res_class: %s\n",xch.res_class);
#endif
    c->clientcmd=strdup(xch.res_name);
    c->clientclass=strdup(xch.res_class);
    c->clientgroup=strdup(xch.res_class);
    if(xch.res_name) XFree(xch.res_name);
    if(xch.res_class) XFree(xch.res_class);
  }
  
  hints=XGetWMHints(dpy, c->window); 
  if(hints) 
  { 
#ifdef SHOW_HINTS
    printf("flags: %#lx\n",hints->flags);
    printf("input: %d\n",hints->input);
    printf("XID  : %#lx\n",hints->window_group);
    printf("state: %#lx\n",hints->initial_state);
#endif
    if(hints->flags&WindowGroupHint) {
      tmpwin=hints->window_group;
// assuming this what the window_group is normally used for -i.e. it points to parent window ID
      c->parent=get_client_from_window((Window) tmpwin,WINDOW_LE);
#ifdef SHOW_HINTS
      printf("Got parent client: %#lx\n",c->parent);
#endif
      if(c->parent!=NULL) {
        add_sibling_client(c);
        if(c->clientgroup) XFree(c->clientgroup);
        c->clientgroup=strdup(c->parent->clientgroup);
      } else {
        head=XGetClassHint(dpy,tmpwin,&xch);
        if(xch.res_class) {
          if(c->clientgroup) XFree(c->clientgroup);
          c->clientgroup=strdup(xch.res_class);
          XFree(xch.res_class);
        }
        if(xch.res_name) XFree(xch.res_name);
      }
    }
    /* Sets the initial window state if set by client */ 
    if((hints->flags & StateHint)!=0 && attr.map_state!=IsViewable) {
#ifdef SHOW_HINTS
      printf("Used initial state hints\n");
#endif
      set_wm_state(c->window, hints->initial_state); 
    }
    if((hints->flags & IconPixmapHint)!=0) {
      MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
      showfront=1;
    }
    XFree(hints); 
  } else {
// set initial client state
    if(attr.map_state!=IsViewable)
      set_wm_state(c->window, NormalState);
// Need to create a new OS X window, so name appears in dock icon's window list
    MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
    //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
// And bring OroborOSX/XDarwin in front so we don't have to go searching for the new window
    showfront=1;
  }

  /* Get MWM hints */ 
  mhints=get_mwm_hints(c->window); 
  if(mhints) 
  { 
#ifdef SHOW_HINTS
    printf("---- MWM HINTS...\n");
    printf("flags      : %#lx\n",mhints->flags);
    printf("functions  : %#lx\n",mhints->functions);
    printf("decorations: %#lx\n",mhints->decorations);
    printf("inputMode  : %#lx\n",mhints->inputMode);
    printf("status     : %#lx\n",mhints->status);
    printf("DECOR_BORDER, DECOR_TITLE: %#lx , %#lx\n",MWM_DECOR_BORDER,MWM_DECOR_TITLE); 
#endif
    if(mhints->flags==0) {
      showfront=1;
      if(c->osxwinfo.winref==0)
        MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
        //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
    }
    if((mhints->flags&MWM_HINTS_DECORATIONS)!=0 && (mhints->decorations&MWM_DECOR_ALL)==0) 
    { 
      /* Window has no border if *neither* MWM_DECOR_BORDER *nor* MWM_DECOR_TITLE are set */ 
      c->has_frame=((mhints->decorations&(MWM_DECOR_BORDER|MWM_DECOR_TITLE))!=0); 

      /* Set if client can be maximized or not */ 
      c->can_max=((mhints->decorations&MWM_DECOR_MAXIMIZE)!=0); 
    }
    // make a note of modal windows... 
    if((mhints->flags&MWM_HINTS_INPUT_MODE)!=0) {
      if((mhints->inputMode&MWM_INPUT_SYSTEM_MODAL)!=0) {
        c->osxwinfo.winref=-1;
        c->modal_state=SYS_MODAL;
        if(c->parent)
          (c->parent)->modal_state|=HAS_SYS_MODAL_CHILD;
      }
      if((mhints->inputMode&MWM_INPUT_PRIMARY_APPLICATION_MODAL)!=0) {
        c->osxwinfo.winref=-1;
        c->modal_state=APP_MODAL;
        if(c->parent)
          (c->parent)->modal_state|=HAS_APP_MODAL_CHILD;
      }
    } 
    if(c->osxwinfo.winref==0) {
      // create OS X window, and bring OroborOSX/XDarwin in front
      MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
      //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
      showfront=1;
    }
    XFree(mhints); 
  } else if(c->osxwinfo.winref==0) {
    // create OS X window, and bring OroborOSX/XDarwin in front
    MakeX11Window(&(c->osxwinfo),c->name,c->clientgroup);
    //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
    showfront=1;
  }

// This is the case when new window should not be listed in the dock's window list
  if(c->osxwinfo.winref==-1)
    c->osxwinfo.winref=0;

#ifdef SHAPE_SUPPORT 
  /* Select shape events */ 
  if(shape) 
    XShapeSelectInput(dpy, c->window, ShapeNotifyMask); 
#endif /* SHAPE */ 
 
  /* Create frame window */ 
  wattr.override_redirect=True;
  // Added the LeaveWindowMask to this, so we know when we've left the window
  wattr.event_mask=SubstructureNotifyMask|SubstructureRedirectMask|ButtonPressMask|ButtonReleaseMask|EnterWindowMask|LeaveWindowMask; 
  c->frame=XCreateWindow(dpy, root, FX, FY, FW, FH, 0, DefaultDepth(dpy, screen), CopyFromParent, DefaultVisual(dpy, screen), CWOverrideRedirect|CWEventMask, &wattr);
  // Set the cursor we want to see when inside the window -would be nice if this was more like the Mac's
  // normal cursor...
  XDefineCursor(dpy, c->frame, normal_curs); 

  /* This ignores the unmap event when reparenting */ 
  if(attr.map_state==IsViewable) 
    c->ignore_unmap++; 
 
  /* Reparent client window */ 
  XReparentWindow(dpy, c->window, c->frame, LW, TH); 
    
  /* Create frame parts if has border and/or title */ 
    // Top frame parts are needed for iconbars, even if no frame
    c->title=create_frame_part(c->frame, TX, TY, TW<1?1:TW, TH, c->has_frame); 
    c->top_left=create_frame_part(c->frame, TLX, TLY, TLW, TLH, c->has_frame); 
    c->top_right=create_frame_part(c->frame, TRX, TRY, TRW, TRH, c->has_frame);  
   
  if(c->has_frame)
  { 
    /* Side frame parts */ 
    c->left=create_frame_part(c->frame, LX, LY, LW, LH, 1); 
    c->right=create_frame_part(c->frame, RX, RY, RW, RH, 1); 
     
    /* Bottom frame parts */ 
    c->bottom=create_frame_part(c->frame, BX, BY, BW<1?1:BW, BH, 1); 
    c->bottom_left=create_frame_part(c->frame, BLX, BLY, BLW, BLH, 1); 
    c->resize=create_frame_part(c->frame, RSX, RSY, RSW, RSH, 1); 
   
    /* Top frame parts 
    c->title=create_frame_part(c->frame, TX, TY, TW<1?1:TW, TH); 
    c->top_left=create_frame_part(c->frame, TLX, TLY, TLW, TLH); 
    c->top_right=create_frame_part(c->frame, TRX, TRY, TRW, TRH);   */
   
    /* Buttons */ 
    c->zoom=create_frame_part(c->frame, HX, HY, HW, HH, 1); 
    c->icon=create_frame_part(c->frame, IX, IY, IW, IH, 1);   
    c->close=create_frame_part(c->frame, CX, CY, CW, CH, 1);
  } 
 
  /* Move the client to position set by hints  -changed this to work more like twm */ 
  if((c->size->flags & USPosition)!=0 || ((c->size->flags & PPosition)!=0 && (c->x!=0 || c->y!=0)) || !reposition) 
  {
    printf("Found hints for geometry...   %d x %d  @ %d,%d\n",c->width,c->height,c->x,c->y);
#ifdef XINERAMA_SUPPORT
    if(xinerama)
    {
      /* If window positioned within 15 pixels of center of display, consider it to be centered on the current head */
      if(abs((c->x+(c->width/2))-(DisplayWidth(dpy, screen)/2))<15 && abs((c->y+(c->height/2))-(DisplayHeight(dpy, screen)/2))<15)
      {
        get_mouse_xy(&x, &y);
        head=find_head(x, y);
        c->x=screen_x(head)+(screen_width(head)/2)-(c->width/2);
        c->y=screen_y(head)+(screen_height(head)/2)-(c->height/2);
        printf("XINERAMA: Found cursor on head %d... set position to %d,%d\n",head,c->x,c->y);
      }
      head=find_head(c->x, c->y);
      printf("XINERAMA: Window top-left is on head %d\n",head);
    }
#endif /* XINERAMA */
 
    xmin=screen_x(head)+LW;
    ymin=screen_y(head)+TH;
    if( screen_y(head)==0 && gMenubarSpace!=0 )
        ymin+=22;
    xmax=(screen_x(head)+screen_width(head))-c->width-RW; 
    ymax=(screen_y(head)+screen_height(head))-c->height-BH; 
   
    printf("Got bounds (min/max): %d/%d , %d/%d\n",xmin,xmax,ymin,ymax);
    gravitate_client(c, APPLY_GRAVITY); 
    wc.x=c->x; 
    wc.y=c->y; 
 
    /* Keep window fully inside desktop */ 
    if(wc.x>xmax) wc.x=xmax; 
    if(wc.y>ymax) wc.y=ymax; 
    if(wc.x<xmin) wc.x=xmin; 
    if(wc.y<ymin) wc.y=ymin; 
  } 
 
  /* Otherwise follow window placement policy */ 
  else 
  { 
#ifdef XINERAMA_SUPPORT
    if(xinerama)
    {
      get_mouse_xy(&x, &y);
      head=find_head(x, y);
      printf("XINERAMA: Found cursor on head %d...\n",head);
    }
#endif /* XINERAMA */

    xmin=screen_x(head)+LW+config.left_margin;
    ymin=screen_y(head)+TH+config.top_margin; 
    if(screen_y(head)==0 && gMenubarSpace!=0)
        ymin+=22;
    xmax=(screen_x(head)+screen_width(head))-c->width-RW-config.right_margin; 
    ymax=(screen_y(head)+screen_height(head))-c->height-BH-config.bottom_margin; 

    printf("Got bounds (min/max): %d/%d , %d/%d\n",xmin,xmax,ymin,ymax);

    if(config.window_placement==CENTER_ROOT) 
    { 
      x=screen_x(head)+(screen_width(head)/2); 
      y=screen_y(head)+(screen_height(head)/2);
      printf("Placing in center: %d,%d\n",x,y);
    } 
    else
    {
      get_mouse_xy(&x, &y); 
      printf("Placing with cursor: %d,%d\n",x,y);
    }
    wc.x=x-(c->width/2); 
    wc.y=y-(c->height/2); 
 
    /* Keep window fully inside desktop */ 
    if(wc.x>xmax) wc.x=xmax; 
    if(wc.y>ymax) wc.y=ymax; 
    if(wc.x<xmin) wc.x=xmin; 
    if(wc.y<ymin) wc.y=ymin; 
  } 
  wc.height=c->height; 
  wc.width=c->width; 
  wc.border_width=0;
// Add one for the 'jig' to work around the blank windows problem
//  wc.x+=1;

  /* Add client to client list */ 
  if(client_list==NULL) 
  { 
    client_list=c; 
    c->prev=c; 
    c->next=c; 
  } 
  else 
  { 
    c->prev=client_list->prev; 
    c->next=client_list; 
    client_list->prev->next=c; 
    client_list->prev=c; 
  } 
  client_count++; 
  update_gnome_client_list();
  reorder_dockicon_winlist(c);
 
  MyDecreaseXServerGrab();

  SendHackCommand(c,1,HACKCMD_INHIBITFULL,0,0,0,0,0);
  configure_client(c, CWHeight|CWWidth|CWX|CWY|CWBorderWidth, &wc, False, True); 
//  redraw_client_frame(c); 
 
  /* Set client to correct state */   
  if(get_wm_state(c->window)==IconicState) {
    show_client(c);
    raise_client(c);
    // turn new client into an iconbar/docked-icon
    toggle_minimise_client(c,0);
    lower_client(c);
  } else {
    show_client(c);
    raise_client(c);
  }
  
  SendHackCommand(c,0,HACKCMD_SETID,0,0,0,0,0); // uses the id stored for client, so args ignored
  SendHackTitleToDockCommand(c,0);
  
  /* Button grab for click-to-focus */ 
  XGrabButton(dpy, AnyButton, AnyModifier, c->frame, False, ButtonPressMask|ButtonReleaseMask, GrabModeSync, GrabModeAsync, None, None); 
 
#ifdef KEYBOARD_SUPPORT 
  grab_keys(c->window); 
#endif /* KEYBOARD */ 
 
#ifdef WORKSPACE_SUPPORT 
  set_gnome_hint(c->window, WIN_WORKSPACE, workspace); 
#endif /* WORKSPACE */ 
 
  SendSimpleHackCommand(HACKCMD_INHIBITINFO,0);

  /* Check if we want to focus new clients */ 
  if(c->is_icon!=NOT_ICONBAR || !config.focus_new || (get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS) || still_in_auto_client()!=0) {
    if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT)
      SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.inactive_translucency,0,0,0,0);
    if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED)
      SendHackCommand(c,1,HACKCMD_DIMWIN,1,0,0,0,0);
    SendHackCommand(c,1,HACKCMD_SETSHADOW,0,0,0,0,0);
    if(lastfocus_client!=NULL)
      if(lastfocus_client->osxwinfo.winref!=0)
         FocusOSXWin(&(lastfocus_client->osxwinfo));
  } else
    set_focus_client(c);

  if(showfront!=0 && gX11InFront==0)
    Bring_To_Front(0);

// 'Jig' it by one pixel (to make sure content appears on secondary screens)...
/*  wc.x=c->x+1; 
  wc.y=c->y; 
  wc.height=c->height; 
  wc.width=c->width; 
  configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, False); 
 
  wc.x=c->x-1; 
  wc.y=c->y; 
  wc.height=c->height; 
  wc.width=c->width; 
  configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, False); 
*/

  sprintf(lastsub,"TRACE1: exit create_new_client\n");
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: exit create_new_client\n");
#endif
} 

/*
 * Initialise the flag for checking if we've done a client during a 'nearest iconbar' search
 */
 
void init_nearest_iconbar()
{
  Client *c=NULL;
  int i;
  
  c=client_list;
  for(i=0;i<client_count;i++) {
    c->done_icon=0;
    c=c->next;
  }
}

/*
 * Find the iconbar nearest (vertically only) to the supplied y co-ordinate (ypos)
 */
 
Client *get_nearest_iconbar(int ypos,int state)
{
  int i, diff, mindif;
  Client *c=NULL, *ac;
  
  mindif=10000;
  ac=client_list;
  for(i=0;i<client_count;i++) {
    if(ac->is_icon==state || ac->is_icon==ICONBAR_RAISED || state==NOT_ICONBAR) {
      // only want clients that have already been an iconbar
      if(ac->icon_y!=-10000) {
        diff=ac->icon_y-ypos;
        if(diff<0) diff=-diff;
        if(diff<=mindif) {
          // only take clients that have not already been found in a previous call
          if(ac->done_icon==0) {
            if(state==NOT_ICONBAR) {
              // want a list of iconbar positions for every window that has been 'iconbar-ised'
              c=ac;
              mindif=diff;
            } else if(state==ICONBAR_HIDDEN) {
              // only want iconbars that are hidden
              c=ac;
              mindif=diff;
            } else if(state==ICONBAR_VISIBLE) {
              // only want iconbars that are visible
              c=ac;
              mindif=diff;
            }
          }
        }
      }
    }
    ac=ac->next;
  }
  
  // Note this client as having been found
  if(c!=NULL)
    c->done_icon=1;
    
  return c;
}

int undock_client(Client *c)
{
    long nowTime;
    
    if(c!=NULL) {
      nowTime=GetCarbonTickCount();
      if(c->dock_timer==0 && c->undock_timer==0)
        nowTime=10;
      if(nowTime<c->dock_timer || nowTime<c->undock_timer) {
        return 1;
      } else {
        c->dock_timer=0;
        c->undock_timer=0;
        printf("******************************** getting client from dock...\n");
        SendHackCommand(c,1,HACKCMD_DOCKWIN,0,0,0,0,0);
        XMapWindow(dpy,c->frame);
        XMapWindow(dpy,c->window);
        set_wm_state(c->frame, NormalState);
        XMoveWindow(dpy,c->frame,FX,FY);
//        show_last_fifteen();
        //sleep(3);
        set_focus_client(c);
        if((c->auto_properties & AUTO_FOCUS)==AUTO_FOCUS || config.focus_style!=FOCUS_CLICK)
          raise_client(c);	// just in case set_focus_client did not raise it
        return 0;
      }
    }
    return 2;
}

void will_undock_client(Client *c)
{
    if(c!=NULL) {
      if(c->undock_timer==0) {
        c->undock_timer=GetCarbonTickCount()+300;
        printf("******************************** will get client from dock...\n");
        SendHackCommand(c,1,HACKCMD_DOCKWIN,2,0,0,0,0);
      } else if(GetCarbonTickCount()>=c->undock_timer) {
        c->dock_timer=0;
        c->undock_timer=0;
        undock_client(c);
      }
    }
}

Client *check_for_undock(int hackid)
{
    Client *c;
    int i;
    
    c=client_list;
    if(c==NULL) return NULL;
    if(hackid==0) return NULL;
    
    for(i=0;i<client_count;i++) {
        if(c->is_docked && hackid==c->hack_id)
            return c;
        c=c->next;
    }
    return NULL;
}

void dock_client(Client *c)
{
    Client *tc;
    
    if(c!=NULL) {
      if(c->dock_timer!=0) return;
        c->is_docked=10;
        tc=get_top_most_client(get_gnome_hint(c->window,WIN_LAYER),False);
        if(tc!=NULL) {
            if(check_client_modality(tc,0)==c) {
                tc=NULL;
                printf("Got to set top to NULL, otherwise window will come back from dock\n");
            }
        }
        c->is_docked=0;
      c->dock_timer=GetCarbonTickCount()+300;
//      tc=get_bottom_most_client(get_gnome_hint(c->window,WIN_LAYER),False);
      printf("sending window to dock... %#lx [%#lx]\n",c->window,c);
//      if(tc!=c)
//        set_focus_client(tc);
//      else
//        set_focus_client(NULL);
      SendHackTitleToDockCommand(c,1);
      //SendHackCommand(c,1,HACKCMD_DOCKWIN,1,0,0,0,0);
      //XMoveWindow(dpy,c->frame,-FW-10,FY);
      //c->ignore_unmap++; 
      //XUnmapWindow(dpy,c->window);
      //XUnmapWindow(dpy,c->frame);
      set_wm_state(c->frame, IconicState);
      //tc=get_top_most_client(get_gnome_hint(c->window,WIN_LAYER),False);
      lower_client(c);
      usleep(300000); // give it 0.3 seconds, so it gets a chance to start sending to the dock
      if(gX11InFront==0) {
        if(tc==NULL)
            set_focus_client(NULL);
      } else if(tc)
        set_focus_client(tc);
      else
        set_focus_client(NULL);
//      SendHackCommand(c,1,HACKCMD_DOCKWIN,1,0,0,0,0);
      //sleep(3);
    }
}

/*
 * Hide/unhide a particular iconbar
 */
 
void toggle_slide_iconbar(Client *c)
{
  XWindowChanges wc;

  sprintf(lastsub,"TRACE1: toggle_slide_iconbar: %#lx\n",c);
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: toggle_slide_iconbar: %#lx\n",c);
#endif

  if(c==NULL) return;
  if(gHiding_Iconbars==0) return;
    
  icontab_width=get_icontab_width();
    
  if(c->is_icon==ICONBAR_HIDDEN) {
#ifdef SHOW_WEIRD
    if(iconbar_state==ICONBAR_VISIBLE)
      printf("%s oddity: Iconbar is hidden, but state is visible!!\n",PROGNAME);
#endif
    if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
      wc.x=c->icon_x+c->icon_w-iconbar_width;
    else
      wc.x=c->icon_x;
    if(c->title) XMapWindow(dpy,c->title);
    if(c->top_left) XMapWindow(dpy,c->top_left);
    if(c->top_right) XMapWindow(dpy,c->top_right);
    wc.y=c->icon_y;
    wc.width=iconbar_width;
    wc.height=c->icon_h;
//    configure_client(c, CWWidth|CWX, &wc, True, False); 
//          printf("doing XMoveResize in toggle_slide_iconbar: %#lx\n",c);
//          XMoveResizeWindow(dpy,c->frame,wc.x,wc.y,wc.width,wc.height);
//    XConfigureWindow(dpy,c->frame,CWWidth|CWX|CWY|CWHeight, &wc);
    configure_client(c, CWWidth|CWX|CWY|CWHeight, &wc, True, False); 
    c->is_icon=ICONBAR_VISIBLE;
  } else if(c->is_icon==ICONBAR_VISIBLE) {
    if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
      wc.x=c->icon_x+c->icon_w-icontab_width;
    else
      wc.x=c->icon_x;
    wc.y=c->icon_y;
    wc.width=icontab_width;
    wc.height=c->icon_h;
//    configure_client(c, CWWidth|CWX, &wc, True, False);
//          XMoveResizeWindow(dpy,c->frame,wc.x,wc.y,wc.width,wc.height);
    configure_client(c, CWWidth|CWX|CWY|CWHeight, &wc, True, False);
    c->is_icon=ICONBAR_HIDDEN; 
    wc.stack_mode=Below; 
//    XConfigureWindow(dpy, c->frame, CWStackMode, &wc);
//    redraw_client_frame(c);
    configure_client(c, CWStackMode, &wc, True, False);
  }
}

/*
 * Hide/unhide all iconbars
 */
 
void slide_all_iconbars(Client *c)
{
  XWindowChanges wc;
  int i, k, neary;
  Client *ac;
  
  sprintf(lastsub,"TRACE1: slide_all_iconbars (client: %#lx)\n",c);
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: slide_all_iconbars (client: %#lx)\n",c);
#endif

  lastIconbarOutTicks=0;
  if(gHiding_Iconbars==0) return;
  
  icontab_width=get_icontab_width();

  if(c==NULL) {
    if(iconbar_state==ICONBAR_HIDDEN) {
#ifdef SHOW_WEIRD
      printf("%s oddity: Iconbars already hidden??\n",PROGNAME);
#endif
      return; }
    iconbar_state=ICONBAR_HIDDEN;
    neary=nexticon_y;
    for(k=icontab_width;k>=icontab_width;k-=SLIDESTEP) {
      init_nearest_iconbar();
      ac=get_nearest_iconbar(neary,ICONBAR_VISIBLE);
      while(ac!=NULL) {
        if(ac->is_icon==ICONBAR_VISIBLE || ac->is_icon==ICONBAR_RAISED) {
          if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
            wc.x=ac->icon_x+ac->icon_w-k;
          else
            wc.x=ac->icon_x;
          if(ac->title) XMapWindow(dpy,ac->title);
          if(ac->top_left) XMapWindow(dpy,ac->top_left);
          if(ac->top_right) XMapWindow(dpy,ac->top_right);
          wc.y=ac->icon_y;
          wc.width=k;
          wc.height=ac->icon_h;
//          wc.stack_mode=Below;
//          configure_client(ac, CWWidth|CWX, &wc, True, False);
//          printf("doing XMoveResize in slide_all_iconbars: %#lx\n",ac);
//          XMoveResizeWindow(dpy,ac->frame,wc.x,wc.y,wc.width,wc.height);
//          XConfigureWindow(dpy,ac->frame,CWWidth|CWX|CWY|CWHeight, &wc);
          configure_client(ac, CWWidth|CWX|CWY|CWHeight, &wc, True, False);
          wc.stack_mode=Below;
//          XConfigureWindow(dpy, ac->frame, CWStackMode, &wc);
//          redraw_client_frame(ac);
          configure_client(ac, CWStackMode, &wc, True, False);
        }
        ac=get_nearest_iconbar(neary,ICONBAR_VISIBLE);
      }
    }
    ac=client_list;
    for(i=0;i<client_count;i++) {
      if(ac->is_icon==ICONBAR_VISIBLE || ac->is_icon==ICONBAR_RAISED)
        ac->is_icon=ICONBAR_HIDDEN; 
      ac=ac->next;
    }
  } else {
    if(iconbar_state==ICONBAR_VISIBLE) {
//#ifdef SHOW_WEIRD
//      printf("Iconbars already visible\n");
//#endif
      return; }
    iconbar_state=ICONBAR_VISIBLE;
//    wc.stack_mode=Above; 
//    XConfigureWindow(dpy, check_win, CWStackMode, &wc);

    if((long) c!=-1) c->is_icon=ICONBAR_RAISED; // special value which means there is no client to start at
    
    for(k=iconbar_width;k<=iconbar_width;k+=SLIDESTEP) {
    
      if((long) c!=-1) {
        if(c->is_icon==ICONBAR_HIDDEN || c->is_icon==ICONBAR_RAISED) {
          wc.stack_mode=Above; 
//          XConfigureWindow(dpy, c->frame, CWStackMode, &wc);
          configure_client(c, CWStackMode, &wc, True, False);
          if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
            wc.x=c->icon_x+c->icon_w-k;
          else
            wc.x=c->icon_x;
          wc.y=c->icon_y;
          wc.width=k;
          wc.height=c->icon_h;
//          configure_client(c, CWWidth|CWX, &wc, True, False);
          configure_client(c, CWWidth|CWX|CWY|CWHeight, &wc, True, False);
//          XMoveResizeWindow(dpy,c->frame,wc.x,wc.y,wc.width,wc.height);
        }
        neary = c->icon_y;
      } else
        neary = nexticon_y;
        
      init_nearest_iconbar();
      ac=get_nearest_iconbar(neary,ICONBAR_HIDDEN);
      
      while(ac!=NULL) {
        if(ac->is_icon==ICONBAR_HIDDEN || ac->is_icon==ICONBAR_RAISED) {
          wc.stack_mode=Above; 
//          XConfigureWindow(dpy, ac->frame, CWStackMode, &wc);
          configure_client(ac, CWStackMode, &wc, True, False);
          if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
            wc.x=ac->icon_x+ac->icon_w-k;
          else
            wc.x=ac->icon_x;
          wc.y=ac->icon_y;
          wc.width=k;
          wc.height=ac->icon_h;
//          configure_client(ac, CWWidth|CWX, &wc, True, False);
          configure_client(ac, CWWidth|CWX|CWY|CWHeight, &wc, True, False);
//          XMoveResizeWindow(dpy,ac->frame,wc.x,wc.y,wc.width,wc.height);
        }
        ac=get_nearest_iconbar(neary,ICONBAR_HIDDEN);
      }

    }
    
    ac=client_list;
    for(i=0;i<client_count;i++) {
      if(ac->is_icon==ICONBAR_HIDDEN || ac->is_icon==ICONBAR_RAISED)
        ac->is_icon=ICONBAR_VISIBLE; 
      ac=ac->next;
    }
    
    if((long) c!=-1) c->is_icon=ICONBAR_RAISED; // watch out for the special value!
  }
}

/*
 * Simple routine to unhide all iconbars
 */
void unhide_all_iconbars()
{
    slide_all_iconbars((Client *) -1); // send special "-1" signal, to say there is no client to start with
}

/*
 * 'Glue-like' routine to allow hiding of iconbars without requiring knowledge of Client structure
 */
void hide_all_iconbars()
{
  if(gHiding_Iconbars==0) return;
  slide_all_iconbars(NULL);
}

/*  
 *  
 * Get client frame from window ID, from window frame ID, title window ID, or from OS X WindowRef
 * New option - send in win with negative value to avoid check on lastenter (should speed it up a bit)
 */ 

Client *get_client_from_window(Window w, int win) 
{ 
  Client *c=NULL; 
  Client *ct=NULL;
  Client *cf=NULL;
  Client *cw=NULL;
  long i;
  int docheck=1; 
 
  sprintf(lastsub,"TRACE3: get_client_from_window (w,win,client_count: %#lx,%d,%d)\n",w,win,client_count);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: get_client_from_window (w,win,client_count: %#lx,%d,%d)\n",w,win,client_count); 
#endif 

  if(w==0 || client_count==0 || w==root || w==check_win ) {
    sprintf(lastsub,"TRACE3: exiting get_client_from_window (got nothing!)\n"); 
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: exiting get_client_from_window (got nothing!)\n"); 
#endif 
    return NULL; }

  if(win<0) {
    win=-win;
    docheck=0; }
    
  c=client_list; 
//#ifdef DEBUG
//   printf("   checking client [%#lx] %#lx _%#lx_\n",c->frame,c->window,c->title);
//#endif

// Just putting the entire loop into these if blocks seems to have brought down the
// CPU time quite noticeably (according to 'top', anyway)...
    if(win==TITLE) 
    { 
      for(i=0;i<client_count;i++) 
      { 
      // No lastenter check here
      if(c->title==w) {
//        ct=c;
  sprintf(lastsub,"TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
#endif 
        return c; 
      }
      c=c->next; 
      } 
    } 
    else if(win==FRAME_LE) 
    { 
      for(i=0;i<client_count;i++) 
      { 
      if(c->frame==w) {
        cf=c;
        // May want to continue and do a lastenter check...
        if(w!=lastenter || docheck==0) {
  sprintf(lastsub,"TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
#endif 
          return c;  }
      }
      c=c->next; 
      } 
    }  
    else  if(win==OSXWINREF)
    { 
      for(i=0;i<client_count;i++) 
      { 
      // No lastenter check here
      if(c->osxwinfo.winref==(long) w) { 
//        cw=c;
  sprintf(lastsub,"TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
#endif 
        return c; 
      }
      c=c->next; 
      } 
    }
    // default is to look for the client that has the actual X window
    else 
    { 
      for(i=0;i<client_count;i++) 
      { 
      if(c->window==w) { 
//        cw=c;
        // May want to continue and do a lastenter check...
//        if(c->frame!=lastenter || docheck==0)
  sprintf(lastsub,"TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
#endif 
          return c; 
      }
      c=c->next; 
      } 
    } 

// do a check for the sanity of lastenter
  if(w==lastenter) {
//    if(win==TITLE) {
//      if(ct==NULL) {
//#ifdef SHOW_WEIRD
//         printf("  failed to find title for lastenter - setting to zero\n");
//#endif
//         lastenter=0; }
//      return ct;
//    } else if(win==FRAME) {
      if(cf==NULL) {
#ifdef SHOW_WEIRD
         printf("%s oddity: failed to find frame for lastenter - setting to zero\n",PROGNAME);
#endif
         lastenter=0; }
//      return cf;
//    } else {
//      if(cw==NULL) {
//#ifdef SHOW_WEIRD
//         printf("  failed to find window for lastenter - setting to zero\n");
//#endif
//         lastenter=0; }
//    }
  }

  sprintf(lastsub,"TRACE3: leaving get_client_from_window (got: %#lx)\n",cf); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: leaving get_client_from_window (got: %#lx)\n",cf); 
#endif

  return cf; 
} 
 
/* 
 * 
 * Get client with focus 
 * 
 */ 

Client *get_focus_client() 
{ 
  Client *c; 
  Window focus_win; 
  int dummy; 
 
  sprintf(lastsub,"TRACE3: get_focus_client (lastfocus: %#lx)\n",lastfocus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_focus_client (lastfocus: %#lx)\n",lastfocus); 
#endif 
 
  XGetInputFocus(dpy, &focus_win, &dummy); 
  c=get_client_from_window(focus_win, WINDOW);

#ifdef DEBUG
  if(c==NULL)
    printf("  got no client for focus_win %#lx (%#lx)\n",focus_win,lastfocus);
  else
    printf("  got focus_win: %#lx (%#lx) [%#lx] %#lx\n",focus_win,lastfocus,c->frame,c->window);
#endif

// hack the focus for when menus are shown from a window -oroborus used to make
// an X11 nedit window 'inactive' when a menu was pulled down from it
  hacking_focus=False;
//  if(gX11InFront==0) return;
  
  if(c!=NULL)
    lastfocus=focus_win;
  else if(focus_win==check_win)
    lastfocus=0;
  else if(lastfocus!=0) {
    c=get_client_from_window(lastfocus, WINDOW);
    if(c!=NULL)
      hacking_focus=True;
    else
      lastfocus=0;
  }

  sprintf(lastsub,"TRACE3: leaving get_focus_client (c: %#lx  hf: %d)\n",c,hacking_focus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: leaving get_focus_client (c: %#lx  hf: %d)\n",c,hacking_focus); 
#endif 

  lastfocus_client=c;
  return c;   
} 
 
/* 
 * 
 * Get next client in client list 
 * 
 */ 

Client *get_next_client(Client *c) 
{ 
  long i; 
 
  sprintf(lastsub,"TRACE3: get_next_client\n"); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_next_client\n"); 
#endif 
 
  if(c!=NULL) 
  { 
    for(i=0;i<client_count;i++) 
    { 
      c=c->next; 
#ifdef WORKSPACE_SUPPORT 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS && get_gnome_hint(c->window, WIN_WORKSPACE)==workspace) 
#else 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS) 
#endif /* WORKSPACE */ 
        return c; 
    } 
  } 
  return NULL; 
} 

/* 
 * 
 * Get previous client in client list 
 * 
 */ 

Client *get_prev_client(Client *c) 
{ 
  long i; 
 
  sprintf(lastsub,"TRACE3: get_prev_client (client: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_prev_client (client: %#lx)\n",c); 
#endif 
 
  if(c!=NULL) 
  { 
    for(i=0;i<client_count;i++) 
    { 
      c=c->prev; 
#ifdef WORKSPACE_SUPPORT 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS && get_gnome_hint(c->window, WIN_WORKSPACE)==workspace) 
#else 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS) 
#endif /* WORKSPACE */ 
        return c; 
    } 
  } 
  return NULL;
}
 
/* 
 * 
 * Get top-most client (and say whether iconbars are to be included in search)
 * 
 */ 

Client *get_top_most_client(int layer, int get_icons) 
{ 
  Window *windows, w1, w2; 
  int count, i; 
  Client *top=NULL; 
  Client *c=NULL; 
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE3: get_top_most_client (layer,get_icons: %d,%d)\n",layer,get_icons); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_top_most_client (layer,get_icons: %d,%d)\n",layer,get_icons); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef LAYER_SUPPORT 
  for(i=0;i<count;i++) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(attr.map_state==IsViewable) 
    { 
      c=get_client_from_window(windows[i], FRAME); 
      if(c!=NULL && (c->is_icon==NOT_ICONBAR || get_icons) && c->is_docked==0) {
        if(get_gnome_hint(c->window, WIN_LAYER)<=layer) 
          top=c; 
        else 
          break; 
      } 
    } 
  } 
#else 
  for(i=count-1;i>=0;i--) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(attr.map_state==IsViewable) { 
      c=get_client_from_window(windows[i], FRAME); 
      if(c!=NULL)
        if((c->modal_state & ANY_FLOATING)==0 && (c->is_icon==NOT_ICONBAR || get_icons) && c->is_docked==0) 
        { 
          top=c;
          break; }
    } 
  } 
#endif /* LAYER */ 
  if(windows) XFree(windows);
  
  if(top==NULL)
    FocusOSXWin(NULL);
  else if( top->is_icon!=NOT_ICONBAR )
    FocusOSXWin(NULL);

  return top; 
} 

/* 
 * 
 * Get top_most client which contains point x,y (and say whether iconbars are to be included)
 * 
 */ 

Client *get_client_container(int x, int y, int get_icons)
{
  Window *windows, w1, w2; 
  int count, i; 
  Client *top=NULL; 
  Client *c=NULL; 
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE2: get_client_container (x,y,get_icons: %d,%d %d)\n",x,y,get_icons); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: get_client_container (x,y,get_icons: %d,%d %d)\n",x,y,get_icons); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  
  for(i=count-1;i<=0;i--) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(attr.map_state==IsViewable) { 
      c=get_client_from_window(windows[i], FRAME); 
      if(c!=NULL && (c->is_icon==NOT_ICONBAR || get_icons) && c->is_docked==0) 
      { 
        if(x>=attr.x)
          if(x<attr.x+attr.width)
            if(y>=attr.y)
              if(y<attr.y+attr.height) {
                top=c;
                break; }
      }
    } 
  } 

  if(windows) XFree(windows);
  
  return top; 
}

Client *get_lowest_entered_client(int get_icons) 
{ 
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c=NULL; 
  Client *bot=NULL;
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE3: get_lowest_entered_client (get_icons: %d)\n",get_icons); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_lowest_entered_client (get_icons: %d)\n",get_icons); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  for(i=0;i<count;i++) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(attr.map_state==IsViewable) 
    { 
      c=get_client_from_window(windows[i], FRAME); 
      if(c!=NULL && ((c->is_icon==NOT_ICONBAR || get_icons) && c->is_docked==0) && (c->is_e_m_r & IS_ENTERED)!=0) 
      { 
        bot=c;
        break; 
      } 
    } 
  } 
  if(windows) XFree(windows);  
  return bot; 
} 

/* 
 * 
 * Get bottom-most client (and say whether iconbars are to be included)
 * 
 */ 

Client *get_bottom_most_client(int layer, int get_icons) 
{ 
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c=NULL; 
  Client *bot=NULL;
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE3: get_bottom_most_client (layer,get_icons: %d,%d)\n",layer,get_icons); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_bottom_most_client (layer,get_icons: %d,%d)\n",layer,get_icons); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  for(i=0;i<count;i++) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(attr.map_state==IsViewable) 
    { 
      c=get_client_from_window(windows[i], FRAME); 
      if(c!=NULL && ((c->is_icon==NOT_ICONBAR || get_icons) && c->is_docked==0)) 
      { 
#ifdef LAYER_SUPPORT 
        if(get_gnome_hint(c->window, WIN_LAYER)>=layer) { 
          bot=c;
          break; }
#else 
        bot=c;
        break; 
#endif /* LAYER */ 
      } 
    } 
  } 
  if(windows) XFree(windows);  
  return bot; 
} 

/*
 * A new window has been chosen from the dock icon's window list, so change the focus in X11
 */
 
void change_focus(long winref)
{
  Client *c=NULL;
  
  sprintf(lastsub,"TRACE1: change_focus (client: %#lx)\n",c);
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: change_focus (client: %#lx)\n",c);
#endif

  c=get_client_from_window((Window) winref,OSXWINREF);
  
  if(c!=NULL) {
    if(c->is_docked)
      will_undock_client(c);
    else if(c->is_icon==NOT_ICONBAR) {
      raise_client(c);
      set_focus_client(c);
      Bring_To_Front(0);
    } else {
      c->is_icon=ICONBAR_RAISED;
      raise_client(c);
      minmax_client(c);
      Bring_To_Front(0);
    }
  }
}

/* 
 *  
 * Raise client window 
 * 
 */ 

void raise_client(Client *c) 
{ 
  Client *orgc;
  XWindowChanges wc; 

  sprintf(lastsub,"TRACE2: raise_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: raise_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
 
  if(c==NULL) 
  { 
#ifdef SHOW_WEIRD 
    printf("%s oddity: client is NULL in raise_client\n",PROGNAME); 
#endif 
    return;
  } 

  if(c->is_icon==NOT_ICONBAR)
    show_client(c);
    
  if(c->is_docked) {
     will_undock_client(c);
     return;
  }

  orgc=c;
  if(c->is_icon==NOT_ICONBAR)
    c=check_client_modality(orgc,1);
  sprintf(lastsub,"  after check_client_modality: c,orgc: %#lx,%#lx)\n",c,orgc); 
  subtrace();
#ifdef DEBUG 
    printf("  raising client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
  if(orgc!=c)
    if(c->is_icon==NOT_ICONBAR)
      show_client(c);
        
//  wc.stack_mode=Above; 
//  XConfigureWindow(dpy, check_win, CWStackMode, &wc);
  wc.stack_mode=Above; 
//  XConfigureWindow(dpy, c->frame, CWStackMode, &wc);
  configure_client(c, CWStackMode, &wc, c->is_icon!=NOT_ICONBAR, False); 
} 

/*
 * Force a particular client to be raised, with no regard for modal children etc.
 */

void force_raise_client(Client *c) 
{ 
//  Client *orgc;
  XWindowChanges wc; 
 
  sprintf(lastsub,"TRACE3: force_raise_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: force_raise_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
 
  if(c==NULL) 
  { 
#ifdef SHOW_WEIRD 
    printf("%s oddity: client is NULL in force_raise_client\n",PROGNAME); 
#endif 
    return;
  } 

#ifdef DEBUG 
    printf("  force-raising client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
  if(c->is_docked) {
    will_undock_client(c);
    return;
  }

  if(c->is_icon==NOT_ICONBAR)
    show_client(c);
  else {
    minmax_client(c);
  }
  
//  wc.stack_mode=Above; 
//  XConfigureWindow(dpy, check_win, CWStackMode, &wc);
  wc.stack_mode=Above; 
//  XConfigureWindow(dpy, c->frame, CWStackMode, &wc);
  configure_client(c, CWStackMode, &wc, c->is_icon!=NOT_ICONBAR, False); 
} 
 
void raise_all_clients()
{
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c=NULL; 
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE1: raise_all_clients\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: raise_all_clients\n"); 
#endif 
  if(can_send_hackcmds) {
    SendHackCommand(client_list,1,HACKCMD_ALLTOFRONT,0,0,0,0,0);
    return;
  }
  
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  for(i=0;i<count;i++) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr); 
    if(attr.map_state==IsViewable) {
      c=get_client_from_window(windows[i], FRAME); 
      if(c!=NULL && (c->is_icon==NOT_ICONBAR)) 
        force_raise_client(c);
    }
  } 
  if(windows) XFree(windows);  
}

/*
 *
 * Entered/Left a client
 *
 */

void entered_client(Client *c, int can_focus)
{
  Client *fc;
  
  sprintf(lastsub,"TRACE1: entered_client %#lx (can_focus: %d)\n",c,can_focus); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: entered_client %#lx (can_focus: %d)\n",c,can_focus); 
#endif 

#ifdef SHOW_ENTER_LEAVE
  printf("entered_client [%#lx] %#lx (%#lx , %#lx)\n",c==NULL?0:c->frame,c==NULL?0:c->window,lastfocus,overicon);
#endif
/*
  if(overicon==NULL && iconbar_state==ICONBAR_VISIBLE) {
#ifdef SHOW_WEIRD
    printf("Overicon is NULL, but iconbar visible? %d\n",iconbar_state);
#endif
    if(c==NULL)
      will_hide_all_iconbars();
    else if(c->is_icon!=NOT_ICONBAR)
      overicon=c;
    else
      overicon=client_list;
  }
*/ 
  if(c!=NULL)
    if(c->is_docked) return;
//      if(undock_client(c)) return;

  if(overicon!=NULL) {
#ifdef SHOW_WEIRD
// we were over an iconbar, but the bars were not visible??
     if(iconbar_state!=ICONBAR_VISIBLE)
       printf("%s oddity: Overicon: %#lx , but not visible? %d\n",PROGNAME,overicon,iconbar_state);
#endif
     if(overicon->is_icon!=ICONBAR_RAISED) {
// we were over an iconbar, but it was not lit up??
#ifdef SHOW_WEIRD
       printf("%s oddity: Overicon: %#lx , but not raised? %d\n",PROGNAME,overicon,overicon->is_icon);
#endif
// if we're not over a new one then hide the iconbars
       if(c==NULL)
         will_hide_all_iconbars();
       else if(c->is_icon==NOT_ICONBAR)
         will_hide_all_iconbars();
       overicon->is_e_m_r &= NOT_ENTERED; // make sure entered bit is not set
       overicon=NULL;
     } else if(overicon!=c) {
// over a different iconbar (or none) to what we were
       if(iconbar_state!=ICONBAR_VISIBLE) {
#ifdef SHOW_WEIRD
          printf("%s oddity: How did that happen? iconbar_state=%d, overicon=%#lx\n",PROGNAME,iconbar_state,overicon);
#endif
          slide_all_iconbars(overicon);
       }
       if(c==NULL) {
// not over a new iconbar, so prepare to hide them
         overicon->is_icon=ICONBAR_VISIBLE;
         will_hide_all_iconbars();
         if( overicon->icon_w!=iconbar_width ) {
           resize_client(overicon,iconbar_width,overicon->icon_h,True);
           raise_client(overicon);
         }
         overicon->is_e_m_r &= NOT_ENTERED;
         redraw_client_frame(overicon);
       } else if(c->is_icon==NOT_ICONBAR) {
// not over a new iconbar, so prepare to hide them
         overicon->is_icon=ICONBAR_VISIBLE;
         will_hide_all_iconbars();
         if( overicon->icon_w!=iconbar_width ) {
           resize_client(overicon,iconbar_width,overicon->icon_h,True);
           raise_client(overicon);
         }
         overicon->is_e_m_r &= NOT_ENTERED;
         redraw_client_frame(overicon);
       } else {
// still over an iconbar, so don't hide them, but make sure last one is redrawn correctly
         overicon->is_icon=ICONBAR_VISIBLE;
         if( overicon->icon_w!=iconbar_width ) {
           resize_client(overicon,iconbar_width,overicon->icon_h,True);
           raise_client(overicon);
         }
         overicon->is_e_m_r &= NOT_ENTERED;
         redraw_client_frame(overicon);
       }
//       lower_client(overicon);
       overicon=NULL;
     }
  } else if(c!=NULL) {
     if(iconbar_state!=ICONBAR_HIDDEN && lastIconbarOutTicks==0 && can_send_hackcmds==0) {
#ifdef SHOW_WEIRD
       printf("Iconbars not ready to be hidden, but overicon NULL?!\n");
#endif
       will_hide_all_iconbars();
     }
  }

  if(c==NULL)
  {
    fc=lastfocus_client;
//    fc=get_focus_client();
    if(config.focus_style==FOCUS_SLOPPY && can_focus)
      set_focus_client(NULL);
    else if(fc!=NULL) {
      if((fc->auto_properties & AUTO_FOCUS)!=0) {
        fc=get_top_most_client(WIN_LAYER_NORMAL,False);
        if(gX11InFront==1) set_focus_client(fc);
      }
    }
  }
  else if(c->is_icon!=NOT_ICONBAR)
  {
// we are over a new iconbar, show them if not already
    slide_all_iconbars(c);
// prepare for new iconbar by changing width (if necessary) and 'light it up'
    c->is_icon=ICONBAR_RAISED;
    if( c->icontxt_w>=TW-6 ) {
      resize_client(c,c->icontxt_w+6-LW-RW+TLW+TRW,c->icon_h,True);
      raise_client(c);
      redraw_client_frame(c);
    } else if(overicon!=c) {
      raise_client(c);
      redraw_client_frame(c);
    }
    overicon=c;
  }
  else if(can_focus) {
    if(config.focus_style==FOCUS_SLOPPY || (c->auto_properties & AUTO_FOCUS)==AUTO_FOCUS) {
      if(gX11InFront==1) set_focus_client(c);
    } else {
      fc=lastfocus_client;
      if(fc!=NULL) {
        if((fc->auto_properties & AUTO_FOCUS)!=0) {
          fc=get_top_most_client(WIN_LAYER_NORMAL,False);
          if(gX11InFront==1) set_focus_client(fc);
        }
      }
    }
  }

}

/*
 * Glue routine to be called from main.c to say we've left a client
 */
void left_client()
{
    if(can_send_hackcmds==0) entered_client(NULL,0);
}

/*
 * Want to check if a client has modal children, and returns the 'newest' child
 * Also, raises modal windows as it findds them if raiseit is non-zero
 */
 
Client *check_client_modality(Client *orgc, int raiseit)
{
  Client *newc,*c,*tmpc;

  c=orgc;
  newc=c;
  if(c->first_child) {
    if(c->modal_state&HAS_ANY_MODAL_CHILD) {
      if( raiseit ) force_raise_client(c);
      c=c->first_child;
      if(c->modal_state&ANY_MODAL) {
        newc=c;
        if( raiseit ) force_raise_client(c);
      }
      if(c->modal_state&HAS_ANY_MODAL_CHILD)
        newc=check_client_modality(c,raiseit);
      while(c->next_sibling) {
        c=c->next_sibling; 
        if(c->modal_state&ANY_MODAL) {
          newc=c;
          if( raiseit ) force_raise_client(c);
        }
        if(c->modal_state&HAS_ANY_MODAL_CHILD)
          newc=check_client_modality(c,raiseit);
      }
    }
  }
  return newc;
}

/* 
 *  
 * Give client window keyboard focus 
 * 
 */ 

void set_focus_client(Client *c) 
{ 
  Client *orgc;
  int can_af=-1, can_ar=-1, should_raise=0, first_click=0, click_raise=0, live_resize=0;
  
  sprintf(lastsub,"TRACE1: set_focus_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: set_focus_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
 
  if(c==NULL) 
  { 
#ifdef DEBUG
    printf("  got NULL client in set_focus_client\n"); 
#endif
    XSetInputFocus(dpy, check_win, RevertToNone, CurrentTime);
//    XGrabPointer(dpy, root, False, ButtonPressMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime); 
    if(lastfocus_client!=NULL) {
      c=lastfocus_client;
      if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT)
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.inactive_translucency,0,0,0,0);
      lastfocus_client=NULL;
      lastfocus=0;
      if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED && c->is_dimmed==0)
        SendHackCommand(c,1,HACKCMD_DIMWIN,1,0,0,0,0); // automatically does a reshape and a frame redraw
      else
        redraw_client_frame(c);
      SendHackCommand(c,1,HACKCMD_SETSHADOW,0,0,0,0,0);
    }
    lastfocus=0; // just to make sure ;-)
    Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0, 0, 0, 0); // No active client, so disable Window & Properties items
    Setup_Properties_Menu_Translucency(NULL); // No active client, so disable translucency items in Properties
    return;
  } 

  orgc=c;
  c=check_client_modality(orgc,0);
  sprintf(lastsub,"  after check_client_modality: c,orgc: %#lx,%#lx)\n",c,orgc); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("  after check_client_modality: c,orgc: %#lx,%#lx)\n",c,orgc); 
#endif

  if(c->is_docked) {
//    will_undock_client(c);
    return;
  }
  
  orgc=c;
    if(lastfocus_client!=NULL && lastfocus_client!=c) {
      c=lastfocus_client;
      if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT && (c->auto_properties & (NONAUTO_FIRSTCLICK | AUTO_FOCUS))!=NONAUTO_FIRSTCLICK)
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.inactive_translucency,0,0,0,0);
      lastfocus_client=NULL;
      if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED && c->is_dimmed==0 && (c->auto_properties & (NONAUTO_FIRSTCLICK | AUTO_FOCUS))!=NONAUTO_FIRSTCLICK)
        SendHackCommand(c,1,HACKCMD_DIMWIN,1,0,0,0,0); // automatically does a reshape and a frame redraw
      else
        redraw_client_frame(c);
      lastfocus_client=c;
      SendHackCommand(c,1,HACKCMD_SETSHADOW,0,0,0,0,0);
      c=orgc;
    }
      
  if(c->is_icon==NOT_ICONBAR)
  { 
#ifdef DEBUG 
    printf("  focusing client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
    XSetInputFocus(dpy, c->window, RevertToNone, CurrentTime);
//    XGrabPointer(dpy, root, False, ButtonPressMask|ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, resize_curs, CurrentTime); 

    if(lastfocus_client!=c || lastfocus!=c->window) {
       lastfocus_client=c;
       lastfocus=c->window;
       if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT)
         SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.translucency,0,0,0,0);
       if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED && c->is_dimmed!=0)
         SendHackCommand(c,1,HACKCMD_DIMWIN,0,0,0,0,0); // automatically does a reshape and a frame redraw
       else
         redraw_client_frame(c);
       SendHackCommand(c,1,HACKCMD_SETSHADOW,1,0,0,0,0);
    }
    if((c->auto_properties & LIVE_RESIZING)==LIVE_RESIZING)
        live_resize=1;
    if((c->auto_properties & NONAUTO_FIRSTCLICK)==NONAUTO_FIRSTCLICK)
        first_click=1;
    if((c->auto_properties & AUTO_CLICKRAISE)==AUTO_CLICKRAISE)
        click_raise=1;
    if((c->auto_properties & AUTO_FOCUS)==AUTO_FOCUS)
       can_af=1;
    if((c->auto_properties & AUTO_RAISE)==AUTO_RAISE) {
       should_raise=1;
       can_ar=1;
    } else if(can_af<1 && config.focus_style==FOCUS_CLICK)
       should_raise=1;
    else if(can_af>0) {
       should_raise=0;
    }
    if(should_raise==1)
        raise_client(c);
//      if(c!=get_top_most_client(WIN_LAYER_NORMAL,False))
    if(c->osxwinfo.winref!=0) {
       FocusOSXWin(&(c->osxwinfo));
       Setup_Window_Menu(c->has_frame, 1, 1, c->can_max, 0, can_af, can_ar, first_click, click_raise, live_resize);
    }
  } 

  sprintf(lastsub,"TRACE2: leaving set_focus_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: leaving set_focus_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
} 

/*
 * When XDarwin is brought in front, we have to also (unfortunately) bring all
 * clients that have auto-focus set (to work around mouse tracking issue in XDarwin)
 * -not any more! (If the 'hacked' version of XDarwin is used this routine is not called)
 */
void raise_auto_clients(Client *skipc)
{
  Client *c;
  int i;
  
  sprintf(lastsub,"TRACE2: raise_auto_clients \n");
  subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: raise_auto_clients \n");
#endif

  c=client_list;
  if(c==NULL) return;
  
  for(i=0;i<client_count;i++) {
    if(c->is_icon==NOT_ICONBAR)
      if((c->auto_properties & AUTO_FOCUS)!=0)
        if(c->is_docked==0)
          if(c!=skipc)
            raise_client(c);
    c=c->next;
  }
}

/*
 * OroborOSX has come in front, so reactivate the focus window
 */
void come_to_front()
{
    Client *c=NULL, *ctop=NULL;

  sprintf(lastsub,"TRACE2: come_to_front \n");
  subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: come_to_front \n");
#endif

    c=get_focus_client();
    if( c ) {
      if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT)
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.translucency,0,0,0,0);
      if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED && c->is_dimmed!=0)
        SendHackCommand(c,1,HACKCMD_DIMWIN,0,0,0,0,0); // automatically does a reshape and a frame redraw
      else
        redraw_client_frame(c);
      SendHackCommand(c,1,HACKCMD_SETSHADOW,1,0,0,0,0);
    }
    
    ctop=get_top_most_client(WIN_LAYER_NORMAL,False);
    if(can_send_hackcmds==0)
        raise_auto_clients(ctop);
    if( ctop )
      raise_client(ctop);
      
#ifdef SHOW_EVENTS
    printf("come_to_front (focus,top: %#lx,%#lx)\n",c,ctop);
#endif
}

/*
 * Another OS X app has come in front, so deactivate the focus window
 */

void gone_to_back()
{
    Client *c=NULL;
    
  sprintf(lastsub,"TRACE2: gone_to_back \n");
  subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: gone_to_back \n");
#endif

    c=get_focus_client();
    if( c!=NULL ) {
      if((gBGWinSetting & BGWIN_TRANSLUCENT)==BGWIN_TRANSLUCENT)
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.inactive_translucency,0,0,0,0);
      if((gBGWinSetting & BGWIN_DIMMED)==BGWIN_DIMMED && c->is_dimmed==0)
        SendHackCommand(c,1,HACKCMD_DIMWIN,1,0,0,0,0); // automatically does a reshape and a frame redraw
      else
        redraw_client_frame(c);
      SendHackCommand(c,1,HACKCMD_SETSHADOW,0,0,0,0,0);
    }
    XUngrabPointer(dpy, CurrentTime); 
    
#ifdef SHOW_EVENTS
    printf("gone_to_back (focus-client: %#lx)\n",c);
#endif
}

/* 
 * 
 * Lower client window 
 * 
 */ 

void lower_client(Client *c) 
{ 
  XWindowChanges wc; 
 
  sprintf(lastsub,"TRACE3: lower_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: lower_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
 
  if(c==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: client is NULL in lower_client\n",PROGNAME);
#endif
    return;
  }
  
  if(c->is_docked) return;
//    if(undock_client(c)) return;

  if(c->is_icon==NOT_ICONBAR) {
    if((c->auto_properties & (AUTO_FOCUS|AUTO_RAISE))!=0) {
//#ifdef SHOW_WEIRD
      printf("  %s question... should an auto-focus/auto-raise client really be lowered?  -hmmmm...\n",PROGNAME);
//#endif
//    return;
    }
  }
  
#ifdef DEBUG
  printf("  lowering client [%#lx] %#lx\n",c->frame,c->window);
#endif

  wc.stack_mode=Below; 
//  XConfigureWindow(dpy, c->frame, CWStackMode, &wc);
//  redraw_client_frame(c);
  configure_client(c, CWStackMode, &wc, c->is_icon!=NOT_ICONBAR, False); 
} 
 
/* 
 *  
 * Show client window 
 * 
 */
 
void show_client(Client *c) 
{ 
  sprintf(lastsub,"TRACE3: show_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: show_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
 
  if(c==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: client is NULL in show_client\n",PROGNAME);
#endif
    return;
  } 

#ifdef DEBUG
  printf("  showing client [%#lx] %#lx\n",c->frame,c->window);
#endif

  if(c->is_docked) {
    will_undock_client(c);
    return;
  }
  
  c->is_icon=NOT_ICONBAR;
  XMapWindow(dpy, c->frame); 
  XMapWindow(dpy, c->window); 
  set_wm_state(c->window, NormalState); 
} 
 
/* 
 * 
 * Hide client window - this is the original iconisation from oroborus which is no longer used
 * 
 */ 

void old_hide_client(Client *c) 
{ 
  sprintf(lastsub,"TRACE3: hide_client (old)\n"); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: hide_client (old)\n"); 
#endif 

  if(c==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: client is NULL in old_hide_client\n",PROGNAME);
#endif
    return;
  }

#ifdef DEBUG 
  printf("  hiding client [%#lx] %#lx\n",c->frame,c->window);
#endif

  /* We need to ignore one of the unmaps, otherwise our client gets deleted */ 
  c->ignore_unmap++; 
  XUnmapWindow(dpy, c->frame); 
  XUnmapWindow(dpy, c->window); 
  set_wm_state(c->window, IconicState); 
} 
 
/*  
 * 
 * Configure client - the main resizing, raising, lowering function 
 * 
 */ 

void configure_client(Client *c, unsigned int mask, XWindowChanges *wc, int frame_only, int config_win) 
{ 
  XConfigureEvent ce; 
  int w, w1, h, h1, orig_is_icon;
  unsigned int orgmask;
  Client *topc;
 
  sprintf(lastsub,"TRACE2: configure_client (client,mask,frame_only,lastfocus: %#lx,%d,%d,%#lx)\n",c,mask,frame_only,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: configure_client (client,mask,frame_only,lastfocus: %#lx,%d,%d,%#lx) %d\n",c,mask,frame_only,lastfocus,CWStackMode); 
#endif 

//  show_last_fifteen();
  
  if(c==NULL) {
#ifdef SHOW_WEIRD
  printf("%s oddity: client is NULL in configure_client\n",PROGNAME);
#endif
    return;
  }
  
  if(c->is_docked) return;
//    if(undock_client(c)) return;

#ifdef DEBUG
  printf("  configuring client [%#lx] %#lx\n",c->frame,c->window);
#endif

  orig_is_icon=c->is_icon;
  
// choose the height and width depending on whether the client is an iconbar or not
  if(c->is_icon!=NOT_ICONBAR) {
    h=c->icon_h;
    w=c->icon_w;
    if(mask & CWX) c->icon_x=wc->x;
    if(mask & CWY) c->icon_y=wc->y;
  } else {
    h=c->height;
    w=c->width;
    if(mask & CWX) c->x=wc->x;
    if(mask & CWY) c->y=wc->y;
  }
/*
  if(mask & CWWidth) {
    if(w==wc->width)
      mask=mask ^ CWWidth;
  }

  if(mask & CWHeight) {
    if(h==wc->height)
      mask=mask ^ CWHeight;
  }
*/

  if(!frame_only) {
// if it wants the whole frame, then it should not be an iconbar!
#ifdef SHOW_WEIRD
    if(c->is_icon!=NOT_ICONBAR)
      printf("%s oddity: Asked to configure an iconbar with whole frame!\n",PROGNAME);
#endif
    c->is_icon=NOT_ICONBAR;
  }

  if(mask & CWWidth) 
  { 
    w=wc->width; 
 
    /*printf("Setting width: %d , %d , %d , %d , %d\n",w,c->icontxt_x,c->name_w,c->width,c->icon_w);*/

    if( c->is_icon!=NOT_ICONBAR && frame_only) {
      // reset the iconbar width
      if(w<1) w=1; 
      c->icon_w=w;
    } else {

      /* Get resize increments */ 
      if(c->size->flags & PResizeInc) 
      { 
        w1=(w-c->size->min_width)/c->size->width_inc; 
        w=c->size->min_width+(w1*c->size->width_inc); 
      } 
      /* Get min size */ 
      if(c->size->flags & PMinSize) 
        if(w<c->size->min_width) w=c->size->min_width; 
      /* Get max size */ 
      if(c->size->flags & PMaxSize) 
        if(w>c->size->max_width) w=c->size->max_width; 
 
      /* Make sure we never set width to <1 */ 
      if(w<1) w=1; 

      c->width=w;
    }
  } 

  if(mask & CWHeight) 
  { 
    h=wc->height; 
 
    if( c->is_icon!=NOT_ICONBAR && frame_only ) {
      // reset the iconbar height
      if(h<1) h=1; 
      c->icon_h=h;
    } else {

      /* Get resize increments */ 
      if(c->size->flags & PResizeInc) 
      { 
        h1=(h-c->size->min_height)/c->size->height_inc; 
        h=c->size->min_height+(h1*c->size->height_inc); 
      } 
      /* Get min size */ 
      if(c->size->flags & PMinSize) 
        if(h<c->size->min_height) h=c->size->min_height; 
      /* Get max size */ 
      if(c->size->flags & PMaxSize) 
        if(h>c->size->max_height) h=c->size->max_height; 
 
      /* Make sure we never set height to <1 */ 
      if(h<1) h=1; 

      c->height=h;
    }
  } 
 
#ifdef LAYER_SUPPORT 
  /* Deal with window layers */ 
  if(mask & CWStackMode) 
  { 
    Client *sibling=NULL; 
 
    /* Takes care of layers */ 
    switch(wc->stack_mode) 
    { 
    case Above: 
      /* Window wants to be raised to top of its layer */ 
      sibling=get_top_most_client(get_gnome_hint(c->window, WIN_LAYER),True); 
      if(sibling==NULL) 
        wc->stack_mode=Below; 
      break; 
    case Below: 
      /* Window wants to be lowered to bottom of its layer */ 
      sibling=get_bottom_most_client(get_gnome_hint(c->window, WIN_LAYER),True); 
      break; 
    } 
    if(sibling==NULL) 
    { 
      if(mask & CWSibling) mask=mask ^ CWSibling; 
    } 
    else if(sibling==c) 
    { 
      if(mask & CWStackMode) mask=mask ^ CWStackMode; 
      if(mask & CWSibling) mask=mask ^ CWSibling; 
    } 
    else 
    { 
      mask=mask | CWSibling; 
      wc->sibling=sibling->frame; 
    } 
#ifdef WORKSPACE_SUPPORT 
    set_gnome_hint(c->window, WIN_WORKSPACE, workspace); 
#endif /* WORKSPACE */ 
  } 
#endif /* LAYER */ 
 

#ifdef SHAPE_SUPPORT
//  if(shape) 
//    XShapeSelectInput(dpy, c->window, ShapeNotifyMask); 
#endif

  orgmask=mask;
//  if(!frame_only && mouse_mode==MOUSE_NORMAL)
  if(!frame_only)
  {
    if((c->is_e_m_r & IS_MOVING)==0 || mouse_mode==0) {
/* We don't want any stacking order changes on our actual client */ 
      XMapWindow(dpy,c->window);	// just to make sure it's back
      if(mask & CWStackMode) mask=mask ^ CWStackMode; 
      if(mask & CWSibling) mask=mask ^ CWSibling;
    
      if((mask & (CWHeight|CWWidth))!=0 && ((c->auto_properties & LIVE_RESIZING)==LIVE_RESIZING || (c->is_e_m_r & IS_RESIZING)==0)) {
//      c->resize_timer=GetCarbonTickCount()+20;
//    {} else if(config_win!=0) {}
        if(config_win!=0) {
          wc->x=LW;
          wc->y=TH;
          wc->width=c->width;
          wc->height=c->height;
          XConfigureWindow(dpy, c->window, mask, wc);
        } else {
          XMoveResizeWindow(dpy,c->window, LW, TH, c->width, c->height);
        }
      }
    }
  } else if(c->is_icon!=NOT_ICONBAR || (get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED==WIN_STATE_SHADED))
    XUnmapWindow(dpy,c->window);	// just to make sure it's gone
  else
    XMapWindow(dpy,c->window);	// just to make sure it's back

  mask=orgmask;


  if((mask & CWStackMode)!=0) {
    if(wc->stack_mode==Above) {
      // This makes sure that the client is brought in front by the XDarwin application, if it wasn't
      // already. Otherwise it can still remain behind other OSX windows if XDarwin was not in front at
      // the time the window was clicked.
      XConfigureWindow(dpy, check_win, CWStackMode, wc);
      wc->stack_mode=Above;
    } else if(wc->stack_mode==Below) {
      wc->stack_mode=Below;
      // Probably no need to do this...
      XConfigureWindow(dpy, check_win, CWStackMode, wc);
      wc->stack_mode=Below;
    }
  }

  wc->x=FX; 
  wc->y=FY; 
  wc->width=w+LW+RW;
  wc->height=FH;
  XConfigureWindow(dpy, c->frame, mask, wc);
  
  /* If size has changed */ 
  if(mask & (CWHeight|CWWidth)) 
  { 
    SendHackCommand(c,1,HACKCMD_INHIBITREFRESH,1,0,0,0,0);
    if( c->has_frame || (c->is_icon!=NOT_ICONBAR) ) 
    { 
      /* Reposition and resize frame parts */ 

      if(c->is_icon!=NOT_ICONBAR && TW<1) {
        if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
           if(c->top_left) {
              XMapWindow(dpy,c->top_left);
              XMoveResizeWindow(dpy, c->top_left, TLX, TLY, TLW, TLH);
//              XMoveResizeWindow(dpy, c->top_left, TLX, TLY, icontab_width, TLH);
              if(c->top_right) XUnmapWindow(dpy, c->top_right);
              if(c->title) XUnmapWindow(dpy, c->title); 
//              if(c->top_right) XMoveResizeWindow(dpy, c->top_right, TRX+TRW-1, TRY, 1, TRH);
//              if(c->title) XMoveResizeWindow(dpy, c->title, 1, TY, 1, TH); 
           } else if(c->top_right) {
              XMapWindow(dpy,c->top_right);
              XMoveResizeWindow(dpy, c->top_right, TRX, TRY, TRW, TRH);
//              XMoveResizeWindow(dpy, c->top_right, TRX, TRY, icontab_width, TRH);
              if(c->top_left) XUnmapWindow(dpy, c->top_left);
              if(c->title) XUnmapWindow(dpy, c->title); 
//              if(c->top_left) XMoveResizeWindow(dpy, c->top_left, TLX, TLY, 1, TLH);
//              if(c->title) XMoveResizeWindow(dpy, c->title, 1, TY, 1, TH); 
           }
        } else {
           if(c->top_right) {
              XMapWindow(dpy,c->top_right);
              XMoveResizeWindow(dpy, c->top_right, TRX, TRY, TRW, TRH);
              if(c->top_left) XUnmapWindow(dpy, c->top_left);
              if(c->title) XUnmapWindow(dpy, c->title); 
//              if(c->top_left) XMoveResizeWindow(dpy, c->top_left, TLX, TLY, 1, TLH);
//              if(c->title) XMoveResizeWindow(dpy, c->title, 1, TY, 1, TH); 
           } else if(c->top_left) {
              XMapWindow(dpy,c->top_left);
              XMoveResizeWindow(dpy, c->top_left, TLX, TLY, TLW, TLH);
              if(c->title) XUnmapWindow(dpy,c->title);
//              if(c->title) XMoveResizeWindow(dpy, c->title, 1, TY, 1, TH); 
           }
        }
      } else {
        if(c->title) XMapWindow(dpy,c->title);
        if(c->top_left) XMapWindow(dpy,c->top_left);
        if(c->top_right) XMapWindow(dpy,c->top_right);
        if(c->title) XMoveResizeWindow(dpy, c->title, TX, TY, TW<1?1:TW, TH); 
        if(c->top_left) XMoveResizeWindow(dpy, c->top_left, TLX, TLY, TLW, TLH); 
        if(c->top_right) XMoveResizeWindow(dpy, c->top_right, TRX, TRY, TRW, TRH);
      }
//      XMoveResizeWindow(dpy, c->top_right, w+LW+RW-TRW, TRY, TRW, TRH);

      if( c->is_icon!=NOT_ICONBAR ) {
        XUnmapWindow(dpy,c->window);
        // window buttons need to be nowehere in sight for iconbars! (These are unmapped now...)
//        XMoveResizeWindow(dpy, c->close, -100, -100, CW, CH);
//        XMoveResizeWindow(dpy, c->icon, -100, -100, IW, IH);
//        XMoveResizeWindow(dpy, c->zoom, -100, -100, HW, HH);
      } else if( frame_only) {
        // only doing frame...
        if(c->close) XMoveResizeWindow(dpy, c->close, CX, CY, CW, CH);
        if(c->icon) XMoveResizeWindow(dpy, c->icon, IX, IY, IW, IH);
        if(c->zoom) XMoveResizeWindow(dpy, c->zoom, HX, HY, HW, HH);
      } else {
        // doing everything...
        if(c->left) XMoveResizeWindow(dpy, c->left, LX, LY, LW, LH);
        if(c->right) XMoveResizeWindow(dpy, c->right, RX, RY, RW, RH);
        if(c->bottom_left) XMoveResizeWindow(dpy, c->bottom_left, BLX, BLY, BLW, BLH);
        if(c->bottom) XMoveResizeWindow(dpy, c->bottom, BX, BY, BW, BH);
        if(c->resize) XMoveResizeWindow(dpy, c->resize, RSX, RSY, RSW, RSH);
        if(c->close) XMoveResizeWindow(dpy, c->close, CX, CY, CW, CH);
        if(c->icon) XMoveResizeWindow(dpy, c->icon, IX, IY, IW, IH);
        if(c->zoom) XMoveResizeWindow(dpy, c->zoom, HX, HY, HW, HH);
      }
//      redraw_client_frame(c);
    }
  }
  
//  c->is_icon=orig_is_icon;

#ifdef SHAPE_SUPPORT
  /* Select shape events */ 
//  if(shape) 
//    XShapeSelectInput(dpy, c->window, 0); 
    if((mask & (CWHeight|CWWidth))!=0) { 
      if(can_send_hackcmds==0) {
        // make sure it's all opaque at the start (shape is inhibited using the second parameter)
        set_client_shape(c,1);
      } else {
        SendHackCommand(c,1,HACKCMD_INHIBITREFRESH,0,0,0,0,0);
        set_client_shape(c,0);
      }
    }
//      set_client_shape(c,c->is_shaped);
#endif /* SHAPE */ 
//    XConfigureWindow(dpy, c->frame, mask, wc); 

//    if( c->has_frame || (c->is_icon!=NOT_ICONBAR) ) 
//      redraw_client_frame(c);

  if((mask & (CWHeight|CWWidth))!=0) {
      if(c->is_icon!=ICONBAR_RAISED)
        c->is_e_m_r &= NOT_ENTERED;
      else if(can_send_hackcmds)
        c->is_e_m_r |= IS_ENTERED;
      redraw_client_frame(c);
  } else if((mask & (CWX|CWY))!=0) {
    if((c->is_e_m_r & IS_MOVING)==IS_MOVING)	// need to show new position in titlebar
      redraw_client_title(c);
    else if((c->is_e_m_r & SHOWING_POSITION)==SHOWING_POSITION)  // put back original titlebar after move
      redraw_client_title(c);
  }
  
    if(c->osxwinfo.winref!=0)
        SetOSXWinPos(&(c->osxwinfo),FX,FY,FW,FH,CX,CY,HX+HW,HY+HH);
        
//#ifdef SHAPE_SUPPORT 
//    set_client_shape(c);
//#endif /* SHAPE */ 
//  }
//  else
//    XConfigureWindow(dpy, c->frame, mask, wc); 
  
//   MyIncreaseXServerGrab();

  /* Notify client of window changes if changes to position or size */ 
//  if(mouse_mode==MOUSE_NORMAL && (mask & (CWX|CWY|CWHeight|CWWidth))!=0 && !frame_only && (c->is_icon==NOT_ICONBAR) ) 
  if(config_win!=2 && (mask & (CWX|CWY|CWHeight|CWWidth))!=0)
  if(((c->is_e_m_r & (IS_MOVING | IS_RESIZING))==0 || mouse_mode==0) && !frame_only && (c->is_icon==NOT_ICONBAR) ) 
  { 
//#ifdef SHOW_EVENTS
    printf("Sending configurenotify event to window: %#lx\n",c->window);
//#endif
    ce.type=ConfigureNotify;
    ce.display=dpy;
    ce.event=c->window; 
    ce.window=c->window; 
    ce.x=c->x; 
    ce.y=c->y; 
    ce.width=c->width; 
    ce.height=c->height;
    ce.border_width=0; 
//    ce.above=c->frame; 
    ce.above=0; 
    ce.override_redirect=0; 
    XSendEvent(dpy, c->window, False, StructureNotifyMask, (XEvent *)&ce);
  }
#ifdef SHAPE_SUPPORT
//  } else if(c->is_shaped==0) {
  // now set the shape (it was made all opaque before, so that it looks a bit better) 
  if((mask & (CWHeight|CWWidth))!=0 && can_send_hackcmds==0)
    set_client_shape(c,0);
//  if(c!=lastfocus_client && !frame_only)
//    XMoveResizeWindow(dpy,c->window, -c->width-10, -c->height-10, c->width, c->height);
#endif /* SHAPE */ 
//  }

//  if((mask & CWStackMode)==0) {
//    if(get_top_most_client(WIN_LAYER_NORMAL,True)==c) {
//      wc->stack_mode=Above;
//      XConfigureWindow(dpy, c->frame, CWStackMode, wc);
//    }
//  }
//  if((mask & (CWHeight|CWWidth))==0)
//    XResizeWindow(dpy, c->frame, wc->width, wc->height);
  
//  MyDecreaseXServerGrab();
} 
 
/* 
 * 
 * Move client to new position 
 * 
 */ 

void move_client(Client *c, int x, int y, int redraw) 
{ 
  XWindowChanges wc;
  int was_entered;
 
  sprintf(lastsub,"TRACE2: move_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: move_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
#endif
#ifdef DEBUG
  printf("  moving client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
  if(c==NULL) return;
  
  if(c->is_docked) return;
//    if(undock_client(c)) return;

  was_entered=c->is_e_m_r;
  if((was_entered & IS_MOVING)==0)
    was_entered &= NOCHANGE_SHOWFLAGS;

  wc.x=x; 
  wc.y=y;
  
  if(redraw!=0) {
    if(c->is_icon==NOT_ICONBAR) {
      wc.width=c->width;
      wc.height=c->height;
    } else {
      wc.width=c->icon_w;
      wc.height=c->icon_h;
    }
    configure_client(c, CWX|CWY|CWWidth|CWHeight, &wc, c->is_icon!=NOT_ICONBAR, 0); 
  } else
    configure_client(c, CWX|CWY, &wc, c->is_icon!=NOT_ICONBAR, 0);
    
  c->is_e_m_r = (c->is_e_m_r & CHANGE_SHOWFLAGS) | was_entered;
} 
 
/* 
 * 
 * Resize client 
 * 
 */ 
void resize_client(Client *c, int w, int h, int frame_only) 
{ 
  XWindowChanges wc;
  int mask, was_entered;
 
  sprintf(lastsub,"TRACE2: resize_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: resize_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
#endif
#ifdef DEBUG
  printf("  resizing client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
  if(c==NULL) return;
  
  if(c->is_docked) return;
//    if(undock_client(c)) return;

  was_entered=c->is_e_m_r;
  if((was_entered & IS_RESIZING)==0)
    was_entered &= NOCHANGE_SHOWFLAGS;

  wc.width=w; 
  wc.height=h;
  mask = CWWidth|CWHeight|CWX|CWY;
  
  if(c->is_icon!=NOT_ICONBAR) {
    // watch out for iconbars on the right!
    if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
//      mask=mask|CWX;
      wc.x=c->icon_x+c->icon_w-w;
      wc.y=c->icon_y;
    } else {
      wc.x=c->icon_x;
      wc.y=c->icon_y;
    }
  } else {
    wc.x=c->x;
    wc.y=c->y;
  }
  
  if(mouse_mode==MOUSE_RESIZE)
    SendHackCommand(c,1,HACKCMD_INHIBITFULL,1,0,0,0,0);

  configure_client(c, mask, &wc, frame_only || (c->is_icon!=NOT_ICONBAR), False);
//  XMoveResizeWindow(dpy,c->frame,wc.x,wc.y,wc.width,wc.height);

  if(mouse_mode==MOUSE_RESIZE)
    SendHackCommand(c,1,HACKCMD_INHIBITFULL,0,0,0,0,0);

  c->is_e_m_r = (c->is_e_m_r & CHANGE_SHOWFLAGS) | was_entered;
} 
 
/* 
 * 
 * Move & Resize client 
 * 
 */ 
void move_resize_client(Client *c, int x, int y, int w, int h, int frame_only) 
{ 
  XWindowChanges wc;
  int mask, was_entered;
 
  sprintf(lastsub,"TRACE2: move_resize_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: move_resize_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
#endif
#ifdef DEBUG
  printf("  moving & resizing client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
  if(c==NULL) return;
  
  if(c->is_docked) return;
//    if(undock_client(c)) return;
    
  was_entered=c->is_e_m_r;

  wc.x=x; 
  wc.y=y;
  wc.width=w; 
  wc.height=h;
  mask = CWX|CWY|CWWidth|CWHeight;
/*  
  if(c->is_icon!=NOT_ICONBAR) {
    // watch out for iconbars on the right!
    if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
      wc.x=c->icon_x+c->icon_w-w;
      wc.y=c->icon_y ;
    } else {
      wc.x=c->icon_x;
      wc.y=c->icon_y;
    }
  } else {
    wc.x=c->x;
    wc.y=c->y;
  }
*/  
  configure_client(c, mask, &wc, frame_only || (c->is_icon!=NOT_ICONBAR), False);
//  XMoveResizeWindow(dpy,c->frame,wc.x,wc.y,wc.width,wc.height);

  c->is_e_m_r=was_entered;
} 
 
/* 
 * 
 * Toggle client maximized 
 * 
 */ 

void toggle_client_maximized(Client *c) 
{ 
  XWindowChanges wc; 
 
  if( c->is_icon!=NOT_ICONBAR )
    return;

  sprintf(lastsub,"TRACE1: toggle_client_maximized (client,can_max: %#lx,%d)\n",c,(c==NULL)?0:c->can_max); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: toggle_client_maximized (client,can_max: %#lx,%d)\n",c,(c==NULL)?0:c->can_max); 
#endif 
 
  if(c->can_max) 
  { 
    if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_MAXIMIZED) 
    { 
      c->x=c->old_x; 
      c->y=c->old_y; 
      c->width=c->old_width; 
      c->height=c->old_height; 
    } 
    else 
    { 
      int head=0; 
 
#ifdef XINERAMA_SUPPORT 
      if(xinerama)
        head=find_head(c->x+c->width/2, c->y+c->height/2); 
#endif /* XINERAMA */ 
 
      c->old_x=c->x; 
      c->old_y=c->y; 
      c->old_width=c->width; 
      c->old_height=c->height; 
      c->x=screen_x(head)+LW+config.left_margin; 
      c->y=screen_y(head)+TH+config.top_margin; 
      c->width=screen_width(head)-LW-RW-config.left_margin-config.right_margin; 
      c->height=screen_height(head)-TH-BH-config.top_margin-config.bottom_margin;
    } 
    wc.x=c->x; 
    wc.y=c->y; 
    wc.height=c->height; 
    wc.width=c->width; 
    configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, c->is_icon!=NOT_ICONBAR, False); 
    set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_MAXIMIZED); 
  } 
} 
 
/* 
 *  
 * Toggle client shaded 
 * 
 */ 

void toggle_client_shaded(Client *c) 
{ 
  int head;
  XWindowChanges wc;
  int mask;

  if(c->is_icon!=NOT_ICONBAR)
    return;

  sprintf(lastsub,"TRACE1: toggle_client_shaded (client: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: toggle_client_shaded (client: %#lx)\n",c); 
#endif 
 
  if(c->has_frame) 
  { 
    set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_SHADED); 

    wc.width=c->width; 
    wc.height=c->height;
    mask = CWWidth|CWHeight;
  
    if(c->is_icon!=NOT_ICONBAR) {
      // watch out for iconbars on the right!
      if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
        mask=mask|CWX;
        wc.x=c->icon_x+c->icon_w-c->width;
      }
    }
    
    if((get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)==WIN_STATE_SHADED) {
      if(c->left) XUnmapWindow(dpy,c->left);
      if(c->right) XUnmapWindow(dpy,c->right);
      if(c->bottom) XUnmapWindow(dpy,c->bottom);
      if(c->bottom_left) XUnmapWindow(dpy,c->bottom_left);
      if(c->resize) XUnmapWindow(dpy,c->resize);
    } else {
      if(c->left) XMapWindow(dpy,c->left);
      if(c->right) XMapWindow(dpy,c->right);
      if(c->bottom) XMapWindow(dpy,c->bottom);
      if(c->bottom_left) XMapWindow(dpy,c->bottom_left);
      if(c->resize) XMapWindow(dpy,c->resize);
    }
  
    configure_client(c, mask, &wc, (get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)==WIN_STATE_SHADED, 1); 
//    resize_client(c, c->width, c->height, True);

// Don't want the titlebar to be off the screen!
    if( xinerama ) {
       head=find_head(c->x,c->y);
       if(c->y-TH<screen_y(head))
         move_client(c, c->x, screen_y(head)+TH, 0);
    } else {
       if(c->y-TH<0)
         move_client(c, c->x, TH, 0);
    }
    
//    redraw_client_frame(c); 
  }

}

void toggle_minimise_client(Client *c, int which)
{
    if((gMinMethod==minmethod_iconbars && which==0) || (gMinMethod==minmethod_dock && which!=0)) {
        minmax_client(c);
    } else if((gMinMethod==minmethod_dock && which==0) || (gMinMethod==minmethod_iconbars && which!=0)) {
        if(c->is_docked==0)
            dock_client(c);
        else
            will_undock_client(c);
    }
}

/*
 * This has turned into a bit of a monster routine to toggle a client into/from an iconbar!
 */
 
void minmax_client(Client *c) 
{ 
  XWindowChanges wc;
  int i, nowt, nowx, nowy, noww, endt, endx, endy, endw;
  int shaded=0;
  Client *newc=NULL, *ac=NULL;
  Window ptrwin;

  sprintf(lastsub,"TRACE1: minmax_client (client: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: minmax_client (client: %#lx)\n",c); 
#endif 
 
//    newc=get_focus_client();
    newc=lastfocus_client;

    if( c->is_icon!=NOT_ICONBAR ) {
       if(gX11InFront==0)
         Bring_X_To_Front(1);
       if(newc) {
         set_focus_client(NULL);
//         redraw_client_frame(newc);
         newc=NULL; }
       if(iconbar_state==ICONBAR_HIDDEN && c->is_icon==ICONBAR_HIDDEN)
         toggle_slide_iconbar(c);
       nowx=c->icon_x;
       nowy=c->icon_y;
       endx=c->x-config.left.width;
       endy=c->y;
       nowt=c->icontxt_x;
       noww=c->icon_w;
       if(shrinking_iconbars!=0) {
         endw=c->width+config.left.width+config.right.width;
         endt=c->name_x;
       } else {
         endt=nowt;
         endw=noww; }
       c->is_icon=ICONBAR_RAISED;
       // move iconbar from its location to where the window should be, changing width as necessary
       for(i=iconbar_steps-1;i>=0;i--) {
         wc.x=endx-(endx-nowx)*i/iconbar_steps;
         wc.y=endy-(endy-nowy)*i/iconbar_steps;
         wc.width=endw-(endw-noww)*i/iconbar_steps;
         wc.height=c->icon_h;
         c->icontxt_x=endt-(endt-nowt)*i/iconbar_steps;
         if(i==iconbar_steps-1)
           configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, True, False);
         else if(shrinking_iconbars!=0)
           configure_client(c, CWWidth|CWX|CWY, &wc, True, False);
         else
           configure_client(c, CWX|CWY, &wc, True, False); 
         //sleep(1);  -heh! interesting to watch it do this slowly...!
       }
       wc.x=c->x;
       wc.y=c->y;
       wc.width=c->width;
       wc.height=c->height;
       c->is_icon=NOT_ICONBAR;
       c->osxwinfo.view_state=VIEWSTATE_VISIBLE;
       c->icon_x=nowx;
       if(noww!=iconbar_width)
         if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
           c->icon_x=nowx+noww-iconbar_width;
       c->icon_y=nowy;
       c->icon_w=iconbar_width;
       if(shrinking_iconbars!=0)
         c->icontxt_x=nowt;
       XMapWindow(dpy,c->window);
       if(c->has_frame==0) {
         if(c->top_left) XUnmapWindow(dpy,c->top_left);
         if(c->top_right) XUnmapWindow(dpy,c->top_right);
         if(c->title) XUnmapWindow(dpy,c->title);
       } else {
         if(c->close) XMapWindow(dpy,c->close);
         if(c->icon) XMapWindow(dpy,c->icon);
         if(c->zoom) XMapWindow(dpy,c->zoom);
         if(c->left) XMapWindow(dpy,c->left);
         if(c->right) XMapWindow(dpy,c->right);
         if(c->bottom) XMapWindow(dpy,c->bottom);
         if(c->bottom_left) XMapWindow(dpy,c->bottom_left);
         if(c->resize) XMapWindow(dpy,c->resize);
       }
    } else {
       // Tell OroborOSX this window is being iconised - doesn't actually do anything, since we still
       // want it to appear in the dock icon's window list, but may be useful at some point...
       if(c->osxwinfo.winref!=0)
         HideOSXWin(&(c->osxwinfo));
       nowx=c->x-config.left.width;
       nowy=c->y;
       noww=c->width+config.left.width+config.right.width;
       nowt=c->name_x;
       if( newc==c ) {
         c->is_icon=ICONBAR_RAISED;
       } else {
         c->is_icon=ICONBAR_VISIBLE; }
       // need these sub-windows back, if there was no frame
       if(c->has_frame==0) {
         if(c->top_left) XMapWindow(dpy,c->top_left);
         if(c->top_right) XMapWindow(dpy,c->top_right);
         if(c->title) XMapWindow(dpy,c->title);
       } else {
         if(c->close) XUnmapWindow(dpy,c->close);
         if(c->icon) XUnmapWindow(dpy,c->icon);
         if(c->zoom) XUnmapWindow(dpy,c->zoom);
         if(c->left) XUnmapWindow(dpy,c->left);
         if(c->right) XUnmapWindow(dpy,c->right);
         if(c->bottom) XUnmapWindow(dpy,c->bottom);
         if(c->bottom_left) XUnmapWindow(dpy,c->bottom_left);
         if(c->resize) XUnmapWindow(dpy,c->resize);
       }
       newc=NULL;
       // Check to see if this window has been 'iconbar-ised' before - if not, set new values
       if(c->icon_x==-10000) {
         c->icon_w=iconbar_width;
         if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
           c->icon_x=nexticon_x-c->icon_w;
         else
           c->icon_x=nexticon_x;
         c->icon_y=nexticon_y;
         c->icon_h=config.top.height;
         c->icontxt_x=3+config.iconbar_left_width;
         c->icontxt_y=c->name_y;
         c->icontxt_w=c->name_w;
         c->icontxt_h=c->name_h;
         // Which way are the iconbars going?
         if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_TOPRIGHT)
           nexticon_y+=config.top.height;
         else if(iconbarpos==ICONBARPOS_BOTLEFT || iconbarpos==ICONBARPOS_BOTRIGHT)
           nexticon_y-=config.top.height;
       }
       endx=c->icon_x;
       endy=c->icon_y;
       endw=c->icon_w;
       endt=c->icontxt_x;
       if(shrinking_iconbars==0) {
         noww=endw;
         nowt=endt; }
       // get rid of the actual window content
       c->ignore_unmap++;
       XUnmapWindow(dpy,c->window);
       // move iconbar from the window to where the it's going, changing width as necessary
       for(i=0;i<iconbar_steps;i++) {
         wc.width=noww-(noww-endw)*i/iconbar_steps;
         wc.height=c->icon_h;
         wc.x=nowx-(nowx-endx)*i/iconbar_steps;
         wc.y=nowy-(nowy-endy)*i/iconbar_steps;
         c->icontxt_x=nowt-(nowt-endt)*i/iconbar_steps;
         if(i==0)
           configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, True, False);
         else if(shrinking_iconbars!=0)
           configure_client(c, CWWidth|CWX|CWY, &wc, True, False);
         else
           configure_client(c, CWX|CWY, &wc, True, False); 
         //sleep(1);
       }
       wc.width=endw;
       wc.height=c->icon_h;
       wc.x=endx;
       wc.y=endy;
       if(shrinking_iconbars!=0)
         c->icontxt_x=endt;
       c->is_icon=ICONBAR_VISIBLE;
       c->osxwinfo.view_state=VIEWSTATE_ICONBAR;
   }

    //shaded=get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED;
    
    // always set the window to shaded -otherwise the window 'flickers' twice when
    // returning from an iconbar, for some reason...
    set_gnome_hint(c->window, WIN_STATE, WIN_STATE_SHADED);
    shaded=WIN_STATE_SHADED;

    if( c->is_icon==NOT_ICONBAR ) {
       overicon=NULL;
       will_hide_all_iconbars();
       raise_client(c);
       set_focus_client(c);
       if(gX11InFront==0)
         Bring_To_Front(0);
    }

    configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, ((shaded==WIN_STATE_SHADED) || (c->is_icon!=NOT_ICONBAR)), False); 

//    redraw_client_frame(c);

    if(c->is_icon!=NOT_ICONBAR) {
//      lower_client(c);
//      printf("Finished iconbar-ing... %d\n",iconbar_state);
      newc=get_top_most_client(WIN_LAYER_NORMAL,False);
      if( newc==NULL ) {
        set_focus_client(NULL);
      } else if( newc->is_icon!=NOT_ICONBAR ) {
        set_focus_client(NULL);
      } else
        set_focus_client(newc);

      ptrwin=get_pointer_win(root);
      
      if(ptrwin==c->frame) {
//        printf("over new iconbar: %#lx\n",c);
        // We've ended up with the cursor over the new iconbar, so make sure iconbars don't disappear!
        if(iconbar_state==ICONBAR_HIDDEN)
          slide_all_iconbars(c);
        else
          lastIconbarOutTicks=0;
        // ...and that it lights up
        c->is_icon=ICONBAR_RAISED;
        // ...and the width expands to see the whole text, if needed
        if( c->icontxt_w>=TW-6 )
          resize_client(c,c->icontxt_w+6-LW-RW+TLW+TRW,c->icon_h,True);
//        redraw_client_frame(c);
        raise_client(c);
        overicon=c;
        if(can_send_hackcmds) c->is_e_m_r |= IS_ENTERED;
      } else {
        ac=get_client_from_window(ptrwin,FRAME_LE);
//        printf("over client: %#lx , ",ac);
        if(ac!=NULL) {
          if(ac->is_icon==NOT_ICONBAR) {
//            printf("not an iconbar...\n");
            if(iconbar_state==ICONBAR_VISIBLE) {
              c->is_icon=ICONBAR_VISIBLE;
              slide_all_iconbars(NULL);
            } else {
              c->is_icon=ICONBAR_VISIBLE;
              toggle_slide_iconbar(c);
            }
          } else {
//            printf("is an iconbar...\n");
            // we've ended up with the cursor over another iconbar, so make sure iconbars don't disappear
            if(iconbar_state!=ICONBAR_VISIBLE) {
              c->is_icon=ICONBAR_VISIBLE;
              slide_all_iconbars(ac);
            } else
              lastIconbarOutTicks=0;
          }
        } else {
//          printf("iconbar_state: %d\n",iconbar_state);
          if(iconbar_state==ICONBAR_VISIBLE) {
            c->is_icon=ICONBAR_VISIBLE;
            slide_all_iconbars(NULL);
          } else {
            c->is_icon=ICONBAR_VISIBLE;
            toggle_slide_iconbar(c);
          }
        }
      }
      
      newc=get_top_most_client(WIN_LAYER_NORMAL,False);

    } else {
//      printf("Finished un-iconbar-ing...  \n");
      raise_client(c);
      set_focus_client(c);
      if( shaded==WIN_STATE_SHADED )
        toggle_client_shaded(c);
// don't need this code any more...?
//      ptrwin=get_pointer_win(root);
//      printf("ptrwin,frame: %#lx %#lx\n",ptrwin,c->frame);
//      if(ptrwin==c->frame) {
//        printf("iconbar_state: %d\n",iconbar_state);
//        will_hide_all_iconbars();
//      }
    }
} 
 
#ifdef LAYER_SUPPORT 
/* 
 * 
 * Set client layer 
 * 
 */ 

void set_client_layer(Client *c, int layer) 
{ 
  sprintf(lastsub,"TRACE2: set_client_layer (client,layer: %#lx,%d)\n",c,layer); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: set_client_layer (client,layer: %#lx,%d)\n",c,layer); 
#endif 
 
  set_gnome_hint(c->window, WIN_LAYER, layer); 
  if(get_wm_state(c->window)==NormalState) 
    if(c->is_icon==NOT_ICONBAR)
      raise_client(c); 
} 
 
/* 
 * 
 * Get client layer 
 * 
 */ 

int get_client_layer(Client *c) 
{ 
  sprintf(lastsub,"TRACE2: get client layer (client: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: get client layer (client: %#lx)\n",c); 
#endif 
 
  return get_gnome_hint(c->window, WIN_LAYER); 
} 
#endif /* LAYER */ 
 
/* 
 * 
 * Delete client 
 * 
 */ 

void delete_client(Client *c, int remap) 
{ 
  sprintf(lastsub,"TRACE1: delete_client (client [window],lastfocus, remap: %#lx [%#lx],%#lx,%d)\n",c,c->window,lastfocus,remap);
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: delete_client (client [window],lastfocus, remap: %#lx [%#lx],%#lx,%d)\n",c,c->window,lastfocus,remap);
#endif

#ifdef SHOW_WEIRD
  if(c==NULL)
    printf("%s oddity: got NULL client in delete_client!\n",PROGNAME);
#endif
  if(c==NULL) return;

#ifdef DEBUG
  printf("  deleting client [%#lx] %#lx\n",c->frame,c->window); 
#endif 

  MyIncreaseXServerGrab();
  
// make sure it has the right state...
  if(c->is_icon!=NOT_ICONBAR && remap==1) {
    if((get_gnome_hint(c->window,WIN_STATE) | WIN_STATE_SHADED)==WIN_STATE_SHADED)
      set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_SHADED); 
    c->ignore_unmap++; 
    XUnmapWindow(dpy, c->frame); 
    XUnmapWindow(dpy, c->window); 
    set_wm_state(c->window, IconicState);
    printf("Set window (%#lx, %s) to IconicState\n",c->window,c->name);
  }
  
  if(remap!=-1) {
    // The get_window_parent produces a BadWindow error, so inhibit it
    inhibit_errors=True;
    /* Check if frame is still the clients parent */
    if(get_window_parent(c->window)==c->frame) 
    { 
      inhibit_errors=False;
      /* Reparent our client back to root if it still exists */ 
      gravitate_client(c, REMOVE_GRAVITY); 
      XReparentWindow(dpy, c->window, root, c->x, c->y);
      if(c->is_icon!=NOT_ICONBAR || c->is_docked) {
        set_wm_state(c->window, IconicState);
        printf("Set window (%#lx, %s) to IconicState\n",c->window,c->name);
      } else {
        set_wm_state(c->window, WithdrawnState); 
        printf("Set window (%#lx, %s) to WithdrawnState\n",c->window,c->name);
      }

#ifdef KEYBOARD_SUPPORT
      /* Remove any key grabs on this window */
      ungrab_keys(c->window);
#endif
 
      /* Remap is 1 if we are quitting OroborOSX (and -1 if window has vanished -but not here) */ 
      if(remap==0 || launched_xdarwin!=0) // don't bother remapping if we launched XDarwin, since it will quit
        XUnmapWindow(dpy, c->window); 
      else
        XMapWindow(dpy, c->window); 
    } else
      inhibit_errors=False;
  }
  
#ifdef DEBUG
  printf("  checking lastenter: [%#lx], [%#lx]\n",c->frame,lastenter);
  if( c->frame==lastenter ) printf("  window was last to be entered: [%#lx]\n",lastenter);
  printf("  checking lastc: [%#lx], [%#lx]\n",c,lastc);
  if( c==lastc ) printf("  window was lastc: [%#lx]\n",lastc);
  if(focus_before_hide==c) printf("  window was focus_before_hide: [%#lx]\n",focus_before_hide);
  if(lastfocus_client==c) printf("  window was lastfocus_client: [%#lx]\n",lastfocus_client);
#endif
  if( c->frame==lastenter ) lastenter=0;
  if( c==last_client_entered_hack ) last_client_entered_hack=NULL;
  if( c==lastc ) lastc=NULL;
  if(lastfocus_client==c) lastfocus_client=NULL;
  // special flag to check if the focus window has vanished while hidden (OK, bad coding, but it works...)
  if(focus_before_hide==c) focus_before_hide=-1;

  if(c->osxwinfo.winref!=0)
    DestroyOSXWindow(&(c->osxwinfo));
    
  if(c->parent)
    remove_sibling_client(c);
    
  if(c->icon_y==nexticon_y-config.top.height)
    nexticon_y-=config.top.height;
	 
  /* Get rid of our frame window */ 
  XDestroyWindow(dpy, c->frame); 
 
  /* Remove reference from our client list */ 
  if(client_count==1) 
    client_list=NULL; 
  else 
  { 
    c->next->prev = c->prev; 
    c->prev->next = c->next; 
    if(c==client_list) 
      client_list=client_list->next; 
  } 
 
  if(c->name) XFree(c->name); 
  if(c->clientcmd) XFree(c->clientcmd); 
  if(c->clientclass) XFree(c->clientclass); 
  if(c->clientgroup) XFree(c->clientgroup); 
  if(c->size) XFree(c->size); 
  free(c); 
 
  client_count--; 
  update_gnome_client_list();

  MyDecreaseXServerGrab();

  sprintf(lastsub,"TRACE2: delete_client -client removed\n");
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: delete_client -client removed\n");
#endif
} 
 
/* 
 * 
 * Send close message to client window 
 * 
 */ 

void close_client(Client *c) 
{ 
  int i, n, found=False; 
  Atom *protocols=NULL; 
  XEvent e; 
 
  sprintf(lastsub,"TRACE1: close_client (client: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: close_client (client: %#lx)\n",c); 
#endif 
 
#ifdef SHOW_WEIRD
  if(c==NULL) printf("%s oddity: got NULL client in close_client!\n",PROGNAME);
#endif
  if(c==NULL) return;
  
  if(XGetWMProtocols(dpy, c->window, &protocols, &n)) 
  { 
    for(i=0; i<n; i++) 
      if(protocols[i]==icccm[WM_DELETE]) found=True; 
    if(protocols) XFree(protocols);  
  }
  
  if(found) 
  { 
#ifdef DEBUG
    printf("  sending delete event to client [%#lx] %#lx\n",c->frame,c->window);
#endif
    e.type=ClientMessage; 
    e.xclient.window=c->window; 
    e.xclient.message_type=icccm[WM_PROTOCOLS]; 
    e.xclient.format=32; 
    e.xclient.data.l[0]=icccm[WM_DELETE]; 
    e.xclient.data.l[1]=CurrentTime; 
    XSendEvent(dpy, c->window, False, NoEventMask, &e); 
  } 
  else  {
#ifdef DEBUG
    printf("  killing client [%#lx] %#lx\n",c->frame,c->window);
#endif
    XKillClient(dpy, c->window); 
  }
} 
 
/* 
 * 
 * Redraw client borders 
 * 
 */ 

void redraw_client_frame(Client *c) 
{ 
  int state=FRAME_INACTIVE, shaded=0, wasalpha; 
  
  sprintf(lastsub,"TRACE3: redraw_client_frame (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: redraw_client_frame (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
#endif 

#ifdef SHOW_WEIRD
  if(c==NULL) printf("%s oddity: got NULL client in redraw_client_frame!\n",PROGNAME);
#endif
  if(c==NULL) return;

#ifdef DEBUG 
  printf("  redraw_client_frame...  [frame: %#lx] window: %#lx\n",c->frame,c->window); 
#endif 
 
#ifdef SHOW_WEIRD 
  if(!c->has_frame) printf("%s oddity: client/window (%#lx,%#lx) has no frame - no frame drawing done\n",PROGNAME,c,c->window); 
#endif 
  if(!c->has_frame) return; 
  
  MyIncreaseXServerGrab();
//  if(c==get_focus_client() && gX11InFront!=0)
  wasalpha=c->osxwinfo.drawalpha;
#ifdef DEBUG
  printf("wasalpha at start of frame: %d\n",wasalpha);
#endif
  if(c==lastfocus_client && gX11InFront!=0) {
    state=FRAME_ACTIVE;
#ifdef DEBUG
    printf("redraw_client_frame telling XDarwin to set alpha to %d\n",config.active_titlebar_alpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,config.active_titlebar_alpha,0,0,0,0);
  } else if(c->is_icon==NOT_ICONBAR) {
#ifdef DEBUG
    printf("redraw_client_frame telling XDarwin to set alpha to %d\n",config.inactive_titlebar_alpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,config.inactive_titlebar_alpha,0,0,0,0);
  }
  
/*  if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED)*/
  if( c->is_icon==ICONBAR_RAISED ) {
    shaded=4;
    state=FRAME_ACTIVE;
  } else if( c->is_icon==ICONBAR_HIDDEN || c->is_icon==ICONBAR_VISIBLE ) {
    shaded=4;
    state=FRAME_INACTIVE;
  } else if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED) {
    shaded=2;
  }
 
  if(c->is_dimmed==1)
    SendHackCommand(c,1,HACKCMD_DIMWIN,2,0,0,0,0); // temporarily switch off dimming -no reshape nor redraw!
  
  if(c->is_icon!=NOT_ICONBAR && TW<1) {
    // this iconbar is 'hidden' (i.e. it is shrunk to a small tab on one side of the screen)
    if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_BOTLEFT) {
      // iconbars on left, so try to only do the right side of the iconbar tab
      if(c->top_right) {
        XSetWindowBackgroundPixmap(dpy, c->top_right, config.top_right.pixmap[state+shaded]);
        XClearWindow(dpy, c->top_right);
      } else if(c->top_left) {
        XSetWindowBackgroundPixmap(dpy, c->top_left, config.top_left.pixmap[state+shaded]);
        XClearWindow(dpy, c->top_left);
      }
    } else {
      // iconbars on right, so try to only do the left side of the iconbar tab
      if(c->top_left) {
        XSetWindowBackgroundPixmap(dpy, c->top_left, config.top_left.pixmap[state+shaded]);
        XClearWindow(dpy, c->top_left);
      } else if(c->top_right) {
        XSetWindowBackgroundPixmap(dpy, c->top_right, config.top_right.pixmap[state+shaded]);
        XClearWindow(dpy, c->top_right);
      }
    }
  } else {
    if(c->top_left) {
      XSetWindowBackgroundPixmap(dpy, c->top_left, config.top_left.pixmap[state+shaded]); 
      XClearWindow(dpy, c->top_left); }
    if(c->top_right) {
      XSetWindowBackgroundPixmap(dpy, c->top_right, config.top_right.pixmap[state+shaded]);
      XClearWindow(dpy, c->top_right); }
  }
  
  if( c->is_icon==NOT_ICONBAR ) {

    if(c->left) { XSetWindowBackgroundPixmap(dpy, c->left, config.left.pixmap[state]); 
    XClearWindow(dpy, c->left); }
    if(c->right) { XSetWindowBackgroundPixmap(dpy, c->right, config.right.pixmap[state]); 
    XClearWindow(dpy, c->right); }
    if(c->bottom_left) { XSetWindowBackgroundPixmap(dpy, c->bottom_left, config.bottom_left.pixmap[state]); 
    XClearWindow(dpy, c->bottom_left); }
    if(c->bottom) { XSetWindowBackgroundPixmap(dpy, c->bottom, config.bottom.pixmap[state]); 
    XClearWindow(dpy, c->bottom); }
    if(c->resize) { XSetWindowBackgroundPixmap(dpy, c->resize, config.resize.pixmap[state]); 
    XClearWindow(dpy, c->resize); }
 
    redraw_client_buttons(c); 
  }

  if( TW>0 ) redraw_client_title(c);
  
  if(c->is_dimmed==1)
    SendHackCommand(c,1,HACKCMD_DIMWIN,3,0,0,0,0); // switch on dimming again -no reshape nor redraw!

//  if(c->osxwinfo.drawalpha!=wasalpha)
#ifdef DEBUG
  printf("wasalpha in frame: %d\n",wasalpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,wasalpha,0,0,0,0);

  MyDecreaseXServerGrab();

  sprintf(lastsub,"   done redraw_client_title\n");
  subtrace();
#ifdef SHOW_TRACE3
  printf("   done redraw_client_title\n");
#endif 

//  XFlush(dpy);
//  XSync(dpy,0);
  
  sprintf(lastsub,"TRACE3: leaving redraw_client_frame\n");
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: leaving redraw_client_frame\n");
#endif 
} 
 
int check_correct_button_state(int but_state, int frame_state)
{
    if(but_state==BUTTON_OVER) {
        if(frame_state==FRAME_INACTIVE)
            return BUTTON_OVERBACK;
    } else if(but_state==BUTTON_OVERBACK) {
        if(frame_state==FRAME_ACTIVE)
            return BUTTON_OVER;
    } else if(but_state==BUTTON_PRESSED) {
        if(frame_state==FRAME_INACTIVE)
            return BUTTON_PRESSEDBACK;
    } else if(but_state==BUTTON_PRESSEDBACK) {
        if(frame_state==FRAME_ACTIVE)
            return BUTTON_PRESSED;
    } else if(but_state==BUTTON_ACTIVEOFF) {
        if(frame_state==FRAME_INACTIVE)
            return BUTTON_INACTIVEOFF;
    } else if(but_state==BUTTON_INACTIVEOFF) {
        if(frame_state==FRAME_ACTIVE)
            return BUTTON_ACTIVEOFF;
    }
    return but_state;
}

/* 
 *  
 * Redraw client buttons 
 * 
 */ 

void redraw_client_buttons(Client *c) 
{ 
  int state=FRAME_INACTIVE, mainstate, wasalpha; 
 
//  sprintf(lastsub,"TRACE3: redraw_client_buttons (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
//  subtrace();
//#ifdef SHOW_TRACE3 
//  printf("TRACE3: redraw_client_buttons (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
//#endif

#ifdef SHOW_WEIRD
  if(c==NULL) printf("%s oddity: got NULL client in redraw_client_buttons!\n",PROGNAME);
#endif
  if(c==NULL) return;

#ifdef DEBUG
  printf("redraw_client_buttons [frame: %#lx] window: %#lx\n",c->frame,c->window); 
#endif 

#ifdef SHOW_WEIRD 
  if(c->is_icon!=NOT_ICONBAR) printf("%s oddity: client/window (%#lx,%#lx) is an iconbar - no button drawing done\n",PROGNAME,c,c->window);
  if(!c->has_frame) printf("%s oddity: client/window (%#lx,%#lx) has no frame - no title drawing done\n",PROGNAME,c,c->window); 
#endif 
  if( c->is_icon!=NOT_ICONBAR ) return;
  if(!c->has_frame) return;
 
  MyIncreaseXServerGrab();

  if(c->is_dimmed==1)
    SendHackCommand(c,1,HACKCMD_DIMWIN,2,0,0,0,0); // temporarily switch off dimming -no reshape nor redraw!

//  if(c==get_focus_client() && gX11InFront!=0) 
  wasalpha=c->osxwinfo.drawalpha;
  if(c==lastfocus_client && gX11InFront!=0) {
    state=FRAME_ACTIVE;
#ifdef DEBUG
    printf("redraw_client_buttons telling XDarwin to set alpha to %d\n", config.active_titlebar_alpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,config.active_titlebar_alpha,0,0,0,0);
  } else {
#ifdef DEBUG
    printf("redraw_client_buttons telling XDarwin to set alpha to %d\n", config.inactive_titlebar_alpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,config.inactive_titlebar_alpha,0,0,0,0);
  }
 
  mainstate=state;
  
  if(c->icon) {
    c->icon_state=check_correct_button_state(c->icon_state,state);
    XSetWindowBackgroundPixmap(dpy, c->icon, config.icon.pixmap[c->icon_state==BUTTON_UNPRESSED?state:c->icon_state]); 
    XClearWindow(dpy, c->icon);
  }
  if(c->close) {
    c->close_state=check_correct_button_state(c->close_state,state);
    XSetWindowBackgroundPixmap(dpy, c->close, config.close.pixmap[c->close_state==BUTTON_UNPRESSED?state:c->close_state]); 
    XClearWindow(dpy, c->close);
  }
  if(c->can_max) {
//    if(c->zoom_state!=BUTTON_UNPRESSED) state=c->zoom_state;
    if(c->zoom_state!=BUTTON_UNPRESSED)
        state=check_correct_button_state(c->zoom_state,state);
  } else if(state==FRAME_ACTIVE)
    state=BUTTON_ACTIVEOFF;
  else if(state==FRAME_INACTIVE)
    state=BUTTON_INACTIVEOFF;
  if(c->zoom) {
    XSetWindowBackgroundPixmap(dpy, c->zoom, config.zoom.pixmap[state]);
    XClearWindow(dpy, c->zoom);
  }

  if(c->is_dimmed==1)
    SendHackCommand(c,1,HACKCMD_DIMWIN,3,0,0,0,0); // switch on dimming again -no reshape nor redraw!

//  if(wasalpha!=c->osxwinfo.drawalpha)
#ifdef DEBUG
  printf("wasalpha in buttons: %d\n",wasalpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,wasalpha,0,0,0,0);

  MyDecreaseXServerGrab();
} 
 
/* 
 * 
 * Redraw client name in the title bar 
 * 
 */ 

void redraw_client_title(Client *c) 
{ 
  int state=FRAME_INACTIVE, shaded=0; 
  Pixmap temp; 
  GC gc; 
  XGCValues gv;
  char *name_string=NULL;
#ifdef I18N_SUPPORT 
  XRectangle ink, logical; 
  int ascent=0, descent=0; 
#endif /* I18N */ 
  int width=0, leftext, rightext, left_add=0, right_add=0, wasalpha; 
 
  sprintf(lastsub,"TRACE3: redraw_client_title (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: redraw_client_title (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
#endif

#ifdef SHOW_WEIRD
  if(c==NULL)
    printf("%s oddity: got NULL client in redraw_client_title!\n",PROGNAME);
#endif
  if(c==NULL) return;

#ifdef SHOW_WEIRD
  if(c->title==0) printf("%s oddity: got zero title window in redraw_client_title!\n",PROGNAME);
#endif
  if(c->title==0) return;
  
#ifdef DEBUG 
  printf("redraw_client_title [%#lx] %#lx (%#lx)\n",c->frame,c->window,lastfocus); 
#endif 
 
#ifdef SHOW_WEIRD 
  if(!c->has_frame && (c->is_icon==NOT_ICONBAR)) printf("%s oddity: client/window (%#lx,%#lx) has no frame - no title drawing done\n",PROGNAME,c,c->window); 
#endif 
  if(!c->has_frame && (c->is_icon==NOT_ICONBAR)) return;

  if(TW<1) printf("Not drawn title of client %#lx: %d\n",c,TW);
  if(TW<1) return;
  
  MyIncreaseXServerGrab();

  if(c->is_dimmed==1)
    SendHackCommand(c,1,HACKCMD_DIMWIN,2,0,0,0,0); // temporarily switch off dimming -no reshape nor redraw!

//  if(c==get_focus_client() && gX11InFront!=0) 
  wasalpha=c->osxwinfo.drawalpha;
  if(c==lastfocus_client && gX11InFront!=0) {
    state=FRAME_ACTIVE;
#ifdef DEBUG
    printf("redraw_client_title telling XDarwin to set alpha to %d\n", config.active_titlebar_alpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,config.active_titlebar_alpha,0,0,0,0);
  } else if(c->is_icon==NOT_ICONBAR) {
#ifdef DEBUG
    printf("redraw_client_title telling XDarwin to set alpha to %d\n", config.inactive_titlebar_alpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,config.inactive_titlebar_alpha,0,0,0,0);
  }
  
/*  if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED)*/
  if( c->is_icon==ICONBAR_VISIBLE || c->is_icon==ICONBAR_HIDDEN ) {
    shaded=4;
    state=FRAME_INACTIVE;
  } else if( c->is_icon==ICONBAR_RAISED ) {
    shaded=4;
    state=FRAME_ACTIVE;
  } else if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED) {
    shaded=2;
  }
 
  gv.function=GXcopy; 
  gv.tile=config.top.pixmap[state+shaded]; 
  gv.fill_style=FillTiled; 
  gc=XCreateGC(dpy, root, GCFunction|GCFillStyle|GCTile, &gv); 
  temp=XCreatePixmap(dpy, root, TW, TH, DefaultDepth(dpy, screen)); 
  XFillRectangle(dpy, temp, gc, 0, 0, TW, TH); 
 
  if((c->is_e_m_r & IS_RESIZING)==IS_RESIZING) {
    if(c->name) {
      name_string=malloc(strlen(c->name)+20);
      sprintf(name_string,"%s [%dx%d]",c->name,c->width,c->height);
    } else {
      name_string=malloc(20);
      sprintf(name_string,"[%dx%d]",c->width,c->height);
    }
    c->is_e_m_r |= SHOWING_SIZE;
  } else if((c->is_e_m_r & IS_MOVING)==IS_MOVING) {
    if(c->name) {
      name_string=malloc(strlen(c->name)+20);
      sprintf(name_string,"%s (%d,%d)",c->name,FX,FY);
    } else {
      name_string=malloc(20);
      sprintf(name_string,"(%d,%d)",FX,FY);
    }
    c->is_e_m_r |= SHOWING_POSITION;
  } else if(c->name) {
    name_string=strdup(c->name);
    c->is_e_m_r &= NOT_SHOWING_SIZEPOS;
  }
   
  if(name_string) 
  { 
    int posx=0;
    int posy=0; 
    
    while(width==0) {
    
#ifdef I18N_SUPPORT
    if(use_fontset)
    {
      XmbTextExtents(config.fontset, name_string, strlen(name_string), &ink, &logical); 
      ascent=-logical.y; 
      descent=logical.height+logical.y; 
      width=logical.width; 
    }
    else
      width=XTextWidth(config.font, name_string, strlen(name_string));
#else  
    width=XTextWidth(config.font, name_string, strlen(name_string)); 
#endif /* I18N */ 
 
//    if( c->is_icon==NOT_ICONBAR ) {
      left_add=TLW;
      right_add=TRW;
//    }
    
    switch(config.text_align) 
    { 
    case ALIGNMENT_MIDDLE: 
//      posx=width>TW?0:(FW-width)/2; 
      posx=(FW-width)/2; 
      break; 
    case ALIGNMENT_RIGHT: 
//      posx=width>TW?0:FW-right_add-width; 
      posx=FW-right_add-width; 
      break; 
    case ALIGNMENT_LEFT: 
      posx=2+left_add; 
      break; 
    }
    
    leftext=left_add;
    rightext=FW-right_add;
    if(config.close.x>0) {
      if(config.close.x+config.close.width>=leftext)
        leftext=config.close.x+config.close.width-1;
    } else if(CX<rightext)
      rightext=CX;
    if(config.icon.x>0) {
      if(config.icon.x+config.icon.width>=leftext)
        leftext=config.icon.x+config.icon.width-1;
    } else if(IX<rightext)
      rightext=IX;
    if(config.zoom.x>0) {
      if(config.zoom.x+config.zoom.width>=leftext)
        leftext=config.zoom.x+config.zoom.width-1;
    } else if(HX<rightext)
      rightext=HX;
    
//    leftext=leftext-TLW;
//    rightext=rightext-TLW;
    
    // iconbars have text on the left - also, deal with case where there is not enough space
    if( c->is_icon!=NOT_ICONBAR )
       posx=2+left_add;
    else {
      if(posx>rightext-width)
        posx=rightext-width;
      if( posx<leftext )
        posx=leftext;
    }
        
    if(strlen(name_string)>12 && posx==leftext && posx+width>rightext) {
      if((c->is_e_m_r & IS_RESIZING)==IS_RESIZING) {
        free(name_string);
        name_string=malloc(20);
        sprintf(name_string,"[%dx%d]",c->width,c->height);
        width=0; // make it go back and check width again
      } else if((c->is_e_m_r & IS_MOVING)==IS_MOVING) {
        free(name_string);
        name_string=malloc(20);
        sprintf(name_string,"(%d,%d)",FX,FY);
        width=0; // make it go back and check width again
      }
    }
    
    } /* while(width) */
    
#ifdef I18N_SUPPORT
    if(use_fontset)
      posy=(TH-descent)-((TH-ascent-descent)/2);
    else
      posy=(TH-config.font->descent)-((TH-config.font->ascent-config.font->descent)/2);
#else 
    posy=(TH-config.font->descent)-((TH-config.font->ascent-config.font->descent)/2); 
#endif /* I18N */ 

    if( c->is_icon!=NOT_ICONBAR ) {
      if( c->icontxt_w==width ) {
        // text width has not changed, so leave position as it was
         posx=c->icontxt_x;
         posy=c->icontxt_y;
      } else {
         c->icontxt_x=posx;
         c->icontxt_y=posy;
         c->icontxt_w=width; }
    } else if( c->name_w==width && c->name_x==posx && c->name_y==posy ) {
        // text width has not changed, so leave position as it was
         posx=c->name_x;
         posy=c->name_y;
    } else {
         c->name_x=posx;
         c->name_y=posy;
         c->name_w=width;
    }

    posx -= left_add;
//    if(width+3<TW) {

//      posx -= TLW;
      
      if(config.title.height>0 && config.title.width>0) {
        if(config.title_left.width==0 || config.title_right.width==0 || config.title_left.height==0 || config.title_right.height==0) {
          XFreeGC(dpy, gc); 
          gv.function=GXcopy; 
          gv.tile=config.title.pixmap[state+shaded]; 
          gv.fill_style=FillTiled; 
          gc=XCreateGC(dpy, root, GCFunction|GCFillStyle|GCTile, &gv);
          XFillRectangle(dpy, temp, gc, posx-2, (config.top.height-config.title.height)/2, width+4, config.title.height);
        } else {
          XFreeGC(dpy, gc); 
          gv.function=GXcopy; 
          gv.tile=config.title_left.pixmap[state+shaded]; 
          gv.fill_style=FillTiled;
          gv.ts_x_origin=posx-config.title_left.width;
          gv.ts_y_origin=(config.top.height-config.title_left.height)/2;
          gc=XCreateGC(dpy, root, GCFunction|GCFillStyle|GCTile|GCTileStipXOrigin|GCTileStipYOrigin, &gv);
          XFillRectangle(dpy, temp, gc, gv.ts_x_origin, gv.ts_y_origin, config.title_left.width, config.title_left.height);          
          XFreeGC(dpy, gc); 
          gv.function=GXcopy; 
          gv.tile=config.title.pixmap[state+shaded]; 
          gv.fill_style=FillTiled; 
          gv.ts_x_origin=posx;
          gv.ts_y_origin=(config.top.height-config.title.height)/2;
          gc=XCreateGC(dpy, root, GCFunction|GCFillStyle|GCTile|GCTileStipXOrigin|GCTileStipYOrigin, &gv);
          XFillRectangle(dpy, temp, gc, gv.ts_x_origin, gv.ts_y_origin, width, config.title.height);          
          XFreeGC(dpy, gc); 
          gv.function=GXcopy; 
          gv.tile=config.title_right.pixmap[state+shaded]; 
          gv.fill_style=FillTiled; 
          gv.ts_x_origin=posx+width;
          gv.ts_y_origin=(config.top.height-config.title_right.height)/2;
          gc=XCreateGC(dpy, root, GCFunction|GCFillStyle|GCTile|GCTileStipXOrigin|GCTileStipYOrigin, &gv);
          XFillRectangle(dpy, temp, gc, gv.ts_x_origin, gv.ts_y_origin, config.title_right.width, config.title_right.height);          
        }
      }
    
#ifdef I18N_SUPPORT
      if(use_fontset)
        XmbDrawString(dpy, temp, config.fontset, config.font_gc[state], posx, posy, name_string, strlen(name_string));
      else
        XDrawString(dpy, temp, config.font_gc[state], posx, posy, name_string, strlen(name_string));
#else 
      XDrawString(dpy, temp, config.font_gc[state], posx, posy, name_string, strlen(name_string)); 
#endif /* I18N */ 

//    }
    
 // need to make sure dock icon's window list is up to date
    if(c->osxwinfo.winref!=0 && mouse_mode!=MOUSE_MOVE && mouse_mode!=MOUSE_RESIZE)
      CheckOSXWinTitle(&(c->osxwinfo),name_string,c->clientgroup);
      
    free(name_string);
  }
  
  XSetWindowBackgroundPixmap(dpy, c->title, temp); 
  XClearWindow(dpy, c->title); 
 
  XFreeGC(dpy, gc); 
  XFreePixmap(dpy, temp); 

  if(c->is_dimmed==1)
    SendHackCommand(c,1,HACKCMD_DIMWIN,3,0,0,0,0); // switch on dimming again -no reshape nor redraw!

//  if(wasalpha!=c->osxwinfo.drawalpha)
#ifdef DEBUG
  printf("wasalpha in title: %d\n",wasalpha);
#endif
    SendHackCommand(c,1,HACKCMD_SETDRAWALPHA,wasalpha,0,0,0,0);

  MyDecreaseXServerGrab();
} 
 
#ifdef SHAPE_SUPPORT 
/* 
 * 
 * Set client window shape 
 * 
 */ 

void set_client_shape(Client *c, int inhibit) 
{ 
  Window temp,wintmp;
  int i, doneset=0;	// doneset makes sure that ShapeSet is applied the first time, then ShapeUnion
 
  sprintf(lastsub,"TRACE3: set_client_shape (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: set_client_shape (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
#endif

#ifdef SHOW_WEIRD
  if(c==NULL) printf("%s oddity: got NULL client in set_client_shape!\n",PROGNAME);
#endif
  if(c==NULL) return;

#ifdef DEBUG
  printf("set_client_shape [frame: %#lx] window: %#lx\n",c->frame,c->window); 
#endif 
 
  if(c->is_docked) return;
//    if(undock_client(c)) return;
 
  if(shape) {
  
    MyIncreaseXServerGrab();

    if(inhibit==0 || inhibit==1)
    {

      temp=XCreateSimpleWindow(dpy, root, 0, 0, FW, FH, 0, 0, 0);
      doneset=ShapeSet;

      if(!c->has_frame && (c->is_icon==NOT_ICONBAR))
      {
        if(inhibit==0 && c->is_shaped) {
          XShapeCombineShape(dpy, temp, ShapeBounding, LW, TH, c->window, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
      }
      else if( c->is_icon!=NOT_ICONBAR)
      {
        // Just shape titlebar parts for an iconbar
        if(TW<1) { // this probably implies it's only an iconbar tab...
          if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
            if(c->top_left) { // try to apply top-left shape
              XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[2], ShapeSet);
              XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, doneset);
              doneset=ShapeUnion;
            } else if(c->top_right) { // failed -try to use top-right
              XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[2], ShapeSet);
              XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, doneset);
              doneset=ShapeUnion;
            }
          } else {
            if(c->top_right) { // try to apply top-right shape
              XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[2], ShapeSet);
              XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, doneset);
              doneset=ShapeUnion;
            } else if(c->top_left) { // failed -try to use top-left
              XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[2], ShapeSet);
              XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, doneset);
              doneset=ShapeUnion;
            }
          }
        } else {
          if(c->top_left) { // try to apply top-left shape
            XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[2], ShapeSet);
            XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, doneset);
            doneset=ShapeUnion;
            if(c->top_right) { // and top-right shape
              XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[2], ShapeSet);
              XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, doneset);
              doneset=ShapeUnion;
            }
          } else if(c->top_right) { // no top-left, so just try to apply top-right shape
            XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[2], ShapeSet);
            XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, doneset);
            doneset=ShapeUnion; 
          }
          if(c->title && config.top.width>0 && config.top.height>0) { // try to apply title shape
            XShapeCombineMask(dpy, c->title, ShapeBounding, 0, 0, config.top.mask[2], ShapeSet);
            for(i=config.top.width;i<TRX-TX;i+=config.top.width)
              XShapeCombineMask(dpy, c->title, ShapeBounding, i, 0, config.top.mask[2], ShapeUnion);
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, TRX-TX, TH, 0, 0, 0);
            XShapeCombineShape(dpy, c->title, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
            XShapeCombineShape(dpy, temp, ShapeBounding, TX, TY, c->title, ShapeBounding, doneset);
            doneset=ShapeUnion;
          }
        }
      }
      else if( (get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)==WIN_STATE_SHADED )
      { // window is shaded, so only shape titlebar, starting with the actual subwindows
        XShapeCombineMask(dpy, c->zoom, ShapeBounding, 0, 0, config.zoom.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->icon, ShapeBounding, 0, 0, config.icon.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->close, ShapeBounding, 0, 0, config.close.mask[0], ShapeSet);
        if(c->top_left) XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[1], ShapeSet);
        if(c->top_right) XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[1], ShapeSet);
        if(config.top.width>0 && config.top.height>0 && TRX>TX) {
          XShapeCombineMask(dpy, c->title, ShapeBounding, 0, 0, config.top.mask[1], ShapeSet);
          for(i=config.top.width;i<TRX-TX;i+=config.top.width)
            XShapeCombineMask(dpy, c->title, ShapeBounding, i, 0, config.top.mask[1], ShapeUnion);
          wintmp=XCreateSimpleWindow(dpy, root, 0, 0, TRX-TX, TH, 0, 0, 0);
          XShapeCombineShape(dpy, c->title, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
          XDestroyWindow(dpy, wintmp);
        }
        // now try to apply subwindow shapes to the entire window
        if(c->top_left) {
          XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
        if(c->top_right) {
          XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
        if(c->close) {
          XShapeCombineShape(dpy, temp, ShapeBounding, CX, CY, c->close, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
        if(c->icon) {
          XShapeCombineShape(dpy, temp, ShapeBounding, IX, IY, c->icon, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
        if(c->zoom) {
          XShapeCombineShape(dpy, temp, ShapeBounding, HX, HY, c->zoom, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
        if(c->title) {
          XShapeCombineShape(dpy, temp, ShapeBounding, TX, TY, c->title, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
      }
      else if(inhibit==0 || inhibit==1)
      { // The entire window is visible (i.e. not shaded nor an iconbar)
        if(inhibit==0) { // Make sure the client window's shape is applied
          XShapeCombineShape(dpy, temp, ShapeBounding, LW, TH, c->window, ShapeBounding, doneset);
          doneset=ShapeUnion;
        } else { // don't apply the client window's shape, instead fill in the whole rect
          wintmp=XCreateSimpleWindow(dpy, root, 0, 0, c->width, c->height, 0, 0, 0);
          XShapeCombineShape(dpy, temp, ShapeBounding, LW, TH, wintmp, ShapeBounding, doneset);
          XDestroyWindow(dpy, wintmp);
          doneset=ShapeUnion;
        }
        // Shape the corner subwindows in case there are 'internal' shapes (eg, triangular resize widget)
        if(c->bottom_left) XShapeCombineMask(dpy, c->bottom_left, ShapeBounding, 0, 0, config.bottom_left.mask[0], ShapeSet);
        if(c->resize) XShapeCombineMask(dpy, c->resize, ShapeBounding, 0, 0, config.resize.mask[0], ShapeSet);
        if(c->top_left) XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[0], ShapeSet);
        if(c->top_right) XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[0], ShapeSet);
        if(c->bottom_left) XShapeCombineMask(dpy, c->bottom_left, ShapeBounding, 0, 0, config.bottom_left.mask[0], ShapeSet);
        // Shape the contents of the widget subwindows
        XShapeCombineMask(dpy, c->zoom, ShapeBounding, 0, 0, config.zoom.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->icon, ShapeBounding, 0, 0, config.icon.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->close, ShapeBounding, 0, 0, config.close.mask[0], ShapeSet);
        // Only shape the frame if we're using a shaped theme...
        if(gUseShapedTheme!=0) {
          // Shape the left edge frame subwindow
          if(config.left.width>0 && config.left.height>0 && BLY>LY && c->left) {
            XShapeCombineMask(dpy, c->left, ShapeBounding, 0, 0, config.left.mask[0], ShapeSet);
            for(i=config.left.height;i<BLY-LY;i+=config.left.height)
              XShapeCombineMask(dpy, c->left, ShapeBounding, 0, i, config.left.mask[0], ShapeUnion);
            // make sure the above shape does not extend beyond the left edge (i.e. not into bottom-left)
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, LW, BLY-LY, 0, 0, 0);
            XShapeCombineShape(dpy, c->left, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
          }
          // Shape the right edge frame window
          if(config.right.width>0 && config.right.height>0 && RSY>RY && c->right) {
            XShapeCombineMask(dpy, c->right, ShapeBounding, 0, 0, config.right.mask[0], ShapeSet);
            for(i=config.right.height;i<RSY-RY;i+=config.right.height)
              XShapeCombineMask(dpy, c->right, ShapeBounding, 0, i, config.right.mask[0], ShapeUnion);
            // make sure the above shape does not extend beyond the right edge (i.e. not into resize)
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, RW, RSY-RY, 0, 0, 0);
            XShapeCombineShape(dpy, c->right, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
          }
          // Shape the bottom edge frame window
          if(config.bottom.width>0 && config.bottom.height>0 && RSX>BX && c->bottom) {
            XShapeCombineMask(dpy, c->bottom, ShapeBounding, 0, 0, config.bottom.mask[0], ShapeSet);
            for(i=config.bottom.width;i<RSX-BX;i+=config.bottom.width)
              XShapeCombineMask(dpy, c->bottom, ShapeBounding, i, 0, config.bottom.mask[0], ShapeUnion);
            // make sure the above shape does not extend beyond the bottom edge (i.e. not into resize)
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, RSX-BX, BH, 0, 0, 0);
            XShapeCombineShape(dpy, c->bottom, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
          }
          // Shape the title frame window
          if(config.top.width>0 && config.top.height>0 && TRX>TX) {
            XShapeCombineMask(dpy, c->title, ShapeBounding, 0, 0, config.top.mask[0], ShapeSet);
            for(i=config.top.width;i<TRX-TX;i+=config.top.width)
              XShapeCombineMask(dpy, c->title, ShapeBounding, i, 0, config.top.mask[0], ShapeUnion);
            // make sure the above shape does not extend beyond the title (i.e. not into top-right)
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, TRX-TX, TH, 0, 0, 0);
            XShapeCombineShape(dpy, c->title, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
          }
          // try to apply the widget shapes to the actual window
          if(c->close) {
            XShapeCombineShape(dpy, temp, ShapeBounding, CX, CY, c->close, ShapeBounding, doneset);
            doneset=ShapeUnion;
          }
          if(c->icon) {
            XShapeCombineShape(dpy, temp, ShapeBounding, IX, IY, c->icon, ShapeBounding, doneset);
            doneset=ShapeUnion;
          }
          if(c->zoom) {
            XShapeCombineShape(dpy, temp, ShapeBounding, HX, HY, c->zoom, ShapeBounding, doneset);
            doneset=ShapeUnion;
          }
        }
        else
        {
          if(config.left.width>0 && config.left.height>0 && BLY>LY && c->left) {
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, LW, BLY-LY, 0, 0, 0);
            XShapeCombineShape(dpy, c->left, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeSet); 
            XDestroyWindow(dpy, wintmp);
          }
          if(config.right.width>0 && config.right.height>0 && RSY>RY && c->right) {
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, RW, RSY-RY, 0, 0, 0);
            XShapeCombineShape(dpy, c->right, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeSet); 
            XDestroyWindow(dpy, wintmp);
          }
          if(config.bottom.width>0 && config.bottom.height>0 && RSX>BX && c->bottom) {
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, RSX-BX, BH, 0, 0, 0);
            XShapeCombineShape(dpy, c->bottom, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeSet); 
            XDestroyWindow(dpy, wintmp);
          }
          if(config.top.width>0 && config.top.height>0 && TRX>TX) {
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, TRX-TX, TH, 0, 0, 0);
            XShapeCombineShape(dpy, c->title, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeSet); 
            XDestroyWindow(dpy, wintmp);
          }
        }
        // Now apply all of the shaped frame subwindows to the actual window
        if(c->title) {
          XShapeCombineShape(dpy, temp, ShapeBounding, TX, TY, c->title, ShapeBounding, doneset);  
          doneset=ShapeUnion;
        }
        if(c->left) {
          XShapeCombineShape(dpy, temp, ShapeBounding, LX, LY, c->left, ShapeBounding, doneset); 
          doneset=ShapeUnion;
        }
        if(c->right) {
          XShapeCombineShape(dpy, temp, ShapeBounding, RX, RY, c->right, ShapeBounding, doneset); 
          doneset=ShapeUnion;
        }
        if(c->bottom) {
          XShapeCombineShape(dpy, temp, ShapeBounding, BX, BY, c->bottom, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
        if(c->top_left) {
          XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, doneset);    
          doneset=ShapeUnion;
        }
        if(c->top_right) {
          XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, doneset);    
          doneset=ShapeUnion;
        }
        if(c->resize) {
          XShapeCombineShape(dpy, temp, ShapeBounding, RSX, RSY, c->resize, ShapeBounding, doneset);    
          doneset=ShapeUnion;
        }
        if(c->bottom_left) {
          XShapeCombineShape(dpy, temp, ShapeBounding, BLX, BLY, c->bottom_left, ShapeBounding, doneset);
          doneset=ShapeUnion;
        }
      } 
    }
    else
      temp=XCreateSimpleWindow(dpy, root, 0, 0, XDisplayWidth(dpy, screen), XDisplayHeight(dpy, screen), 0, 0, 0);
//      temp=XCreateSimpleWindow(dpy, root, 0, 0, FW, FH, 0, 0, 0);

    XShapeCombineShape(dpy, c->frame, ShapeBounding, 0, 0, temp, ShapeBounding, ShapeSet); 
//    XShapeCombineMask(dpy, c->frame, ShapeBounding, 0, 0, temp, ShapeSet); 
    XDestroyWindow(dpy, temp);

    MyDecreaseXServerGrab();
  }
} 
#endif /* SHAPE */ 
 
#ifdef WORKSPACE_SUPPORT 
/* 
 * 
 * Toggle sticky/unsticky 
 * 
 */ 

void toggle_client_sticky(Client *c) 
{ 
  sprintf(lastsub,"TRACE2: toggle_client_sticky (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: toggle_client_sticky (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
 
#ifdef DEBUG 
  printf("toggle_client_sticky [frame: %#lx] window:%#lx (%#lx)\n",c->frame,c->window); 
#endif 
 
  set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_STICKY); 
} 
 
/* 
 * 
 * Move client to new workspace 
 * 
 */ 

void move_client_to_workspace(Client *c, int ws) 
{ 
  sprintf(lastsub,"TRACE2: move_client_to_workspace (client,ws: %#lx,%d)\n",c,ws); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: move_client_to_workspace (client,ws: %#lx,%d)\n",c,ws); 
#endif 

#ifdef SHOW_WEIRD
  if(ws==workspace || ws<0 || ws>=config.workspace_count)
    printf("%s oddity: got invalid workspace in move_client_to_workspace: %d\n",PROGNAME,ws);
#endif

  if(ws==workspace || ws<0 || ws>=config.workspace_count) return; 
  set_gnome_hint(c->window, WIN_WORKSPACE, ws); 
  change_to_workspace(ws); 
  if(c->is_icon==NOT_ICONBAR) {
    set_focus_client(c); 
    raise_client(c); }
} 
#endif /* WORKSPACE */ 

void manipulate_focus_client(int action)
{
  Client *c, *tc;
  
  sprintf(lastsub,"TRACE1: manipulate_focus_client (action: %d)\n",action); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: manipulate_focus_client (action: %d)\n",action); 
#endif 

//  c=get_focus_client();
  c=lastfocus_client;
  if(c==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: Got NULL on Window menu: %d\n",PROGNAME,action);
#endif
    return; }
  
  switch(action)
  {
    case itemWindowshade:
      toggle_client_shaded(c);
      break;
      
    case itemMinimise:
      if(c->is_icon==NOT_ICONBAR)
        toggle_minimise_client(c,0);
#ifdef SHOW_WEIRD
      else
        printf("%s oddity: Tried to minimise an iconbar!\n",PROGNAME);
#endif
      break;
      
    case itemZoom:
      toggle_client_maximized(c);
      break;
      
    case itemClose:
      close_client(c);
      break;
      
    case itemLower:
      tc=get_bottom_most_client(get_gnome_hint(c->window,WIN_LAYER),False);
      if(tc!=c) {
        lower_client(c);
        tc=get_top_most_client(get_gnome_hint(c->window,WIN_LAYER),False);
        if(tc!=c && tc!=NULL) {
          if(tc->is_docked)
            will_undock_client(tc);
          else
            set_focus_client(tc);
        }
      }
      break;

    case itemNext:
      get_next_osxwinref(1);
//      printf("Next is not yet functional\n");
//      lower_client(c);
      break;

    case itemFirstClick+100:
        c->auto_properties ^= NONAUTO_FIRSTCLICK;
        set_focus_client(c);
        break;
        
    case itemAutoFocus+100:
//#ifdef SHOW_WEIRD
//      printf("Autofocus not yet available\n");
//#endif
      c->auto_properties ^= AUTO_FOCUS;
      set_focus_client(c);
      break;
      
    case itemAutoRaise+100:
//#ifdef SHOW_WEIRD
//      printf("Autoraise not yet available\n");
//#endif
      c->auto_properties ^= AUTO_RAISE;
      set_focus_client(c);
      break;
      
    case itemClickRaise+100:
        c->auto_properties ^= AUTO_CLICKRAISE;
        set_focus_client(c);
        break;
        
    case itemFloat+100:
#ifdef SHOW_WEIRD
      printf("Floating palettes not yet available\n");
#endif
      break;

    case itemPopup+100:
#ifdef SHOW_WEIRD
      printf("Popup windows not yet available\n");
#endif
      break;
      
    case itemLiveResize+100:
      c->auto_properties ^= LIVE_RESIZING;
      set_focus_client(c);
      break;

    case itemTranslucent0+100:
      if(c->osxwinfo.translucency!=ACTIVE_TRANSLUCENT0) {
        printf("Making window non-translucent...\n");
        c->osxwinfo.translucency=ACTIVE_TRANSLUCENT0;
        c->osxwinfo.inactive_translucency=INACTIVE_TRANSLUCENT0;
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,ACTIVE_TRANSLUCENT0,0,0,0,0);
        Setup_Properties_Menu_Translucency(&c->osxwinfo);
      }
      break;

    case itemTranslucent1+100:
      if(c->osxwinfo.translucency!=ACTIVE_TRANSLUCENT1) {
        printf("Making window translucent...\n");
        c->osxwinfo.translucency=ACTIVE_TRANSLUCENT1;
        c->osxwinfo.inactive_translucency=INACTIVE_TRANSLUCENT1;
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,ACTIVE_TRANSLUCENT1,0,0,0,0);
        Setup_Properties_Menu_Translucency(&c->osxwinfo);
      }
      break;

    case itemTranslucent2+100:
      if(c->osxwinfo.translucency!=ACTIVE_TRANSLUCENT2) {
        printf("Making window very translucent...\n");
        c->osxwinfo.translucency=ACTIVE_TRANSLUCENT2;
        c->osxwinfo.inactive_translucency=INACTIVE_TRANSLUCENT2;
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,ACTIVE_TRANSLUCENT2,0,0,0,0);
        Setup_Properties_Menu_Translucency(&c->osxwinfo);
      }
      break;

/*    case itemSendToDock+100:
      dock_client(c);
      break;

    case itemGetFromDock+100:
      will_undock_client(c);
      break;
*/
    case itemNoframes+100:
#ifdef SHOW_WEIRD
      printf("No frames toggle not yet available\n");
#endif
      break;

    default:
#ifdef SHOW_WEIRD
      printf("%s oddity: Got funny action in manipulate_focus_client: %d\n",PROGNAME,action);
#endif
      break;
  }
}

void change_iconbar_menuspace()
{
  Client *c;
  int neary, newx, newy;
  long i;
  
  sprintf(lastsub,"TRACE1: change_iconbar_menuspace (gMenubarSpace: %d)\n",gMenubarSpace); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: change_iconbar_menuspace (gMenubarSpace: %d)\n",gMenubarSpace); 
#endif 

  if(gMenubarSpace==0)
    gMenubarSpace=21;
  else
    gMenubarSpace=-21;

  if(iconbarpos==ICONBARPOS_BOTLEFT || iconbarpos==ICONBARPOS_BOTRIGHT) {
    if(gMenubarSpace==-21) gMenubarSpace=0;
    return;
  }

  neary=nexticon_y;
  
  init_nearest_iconbar();
  c=get_nearest_iconbar(neary,NOT_ICONBAR);
  while(c!=NULL) {
    newx=c->icon_x;
    newy=c->icon_y+gMenubarSpace;
    if(c->is_icon!=NOT_ICONBAR)
      move_client(c,newx,newy,0);
    else {
      c->icon_x=newx;
      c->icon_y=newy;
    }
    c=get_nearest_iconbar(neary,NOT_ICONBAR);
  }
    
  nexticon_y+=gMenubarSpace;
  
  if(gMenubarSpace==-21) gMenubarSpace=0;
}

void change_iconbar_screen(int newhead)
{
  Client *c;
  int neary, newx, newy, dx, dy;
  
  sprintf(lastsub,"TRACE1: change_iconbar_screen (newhead,current: %d,%d)\n",newhead,iconbar_screen); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: change_iconbar_screen (newhead,current: %d,%d)\n",newhead,iconbar_screen); 
#endif 

  if(iconbar_screen==newhead) return;
  
  neary=nexticon_y;
    
  dx = screen_x(newhead) - screen_x(iconbar_screen);
  dy = screen_y(newhead) - screen_y(iconbar_screen);
  if(iconbarpos==ICONBARPOS_TOPRIGHT) {
    dx += screen_width(newhead) - screen_width(iconbar_screen);
  } else if(iconbarpos==ICONBARPOS_BOTLEFT) {
    dy += screen_height(newhead) - screen_height(iconbar_screen);
  } else if(iconbarpos==ICONBARPOS_BOTRIGHT) {
    dx += screen_width(newhead) - screen_width(iconbar_screen);
    dy += screen_height(newhead) - screen_height(iconbar_screen);
  }
  
  nexticon_x+=dx;
  nexticon_y+=dy;
  
  iconbar_screen=newhead;
  
  init_nearest_iconbar();
  c=get_nearest_iconbar(neary,NOT_ICONBAR);
  
  while(c!=NULL) {
  
    newx=c->icon_x+dx;
    newy=c->icon_y+dy;
    
    if(c->is_icon!=NOT_ICONBAR)
      move_client(c,newx,newy,0);
    else {
      c->icon_x=newx;
      c->icon_y=newy;
    }
      
    c=get_nearest_iconbar(neary,NOT_ICONBAR);
  }

}

void change_iconbarpos(int newpos)
{
  Client *c;
  int head=0,swapr2l=0, swapl2r=0, swapt2b=0, swapb2t=0, neary, newx, newy, old_icontab_width;
  
  sprintf(lastsub,"TRACE1: change_iconbarpos (newpos: %d)\n",newpos); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: change_iconbarpos (newpos: %d)\n",newpos); 
#endif 

  if(iconbarpos==newpos) return;
  
  head=iconbar_screen;

//#ifdef XINERAMA_SUPPORT
//    if(xinerama)
//      head=find_head(50, 50);
//#endif /* XINERAMA */

  if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_TOPLEFT) {
    if(newpos==ICONBARPOS_BOTRIGHT || newpos==ICONBARPOS_BOTLEFT)
      swapt2b=1;
  } else if(newpos==ICONBARPOS_TOPRIGHT || newpos==ICONBARPOS_TOPLEFT)
    swapb2t=1;
  
  if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_BOTLEFT) {
    if(newpos==ICONBARPOS_TOPRIGHT || newpos==ICONBARPOS_BOTRIGHT)
      swapl2r=1;
  } else if(newpos==ICONBARPOS_TOPLEFT || newpos==ICONBARPOS_BOTLEFT)
    swapr2l=1;

  iconbarpos=newpos;
  
  neary=nexticon_y;
  if(swapt2b==1) {
    nexticon_y=screen_y(head)+screen_height(head)-neary+screen_y(head);
    if(gMenubarSpace==0) nexticon_y-=22;
  } else if(swapb2t==1) {
    nexticon_y=screen_y(head)+screen_height(head)-neary+screen_y(head);
    if(gMenubarSpace==0) nexticon_y-=22;
  }
  
  if(swapl2r==1) {
    nexticon_x=screen_x(head)+screen_width(head);
  } else if(swapr2l==1) {
    nexticon_x=screen_x(head);
  }
  
  init_nearest_iconbar();
  c=get_nearest_iconbar(neary,NOT_ICONBAR);
  
  old_icontab_width=icontab_width;
  icontab_width=get_icontab_width();
    
  while(c!=NULL) {
  
    newx=c->icon_x;
    newy=c->icon_y;
    
    if(swapl2r==1)
      newx=screen_x(head)+screen_width(head)-c->icon_w;
    else if(swapr2l==1)
      newx=screen_x(head);
    
    if(swapt2b==1) {
      newy=screen_y(head)+screen_height(head)-c->icon_y+screen_y(head);
      if(gMenubarSpace==0) newy-=22;
    } else if(swapb2t==1) {
      newy=screen_y(head)+screen_height(head)-c->icon_y+screen_y(head);
      if(gMenubarSpace==0) newy-=22;
    }

    if(c->is_icon!=NOT_ICONBAR) {
      if(c->icon_w==old_icontab_width) {
        move_resize_client(c,newx,newy,icontab_width,TH,1);
        lower_client(c);
      } else
        move_client(c,newx,newy,1);
    } else {
      c->icon_x=newx;
      c->icon_y=newy;
    }
      
    c=get_nearest_iconbar(neary,NOT_ICONBAR);
  }
  
}

void compress_iconbars()
{
  Client *c;
  int neary, newx, newy, yadd;
    
  if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_TOPLEFT) {
    newy=screen_y(iconbar_screen);
    yadd=config.top.height;
    if(gMenubarSpace!=0) newy+=22;
  } else {
    newy=screen_y(iconbar_screen)+screen_height(iconbar_screen)-config.top.height;
    yadd=-config.top.height;
  }
  
  neary=newy;
  init_nearest_iconbar();
  c=get_nearest_iconbar(neary,NOT_ICONBAR);

  while(c!=NULL) {
  
    newy+=yadd;
    newx=c->icon_x;
    
    if(c->is_icon!=NOT_ICONBAR) {
      move_client(c,newx,newy,1);
    } else {
      c->icon_x=newx;
      c->icon_y=newy;
    }
    
    c=get_nearest_iconbar(neary,NOT_ICONBAR);
  }

  nexticon_y=newy+yadd;
  
}

void reset_all_inactive_clients(int newset, int which)
{
    Client *c;
    int i;
    
    c=client_list;
    if(c==NULL) return;
    
    for(i=0;i<client_count;i++) {
      if(c!=lastfocus_client) {
        if(which==BGWIN_DIMMED) {
          // check if client is becoming dimmed
          if((newset & which)!=0 && (c->auto_properties & (NONAUTO_FIRSTCLICK | AUTO_FOCUS))!=NONAUTO_FIRSTCLICK)
            SendHackCommand(c,1,HACKCMD_DIMWIN,1,0,0,0,0);
          else
            SendHackCommand(c,1,HACKCMD_DIMWIN,0,0,0,0,0);
        } else if(which==BGWIN_TRANSLUCENT) {
          // check if client is becoming translucent
          if((newset & which)!=0 && (c->auto_properties & (NONAUTO_FIRSTCLICK | AUTO_FOCUS))!=NONAUTO_FIRSTCLICK)
            SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.inactive_translucency,0,0,0,0);
          else
            SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c->osxwinfo.translucency,0,0,0,0);
        }
      }
      c=c->next;
    }
}

void change_options(int newattr)
{
  switch(newattr)
  {
    case itemBGWinDimmed:
      gBGWinSetting ^= BGWIN_DIMMED;
      Setup_Options_Menu(0);
      reset_all_inactive_clients(gBGWinSetting,BGWIN_DIMMED); // toggle dimness of inactive clients
      break;
      
    case itemBGWinTranslucent:
      gBGWinSetting ^= BGWIN_TRANSLUCENT;
      Setup_Options_Menu(0);
      reset_all_inactive_clients(gBGWinSetting,BGWIN_TRANSLUCENT); // toggle translucency of inactive clients
      break;
      
    case itemMinToDock:
#ifdef SHOW_WEIRD
      if(can_send_hackcmds==0)
        printf("Should not get itemMinToDock!\n");
#endif
      if(gMinMethod==minmethod_dock)
        gMinMethod=minmethod_iconbars;
      else if(gMinMethod==minmethod_iconbars)
        gMinMethod=minmethod_dock;
      Setup_Options_Menu(0);
      break;        

    case itemHideIconbars:
#ifdef SHOW_WEIRD
      if(can_send_hackcmds==0)
        printf("Should not get itemHideIconbars!\n");
#endif
      if(gHiding_Iconbars==0) {
        gHiding_Iconbars=1;
        will_hide_all_iconbars();
      } else {
        unhide_all_iconbars();
        gHiding_Iconbars=0;
      }
      Setup_Options_Menu(0);
      break;
      
    case itemTopLeft:
      change_iconbarpos(ICONBARPOS_TOPLEFT);
      Setup_Options_Menu(newattr);
      break;

    case itemTopRight:
      change_iconbarpos(ICONBARPOS_TOPRIGHT);
      Setup_Options_Menu(newattr);
      break;

    case itemBotLeft:
      change_iconbarpos(ICONBARPOS_BOTLEFT);
      Setup_Options_Menu(newattr);
      break;

    case itemBotRight:
      change_iconbarpos(ICONBARPOS_BOTRIGHT);
      Setup_Options_Menu(newattr);
      break;
    
    case itemSlowIconbars:
      iconbar_steps=ICONSTEPS_SLOW;
      Setup_Options_Menu(newattr);
      break;
      
    case itemMedIconbars:
      iconbar_steps=ICONSTEPS_MED;
      Setup_Options_Menu(newattr);
      break;
      
    case itemFastIconbars:
      iconbar_steps=ICONSTEPS_FAST;
      Setup_Options_Menu(newattr);
      break;
      
    case itemFancyIconbars:
      if(shrinking_iconbars==0)
        shrinking_iconbars=1;
      else
        shrinking_iconbars=0;
      Setup_Options_Menu(0);
      break;
      
    case itemInterleave:
#ifdef SHOW_WEIRD
      printf("Automatic interleave control not yet available...\n");
#endif
      Setup_Options_Menu(0);
      break;

    case itemLeaveMenubar:
      change_iconbar_menuspace();
      Setup_Options_Menu(0);
      break;
    
    default:
      if(newattr>=itemScreenListStart) {
        change_iconbar_screen(newattr-itemScreenListStart);
        Setup_Options_Menu(newattr);
      } else {
#ifdef SHOW_WEIRD
      printf("%s oddity: Got unknown attr in change_options\n",PROGNAME);
#endif
      }
      break;
  }

}

void check_missing_clients()
{
  Client *c, *orgc;
  int i, done=0;

  while(done==0) {
    c=client_list;
    for(i=0;i<client_count;i++) {
      if(c->osxwinfo.view_state==VIEWSTATE_HIDDEN) {
        printf("client has disappeared: %#lx (%s)\n",c,c->name);
        delete_client(c,-1);
        done=1;
        break;
      } else
        c=c->next;
        if(c==NULL) {
          done=1;
          break;
        } else if(i==client_count-1)
          done=1;
    }
  }
  
}

void map_all_clients() {
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c;

//  if(can_send_hackcmds!=0) return;
  
  sprintf(lastsub,"TRACE1: map_all_clients\n"); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE1: map_all_clients\n"); 
#endif
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef DEBUG
  printf("  children of root: %d\n",count);
#endif

  // Search through the windows that actually exist. (It's possible that some might have
  // disappeared while hidden, but they remain in the client list. This is because the window
  // manager had no notification they were unmapped as they already were unmapped when the
  // user chose to hide OroborOSX.)
  for(i=0;i<count;i++) { 
#ifdef DEBUG
    printf("  child %d (%#lx)\n",i,windows[i]);
#endif
    c=get_client_from_window(windows[i],FRAME);
    if(c!=NULL) {
      if(c->is_docked==0) {
      XMapWindow(dpy, c->frame); 
      if(c->is_icon!=NOT_ICONBAR)
        c->osxwinfo.view_state=VIEWSTATE_ICONBAR;
      else
        c->osxwinfo.view_state=VIEWSTATE_VISIBLE;
      XMapWindow(dpy, c->window); 
      set_wm_state(c->window, NormalState);
      SendHackCommand(c,0,HACKCMD_SETID,0,0,0,0,0); // uses the id stored for client, so args ignored
      if(gBGWinSetting & BGWIN_TRANSLUCENT)
        SendHackCommand(c,1,HACKCMD_TRANSLUCENT,c==focus_before_hide?c->osxwinfo.translucency:c->osxwinfo.inactive_translucency,0,0,0,0);
      if(gBGWinSetting & BGWIN_DIMMED)
        SendHackCommand(c,1,HACKCMD_DIMWIN,c==focus_before_hide?1:0,0,0,0,0);
      }
    }
  } 

#ifdef DEBUG
  printf("focus client before: %#lx\n",focus_before_hide);
#endif

  if(windows) XFree(windows);

  // loop through possibilities until there is nothing left (got NULL) or find OK client (then set NULL)
  while(focus_before_hide!=NULL) {

    if(focus_before_hide==-1) {	// focus client disappeared while hidden, so get topmost client and try again
      printf("%s warning: focus client disappeared while hidden\n",PROGNAME);
      focus_before_hide=get_top_most_client(WIN_LAYER_NORMAL,False);
    } else if(focus_before_hide->osxwinfo.view_state==VIEWSTATE_HIDDEN) {	// client is still hidden?!
#ifdef SHOW_WEIRD
      printf("focus client is hidden!! %#lx (%s)\n",focus_before_hide,focus_before_hide->name);
#endif
      delete_client(focus_before_hide,-1);
      focus_before_hide=get_top_most_client(WIN_LAYER_NORMAL,False);
    } else if(focus_before_hide->osxwinfo.view_state==VIEWSTATE_ICONBAR) {	// focus client is an iconbar!?
#ifdef SHOW_WEIRD
      printf("focus client is an iconbar!! %#lx (%s)\n",focus_before_hide,focus_before_hide->name);
#endif
      focus_before_hide=get_top_most_client(WIN_LAYER_NORMAL,False);		// go for topmost instead
    } else {	// everything is OK now, so set the focus client and get out
#ifdef DEBUG
      printf("focus client OK: %#lx\n",focus_before_hide);
#endif
      set_focus_client(focus_before_hide);
      focus_before_hide=NULL;
    }

  }

  check_missing_clients();

}

void unmap_all_clients() {
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c;
  
  focus_before_hide=get_focus_client();
  
//  if(can_send_hackcmds!=0) {
//    SendSimpleHackCommand(HACKCMD_HIDE_ALLXWIN,1);
//    return;
//  }
  
  sprintf(lastsub,"TRACE1: unmap_all_clients\n"); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE1: unmap_all_clients\n"); 
#endif
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef DEBUG
  printf("  children of root: %d\n",count);
#endif

  for(i=0;i<count;i++) 
  { 
#ifdef DEBUG
    printf("  child %d (%#lx)\n",i,windows[i]);
#endif
    c=get_client_from_window(windows[i],FRAME);
    if(c!=NULL)
      if(c->is_docked==0) {
        c->ignore_unmap++; 
        XUnmapWindow(dpy, c->frame); 
        c->osxwinfo.view_state=VIEWSTATE_HIDDEN;
        XUnmapWindow(dpy, c->window); 
        set_wm_state(c->window, WithdrawnState); 
      }
  } 

  if(windows) XFree(windows);
  
//  c=client_list;
//  for(i=0;i<client_count;i++) {
//    c=c->next;
//  }
}

/*
 * For debugging purposes
 */
void list_all_windows()
{
  Window *windows=NULL, w1, w2; 
  int count, i, hints, argc; 
  char **argv;
  Client *c;
  XWindowAttributes attr; 
  XClassHint xch;

  XGetWindowAttributes(dpy, root, &attr); 

  XQueryTree(dpy, root, &w1, &w2, &windows, &count);
  if(count==1)
    printf("%s info: there is %d top level X11 window...\n",PROGNAME,count);
  else
    printf("%s info: there are %d top level X11 windows...\n",PROGNAME,count);

  for(i=count-1;i>=0;i--) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr);
    if(attr.map_state==IsViewable)
      printf("  window %d (%#lx) viewable, (X,Y): %d,%d (W,H): %d,%d\n",i,windows[i],attr.x,attr.y,attr.width,attr.height);
    else
      printf("  window %d (%#lx) not viewable, (X,Y): %d,%d (W,H): %d,%d\n",i,windows[i],attr.x,attr.y,attr.width,attr.height);
    c=get_client_from_window(windows[i], FRAME);
    if(c==NULL) {
      if(windows[i]==check_win)
        printf("     this is the oroborus 'check_win' created for use with gnome...\n");
      else
        printf("     this window is not recognised as a client by OroborOSX!\n");
      hints=XGetClassHint(dpy,windows[i],&xch);
      if(hints!=0) {
        printf("     window class & name: '%s', '%s'\n",xch.res_class,xch.res_name);
        XFree(xch.res_name);
        XFree(xch.res_class);
      }
      if(XGetCommand(dpy,windows[i],&argv,&argc)!=0) {
        printf("     window command (%d args): %s\n",argc,*argv);
        if(*argv!=NULL) XFreeStringList(argv);
      }
    } else {
      printf("     client id, title, class, name, group: %#lx , '%s' , '%s' , '%s' ,  '%s'\n",c,c->name,c->clientclass,c->clientcmd,c->clientgroup);
      printf("     client frame, window, osxwinref, iconbarstate: %#lx, %#lx, %#lx, %d\n",c->frame,c->window,c->osxwinfo.winref,c->is_icon);
      if(XGetCommand(dpy,windows[i],&argv,&argc)!=0) {
        printf("     window command (%d args): %s\n",argc,*argv);
        if(*argv!=NULL) XFreeStringList(argv);
      }
    }
  } 

  if(windows) XFree(windows);

  c=client_list;
  if(c==NULL) {
    printf("%s info: no clients\n",PROGNAME);
    return;
  }
  if(client_count==0) {
    printf("%s oddity: No clients in list, but client_list is %#lx !?\n",PROGNAME,client_list);
    return;
  }
  
}

void list_all_clients()
{
  Client *c;
  int i;
  
  c=client_list;
  if(c==NULL) {
    printf("%s info: no clients\n",PROGNAME);
    return;
  }
  if(client_count==0) {
    printf("%s oddity: No clients in list, but client_list is %#lx !?\n",PROGNAME,client_list);
    return;
  }

  printf("%s info: List of clients:\n",PROGNAME);
  for(i=0;i<client_count;i++) {
    printf("  %#lx: title, class, name, grp: '%s' , '%s' , '%s' , '%s'\n",c,c->name,c->clientclass,c->clientcmd,c->clientgroup);
    printf("            frame, window, osxwinref, iconbarstate: %#lx, %#lx, %#lx, %d\n",c->frame,c->window,c->osxwinfo.winref,c->is_icon);
    c=c->next;
  }

}

/*
 * Check that OSX windows are off-screen
 */
void check_osx_windows_bounds() {
  Window *windows=NULL; 
  int count, i; 
  Client *c;

    sprintf(lastsub,"TRACE3: check_osx_windows_bounds\n");
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: check_osx_windows_bounds\n");
#endif

  c=client_list;
  if(c==NULL) return;
  if(client_count==0) {
    printf("%s oddity: No clients in list, but client_list is %#lx !?\n",PROGNAME,client_list);
    return;
  }
  
  for(i=0;i<client_count;i++) {
    if(c->osxwinfo.winref!=0)
        Check_OSX_Window_Bounds(&c->osxwinfo);
    c=c->next;
  }
}

/*
 * Find the OSX WindowRef (returned as a long) for the bottom-most client (for the "Next" menu item)
 */
long get_next_osxwinref(int raiseit)
{
    Client *c, *fc;
    long osxwr=0;
    
    fc=get_focus_client();
    
#ifdef DEBUG
    if(fc==NULL)
      printf("Got NULL focus client in get_next_osxwinref.\n");
#endif
    if(fc==NULL) return 0;
    
    c=get_bottom_most_client(get_gnome_hint(fc->window,WIN_LAYER),False);
    
#ifdef DEBUG
    if(c==fc)
      printf("Got focus client is next client in get_next_osxwinref.\n");
#endif
    if(c==fc) return 0;
    
    if(c!=NULL) {
      if(raiseit!=0) {
        if(c->is_docked)
          will_undock_client(c);
        else {
          set_focus_client(c);
          raise_client(c);
        }
      }
      osxwr=c->osxwinfo.winref;
    }
    
    sprintf(lastsub,"TRACE2: get_next_osxwinref -got next client: %#lx (winref: %d)\n",c,osxwr);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE3: get_next_osxwinref -got next client: %#lx (winref: %d)\n",c,osxwr);
#endif
  
    return osxwr;
}

